Package: MetaboAnnotation
Title: Utilities for Annotation of Metabolomics Data
Version: 1.14.0
Description: 
    High level functions to assist in annotation of (metabolomics) data sets.
    These include functions to perform simple tentative annotations based on
    mass matching but also functions to consider m/z and retention times for
    annotation of LC-MS features given that respective reference values are
    available. In addition, the function provides high-level functions to
    simplify matching of LC-MS/MS spectra against spectral libraries and objects
    and functionality to represent and manage such matched data.
Authors@R: 
    c(person(given = "Michael", family = "Witting",
             role = c("aut"),
             email = "michael.witting@helmholtz-muenchen.de",
             comment = c(ORCID = "0000-0002-1462-4426")),
      person(given = "Johannes", family = "Rainer",
             email = "Johannes.Rainer@eurac.edu",
             role = c("aut", "cre"),
             comment = c(ORCID = "0000-0002-6977-7147")),
      person(given = "Andrea", family = "Vicini",
             email = "andrea.vicini@eurac.edu",
             role = c("aut"),
             comment = c(ORCID = "0000-0001-9438-6909")),
      person(given = "Carolin", family = "Huber",
             email = "carolin-elisabeth.huber@ufz.de",
             role = c("aut"),
             comment = c(ORCID = "0000-0002-9355-8948")),
      person(given = "Philippine", family = "Louail",
             email = "philippine.louail@eurac.edu",
             role = c("aut"),
             comment = c(ORCID = "0009-0007-5429-6846")),
      person(given = "Nir", family = "Shachaf",
             email = "nir.shachaf@weizmann.ac.il",
	     role = "ctb"))
Depends: R (>= 4.0.0)
Imports: BiocGenerics, MsCoreUtils, MetaboCoreUtils, ProtGenerics,
        methods, S4Vectors, Spectra (>= 1.17.6), BiocParallel,
        SummarizedExperiment, QFeatures, AnnotationHub, graphics,
        CompoundDb
Suggests: testthat, knitr, msdata, BiocStyle, rmarkdown, plotly, shiny,
        shinyjs, msentropy, DT, microbenchmark, mzR
Enhances: RMariaDB, RSQLite
License: Artistic-2.0
VignetteBuilder: knitr
BugReports: https://github.com/RforMassSpectrometry/MetaboAnnotation/issues
URL: https://github.com/RforMassSpectrometry/MetaboAnnotation
biocViews: Infrastructure, Metabolomics, MassSpectrometry
Roxygen: list(markdown=TRUE)
RoxygenNote: 7.3.3
Encoding: UTF-8
Collate: 'AllClassUnions.R' 'AllGenerics.R' 'CompAnnotationSource.R'
        'CompDbSource.R' 'Matched.R' 'MatchedSpectra.R'
        'group_standards.R' 'hidden-aliases.R' 'matchFormula.R'
        'matchSpectra.R' 'matchValues.R' 'validateMatchedSpectra.R'
git_url: https://git.bioconductor.org/packages/MetaboAnnotation
git_branch: RELEASE_3_22
git_last_commit: 790d88f
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-10-29
NeedsCompilation: no
Packaged: 2025-10-30 04:56:24 UTC; biocbuild
Author: Michael Witting [aut] (ORCID: <https://orcid.org/0000-0002-1462-4426>),
  Johannes Rainer [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6977-7147>),
  Andrea Vicini [aut] (ORCID: <https://orcid.org/0000-0001-9438-6909>),
  Carolin Huber [aut] (ORCID: <https://orcid.org/0000-0002-9355-8948>),
  Philippine Louail [aut] (ORCID:
    <https://orcid.org/0009-0007-5429-6846>),
  Nir Shachaf [ctb]
Maintainer: Johannes Rainer <Johannes.Rainer@eurac.edu>

# Generated by roxygen2: do not edit by hand

export(CompDbSource)
export(CompareSpectraParam)
export(Mass2MzParam)
export(Mass2MzRtParam)
export(MassBankSource)
export(MatchForwardReverseParam)
export(Matched)
export(MatchedSpectra)
export(Mz2MassParam)
export(Mz2MassRtParam)
export(MzParam)
export(MzRtParam)
export(ScoreThresholdParam)
export(SelectMatchesParam)
export(SingleMatchParam)
export(TopRankedMatchesParam)
export(ValueParam)
export(createStandardMixes)
export(matchFormula)
export(matchMz)
export(matchSpectra)
export(matchValues)
export(matches)
export(pruneTarget)
export(queryIndex)
export(scoreVariables)
export(target)
export(targetIndex)
export(validateMatchedSpectra)
export(whichQuery)
export(whichTarget)
exportClasses(CompAnnotationSource)
exportClasses(CompDbSource)
exportClasses(CompareSpectraParam)
exportClasses(Matched)
exportClasses(MatchedSpectra)
exportMethods("$")
exportMethods("[")
exportMethods(addMatches)
exportMethods(addProcessing)
exportMethods(colnames)
exportMethods(endoapply)
exportMethods(filterMatches)
exportMethods(lapply)
exportMethods(length)
exportMethods(matchSpectra)
exportMethods(matchedData)
exportMethods(metadata)
exportMethods(plotSpectraMirror)
exportMethods(query)
exportMethods(queryVariables)
exportMethods(setBackend)
exportMethods(show)
exportMethods(spectraVariables)
exportMethods(targetVariables)
importClassesFrom(CompoundDb,CompDb)
importClassesFrom(ProtGenerics,Param)
importClassesFrom(QFeatures,QFeatures)
importClassesFrom(S4Vectors,DataFrame)
importClassesFrom(Spectra,MsBackendDataFrame)
importClassesFrom(Spectra,Spectra)
importClassesFrom(SummarizedExperiment,SummarizedExperiment)
importFrom(BiocGenerics,colnames)
importFrom(BiocGenerics,lapply)
importFrom(BiocParallel,SerialParam)
importFrom(BiocParallel,bplapply)
importFrom(BiocParallel,bpmapply)
importFrom(BiocParallel,bpparam)
importFrom(BiocParallel,register)
importFrom(CompoundDb,CompDb)
importFrom(MetaboCoreUtils,adductNames)
importFrom(MetaboCoreUtils,mass2mz)
importFrom(MetaboCoreUtils,mz2mass)
importFrom(MetaboCoreUtils,standardizeFormula)
importFrom(MsCoreUtils,between)
importFrom(MsCoreUtils,common)
importFrom(MsCoreUtils,ndotproduct)
importFrom(MsCoreUtils,ppm)
importFrom(MsCoreUtils,rbindFill)
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,endoapply)
importFrom(Spectra,MsBackendDataFrame)
importFrom(Spectra,Spectra)
importFrom(Spectra,joinPeaks)
importFrom(Spectra,scalePeaks)
importFrom(grDevices,n2mfrow)
importFrom(graphics,par)
importFrom(methods,"slot<-")
importFrom(methods,.hasSlot)
importFrom(methods,as)
importFrom(methods,callNextMethod)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,slotNames)
importFrom(methods,validObject)
importMethodsFrom(AnnotationHub,query)
importMethodsFrom(BiocGenerics,colnames)
importMethodsFrom(ProtGenerics,as.list)
importMethodsFrom(ProtGenerics,setBackend)
importMethodsFrom(ProtGenerics,spectraVariables)
importMethodsFrom(S4Vectors,"$")
importMethodsFrom(S4Vectors,cbind)
importMethodsFrom(S4Vectors,metadata)
importMethodsFrom(Spectra,"spectraNames<-")
importMethodsFrom(Spectra,addProcessing)
importMethodsFrom(Spectra,backendBpparam)
importMethodsFrom(Spectra,compareSpectra)
importMethodsFrom(Spectra,containsMz)
importMethodsFrom(Spectra,filterPrecursorMzRange)
importMethodsFrom(Spectra,filterRt)
importMethodsFrom(Spectra,peaksData)
importMethodsFrom(Spectra,peaksVariables)
importMethodsFrom(Spectra,plotSpectraMirror)
importMethodsFrom(Spectra,precursorMz)
importMethodsFrom(Spectra,rtime)
importMethodsFrom(Spectra,spectraData)
importMethodsFrom(Spectra,spectraNames)
importMethodsFrom(SummarizedExperiment,rowData)
importMethodsFrom(methods,show)

#' @title Representation of Spectra matches
#'
#' @name MatchedSpectra
#'
#' @aliases MatchedSpectra MatchedSpectra-class
#'
#' @description
#'
#' Matches between query and target spectra can be represented by the
#' `MatchedSpectra` object. Functions like the [matchSpectra()] function will
#' return this type of object. By default, all data accessors work as
#' *left joins* between the *query* and the *target* spectra, i.e. values are
#' returned for each *query* spectrum with eventual duplicated entries (values)
#' if the query spectrum matches more than one target spectrum.
#'
#' @section Creation, subset and filtering:
#'
#' `MatchedSpectra` objects are the result object from the [matchSpectra()].
#' While generally not needed, `MatchedSpectra` objects can also be created
#' with the `MatchedSpectra` function providing the `query` and `target`
#' `Spectra` objects as well as a `data.frame` with the *matches* between
#' query and target elements. This data frame is expected to have columns
#' `"query_idx"`, `"target_idx"` with the `integer` indices of query and
#' target objects that are *matched* and a column `"score"` with a `numeric`
#' score for the match.
#'
#' `MatchedSpectra` objects can be subset using:
#'
#' - `[` subset the `MatchedSpectra` selecting `query` spectra to keep with
#'   parameter `i`. The `target` spectra will by default be returned as-is.
#'
#' - `pruneTarget` *cleans* the `MatchedSpectra` object by removing non-matched
#'   target spectra.
#'
#' In addition, `MatchedSpectra` can be filtered with any of the filtering
#' approaches defined for [Matched()] objects: [SelectMatchesParam()],
#' [TopRankedMatchesParam()] or [ScoreThresholdParam()].
#'
#'
#' @section Extracting data:
#'
#' - `$` extracts a single spectra variable from the `MatchedSpectra` `x`. Use
#'   `spectraVariables` to get all available spectra variables. Prefix
#'   `"target_"` is used for spectra variables from the *target* `Spectra`. The
#'   matching scores are available as *spectra variable* `"score"`.
#'   Similar to a left join between the query and target spectra, this function
#'   returns a value for each query spectrum with eventual duplicated values for
#'   query spectra matching more than one target spectrum. If spectra variables
#'   from the target spectra are extracted, an `NA` is reported for *query*
#'   spectra that don't match any target spectra. See examples below for more
#'   details.
#'
#' - `length` returns the number of **query** spectra.
#'
#' - `matchedData` same as `spectraData` below.
#'
#' - `query` returns the *query* `Spectra`.
#'
#' - `queryVariables` returns the `spectraVariables` of *query*.
#'
#' - `spectraData` returns spectra variables from the query and/or target
#'   `Spectra` as a `DataFrame`. Parameter `columns` allows to define which
#'   variables should be returned (defaults to
#'   `columns = spectraVariables(object)`), spectra variable names of the target
#'   spectra need to be prefixed with `target_` (e.g. `target_msLevel` to get
#'   the MS level from target spectra). The score from the matching function is
#'   returned as spectra variable `"score"`. Similar to `$`, this function
#'   performs a *left join* of spectra variables from the *query* and *target*
#'   spectra returning all values for all query spectra (eventually returning
#'   duplicated elements for query spectra matching multiple target spectra)
#'   and the values for the target spectra matched to the respective query
#'   spectra. See help on `$` above or examples below for details.
#'
#' - `spectraVariables` returns all available spectra variables in the *query*
#'   and *target* spectra. The prefix `"target_"` is used to label spectra
#'   variables of target spectra (e.g. the name of the spectra variable for the
#'   MS level of target spectra is called `"target_msLevel"`).
#'
#' - `target` returns the *target* `Spectra`.
#'
#' - `targetVariables` returns the `spectraVariables` of *target* (prefixed
#'    with `"target_"`).
#'
#' - `whichTarget` returns an `integer` with the indices of the spectra in
#'   *target* that match at least on spectrum in *query*.
#'
#' - `whichQuery` returns an `integer` with the indices of the spectra in
#'   *query* that match at least on spectrum in *target*.
#'
#' @section Data manipulation and plotting:
#'
#' - `addProcessing`: add a processing step to both the *query* and *target*
#'   `Spectra` in `object`. Additional parameters for `FUN` can be passed *via*
#'   `...`. See `addProcessing` documentation in [Spectra::Spectra()] for more
#'   information.
#'
#' - `plotSpectraMirror`: creates a mirror plot between the query and each
#'   matching target spectrum. Can only be applied to a `MatchedSpectra` with a
#'   single query spectrum. Setting parameter `scalePeaks = TRUE` will scale
#'   the peak intensities per spectrum to a total sum of one for a better
#'   graphical visualization. Additional plotting parameters can be passed
#'   through `...`. The parameters `ppm` and `tolerance` can be used to
#'   define the m/z tolerance for matching peaks between the query and target
#'   spectra. If not provided by the user, the values from the `param`
#'   object used to create the `MatchedSpectra` object are used; if these are
#'   missing, the default values (`ppm =20` and `tolerance = 0`) are used.
#'
#' - `setBackend`: allows to change the *backend* of both the query and target
#'   [Spectra::Spectra()] object. The function will return a `MatchedSpectra`
#'   object with the query and target `Spectra` changed to the specified
#'   `backend`, which can be any backend extending [Spectra::MsBackend].
#'
#' @param backend for `setBackend`: instance of an object extending
#'   [Spectra::MsBackend]. See help for [Spectra::setBackend()]
#'   for more details.
#'
#' @param columns for `spectraData`: `character` vector with spectra variable
#'   names that should be extracted.
#'
#' @param FUN for `addProcessing`: function to be applied to the peak matrix
#'   of each spectrum in `object`. See [Spectra::Spectra()] for more details.
#'
#' @param main for `plotSpectraMirror`: an optional title for each plot.
#'
#' @param matches `data.frame` with columns `"query_idx"` (`integer`),
#'   `"target_idx"` (`integer`) and `"score"` (`numeric`) representing the
#'   *n:m* mapping of elements between the `query` and the `target` `Spectra`.
#'
#' @param name for `$`: the name of the spectra variable to extract.
#'
#' @param object `MatchedSpectra` object.
#'
#' @param scalePeaks for `plotSpectraMirror`: `logical(1)` if peak intensities
#'   (per spectrum) should be scaled to a total sum of one (per spectrum) prior
#'   to plotting.
#'
#' @param spectraVariables for `addProcessing`: `character` with additional
#'   spectra variables that should be passed along to the function defined
#'   with `FUN`. See [Spectra::Spectra()] for details.
#'
#' @param target `Spectra` with the spectra against which `query` has been
#'   matched.
#'
#' @param query `Spectra` with the query spectra.
#'
#' @param x `MatchedSpectra` object.
#'
#' @param xlab for `plotSpectraMirror`: the label for the x-axis.
#'
#' @param ylab for `plotSpectraMirror`: the label for the y-axis.
#'
#' @param ... for `addProcessing`: additional parameters for the function `FUN`.
#'   For `plotSpectraMirror`: additional parameters passed to the plotting
#'   functions.
#'
#' @return See individual method desciption above for details.
#'
#' @importClassesFrom Spectra Spectra
#'
#' @importFrom Spectra Spectra
#'
#' @exportClass MatchedSpectra
#'
#' @author Johannes Rainer
#'
#' @rdname MatchedSpectra
#'
#' @seealso [Matched()] for additional functions available for `MatchedSpectra`.
#'
#' @examples
#'
#' ## Creating a dummy MatchedSpectra object.
#' library(Spectra)
#' df1 <- DataFrame(
#'     msLevel = 2L, rtime = 1:10,
#'     spectrum_id = c("a", "b", "c", "d", "e", "f", "g", "h", "i", "j"))
#' df2 <- DataFrame(
#'     msLevel = 2L, rtime = rep(1:10, 20),
#'     spectrum_id = rep(c("A", "B", "C", "D", "E"), 20))
#' sp1 <- Spectra(df1)
#' sp2 <- Spectra(df2)
#' ## Define matches between query spectrum 1 with target spectra 2 and 5,
#' ## query spectrum 2 with target spectrum 2 and query spectrum 4 with target
#' ## spectra 8, 12 and 15.
#' ms <- MatchedSpectra(
#'     sp1, sp2, matches = data.frame(query_idx = c(1L, 1L, 2L, 4L, 4L, 4L),
#'                                    target_idx = c(2L, 5L, 2L, 8L, 12L, 15L),
#'                                    score = 1:6))
#'
#' ## Which of the query spectra match at least one target spectrum?
#' whichQuery(ms)
#'
#' ## Extracting spectra variables: accessor methods for spectra variables act
#' ## as "left joins", i.e. they return a value for each query spectrum, with
#' ## eventually duplicated elements if one query spectrum matches more than
#' ## one target spectrum.
#'
#' ## Which target spectrum matches at least one query spectrum?
#' whichTarget(ms)
#'
#' ## Extracting the retention times of the query spectra.
#' ms$rtime
#'
#' ## We have duplicated retention times for query spectrum 1 (matches 2 target
#' ## spectra) and 4 (matches 3 target spectra). The retention time is returned
#' ## for each query spectrum.
#'
#' ## Extracting retention times of the target spectra. Note that only retention
#' ## times for target spectra matching at least one query spectrum are returned
#' ## and an NA is reported for query spectra without matching target spectrum.
#' ms$target_rtime
#'
#' ## The first query spectrum matches target spectra 2 and 5, thus their
#' ## retention times are returned as well as the retention time of the second
#' ## target spectrum that matches also query spectrum 2. The 3rd query spectrum
#' ## does match any target spectrum, thus `NA` is returned. Query spectrum 4
#' ## matches target spectra 8, 12, and 15, thus the next reported retention
#' ## times are those from these 3 target spectra. None of the remaining 6 query
#' ## spectra matches any target spectra and thus `NA` is reported for each of
#' ## them.
#'
#' ## With `queryIndex` and `targetIndex` it is possible to extract the indices
#' ## of the matched query-index pairs
#' queryIndex(ms)
#' targetIndex(ms)
#'
#' ## The first match is between query index 1 and target index 2, the second
#' ## match between query index 1 and target index 5 and so on.
#' ## We could use these indices to extract a `Spectra` object containing only
#' ## matched target spectra and assign a spectra variable with the indices of
#' ## the query spectra
#' matched_target <- target(ms)[targetIndex(ms)]
#' matched_target$query_index <- queryIndex(ms)
#'
#' ## This `Spectra` object thus contains information from the matching, but
#' ## is a *conventional* `Spectra` object that could be used for further
#' ## analyses.
#'
#' ## `spectraData` can be used to extract all (or selected) spectra variables
#' ## from the object. Same as with `$`, a left join between the specta
#' ## variables from the query spectra and the target spectra is performed. The
#' ## prefix `"target_"` is used to label the spectra variables from the target
#' ## spectra. Below we extract selected spectra variables from the object.
#' res <- spectraData(ms, columns = c("rtime", "spectrum_id",
#'     "target_rtime", "target_spectrum_id"))
#' res
#' res$spectrum_id
#' res$target_spectrum_id
#'
#' ## Again, all values for query spectra are returned and for query spectra not
#' ## matching any target spectrum NA is reported as value for the respecive
#' ## variable.
#'
#' ## The example matched spectra object contains all query and all target
#' ## spectra. Below we subset the object keeping only query spectra that are
#' ## matched to at least one target spectrum.
#' ms_sub <- ms[whichQuery(ms)]
#'
#' ## ms_sub contains now only 3 query spectra:
#' length(query(ms_sub))
#'
#' ## while the original object contains all 10 query spectra:
#' length(query(ms))
#'
#' ## Both object contain however still the full target `Spectra`:
#' length(target(ms))
#' length(target(ms_sub))
#'
#' ## With the `pruneTarget` we can however reduce also the target spectra to
#' ## only those that match at least one query spectrum
#' ms_sub <- pruneTarget(ms_sub)
#' length(target(ms_sub))
NULL

#' @importFrom Spectra Spectra
#'
#' @noRd
setClass(
    "MatchedSpectra",
    contains = "Matched",
    prototype = prototype(
        query = Spectra(),
        target = Spectra(),
        matches = data.frame(query_idx = integer(),
                             target_idx = integer(),
                             score = numeric()),
        metadata = list(),
        version = "0.1")
)

#' @importFrom Spectra Spectra
#'
#' @export
#'
#' @rdname MatchedSpectra
MatchedSpectra <- function(query = Spectra(), target = Spectra(),
                           matches = data.frame(query_idx = integer(),
                                                target_idx = integer(),
                                                score = numeric())) {
    .matched_spectra(query = query, target = target, matches = matches)
}

.matched_spectra <- function(query = Spectra(), target = Spectra(),
                             matches = data.frame(query_idx = integer(),
                                                  target_idx = integer(),
                                                  score = numeric()),
                             metadata = list(), validate = TRUE) {
    res <- new("MatchedSpectra")
    slot(res, "query", check = FALSE) <- query
    slot(res, "target", check = FALSE) <- target
    slot(res, "matches", check = FALSE) <- matches
    slot(res, "metadata", check = FALSE) <- metadata
    if (validate)
        validObject(res)
    res
}

setValidity("MatchedSpectra", function(object) {
    msg <- .validate_matches_format(object@matches)
    if (length(msg)) return(msg)
    msg <- .validate_matches_content(object@matches, length(object@query),
                                     length(object@target))
    if (length(msg)) return(msg)
    TRUE
})

#' @importMethodsFrom ProtGenerics spectraVariables
#'
#' @importMethodsFrom BiocGenerics colnames
#'
#' @rdname MatchedSpectra
#'
#' @export
setMethod("spectraVariables", "MatchedSpectra", function(object) {
    svq <- spectraVariables(query(object))
    svt <- spectraVariables(target(object))
    cns <- colnames(object@matches)
    c(svq, paste0("target_", svt), cns[!cns %in% c("query_idx", "target_idx")])
})

#' @rdname MatchedSpectra
setMethod("queryVariables", "MatchedSpectra", function(object) {
    spectraVariables(query(object))
})

#' @rdname MatchedSpectra
setMethod("targetVariables", "MatchedSpectra", function(object) {
    paste0("target_", spectraVariables(target(object)))
})

#' @exportMethod colnames
#'
#' @rdname MatchedSpectra
setMethod("colnames", "MatchedSpectra", function(x) {
    spectraVariables(x)
})

#' @rdname MatchedSpectra
setMethod("$", "MatchedSpectra", function(x, name) {
    if (name %in% spectraVariables(x@query))
        qry <- spectraData(x@query, name)
    else qry <- spectraData(x@query, "msLevel")
    if (name %in% paste0("target_", spectraVariables(x@target)))
        trg <- spectraData(x@target, sub("^target_", "", name))
    else trg <- data.frame()
    .dollar(qry, trg, x@matches, name)
})

#' @importMethodsFrom Spectra spectraData
#'
#' @rdname MatchedSpectra
setMethod("spectraData", "MatchedSpectra",
          function(object, columns = spectraVariables(object)) {
              mtches <- object@matches
              ## Better to subset target Spectra before passing it on.
              cols_trg <- spectraVariables(object@target)
              cols_trg <- cols_trg[paste0("target_", cols_trg) %in% columns]
              if (length(cols_trg)) {
                  trg <- as.data.frame(
                      spectraData(object@target[mtches$target_idx], cols_trg))
                  mtches$target_idx <- seq_len(nrow(mtches))
              } else
                  trg <- data.frame()
              cols_qry <- columns[columns %in% spectraVariables(object@query)]
              .matchedData(
                  spectraData(object@query, unique(c("msLevel", cols_qry))),
                  trg, mtches, columns)
          })

#' @rdname MatchedSpectra
setMethod("matchedData", "MatchedSpectra",
          function(object, columns = spectraVariables(object), ...) {
              spectraData(object, columns)
          })

#' @importMethodsFrom Spectra addProcessing
#'
#' @rdname MatchedSpectra
#'
#' @export
setMethod(
    "addProcessing", "MatchedSpectra",
    function(object, FUN, ..., spectraVariables = character()) {
        if (missing(FUN))
            return(object)
        object@query <- addProcessing(object@query, FUN = FUN, ...,
                                      spectraVariables = spectraVariables)
        object@target <- addProcessing(object@target, FUN = FUN, ...,
                                       spectraVariables = spectraVariables)
        object
    })

#' @importMethodsFrom Spectra plotSpectraMirror
#'
#' @rdname MatchedSpectra
#'
#' @importFrom graphics par
#'
#' @importFrom grDevices n2mfrow
#'
#' @importFrom Spectra scalePeaks
#'
#' @importMethodsFrom ProtGenerics as.list
#'
#' @export
setMethod(
    "plotSpectraMirror", "MatchedSpectra",
    function(x, xlab = "m/z", ylab = "intensity", main = "",
             scalePeaks = FALSE, ...) {
        if (length(query(x)) != 1)
            stop("Length of 'query(x)' has to be 1.")
        dots <- list(...)
        pl <- as.list(x@metadata[["param"]])
        ppm_res <- .res_setting(dots = dots, param_list = pl,
                                 name = "ppm", default = 20)
        tol_res <- .res_setting(dots = dots, param_list = pl,
                                 name = "tolerance",default = 10)
        y <- x@target[x@matches$target_idx]
        if (!length(y))
            y <- Spectra(DataFrame(msLevel = 2L))
        if (scalePeaks) {
            x <- scalePeaks(query(x)[1L])
            y <- scalePeaks(y)
        } else x <- query(x)[1L]
        par(mfrow = n2mfrow(length(y)))
        for (i in seq_along(y))
            plotSpectraMirror(x = x, y = y[i],
                              xlab = xlab, ylab = ylab, main = main,
                              ppm = ppm_res, tolerance = tol_res,...)
    })

#' Helper for ppm and tolerance setting.
#' @noRd
.res_setting <- function(dots = list(), param_list, name, default) {
    if (!is.null(dots[[name]]))
        return(dots[[name]])
    if (!is.null(param_list[[name]]))
        return(param_list[[name]])
    return(default)
}

#' @importMethodsFrom ProtGenerics setBackend
#'
#' @rdname MatchedSpectra
#'
#' @export
setMethod("setBackend", c("MatchedSpectra", "MsBackend"),
          function(object, backend, ...) {
              object@query <- setBackend(object@query, backend = backend, ...)
              object@target <- setBackend(object@target, backend = backend, ...)
              object
          })

#' @title Validating MatchedSpectra
#'
#' @description
#'
#' The `validateMatchedSpectra()` function opens a simple shiny application
#' that allows to browse results stored in a `MatchedSpectra` object and to
#' *validate* the presented matches. For each query spectrum a table with
#' matched target spectra are shown (if available) and an interactive mirror
#' plot is generated. Valid matches can be selected using a check box which is
#' displayed below the mirror plot. Upon pushing the "Save & Close"
#' button the app is closed and a filtered `MatchedSpectra` is returned,
#' containing only *validated* matches.
#'
#' Note that column `"query_index_"` and `"target_index_"` are temporarily
#' added to the query and target `Spectra` object to display them in the
#' interactive graphics for easier identification of the compared spectra.
#'
#' @param object A non-empty instance of class `MatchedSpectra`.
#'
#' @return A `MatchedSpectra` with validated results.
#'
#' @export
#'
#' @importFrom BiocParallel bpparam SerialParam register
#'
#' @importFrom methods slotNames
#'
#' @author Carolin Huber, Michael Witting, Johannes Rainer
#'
#' @examples
#'
#' library(Spectra)
#' ## Load test data
#' fl <- system.file("TripleTOF-SWATH", "PestMix1_DDA.mzML", package = "msdata")
#' pest_ms2 <- filterMsLevel(Spectra(fl), 2L)
#' pest_ms2 <- pest_ms2[c(808, 809, 945:955)]
#' load(system.file("extdata", "minimb.RData", package = "MetaboAnnotation"))
#'
#' ## Normalize intensities and match spectra
#' csp <- CompareSpectraParam(requirePrecursor = TRUE,
#'                            THRESHFUN = function(x) x >= 0.7)
#' norm_int <- function(x) {
#'     x[, "intensity"] <- x[, "intensity"] / max(x[, "intensity"]) * 100
#'     x
#' }
#' ms <- matchSpectra(addProcessing(pest_ms2, norm_int),
#'                    addProcessing(minimb, norm_int), csp)
#'
#' ## validate matches using the shiny app. Note: the call is only executed
#' ## in interactive mode.
#' if (interactive()) {
#'     res <- validateMatchedSpectra(ms)
#' }
validateMatchedSpectra <- function(object) {
    if (!.is_shiny_available())
        stop("The use of 'validateMatchedSpectra' requires package 'shiny'.",
             " Please install with 'BiocManager::install(\"shiny\")'")
    if (!.is_shinyjs_available())
        stop("The use of 'validateMatchedSpectra' requires package 'shinyjs'.",
             " Please install with 'BiocManager::install(\"shinyjs\")'")
    if (!.is_dt_available())
        stop("The use of 'validateMatchedSpectra' requires package 'DT'.",
             " Please install with 'BiocManager::install(\"DT\")'")

    stopifnot(inherits(object, "MatchedSpectra"))
    if (!length(object))
        stop("The 'MatchedSpectra' object is empty.")

    ## Add query and target index
    object@query$query_index_ <- seq_along(object@query)
    object@target$index_ <- seq_along(object@target)

    bpp <- bpparam()
    on.exit(register(bpp))
    register(SerialParam())
    ui <- shiny::fluidPage(shinyjs::useShinyjs(),
                    shiny::titlePanel(shiny::div("validateMatchedSpectra")),
                    shiny::sidebarLayout(shiny::sidebarPanel(
                        shiny::selectInput("query", "Query Spectra:",
                                    choices = list(),
                                    multiple = FALSE,
                                    selectize = FALSE,
                                    size = 25),
                        shiny::actionButton("b_store", "Save & Close")
                    ),
                    shiny::mainPanel(
                        plotly::plotlyOutput("plot"),
                        shiny::checkboxInput("valid", "Current match OK?",
                                             value = TRUE, width = NULL),
                        DT::DTOutput("targets")
                        )
                    ))
    server <- function(input, output, session) {
        query_ids <- .createChoices(object)
        dt <- .create_dt(object)
        ppm <- 20
        tolerance <- 0
        if (length(prm <- object@metadata$param)) {
            if ("ppm" %in% slotNames(prm))
                ppm <- prm@ppm
            if ("tolerance" %in% slotNames(prm))
                tolerance <- prm@tolerance
        }
        if (nrow(dt)) {
            dt <- data.frame(valid = TRUE, dt)
        } else dt <- data.frame(valid = logical(), dt)
        dtl <- split(dt, factor(dt$query_index_, seq_along(object)))
        rv <- shiny::reactiveValues(
            queries = query_ids,
            dtl = dtl
            )
        rv_query <- shiny::reactiveValues(idx = 1L)
        rv_target <- shiny::reactiveValues(idx = 1L)
        shiny::observe(shiny::updateSelectInput(inputId = "query",
                                  choices = rv$queries))
        shinyjs::disable("valid")
        ## Choose query spectrum
        shiny::observeEvent(input$query, {
            rv_query$idx <- as.integer(input$query)
            rv_target$idx <- 1L
            current_match <- object[rv_query$idx]
            output$targets <- DT::renderDT(
                DT::datatable(rv$dtl[[rv_query$idx]],
                          selection = list(mode = "single",
                                           selected = rv_target$idx),
                          options = list(dom = "t", ordering = FALSE),
                          escape = FALSE, rownames = FALSE),
                server = TRUE)
            if (nrow(rv$dtl[[rv_query$idx]]) >= rv_target$idx) {
                ## rv_target$idx <- input$targets_rows_selected
                shinyjs::enable("valid")
                current_valid <- rv$dtl[[rv_query$idx]]$valid[rv_target$idx]
                shiny::updateCheckboxInput(session, "valid",
                                           value = current_valid)
            } else
                shinyjs::disable("valid")
            output$plot <- plotly::renderPlotly(
                        .plotlySpectraMirror(query(current_match), Spectra(),
                                             xLabel = "query",
                                             xColor = "#E41A1C",
                                             yLabel = "target",
                                             yColor = "#377EB8"))
        })
        ## Choose target spectrum
        shiny::observeEvent(input$targets_rows_selected, {
            rv_target$idx <- input$targets_rows_selected
            current_match <- object[rv_query$idx]
            if (nrow(current_match@matches)) {
                tidx <- current_match@matches$target_idx[rv_target$idx]
                current_valid <- rv$dtl[[rv_query$idx]]$valid[rv_target$idx]
                shiny::updateCheckboxInput(session, "valid",
                                           value = current_valid)
                output$plot <- plotly::renderPlotly(
                        .plotlySpectraMirror(query(current_match),
                                             target(current_match)[tidx],
                                             ppm = ppm, tolerance = tolerance,
                                             xLabel = "query",
                                             xColor = "#E41A1C",
                                             yLabel = "target",
                                             yColor = "#377EB8"))
            }
        })
        shiny::observeEvent(input$valid, {
            if (length(rv_target$idx) &&
                nrow(rv$dtl[[rv_query$idx]]) >= rv_target$idx) {
                rv$dtl[[rv_query$idx]]$valid[rv_target$idx] <- input$valid
            }
        })
        shiny::observeEvent(input$b_store, {
            ## Collect all the selections from all data frames
            idx <- which(do.call(rbind, rv$dtl)$valid)
            shiny::stopApp(filterMatches(object, index = idx))
        })
    }
    shiny::runApp(shiny::shinyApp(ui, server))
}

#' function to create interactive plot
#'
#' This is `SpectraVis::plotlySpectraMirror` - replace with import or require
#' once `SpectraVis` is in Bioconductor.
#'
#' @importMethodsFrom Spectra peaksData
#'
#' @importFrom MsCoreUtils common
#'
#' @noRd
.plotlySpectraMirror <- function(x, y, xLabel = "", xColor = "#737373",
                                yLabel = "", yColor = "#737373", matchSize = 5,
                                ppm = 20, tolerance = 0) {
    stopifnot(inherits(x, "Spectra"))
    stopifnot(inherits(y, "Spectra"))
    if (!.is_plotly_available())
        stop("The use of '.plotlySpectraMirror' requires package 'plotly'. ",
             "Please install with 'BiocInstaller::install(\"plotly\")'")
    p <- plotly::plot_ly()
    if (length(x) > 1 || length(y) > 1)
        stop("'x' and 'y' have to be of length 1")
    if (length(x))
        x_peaks <- as.data.frame(peaksData(x)[[1L]])
    else x_peaks <- data.frame(mz = numeric(), intensity = numeric(),
                               match = character())
    if (length(y))
        y_peaks <- as.data.frame(peaksData(y)[[1L]])
    else y_peaks <- data.frame(mz = numeric(), intensity = numeric(),
                               match = character())

    x_range <- range(x_peaks$mz, y_peaks$mz, na.rm = TRUE) + c(-1, 1)
    y_max <- max(x_peaks$intensity, y_peaks$intensity, na.rm = TRUE)
    y_peaks$intensity <- -y_peaks$intensity

    ht <- "<b>%{text}</b><br>mz: %{x}<br>int: %{y}"
    if (nrow(x_peaks)) {
        x_peaks$zero <- 0.0
        x_peaks$match <- ""
        x_peaks$color <- xColor[1L]
        idx <- which(common(x_peaks$mz, y_peaks$mz, tolerance, ppm))
        if (length(idx))
            x_peaks$match[idx] <- "matched"
        p <- .plotly_peaks(p, x_peaks, name = xLabel, col = xColor[1L],
                           hovertemplate = ht, text = ~match)
    }
    if (nrow(y_peaks)) {
        y_peaks$zero <- 0.0
        y_peaks$match <- ""
        y_peaks$color <- yColor[1L]
        idx <- which(common(y_peaks$mz, x_peaks$mz, tolerance, ppm))
        if (length(idx))
            y_peaks$match[idx] <- "matched"
        p <- .plotly_peaks(p, y_peaks, name = yLabel, col = yColor[1L],
                           hovertemplate = ht, text = ~match)
    }
    pks <- rbind(x_peaks, y_peaks)
    pks <- pks[pks$match != "", , drop = FALSE]
    if (nrow(pks))
        p <- plotly::add_trace(p, data = pks, x = ~mz, y = ~intensity,
                       type = "scatter", mode = "markers",
                       hoverinfo = "none", name = "matched",
                       marker = list(size = matchSize[1L],
                                     color = ~color))
    plotly::layout(p, xaxis = list(title = "m/z", zeroline = FALSE),
           yaxis = list(title = "intensity", zeroline = TRUE),
           hovermode = "x", hoverdistance = 1)
}

#' That's also from `SpectraVis` - remove if `SpectraVis` is used instead.
#'
#' @noRd
.plotly_peaks <- function(p, data, col = "#737373", name = "",
                          hovertemplate = "<br>mz: %{x}<br>int: %{y}<br>",
                          ...) {
    plotly::add_segments(p, data = data, x = ~mz, y = ~zero, xend = ~mz,
                 yend = ~intensity, line = list(color = col),
                 name = name, hovertemplate = hovertemplate, ...)
}


## .plotlySpectraMirror <- function(x, y, col = c("#E41A1C", "#377EB8"),
##                                  main = "") {
##     if (!requireNamespace("plotly", quietly = TRUE))
##         stop("The use of '.plotlySpectraMirror' requires package 'plotly'. ",
##              "Please install with 'BiocInstaller::install(\"plotly\")'")
##     if (length(col) != 2)
##         col <- col[c(1, 1)]
##     if (length(x)) {
##         upper <- as.data.frame(peaksData(x)[[1L]])
##     } else upper <- data.frame(mz = numeric(), intensity = numeric())
##     if (length(y)) {
##         lower <- as.data.frame(peaksData(y)[[1L]])
##     } else lower <- data.frame(mz = numeric(), intensity = numeric())
##     if (nrow(upper))
##         upper$zero <- 0.0
##     else upper$zero <- numeric()
##     if (nrow(lower))
##         lower$zero <- 0.0
##     else lower$zero <- numeric()

##     mz_range <- range(upper$mz, lower$mz) + c(-1, 1)
##     maxy <- max(upper$intensity, lower$intensity, na.rm = TRUE)
##     int_range <- list(-maxy, maxy)

##     lower$intensity <- -lower$intensity
##     p <- plotly::plot_ly()
##     p <- plotly::add_segments(p, data = upper, x = ~mz, y = ~zero, xend = ~mz,
##                               yend = ~intensity, line = list(color = col[1L]),
##                               name = "query",
##                               hovertemplate = "<br>mz: %{x}<br>int: %{y}<br>")
##     p <- plotly::add_segments(p, data = lower, x = ~mz, y = ~zero, xend = ~mz,
##                               yend = ~intensity, line = list(color = col[2L]),
##                               name = "target",
##                               hovertemplate = "<br>mz: %{x}<br>int: %{y}<br>")
##     p <- plotly::layout(p, title = main,
##                         xaxis = list(title = "m/z", range = mz_range),
##                         yaxis = list(title = "intensity", zeroline = TRUE,
##                                      range = int_range),
##                         hovermode = "x", hoverdistance = 1)
##     p
## }

#' isolate entries for feature selection
#' @noRd
.createChoices <- function(x) {
    sl <- seq_along(x)
    l_choices <- as.list(sl)
    names(l_choices) <- paste0(sl, " - MZ",
                               round(query(x)$precursorMz, 4), "@RT",
                               round(query(x)$rtime / 60, 2), " min")
    l_choices
}

.create_dt <- function(x){
    .sel_cols <- c("query_index_", "target_index_", "precursorMz",
                   "target_precursorMz", "rtime", "target_rtime",
                   "target_name", "target_compound_name",
                   "score", "reverse_score", "presence_ratio")
    cols <- .sel_cols[.sel_cols %in% spectraVariables(x)]
    tbl <- as.data.frame(matchedData(x, cols))
    tbl$score <- round(tbl$score, 3)
    if (any(colnames(tbl) == "precursorMz"))
        tbl$precursorMz <- round(tbl$precursorMz, 3)
    if (any(colnames(tbl) == "reverse_score"))
        tbl$reverse_score <- round(tbl$reverse_score, 3)
    if (any(colnames(tbl) == "presence_ratio"))
        tbl$presence_ratio <- round(tbl$presence_ratio, 3)
    if (any(colnames(tbl) == "rtime"))
        tbl$rtime <- round(tbl$rtime, 1)
    if (any(colnames(tbl) == "target_rtime"))
        tbl$target_rtime <- round(tbl$target_rtime, 1)
    tbl[!is.na(tbl$score), , drop = FALSE]
}

.is_shiny_available <- function() {
    requireNamespace("shiny", quietly = TRUE)
}

.is_shinyjs_available <- function() {
    requireNamespace("shinyjs", quietly = TRUE)
}

.is_dt_available <- function() {
    requireNamespace("DT", quietly = TRUE)
}

.is_plotly_available <- function() {
    requireNamespace("plotly", quietly = TRUE)
}

csp <- CompareSpectraParam(requirePrecursor = TRUE,
                           THRESHFUN = function(x) x >= 0.7)
norm_int <- function(x) {
    x[, "intensity"] <- x[, "intensity"] / max(x[, "intensity"]) * 100
    x
}
ms <- matchSpectra(addProcessing(pest_ms2, norm_int),
                   addProcessing(minimb, norm_int), csp)
ms2 <- ms[whichQuery(ms)]

test_that(".createChoices works", {
    res <- .createChoices(ms)
    expect_true(is.list(res))
    expect_equal(unname(res), as.list(seq_along(ms)))
})

test_that(".create_dt works", {
    res <- .create_dt(ms)
    expect_true(is.data.frame(res))
    expect_true(all(c("precursorMz", "target_precursorMz", "rtime", "score")
                    %in% colnames(res)))

    ms3 <- ms
    ms3@matches$reverse_score <- 3.2
    ms3@matches$presence_ratio <- 32.1
    res <- .create_dt(ms3)
    expect_true(all(res$reverse_score == 3.2))
    expect_true(all(res$presence_ratio == 32.1))
})

test_that("validateMatchedSpectra works", {
    with_mocked_bindings(
        ".is_shiny_available" = function() FALSE,
        code = expect_error(validateMatchedSpectra(ms), "requires package")
    )
    with_mocked_bindings(
        ".is_shinyjs_available" = function() FALSE,
        code = expect_error(validateMatchedSpectra(ms), "requires package")
    )
    with_mocked_bindings(
        ".is_dt_available" = function() FALSE,
        code = expect_error(validateMatchedSpectra(ms), "requires package")
    )

    expect_error(validateMatchedSpectra(3), "is not TRUE")
    expect_error(validateMatchedSpectra(MatchedSpectra()), "is empty")

    ## validateMatchedSpectra(ms)
})

test_that("packages available", {
    expect_true(.is_shiny_available())
    expect_true(.is_shinyjs_available())
    expect_true(.is_dt_available())
    expect_true(.is_plotly_available())
})

test_that("plotlySpectraMirror works", {
    with_mocked_bindings(
        ".is_plotly_available" = function() FALSE,
        code = expect_error(.plotlySpectraMirror(pest_ms2[1], pest_ms2[1]),
                            "'plotly'")
    )

    expect_error(.plotlySpectraMirror(3, pest_ms2[2]), "not TRUE")
    expect_error(.plotlySpectraMirror(pest_ms2[2], 3), "not TRUE")
    expect_error(.plotlySpectraMirror(pest_ms2, pest_ms2[1]), "length 1")
    expect_error(.plotlySpectraMirror(pest_ms2[1], pest_ms2), "length 1")
    p <- .plotlySpectraMirror(pest_ms2[1], pest_ms2[2])
    expect_true(is(p, "plotly"))
    p <- .plotlySpectraMirror(pest_ms2[6], pest_ms2[8], xLabel = "query",
                              yLabel = "target")

    ## One Spectra empty.
    p <- .plotlySpectraMirror(pest_ms2[2], Spectra())
    expect_true(is(p, "plotly"))
    p <- .plotlySpectraMirror(Spectra(), pest_ms2[2])
    expect_true(is(p, "plotly"))
})

#' @title Chemical Formula Matching
#'
#' @name matchFormula
#'
#' @description
#'
#' The `matchFormula` method matches chemical formulas from different inputs
#' (parameter `query` and `target`). Before comparison all formulas are
#' normalized using [MetaboCoreUtils::standardizeFormula()]. Inputs can be
#' either a `character` or `data.frame` containing a column with formulas.
#' In case of `data.frame`s parameter `formulaColname` needs to be used to
#' specify the name of the column containing the chemical formulas.
#'
#' @param BPPARAM parallel processing setup. See `BiocParallel::bpparam()` for
#'     details.
#'
#' @param formulaColname `character` with the name of the column containing
#'     chemical formulas. Can be of length 1 if both `query` and `target` are
#'     `data.frame`s and the name of the column with chemical formulas is the
#'     same for both. If different columns are used, `formulaColname[1]` can be
#'     used to define the column name in `query` and `formulaColname[2]` the
#'     one of `target`.
#'
#' @param target `character` or `data.frame` with chemical formulas to compare
#'     against.
#'
#' @param query `character` or `data.frame` with chemical formulas to search.
#'
#' @param ... currently ignored
#'
#' @return [Matched] object representing the result.
#'
#' @author Michael Witting
#'
#' @examples
#'
#' ## input formula
#' query <- c("H12C6O6", "C11H12O2", "HN3")
#' target <- c("HCl", "C2H4O", "C6H12O6")
#'
#' query_df <- data.frame(
#'     formula = c("H12C6O6", "C11H12O2", "HN3"),
#'     name = c("A", "B", "C")
#' )
#' target_df <- data.frame(
#'     formula = c("HCl", "C2H4O", "C6H12O6"),
#'     name = c("D", "E", "F")
#' )
#'
#' ## character vs character
#' matches <- matchFormula(query, target)
#' matchedData(matches)
#'
#' ## data.frame vs data.frame
#' matches <- matchFormula(query_df, target_df)
#' matchedData(matches)
#' ## data.frame vs character
#' matches <- matchFormula(query_df, target)
#' matchedData(matches)
#' ## character vs data.frame
#' matches <- matchFormula(query, target_df)
#' matchedData(matches)
NULL

#' @rdname matchFormula
#'
#' @export
setGeneric("matchFormula", function(query, target, ...)
  standardGeneric("matchFormula"))

#' @rdname matchFormula
#'
#' @importFrom MetaboCoreUtils standardizeFormula
#' 
#' @importFrom BiocParallel bpmapply SerialParam
setMethod(
    "matchFormula",
    signature = c(query = "character",
                  target = "character"),
    function(query, target, BPPARAM = SerialParam()) {
        matches <- do.call(
            rbind, bpmapply(seq_along(query), standardizeFormula(query),
                            FUN = .getFormulaMatches,
                            MoreArgs =list(target = standardizeFormula(target)),
                            BPPARAM = BPPARAM, SIMPLIFY = FALSE))
        Matched(query = query, target = target, matches = matches)
    })

#' @rdname matchFormula
setMethod("matchFormula",
          signature = c(query = "data.frameOrSimilar",
                        target = "data.frameOrSimilar"),
          function(query, target, formulaColname = c("formula", "formula"),
                   BPPARAM = SerialParam()) {
            if(length(formulaColname) == 1L)
              formulaColname <- rep(formulaColname, 2L)
            if(!formulaColname[1] %in% colnames(query))
              stop("Missing column \"", formulaColname[1L], "\" in query")
            if(!formulaColname[2L] %in% colnames(target))
              stop("Missing column \"", formulaColname[2L], "\" in target")
            res <- matchFormula(query[, formulaColname[1L]],
                                target[, formulaColname[2L]],
                                BPPARAM = BPPARAM)
            res@query <- query
            res@target <- target
            res
          })

#' @rdname matchFormula
setMethod("matchFormula",
          signature = c(query = "character",
                        target = "data.frameOrSimilar"),
          function(query, target, formulaColname = "formula",
                   BPPARAM = SerialParam()) {
            if(!formulaColname %in% colnames(target))
              stop("Missing column \"", formulaColname, "\" in target")
            res <- matchFormula(query, target[, formulaColname],
                                BPPARAM = BPPARAM)
            res@query <- query
            res@target <- target
            res
          })

#' @rdname matchFormula
setMethod("matchFormula",
          signature = c(query = "data.frameOrSimilar",
                        target = "character"),
          function(query, target, formulaColname = "formula",
                   BPPARAM = SerialParam()) {
            if(!formulaColname %in% colnames(query))
              stop("Missing column \"", formulaColname, "\" in query")
            res <- matchFormula(query[, formulaColname], target,
                                BPPARAM = BPPARAM)
            res@query <- query
            res@target <- target
            res
          })


#' @author Michael Witting
#'
#' @param queryIndex `integer(1)` with the index of the query
#'
#' @param queryFormula `character(1)` with the formula of the query
#'
#' @param target `character` with target formulas
#'
#' @noRd
.getFormulaMatches <- function(queryIndex, queryFormula, target) {
    cls <- which(target == queryFormula)
    if(length(cls)) {
        data.frame(query_idx = queryIndex,
                   target_idx = cls,
                   score = 1)
    } else {
        data.frame(query_idx = integer(),
                   target_idx = integer(),
                   score = numeric())
    }
}

test_that("matchFormula works", {
  # input formula
  x <- c("H12C6O6", "C11H12O2", "HN3", "ClH", "HCl")
  y <- c("HCl", "C2H4O", "C6H12O6", "N3H")

  # character vs character
  res <- matchFormula(x, y)
  expect_s4_class(res, "Matched")
  expect_equal(query(res), x)
  expect_equal(target(res), y)
  expect_equal(res@matches$query_idx, c(1L, 3L, 4L, 5L))
  expect_equal(res@matches$target_idx, c(3L, 4L, 1L, 1L))
  expect_equal(res@matches$score, rep(1L, 4))

  x_df <- data.frame(chem_formula = x,
                     name = c("A", "B", "C", "D", "E"))

  y_df <- data.frame(formula = y,
                     name = c("D", "E", "F", "G"))

  ## data.frame vs data.frame
  expect_error(matchFormula(x_df, y_df), "Missing column")
  expect_error(matchFormula(x_df, y_df, formulaColname = c("chem_formula")),
               "Missing column")
  res <- matchFormula(x_df, y_df, formulaColname = c("chem_formula", "formula"))
  expect_equal(query(res), x_df)
  expect_equal(target(res), y_df)
  expect_equal(res@matches$query_idx, c(1L, 3L, 4L, 5L))
  expect_equal(res@matches$target_idx, c(3L, 4L, 1L, 1L))
  expect_equal(res@matches$score, rep(1L, 4))

  # data.frame vs character
  res <- matchFormula(x_df, y, formulaColname = "chem_formula")
  expect_equal(query(res), x_df)
  expect_equal(target(res), y)
  expect_equal(res@matches$query_idx, c(1L, 3L, 4L, 5L))
  expect_equal(res@matches$target_idx, c(3L, 4L, 1L, 1L))
  expect_equal(res@matches$score, rep(1L, 4))
  expect_error(matchFormula(x_df, y, formulaColname = "other"),
               "Missing column")

  # character vs data.frame
  res <- matchFormula(x, y_df)
  expect_equal(query(res), x)
  expect_equal(target(res), y_df)
  expect_equal(res@matches$query_idx, c(1L, 3L, 4L, 5L))
  expect_equal(res@matches$target_idx, c(3L, 4L, 1L, 1L))
  expect_equal(res@matches$score, rep(1L, 4))
  expect_error(matchFormula(x, y_df, formulaColname = "other"),
               "Missing column")
})

test_that(".getFormulaMatches works", {
    res <- MetaboAnnotation:::.getFormulaMatches(3L, "A", c("A", "B", "A"))
    expect_true(is.data.frame(res))
    expect_equal(colnames(res), c("query_idx", "target_idx", "score"))
    expect_equal(res$query_idx, c(3L, 3L))
    expect_equal(res$target_idx, c(1L, 3L))
})

#' @title Representation of generic objects matches
#'
#' @name Matched
#'
#' @aliases Matched Matched-class [,Matched-method
#'
#' @description
#'
#' Matches between *query* and *target* generic objects can be represented by
#' the `Matched` object. By default, all data accessors work as
#' *left joins* between the *query* and the *target* object, i.e. values are
#' returned for each *query* object with eventual duplicated entries (values)
#' if the *query* object matches more than one *target* object. See also
#' *Creation and subsetting* as well as *Extracting data* sections below for
#' details and more information.
#'
#' The `Matched` object allows to represent matches between one-dimensional
#' `query` and `target` objects (being e.g. `numeric` or `list`),
#' two-dimensional objects (`data.frame` or `matrix`) or more complex
#' structures such as `SummarizedExperiments` or `QFeatures`. Combinations of
#' all these different data types are also supported. Matches are represented
#' between elements of one-dimensional objects, or rows for two-dimensional
#' objects (including `SummarizedExperiment` or `QFeatures`). For
#' [QFeatures::QFeatures()] objects matches to only one of the *assays*
#' within the object is supported.
#'
#' @section Creation and general handling:
#'
#' `Matched` object is returned as result from the [matchValues()] function.
#'
#' Alternatively, `Matched` objects can also be created with the `Matched`
#' function providing the `query` and `target` objects as well as the `matches`
#' `data.frame` with two columns of integer indices defining which elements
#' from *query* match which element from *target*.
#'
#' - `addMatches`: add new matches to an existing object. Parameters
#'   `queryValue` and `targetValue` allow to define which element(s) in
#'   `query` and `target` should be considered matching. If `isIndex = TRUE`,
#'   both `queryValue` and `targetValue` are considered to be integer indices
#'   identifying the matching elements in `query` and `target`, respectively.
#'   Alternatively (with `isIndex = FALSE`) `queryValue` and `targetValue` can
#'   be elements in columns `queryColname` or `targetColname` which can be used
#'   to identify the matching elements. Note that in this case
#'   **only the first** matching pair is added. Parameter `score` allows to
#'   provide the score for the match. It can be a numeric with the score or a
#'   `data.frame` with additional information on the manually added matches. In
#'   both cases its length (or number of rows) has to match the length of
#'   `queryValue`. See examples below for more information.
#'
#' - `endoapply`: applies a user defined function `FUN` to each subset of
#'   matches in a `Matched` object corresponding to a `query` element (i.e. for
#'   each `x[i]` with `i` being 1 to `length(x)`). The results are then combined
#'   in a single `Matched` object representing updated matches. Note that `FUN`
#'   has to return a `Matched` object.
#'
#' - `lapply`: applies a user defined function `FUN` to each subset of
#'   matches in a `Matched` object for each `query` element (i.e. to each `x[i]`
#'   with `i` from `1` to `length(x)`). It returns a `list` of `length(object)`
#'   elements where each element is the output of `FUN` applied to each subset
#'   of matches.
#'
#'
#' @section Filtering and subsetting:
#'
#' - `[`: subset the object selecting `query` object elements to keep with
#'   parameter `i`. The resulting object will contain all the matches
#'   for the selected query elements. The `target` object will by default be
#'   returned as-is.
#'
#' - `filterMatches`: filter matches in a `Matched` object using different
#'    approaches depending on the class of `param`:
#'
#'   - `ScoreThresholdParam`: keeps only the matches whose score is strictly
#'     above or strictly below a certain threshold (respectively when parameter
#'     `above = TRUE` and `above = FALSE`). The name of the column containing
#'     the scores to be used for the filtering can be specified with parameter
#'     `column`. The default for `column` is `"score"`. Such variable is present
#'     in each `Matched` object. The name of other score variables (if present)
#'     can be provided (the names of all score variables can be obtained with
#'     `scoreVariables()` function). For example `column = "score_rt"` can be
#'     used to filter matches based on retention time scores for `Matched`
#'     objects returned by [matchValues()] when `param` objects involving a
#'     retention time comparison are used.
#'
#'   - `SelectMatchesParam`: keeps or removes (respectively when parameter
#'     `keep = TRUE` and `keep = FALSE`) matches corresponding to certain
#'     indices or values of `query` and `target`. If `queryValue` and
#'     `targetValue` are provided, matches for these value pairs are kept or
#'     removed. Parameter index` allows to filter matches providing their index
#'     in the [matches()] matrix. Note that `filterMatches` removes only matches
#'     from the [matches()] matrix from the `Matched` object but thus not alter
#'     the `query` or `target` in the object. See examples below for more
#'     information.
#'
#'   - `SingleMatchParam`: reduces matches to keep only (at most) a
#'     single match per query. The deduplication strategy can be defined with
#'     parameter `duplicates`:
#'     - `duplicates = "remove"`: all matches for query elements matching more
#'       than one target element will be removed.
#'     - `duplicates = "closest"`: keep only the *closest* match for each
#'       query element. The closest match is defined by the value(s) of
#'       *score* (and eventually *score_rt*, if present). The one match with
#'       the smallest value for this (these) column(s) is retained. This is
#'       equivalent to `TopRankedMatchesParam(n = 1L, decreasing = FALSE)`.
#'     - `duplicates = "top_ranked"`: select the *best ranking* match for each
#'       query element. Parameter `column` allows to specify the column by
#'       which matches are ranked (use `targetVariables(object)` or
#'       `scoreVariables(object)` to list possible columns). Parameter
#'       `decreasing` allows to define whether the match with the highest
#'       (`decreasing = TRUE`) or lowest (`decreasing = FALSE`) value in
#'       `column` for each *query* will be selected.
#'
#'   - `TopRankedMatchesParam`: for each query element the matches are ranked
#'     according to their score and only the `n` best of them are kept (if `n`
#'     is larger than the number of matches for a given query element all the
#'     matches are returned). For the ranking (ordering) R's `rank` function is
#'     used on the absolute values of the scores (variable `"score"`), thus,
#'     smaller score values (representing e.g. smaller differences between
#'     expected and observed m/z values) are considered *better*. By
#'     setting parameter `decreasing = TRUE` matches can be ranked in decreasing
#'     order (i.e. higher scores are ranked higher and are thus selected).
#'     If besides variable `"score"` also variable `"score_rt"` is available in
#'     the `Matched` object (which is the case for the `Matched` object
#'     returned by [matchValues()] for `param` objects involving a retention
#'     time comparison), the ordering of the matches is based on the product of
#'     the ranks of the two variables (ranking of retention time differences
#'     is performed on the absolute value of `"score_rt"`). Thus, matches with
#'     small (or, depending on parameter `decreasing`, large) values for
#'     `"score"` **and** `"score_rt"` are returned.
#'
#' - `pruneTarget`: *cleans* the object by removing non-matched
#'   **target** elements.
#'
#' @section Extracting data:
#'
#' - `$` extracts a single variable from the `Matched` `x`. The variables that
#'   can be extracted can be listed using `colnames(x)`. These variables can
#'   belong to *query*, *target* or be related to the matches (e.g. the
#'   score of each match). If the *query* (*target*) object is two dimensional,
#'   its columns can be extracted (prefix` "target_"` is used for columns in the
#'   *target* object) otherwise if *query* (*target*) has only a single
#'   dimension (e.g. is a `list` or a `character`) the whole object can be
#'   extracted with `x$query` (`x$target`). More precisely, when
#'   *query* (*target*) is a `SummarizedExperiment` the columns from
#'   `rowData(query)` (rowData(`target`)) are extracted; when *query* (*target*)
#'   is a [QFeatures::QFeatures()] the columns from `rowData` of the assay
#'   specified in the `queryAssay` (`targetAssay`) slot are extracted.
#'   The matching scores
#'   are available as *variable* `"score"`. Similar to a left join between the
#'   query and target elements, this function returns a value for each query
#'   element, with eventual duplicated values for query elements matching more
#'   than one target element. If variables from the target `data.frame` are
#'   extracted, an `NA` is reported for the entries corresponding to *query*
#'   elements that don't match any target element. See examples below for
#'   more details.
#'
#' - `length` returns the number of **query** elements.
#'
#' - `matchedData` allows to extract multiple variables contained in the
#'   `Matched` object as a `DataFrame`. Parameter `columns` allows to
#'   define which columns (or variables) should be returned (defaults to
#'   `columns = colnames(object)`). Each single column in the returned
#'   `DataFrame` is constructed in the same way as in `$`. That is, like `$`,
#'   this function performs a *left join* of variables from the *query* and
#'   *target* objects returning all values for all *query* elements
#'   (eventually returning duplicated elements for query elements matching
#'   multiple target elements) and the values for the target elements matched
#'   to the respective query elements (or `NA` if the target element is not
#'   matched to any query element).
#'
#' - `matches` returns a `data.frame` with the actual matching information with
#'   columns `"query_idx"` (index of the element in `query`), `"target_idx"`
#'   (index of the element in `target`) `"score"` (the score of the match) and
#'   eventual additional columns.
#'
#' - `target` returns the *target* object.
#'
#' - `targetIndex` returns the indices of the matched targets in the order they
#'   are assigned to the query elements. The length of the returned `integer`
#'   vector is equal to the total number of matches in the object. `targetIndex`
#'   and `queryIndex` are aligned, i.e. each element in them represent a matched
#'   query-target pair.
#'
#' - `query` returns the *query* object.
#'
#' - `queryIndex` returns the indices of the query elements with matches to
#'   target elements. The length of the returned `integer` vector is equal to
#'   the total number of matches in the object. `targetIndex` and `queryIndex`
#'   are aligned, i.e. each element in them represent a matched query-target
#'   pair.
#'
#' - `queryVariables` returns the names of the variables (columns) in *query*.
#'
#' - `scoreVariables` returns the names of the score variables stored in the
#'   `Matched` object (precisely the names of the variables in `matches(object)`
#'   containing the string "score" in their name ignoring the case).
#'
#' - `targetVariables` returns the names of the variables (columns) in *target*
#'   (prefixed with `"target_"`).
#'
#' - `whichTarget` returns an `integer` with the indices of the elements in
#'   *target* that match at least one element in *query*.
#'
#' - `whichQuery` returns an `integer` with the indices of the elements in
#'   *query* that match at least one element in *target*.
#'
#' @param above for `ScoreThresholdParam` : `logical(1)` specifying whether
#'   to keep matches above (`above = TRUE`) or below (`above = FALSE`) a certain
#'   threshold.
#'
#' @param column for `ScoreThresholdParam`: `character(1)` specifying the name
#'   of the score variable to consider for the filtering (the default is
#'   `column = "score"`). For `SingleMatchParam`: `character(1)` defining the
#'   name of the column to be used for de-duplication. See description of
#'   `SingleMatchParam` in the *Filtering and subsetting* section for details.
#'
#' @param columns for `matchedData`: `character` vector with column names of
#'   variables that should be extracted.
#'
#' @param decreasing for `TopRankedMatchesParam`: `logical(1)` whether scores
#'   should be ordered increasing or decreasing. Defaults to
#'   `decreasing = FALSE`.
#'
#' @param drop for `[`: ignored.
#'
#' @param duplicates for `SingleMatchParam`: `character(1)` defining the
#'   *de-duplication* strategy. See the description of `SingleMatchParam` in
#'   the *Filtering and subsetting* subsection for choices and details.
#'
#' @param FUN for `lapply` and `endoapply`: user defined `function` that takes a
#'   `Matched` object as a first parameter and possibly additional parameters
#'   (that need to be provided in the `lapply` or `endoapply` call. For lapply
#'   `FUN` can return any object while for `endoapply` it must return a
#'   `Matched` object.
#'
#' @param i `integer` or `logical` defining the `query` elements to keep.
#'
#' @param index for `SelectMatchesParam`: indices of the matches to keep (if
#'   `keep = TRUE`) or to drop if (`keep = FALSE`).
#'
#' @param isIndex for `addMatches`: specifies if `queryValue` and
#'   `targetValue` are expected to be vectors of indices.
#'
#' @param j for `[`: ignored.
#'
#' @param keep for `SelectMatchesParam`: `logical`. If `keep = TRUE` the matches
#'   are kept, if `keep = FALSE` they are removed.
#'
#' @param matches `data.frame` with columns `"query_idx"` (`integer`),
#'   `"target_idx"` (`integer`) and `"score"` (`numeric`) representing the n:m
#'   mapping of elements between the `query` and the `target` objects.
#'
#' @param metadata `list` with optional additional metadata.
#'
#' @param n for `TopRankedMatchesParam`: `integer(1)` with number of best
#'   ranked matches to keep for each `query` element.
#'
#' @param name for `$`: the name of the column (or variable) to extract.
#'
#' @param object a `Matched` object.
#'
#' @param param for `filterMatches`: parameter object to select and customize
#'   the filtering procedure.
#'
#' @param pattern for `query`: ignored.
#'
#' @param score for `addMatches`: `numeric` (same length than `queryValue`) or
#'   `data.frame` (same number of rows than `queryValue`) specifying the scores
#'   for the matches to add. If not specified, a `NA` will be used as score.
#'
#' @param target object with the elements against which `query` has been
#'   matched.
#'
#' @param targetAssay `character` that needs to be specified when `target` is
#'   a `QFeatures`. In this case, `targetAssay` is expected to be the name of
#'   one of the assays in `target` (the one on which the matching was
#'   performed).
#'
#' @param targetColname for `SelectMatchesParam`: if `query` is 2-dimensional it
#'   represents the column of `target` against which elements of `targetValue`
#'   are compared.
#'
#' @param targetValue for `SelectMatchesParam`: vector of values to search for
#'   in `target` (if `target` is 1-dimensional) or in column `targetColname` of
#'   `target` (if `target` is 2-dimensional). For `addMatches`: either an
#'   index in `target` or value in column `targetColname` of `target` defining
#'   (together with `queryValue`) the pair of query and target elements for
#'   which a match should be manually added. Lengths of `queryValue` and
#'   `targetValue` have to match.
#'
#' @param threshold for `ScoreThresholdParam` : `numeric(1)` specifying the
#'   threshold to consider for the filtering.
#'
#' @param query object with the query elements.
#'
#' @param queryAssay `character` that needs to be specified when `query` is
#'   a `QFeatures`. In this case, `queryAssay` is expected to be the name of
#'   one of the assays in `query` (the one on which the matching was performed).
#'
#' @param queryColname for `SelectMatchesParam`: if `query` is 2-dimensional it
#'   represents the column of `query` against which elements of `queryValue`
#'   are compared.
#'
#' @param queryValue for `SelectMatchesParam`: vector of values to search for in
#'   `query` (if `query` is 1-dimensional) or in column `queryColname` of
#'   `query` (if `query` is 2-dimensional). For `addMatches`: either an index
#'   in `query` or value in column `queryColname` of `query` defining (together
#'   with `targetValue`) the pair of query and target elements for which a
#'   match should be manually added. Lengths of `queryValue` and
#'   `targetValue` have to match.
#'
#' @param x `Matched` object.
#'
#' @param X `Matched` object.
#'
#' @param ... additional parameters.
#'
#' @return See individual method description above for details.
#'
#' @seealso [MatchedSpectra()] for matched [Spectra::Spectra()] objects.
#'
#' @exportClass Matched
#'
#' @author Andrea Vicini, Johannes Rainer
#'
#' @rdname Matched
#'
#' @examples
#'
#' ## Creating a `Matched` object.
#' q1 <- data.frame(col1 = 1:5, col2 = 6:10)
#' t1 <- data.frame(col1 = 11:16, col2 = 17:22)
#' ## Define matches between query row 1 with target row 2 and, query row 2
#' ## with target rows 2,3,4 and query row 5 with target row 5.
#' mo <- Matched(
#'     q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
#'                                  target_idx = c(2L, 2L, 3L, 4L, 5L),
#'                                  score = seq(0.5, 0.9, by = 0.1)))
#' mo
#'
#' ## Which of the query elements (rows) match at least one target
#' ## element (row)?
#' whichQuery(mo)
#'
#' ## Which target elements (rows) match at least one query element (row)?
#' whichTarget(mo)
#'
#' ## Extracting variable "col1" from query object .
#' mo$col1
#'
#' ## We have duplicated values for the entries of `col1` related to query
#' ## elements (rows) matched to multiple rows of the target object). The
#' ## value of `col1` is returned for each element (row) in the query.
#'
#' ## Extracting variable "col1" from target object. To access columns from
#' ## target we have to prefix the name of the column by `"target_"`.
#' ## Note that only values of `col1` for rows matching at least one query
#' ## row are returned and an NA is reported for query rows without matching
#' ## target rows.
#' mo$target_col1
#'
#' ## The 3rd and 4th query rows do not match any target row, thus `NA` is
#' ## returned.
#'
#' ## `matchedData` can be used to extract all (or selected) columns
#' ## from the object. Same as with `$`, a left join between the columns
#' ## from the query and the target is performed. Below we extract selected
#' ## columns from the object as a DataFrame.
#' res <- matchedData(mo, columns = c("col1", "col2", "target_col1",
#'                                    "target_col2"))
#' res
#' res$col1
#' res$target_col1
#'
#' ## With the `queryIndex` and `targetIndex` it is possible to extract the
#' ## indices of the matched query-target pairs:
#' queryIndex(mo)
#' targetIndex(mo)
#'
#' ## Hence, the first match is between the query with index 1 to the target
#' ## with index 2, then, query with index 2 is matched to target with index 2
#' ## and so on.
#'
#' ## The example matched object contains all query and all target
#' ## elements (rows). Below we subset the object keeping only query rows that
#' ## are matched to at least one target row.
#' mo_sub <- mo[whichQuery(mo)]
#'
#' ## mo_sub contains now only 3 query rows:
#' nrow(query(mo_sub))
#'
#' ## while the original object contains all 5 query rows:
#' nrow(query(mo))
#'
#' ## Both objects contain however still the full target object:
#' nrow(target(mo))
#' nrow(target(mo_sub))
#'
#' ## With the `pruneTarget` we can however reduce also the target rows to
#' ## only those that match at least one query row
#' mo_sub <- pruneTarget(mo_sub)
#' nrow(target(mo_sub))
#'
#' ########
#' ## Creating a `Matched` object with a `data.frame` for `query` and a `vector`
#' ## for `target`. The matches are specified in the same way as the example
#' ## before.
#'
#' q1 <- data.frame(col1 = 1:5, col2 = 6:10)
#' t2 <- 11:16
#' mo <- Matched(q1, t2, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
#'     target_idx = c(2L, 2L, 3L, 4L, 5L), score = seq(0.5, 0.9, by = 0.1)))
#'
#' ## *target* is a simple vector and has thus no columns. The matched values
#' ## from target, if it does not have dimensions and hence column names, can
#' ## be retrieved with `$target`
#' mo$target
#'
#' ## Note that in this case "target" is returned by the function `colnames`
#' colnames(mo)
#'
#' ## As before, we can extract all data as a `DataFrame`
#' res <- matchedData(mo)
#' res
#'
#' ## Note that the columns of the obtained `DataFrame` are the same as the
#' ## corresponding vectors obtained with `$`
#' res$col1
#' res$target
#'
#' ## Also subsetting and pruning works in the same way as the example above.
#'
#' mo_sub <- mo[whichQuery(mo)]
#'
#' ## mo_sub contains now only 3 query rows:
#' nrow(query(mo_sub))
#'
#' ## while the original object contains all 5 query rows:
#' nrow(query(mo))
#'
#' ## Both object contain however still the full target object:
#' length(target(mo))
#' length(target(mo_sub))
#'
#' ## Reducing the target elements to only those that match at least one query
#' ## row
#' mo_sub <- pruneTarget(mo_sub)
#' length(target(mo_sub))
#'
#' ########
#' ## Filtering `Matched` with `filterMatches`
#'
#' ## Inspecting the matches in `mo`:
#' mo$col1
#' mo$target
#'
#' ## We have thus target *12* matched to both query elements with values 1 and
#' ## 2, and query element 2 is matching 3 target elements. Let's assume we want
#' ## to resolve this multiple mappings to keep from them only the match between
#' ## query 1 (column `"col1"` containing value `1`) with target 1 (value `12`)
#' ## and query 2 (column `"col1"` containing value `2`) with target 2 (value
#' ## `13`). In addition we also want to keep query element 5 (value `5` in
#' ## column `"col1"`) with the target with value `15`:
#' mo_sub <- filterMatches(mo,
#'     SelectMatchesParam(queryValue = c(1, 2, 5), queryColname = "col1",
#'                        targetValue = c(12, 13, 15)))
#' matchedData(mo_sub)
#'
#' ## Alternatively to specifying the matches to filter with `queryValue` and
#' ## `targetValue` it is also possible to specify directly the index of the
#' ## match(es) in the `matches` `data.frame`:
#' matches(mo)
#'
#' ## To keep only matches like in the example above we could use:
#' mo_sub <- filterMatches(mo, SelectMatchesParam(index = c(1, 3, 5)))
#' matchedData(mo_sub)
#'
#' ## Note also that, instead of keeping the specified matches, it would be
#' ## possible to remove them by setting `keep = FALSE`. Below we remove
#' ## selected matches from the object:
#' mo_sub <- filterMatches(mo,
#'     SelectMatchesParam(queryValue = c(2, 2), queryColname = "col1",
#'                        targetValue = c(12, 14), keep = FALSE))
#' mo_sub$col1
#' mo_sub$target
#'
#' ## As alternative to *manually* selecting matches it is also possible to
#' ## filter matches keeping only the *best matches* using the
#' ## `TopRankedMatchesParam`. This will rank matches for each query based on
#' ## their *score* value and select the best *n* matches with lowest score
#' ## values (i.e. smallest difference in m/z values).
#' mo_sub <- filterMatches(mo, TopRankedMatchesParam(n = 1L))
#' matchedData(mo_sub)
#'
#' ## Additionally it is possible to select matches based on a threshold
#' ## for their *score*. Below we keep matches with score below 0.75 (one
#' ## could select matches with *score* greater than the threshold by setting
#' ## `ScoreThresholdParam` parameter `above = TRUE`.
#' mo_sub <- filterMatches(mo, ScoreThresholdParam(threshold = 0.75))
#' matchedData(mo_sub)
#'
#' ########
#' ## Selecting the best match for each `query` element with `endoapply`
#'
#' ## It is also possible to select for each `query` element the match with the
#' ## lowest score using `endoapply`. We manually define a function to select
#' ## the best match for each query and give it as input to `endoapply`
#' ## together with the `Matched` object itself. We obtain the same results as
#' ## in the `filterMatches` example above.
#'
#' FUN <- function(x) {
#'     if(nrow(x@matches) > 1)
#'         x@matches <- x@matches[order(x@matches$score)[1], , drop = FALSE]
#'     x
#' }
#'
#' mo_sub <- endoapply(mo, FUN)
#' matchedData(mo_sub)
#'
#' ########
#' ## Adding matches using `addMatches`
#'
#' ## `addMatches` allows to manually add matches. Below we add a new match
#' ## between the `query` element with a value of `1` in column `"col1"` and
#' ## the target element with a value of `15`. Parameter `score` allows to
#' ## assign a score value to the match.
#' mo_add <- addMatches(mo, queryValue = 1, queryColname = "col1",
#'     targetValue = 15, score = 1.40)
#' matchedData(mo_add)
#' ## Matches are always sorted by `query`, thus, the new match is listed as
#' ## second match.
#'
#' ## Alternatively, we can also provide a `data.frame` with parameter `score`
#' ## which enables us to add additional information to the added match. Below
#' ## we define the score and an additional column specifying that this match
#' ## was added manually. This information will then also be available in the
#' ## `matchedData`.
#' mo_add <- addMatches(mo, queryValue = 1, queryColname = "col1",
#'     targetValue = 15, score = data.frame(score = 5, manual = TRUE))
#' matchedData(mo_add)
#'
#' ## The match will get a score of NA if we're not providing any score.
#' mo_add <- addMatches(mo, queryValue = 1, queryColname = "col1",
#'     targetValue = 15)
#' matchedData(mo_add)
#'
#' ## Creating a `Matched` object with a `SummarizedExperiment` for `query` and
#' ## a `vector` for `target`. The matches are specified in the same way as
#' ## the example before.
#' library(SummarizedExperiment)
#' q1 <- SummarizedExperiment(
#'   assays = data.frame(matrix(NA, 5, 2)),
#'   rowData = data.frame(col1 = 1:5, col2 = 6:10),
#'   colData = data.frame(cD1 = c(NA, NA), cD2 = c(NA, NA)))
#' t1 <- data.frame(col1 = 11:16, col2 = 17:22)
#' ## Define matches between row 1 in rowData(q1) with target row 2 and,
#' ## rowData(q1) row 2 with target rows 2,3,4 and rowData(q1) row 5 with target
#' ## row 5.
#' mo <- Matched(
#'     q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
#'                                 target_idx = c(2L, 2L, 3L, 4L, 5L),
#'                                  score = seq(0.5, 0.9, by = 0.1)))
#' mo
#'
#' ## Which of the query elements (rows) match at least one target
#' ## element (row)?
#' whichQuery(mo)
#'
#' ## Which target elements (rows) match at least one query element (row)?
#' whichTarget(mo)
#'
#' ## Extracting variable "col1" from rowData(q1).
#' mo$col1
#'
#' ## We have duplicated values for the entries of `col1` related to rows of
#' ## rowData(q1) matched to multiple rows of the target data.frame t1. The
#' ## value of `col1` is returned for each row in the rowData of query.
#'
#' ## Extracting variable "col1" from target object. To access columns from
#' ## target we have to prefix the name of the column by `"target_"`.
#' ## Note that only values of `col1` for rows matching at least one row in
#' ## rowData of query are returned and an NA is reported for those without
#' ## matching target rows.
#' mo$target_col1
#'
#' ## The 3rd and 4th query rows do not match any target row, thus `NA` is
#' ## returned.
#'
#' ## `matchedData` can be used to extract all (or selected) columns
#' ## from the object. Same as with `$`, a left join between the columns
#' ## from the query and the target is performed. Below we extract selected
#' ## columns from the object as a DataFrame.
#' res <- matchedData(mo, columns = c("col1", "col2", "target_col1",
#'                                   "target_col2"))
#' res
#' res$col1
#' res$target_col1
#'
#' ## The example `Matched` object contains all rows in the
#' ## `rowData` of the `SummarizedExperiment` and all target rows. Below we
#' ## subset the object keeping only rows that are matched to at least one
#' ## target row.
#' mo_sub <- mo[whichQuery(mo)]
#'
#' ## mo_sub contains now a `SummarizedExperiment` with only 3 rows:
#' nrow(query(mo_sub))
#'
#' ## while the original object contains a `SummarizedExperiment` with all 5
#' ## rows:
#' nrow(query(mo))
#'
#' ## Both objects contain however still the full target object:
#' nrow(target(mo))
#' nrow(target(mo_sub))
#'
#' ## With the `pruneTarget` we can however reduce also the target rows to
#' ## only those that match at least one in the `rowData` of query
#' mo_sub <- pruneTarget(mo_sub)
#' nrow(target(mo_sub))
NULL

setClass(
    "Matched",
    slots = c(
        query = "ANY",
        target = "ANY",
        matches = "data.frame",
        queryAssay = "character",
        targetAssay = "character",
        metadata = "list",
        version = "character"
    ),
    prototype = prototype(
        query = list(),
        target = list(),
        matches = data.frame(query_idx = integer(),
                             target_idx = integer(),
                             score = numeric()),
        queryAssay = character(),
        targetAssay = character(),
        metadata = list(),
        version = "0.1")
)

#' @export
#'
#' @rdname Matched
Matched <- function(query = list(), target = list(),
                    matches = data.frame(query_idx = integer(),
                                         target_idx = integer(),
                                         score = numeric()),
                    queryAssay = character(), targetAssay = character(),
                    metadata = list()) {
    new("Matched", query = query, target = target, matches = matches,
        queryAssay = queryAssay, targetAssay = targetAssay, metadata = metadata)
}

setValidity("Matched", function(object) {
    msg <- .validate_matches_format(object@matches)
    if (length(msg)) return(msg)
    msg <- .validate_matches_content(object@matches, length(object),
                                     .nelements_t(object))
    if (any(c("query", "target") %in% colnames(object@matches)))
        return("\"query\" and \"target\" can't be used as matches column names")
    if (length(msg)) return(msg)
    msg <- .validate_qt(object@query)
    if (length(msg)) return(msg)
    msg <- .validate_qt(object@target)
    if (length(msg)) return(msg)
    msg <- .validate_assay(object@query, object@queryAssay)
    if (length(msg)) return(msg)
    msg <- .validate_assay(object@target, object@targetAssay, "target")
    if (length(msg)) return(msg)
    TRUE
})

#' @export
#'
#' @rdname Matched
setMethod("length", "Matched",
          function(x) .nelements(.objectToMatch(x@query, x@queryAssay)))

.nelements_t <- function(x) .nelements(.objectToMatch(x@target, x@targetAssay))

#' @exportMethod show
#'
#' @importMethodsFrom methods show
#'
#' @rdname Matched
setMethod("show", "Matched", function(object) {
    cat("Object of class", class(object)[1L], "\n")
    cat("Total number of matches:", nrow(object@matches), "\n")
    cat("Number of query objects: ", length(object),
        " (", length(unique(object@matches$query_idx)), " matched)\n", sep = "")
    cat("Number of target objects: ", .nelements_t(object), " (",
        length(unique(object@matches$target_idx)), " matched)\n", sep = "")
})

#' @exportMethod [
#'
#' @rdname Matched
setMethod("[", "Matched", function(x, i, j, ..., drop = FALSE) {
    if (missing(i))
        return(x)
    if (is.logical(i))
        i <- which(i)
    .subset_matches_nodim(x, i)
})

#' @rdname Matched
#'
#' @export
matches <- function(object) {
    object@matches
}

#' @rdname Matched
#'
#' @export
target <- function(object) {
    object@target
}

#' @rdname Matched
#'
#' @importMethodsFrom AnnotationHub query
#'
#' @export
setMethod("query", "Matched", function(x, ...) {
    x@query
})

#' @rdname Matched
#'
#' @export
targetIndex <- function(object) {
    if (!inherits(object, "Matched"))
        stop("'object' is expected to be a or inherit from an object of ",
             "type 'Matched'.")
    matches(object)$target_idx
}

#' @rdname Matched
#'
#' @export
queryIndex <- function(object) {
    if (!inherits(object, "Matched"))
        stop("'object' is expected to be a or inherit from an object of ",
             "type 'Matched'.")
    matches(object)$query_idx
}

#' @rdname Matched
#'
#' @export
whichTarget <- function(object) {
    unique(object@matches$target_idx)
}

#' @rdname Matched
#'
#' @export
whichQuery <- function(object) {
    unique(object@matches$query_idx)
}

#' @importMethodsFrom S4Vectors $
#'
#' @rdname Matched
#'
#' @export
setMethod("$", "Matched", function(x, name) {
  .dollar(.objectToMatch(x@query, x@queryAssay),
          .objectToMatch(x@target, x@targetAssay), x@matches, name)
})

#' @importFrom BiocGenerics colnames
#'
#' @exportMethod colnames
#'
#' @rdname Matched
setMethod("colnames", "Matched", function(x) {
  .colnames(.objectToMatch(x@query, x@queryAssay),
            .objectToMatch(x@target, x@targetAssay), x@matches)
})

#' @rdname Matched
#'
#' @export
scoreVariables <- function(object) {
    matchescols <- colnames(object@matches)
    matchescols[grep("score", matchescols, ignore.case = TRUE)]
}

#' @rdname Matched
setMethod("queryVariables", "Matched", function(object) {
    query <- .objectToMatch(object@query, object@queryAssay)
    cnq <- character()
    if (length(dim(query)) == 2)
        cnq <- colnames(query)
    if (is.null(dim(query)))
        cnq <- "query"
    cnq
})

#' @rdname Matched
setMethod("targetVariables", "Matched", function(object) {
    .cnt(.objectToMatch(object@target, object@targetAssay))
})

#' @importMethodsFrom S4Vectors cbind
#'
#' @importFrom S4Vectors DataFrame
#'
#' @rdname Matched
#'
#' @export
setMethod("matchedData", "Matched", function(object,
                                             columns = colnames(object), ...) {
    .matchedData(.objectToMatch(object@query, object@queryAssay),
                 .objectToMatch(object@target, object@targetAssay),
                 object@matches, columns, ... )
})


.nelements <- function(x) {
    ifelse(is.null(d <- dim(x)), length(x), d[1])
}

#' @importFrom methods is
#'
#' @noRd
.extract_elements <- function(x, i, j, drop = FALSE) {
    if (length(dim(x))) {
        if (missing(j)) j <- seq_len(dim(x)[2])
        res <- x[i, j, drop = drop]
    } else res <- x[i]
    if (is(x, "list") && any(na <- is.na(i)))
        res[na] <- NA
    res
}

#' Subsetting of a matched object with slots query, target and matches.
#'
#' @param i has to be an `integer` vector with the indices of the query elements
#'     to keep.
#'
#' @importFrom methods slot<-
#'
#' @noRd
.subset_matches_nodim <- function(x, i) {
    if (!all(i %in% seq_len(length(x))))
        stop("subscript contains out-of-bounds indices", call. = FALSE)
    slot(x, "query", check = FALSE) <- .subset_qt(x@query, x@queryAssay, i)
    mtches <- x@matches[x@matches$query_idx %in% i, , drop = FALSE]
    if (nrow(mtches)) {
        ## Support handling duplicated indices.
        mtches <- split.data.frame(
            mtches, f = as.factor(mtches$query_idx))[as.character(i)]
        lns <- vapply(mtches, function(z)
            if (length(z)) nrow(z) else 0L, integer(1))
        mtches <- do.call(rbind, mtches[lengths(mtches) > 0])
        rownames(mtches) <- NULL
        mtches$query_idx <- rep(seq_along(i), lns)
    }
    slot(x, "matches", check = FALSE) <- mtches
    x
}

.subset_qt <- function(x, assay, i, j, drop = FALSE) {
    if (is(x, "QFeatures")) {
        if (!assay %in% names(x))
            stop("Invalid assay name.", call. = FALSE)
        x[[assay]] <- .extract_elements(x[[assay]], i, j, drop)
    } else {
        x <- .extract_elements(x, i, j, drop)
    }
    x
}

.fill_index <- function(x, y) {
    sort(c(setdiff(x, y), y))
}

.validate_matches_format <- function(x) {
    msg <- NULL
    if (!is.data.frame(x))
        msg <- c(msg, "'matches' should be a 'data.frame'")
    else {
        if (!all(c("query_idx", "target_idx", "score") %in% colnames(x)))
            return(c(msg, paste0("Not all required column names \"query_idx\",",
                                 " \"target_idx\" and \"score\" found in",
                                 " 'matches'")))
        if (!is.integer(x$target_idx))
            msg <- c(msg,
                     "column \"target_idx\" is expected to be of type integer")
        if (!is.integer(x$query_idx))
            msg <- c(msg,
                     "column \"query_idx\" is expected to be of type integer")
    }
    msg
}

.validate_matches_content <- function(x, nquery, ntarget) {
    msg <- NULL
    if (!all(x$query_idx %in% seq_len(nquery)))
        msg <- c(msg, "indices in \"query_idx\" are out-of-bounds")
    if (!all(x$target_idx %in% seq_len(ntarget)))
        msg <- c(msg, "indices in \"target_idx\" are out-of-bounds")
    msg
}

.validate_qt <- function(x){
    msg <- NULL
    ndim <- length(dim(x))
    if (!(ndim %in% c(0,2)))
        msg <- c(msg,
                 "unsupported dimensions in either \"query\" or \"target\"")
    if (ndim == 2 && !inherits(x, "SummarizedExperiment") &&
        is.null(colnames(x)))
        msg <- c(msg, paste0("either \"query\" or \"target\" have 2",
                             " dimensions but no column names"))
    msg
}

.validate_assay <- function(x, assay, what = "query") {
    if (is(x, "QFeatures")) {
        if (!(len_ass <- length(assay)))
            return(paste0("`", what, "Assay` has to be provided when `", what,
                          "` is `QFeatures`."))
        if (len_ass != 1)
            return(paste0("`", what, "Assay` must be `character(1)`"))
        if (!assay %in% names(x))
            return(paste0("No assay in `", what, "` with name \"", assay, "\""))
    }
}

.cnt <- function(target) {
    ndim <- length(dim(target))
    if (ndim == 2) {
        cnt <- colnames(target)
        if (length(cnt)) cnt <- paste0("target_", cnt)
    } else if (ndim == 0) {
        cnt <- "target"
    } else {
        stop("unsupported dimensions in \"target\"", call. = FALSE)
    }
    cnt
}

.colnames <- function(query, target, matches) {
    cns <- colnames(matches)
    cnq <- NULL
    cnt <- .cnt(target)
    if (length(dim(query)) == 2)
        cnq <- colnames(query)
    if (is.null(dim(query)))
        cnq <- "query"
    c(cnq, cnt, cns[!cns %in% c("query_idx", "target_idx")])
}

.dollar <- function(query, target, matches, name) {
    if (name %in% .colnames(query, target, matches))
    {
        not_mtchd <- setdiff(seq_len(.nelements(query)), matches$query_idx)
        idxs_qry <- c(matches$query_idx, not_mtchd)
        ord <- order(idxs_qry)
        if (name %in% colnames(matches)) {
            idxs_mtch <- c(seq_len(nrow(matches)),
                           rep(NA, length(not_mtchd)))[ord]
            return(matches[idxs_mtch, name])
        }
        if (name %in% .cnt(target)) {
            idxs_trg <- c(matches$target_idx, rep(NA, length(not_mtchd)))[ord]
            .extract_elements(target, idxs_trg, sub("target_", "", name),
                              drop = TRUE)
        }else
            .extract_elements(query, idxs_qry[ord], name, drop = TRUE)
    } else stop("'", name, "' not available", call. = FALSE)
}

.matchedData <- function(query, target, matches, columns, ...) {
    cnms <- .colnames(query, target, matches)
    if (any(!columns %in% cnms))
        stop("column(s) ", paste0(columns[!columns %in% cnms],
                                  collapse = ", "), " not available",
             call. = FALSE)
    not_mtchd <- setdiff(seq_len(.nelements(query)), matches$query_idx)
    idxs_qry <- c(matches$query_idx, not_mtchd)
    ord <- order(idxs_qry)
    from_target <- columns %in% .cnt(target)
    from_matches <- columns %in% colnames(matches)
    from_query <- !(from_target | from_matches)
    res_q <- NULL
    res_t <- NULL
    res_m <- NULL
    if (any(from_query))
        res_q <- .extract_elements(query, idxs_qry[ord], columns[from_query])
    if (any(from_target)) {
        idxs_trg <- c(matches$target_idx, rep(NA, length(not_mtchd)))[ord]
        res_t <- .extract_elements(target, idxs_trg,
                                   sub("target_", "", columns[from_target]))
    }
    if (any(from_matches)) {
        idxs_mtch <- c(seq_len(nrow(matches)), rep(NA, length(not_mtchd)))[ord]
        res_m <- matches[idxs_mtch, columns[from_matches], drop = FALSE]
    }
    if (!is.null(res_q) && is.null(dim(query))) res_q <- I(res_q)
    if (!is.null(res_t) && is.null(dim(target))) res_t <- I(res_t)
    any_qtm <- c(any(from_query), any(from_target), any(from_matches))
    res <- DataFrame(do.call(cbind, list(res_q, res_t, res_m)[any_qtm]))
    colnames(res) <- c(columns[from_query], columns[from_target],
                       columns[from_matches])
    res[, columns, drop = FALSE]
}

#' @importMethodsFrom SummarizedExperiment rowData
#'
#' @noRd
.objectToMatch <- function(x, assayname = character(), colnames = character()) {
    what <- as.character(sys.call()[-1])[1]
    if (is(x, "QFeatures")) {
        msg <- .validate_assay(x, assayname, what)
        if (!is.null(msg)) stop(msg, call. = FALSE)
        else x <- x[[assayname]]
    }
    if (is(x, "SummarizedExperiment"))
        x <- rowData(x)
    if (is(x, "Spectra")) {
        if (any(tmp <- !colnames %in% spectraVariables(x)))
            stop(paste0("Missing spectra variables \"", colnames[tmp],
                        "\" in ", what, collapse = "\n"), call. = FALSE)
        if (length(colnames))
            x <- spectraData(x, colnames)
        else
            x <- spectraData(x)
    }
    if (length(colnames)) {
        if (any(tmp <- !colnames %in% colnames(x)))
            stop(paste0("Missing column \"", colnames[tmp], "\" in ",
                        what, collapse = "\n"), call. = FALSE)
        x <- x[, colnames]
    }
    x
}

#' @rdname Matched
#'
#' @importFrom methods validObject
#'
#' @export
pruneTarget <- function(object) {
    keep <- whichTarget(object)
    object@target <- .subset_qt(object@target, object@targetAssay, keep)
    object@matches$target_idx <- match(object@matches$target_idx, keep)
    validObject(object)
    object
}

.findMatchesIdxs <- function(query, target, matches, queryValue = integer(),
                             targetValue = integer(),
                             queryColname = character(),
                             targetColname = character()) {
    if (length(queryValue) != length(targetValue))
        stop("'queryValue' and 'targetValue' must have the same length",
             call. = FALSE)
    if (length(dim(query)) == 2) {
        if (length(queryColname) == 0)
            stop("'queryColname' must be provided when ",
                 "'query' is 2-dimensional", call. = FALSE)
        if (!queryColname %in% colnames(query))
            stop("\"", queryColname, "\" is not a column of 'query'",
                 call. = FALSE)
    }
    targetColname <- sub("target_", "", targetColname)
    if (length(dim(target)) == 2) {
        if (length(targetColname) == 0)
            stop("'targetColname' must be provided when ",
                 "'target' is 2-dimensional", call. = FALSE)
        if (!targetColname %in% colnames(target))
            stop("\"", targetColname, "\" is not a column of 'target'",
                 call. = FALSE)
    }
    mq <- .extract_elements(query, matches$query_idx, queryColname,
                            drop = TRUE)
    mt <- .extract_elements(target, matches$target_idx, targetColname,
                            drop = TRUE)
    unlist(sapply(seq_along(queryValue), function(i)
        which(mq == queryValue[i] & mt == targetValue[i])))
}

#' @rdname Matched
#'
#' @importFrom methods validObject
#'
#' @export
setMethod("filterMatches",
          c("Matched", "missing"), function (object, queryValue = integer(),
                                             targetValue = integer(),
                                             queryColname = character(),
                                             targetColname = character(),
                                             index = integer(),
                                             keep = TRUE, ...) {
              param <- SelectMatchesParam(queryValue = queryValue,
                                          targetValue = targetValue,
                                          queryColname = queryColname,
                                          targetColname = targetColname,
                                          index = index, keep = keep)
              filterMatches(object, param, ...)
})

#' @noRd
setClass("SelectMatchesParam",
         slots = c(
             queryValue = "characterOrNumeric",
             targetValue = "characterOrNumeric",
             queryColname = "character",
             targetColname = "character",
             index = "integer",
             keep = "logical"),
         contains = "Param",
         prototype = prototype(
             queryValue = numeric(),
             targetValue = numeric(),
             queryColname = character(),
             targetColname = character(),
             index = integer(),
             keep = TRUE),
         validity = function(object) {
             msg <- NULL
             if(length(object@queryValue) != length(object@targetValue))
                 msg <- c(msg, paste0("'queryValue' and 'targetValue' ",
                                      "must have the same length"))
             if (length(object@queryColname) > 1)
                 msg <- c(msg,
                          "'queryColname' cannot be of length greater than 1")
             if (length(object@targetColname) > 1)
                 msg <- c(msg,
                          "'targetColname' cannot be of length greater than 1")
             if (any(object@index <= 0))
                 msg <- c(msg, "'index' must contain positive integers")
             if (length(object@keep) != 1)
                 msg <- c(msg, "'keep' must be a logical of length 1")
             msg
         })

#' @rdname Matched
#'
#' @importFrom methods new
#'
#' @export
SelectMatchesParam <-
    function(queryValue = numeric(), targetValue = numeric(),
             queryColname = character(), targetColname = character(),
             index = integer(), keep = TRUE) {
        new("SelectMatchesParam",
            queryValue = queryValue, targetValue = targetValue,
            queryColname = queryColname, targetColname = targetColname,
            index = as.integer(index), keep = keep)
    }

#' @noRd
setClass("TopRankedMatchesParam",
         slots = c(
             n = "integer",
             decreasing = "logical"),
         contains = "Param",
         prototype = prototype(
             n = 1L,
             decreasing = FALSE),
         validity = function(object) {
             msg <- NULL
             if (length(object@n) != 1 || object@n <= 0)
                 msg <- "'n' must be a length 1 positive integer"
             msg
         })

#' @rdname Matched
#'
#' @importFrom methods new
#'
#' @export
TopRankedMatchesParam <- function(n = 1L, decreasing = FALSE) {
    new("TopRankedMatchesParam", n = as.integer(n), decreasing = decreasing[1L])
}

#' @noRd
setClass("ScoreThresholdParam",
         slots = c(
             threshold = "numeric",
             above = "logical",
             column = "character"),
         contains = "Param",
         prototype = prototype(
             threshold = 0,
             above = FALSE,
             column = "score")
         )

#' @rdname Matched
#'
#' @importFrom methods new
#'
#' @export
ScoreThresholdParam <- function(threshold = 0, above = FALSE,
                                column = "score") {
    new("ScoreThresholdParam", threshold = threshold[1L], above = above[1L],
        column = column[1L])
}

#' @rdname Matched
#'
#' @importFrom methods validObject
#'
#' @export
setMethod("filterMatches", c("Matched", "SelectMatchesParam"), function (object,
                                                                         param,
                                                                         ...) {
    index <- param@index
    if (length(index) && any(!index %in% seq_len(nrow(object@matches))))
        stop("some indices in \"index\" are out-of-bounds", call. = FALSE)
    if (!length(index) && length(param@queryValue))
        index  <- .findMatchesIdxs(.objectToMatch(object@query,
                                                  object@queryAssay),
                                   .objectToMatch(object@target,
                                                  object@targetAssay),
                                   object@matches, param@queryValue,
                                   param@targetValue, param@queryColname,
                                   param@targetColname)
    if (param@keep) to_keep <- seq_len(nrow(object@matches)) %in% index
    else to_keep <- !seq_len(nrow(object@matches)) %in% index
    object@matches <- object@matches[to_keep, , drop = FALSE]
    object@metadata <- c(object@metadata, param = param)
    validObject(object)
    object
})

#' @rdname Matched
#'
#' @importFrom methods validObject
#'
#' @export
setMethod("filterMatches", c("Matched", "TopRankedMatchesParam"),
          function (object, param, ...) {
              sign <- ifelse(param@decreasing, yes = -1, no = 1)
              rank <- rank(abs(object@matches$score) * sign)
              if ("score_rt" %in% colnames(object@matches))
                  rank <- rank * rank(abs(object@matches$score_rt) * sign)
              seq_len_nm <- seq_len(nrow(object@matches))
              tmp <- split.data.frame(cbind(seq_len_nm, rank),
                                      object@matches$query_idx)
              index <- do.call("c", lapply(tmp, function(x)
                  x[order(x[, 2])[seq_len(min(param@n, nrow(x)))], 1]))
              to_keep <- seq_len_nm %in% index
              object@matches <- object@matches[to_keep, , drop = FALSE]
              object@metadata <- c(object@metadata, param = param)
              validObject(object)
              object
          })

#' @rdname Matched
#'
#' @importFrom methods validObject
#'
#' @export
setMethod("filterMatches", c("Matched", "ScoreThresholdParam"),
          function (object, param, ...) {
              if (!param@column %in% colnames(object@matches))
                  stop("\"", param@column,
                       "\" variable not present in `object`")
              if (param@above)
                  to_keep <- object@matches[, param@column] > param@threshold
              else to_keep <- object@matches[, param@column] < param@threshold
              object@matches <- object@matches[to_keep, , drop = FALSE]
              object@metadata <- c(object@metadata, param = param)
              validObject(object)
              object
          })

#' @noRd
setClass("SingleMatchParam",
         slots = c(
             duplicates = "character",
             column = "character",
             decreasing = "logical"),
         contains = "Param",
         prototype = prototype(
             duplicates = "remove",
             column = "score",
             decreasing = TRUE)
         )

#' @rdname Matched
#'
#' @export
SingleMatchParam <- function(duplicates = c("remove", "closest", "top_ranked"),
                             column = "score", decreasing = TRUE) {
    duplicates <- force(match.arg(duplicates))
    new("SingleMatchParam", duplicates = duplicates, column = column[1L],
        decreasing = decreasing[1L])
}

#' @rdname Matched
#'
#' @export
setMethod(
    "filterMatches", c("Matched", "SingleMatchParam"),
    function (object, param, ...) {
        if (!param@column %in% c(scoreVariables(object),
                                 targetVariables(object)))
            stop("Variable \"", param@column, "\" not found. `column` ",
                 "should be one of 'scoreVariables(object)' or ",
                 "'targetVariables(object)'.")
        object@metadata <- c(object@metadata, param = param)
        if (!nrow(object@matches))
            return(object)
        switch(
            param@duplicates[1L],
            "remove" = {
                s <- split(seq_len(nrow(object@matches)),
                           object@matches$query_idx)
                keep <- unlist(s[lengths(s) == 1L], use.names = FALSE)
                object@matches <- object@matches[keep, , drop = FALSE]
            },
            "closest" = {
                object <- filterMatches(
                    object, TopRankedMatchesParam(n = 1L, decreasing = FALSE))
            },
            "top_ranked" = {
                ## Rank matches by "column"
                if (param@column %in% scoreVariables(object))
                    vals <- cbind(seq_len(nrow(object@matches)),
                                  object@matches$query_idx,
                                  object@matches[, param@column])
                else
                    vals <- cbind(
                        seq_len(nrow(object@matches)),
                        object@matches$query_idx,
                        .extract_elements(
                            .objectToMatch(object@target, object@targetAssay),
                            object@matches$target_idx,
                            sub("target_", "", param@column)))
                    vals <- vals[order(vals[, 3L],
                                       decreasing = param@decreasing), ,
                                 drop = FALSE]
                    keep <- vals[match(unique(object@matches$query_idx),
                                       vals[, 2L]), 1L]
                    object@matches <- object@matches[keep, , drop = FALSE]
            },
            stop("'duplicates' has to be one of \"remove\", \"closest\"",
                 " or \"top_ranked\"."))
        validObject(object)
        object
    })

#' @importFrom MsCoreUtils rbindFill
.addMatches <- function(query, target, matches, queryValue = integer(),
                        targetValue = integer(), queryColname = character(),
                        targetColname = character(),
                        score = rep(NA_real_, length(queryValue)),
                        isIndex = FALSE) {
    if (is.numeric(score))
        score <- data.frame(score = score)
    if (!is.data.frame(score))
        stop("'score' needs to be either a 'data.frame' or a numeric",
             call. = FALSE)
    if (length(queryValue) != length(targetValue) ||
        length(queryValue) != nrow(score))
        stop("'queryValue', 'targetValue' and 'score' must have the",
             " same length", call. = FALSE)
    if (isIndex) {
        if (!is.integer(queryValue) || !is.integer(targetValue))
            stop("'queryValue' and 'targetValue' must be integer ",
                 "vectors when 'isIndex = TRUE'", call. = FALSE)
        if (any(!queryValue %in% seq_len(.nelements(query))))
            stop("Provided indices in 'queryValue' are out-of-bounds.",
                 call. = FALSE)
        if (any(!targetValue %in% seq_len(.nelements(target))))
            stop("Provided indices in 'targetValue' are out-of-bounds.",
                 call. = FALSE)
        query_idx <- queryValue
        target_idx <- targetValue
    } else {
        if (length(dim(query)) == 2)
            if (!queryColname %in% colnames(query))
                stop("\"", queryColname,"\" is not a column of 'query'",
                     call. = FALSE)
        if (length(dim(target)) == 2)
            if (!targetColname %in% colnames(target))
                stop("\"", targetColname,"\" is not a column of 'target'",
                     call. = FALSE)
        mq <- match(queryValue, .extract_elements(query, j = queryColname,
                                                  drop = TRUE))
        mt <- match(targetValue, .extract_elements(target, j = targetColname,
                                                   drop = TRUE))
        to_keep <- !is.na(mq) & !is.na(mt)
        query_idx <- mq[to_keep]
        target_idx <- mt[to_keep]
        score <- score[to_keep, ]
    }
    new_matches <- cbind(
        data.frame(query_idx = query_idx, target_idx = target_idx),
        score)
    new_matches <- rbindFill(matches, new_matches)
    ## remove possible matches that were already in matches
    new_matches[!duplicated(new_matches[, c("query_idx", "target_idx")]), ]
}

#' @rdname Matched
#'
#' @importFrom methods validObject
#'
#' @export
setMethod("addMatches", "Matched",
          function(object, queryValue = integer(), targetValue = integer(),
                   queryColname = character(), targetColname = character(),
                   score = rep(NA_real_, length(queryValue)), isIndex = FALSE) {
              object@matches <- .addMatches(.objectToMatch(object@query,
                                                           object@queryAssay),
                                            .objectToMatch(object@target,
                                                           object@targetAssay),
                                            object@matches, queryValue,
                                            targetValue, queryColname,
                                            targetColname, score, isIndex)
              validObject(object)
              object
          })

#' @rdname Matched
#'
#' @importFrom methods validObject
#'
#' @importFrom S4Vectors endoapply
#'
#' @export
setMethod("endoapply", "ANY", function(X, FUN, ...) {
    S4Vectors::endoapply(X, FUN, ...)
})

#' @rdname Matched
#'
#' @importFrom methods validObject
#'
#' @export
setMethod("endoapply", "Matched", function(X, FUN, ...) {
    tmp <- lapply(seq_along(X), function(i) FUN(X[i], ...)@matches)
    matches <- do.call(rbind, tmp)
    matches$query_idx <- rep(seq_along(tmp), vapply(tmp, nrow, integer(1)))
    X@matches <- matches
    validObject(X)
    X
})

#' @rdname Matched
#'
#' @importFrom BiocGenerics lapply
#'
#' @export
setMethod("lapply", "Matched", function(X, FUN, ...) {
    lapply(seq_along(X), function(i) FUN(X[i], ...))
})

library(Spectra)
library(ProtGenerics)
df1 <- DataFrame(
    msLevel = 2L, rtime = 1:10,
    spectrum_id = c("a", "b", "c", "d", "e", "f", "g", "h", "i", "j"))
df2 <- DataFrame(
    msLevel = 2L, rtime = rep(1:10, 20),
    spectrum_id = rep(c("A", "B", "C", "D", "E"), 20))
sp1 <- Spectra(df1)
sp2 <- Spectra(df2)

test_that("MatchedSpectra works", {
    ms <- MatchedSpectra()
    expect_true(validObject(ms))
    expect_output(show(ms), "MatchedSpectra")
    expect_true(is(query(ms), "Spectra"))
    expect_true(is(target(ms), "Spectra"))
    expect_equal(whichQuery(ms), integer())
    expect_equal(whichTarget(ms), integer())
    expect_equal(scoreVariables(ms), "score")
    expect_equal(queryVariables(ms), spectraVariables(ms@query))
    expect_equal(targetVariables(ms),
                 paste0("target_", spectraVariables(ms@target)))

    ## With data
    ms <- MatchedSpectra(
        sp1, sp2, matches = data.frame(query_idx = c(1L, 1L, 2L, 4L, 4L, 4L),
                                       target_idx = c(2L, 5L, 2L, 8L, 12L, 15L),
                                       score = 1:6))
    expect_true(validObject(ms))
    expect_true(length(ms) == 10)
    expect_output(show(ms), "5 matched")
    expect_true(is(query(ms), "Spectra"))
    expect_true(is(target(ms), "Spectra"))
    expect_equal(whichQuery(ms), c(1L, 2L, 4L))
    expect_equal(whichTarget(ms), c(2L, 5L, 8L, 12L, 15L))
    expect_equal(scoreVariables(ms), "score")
    expect_equal(queryVariables(ms), spectraVariables(ms@query))
    expect_equal(targetVariables(ms),
                 paste0("target_", spectraVariables(ms@target)))

    ms <- .matched_spectra(
        sp1, sp2, matches = data.frame(query_idx = c(1L, 1L, 2L, 4L, 4L, 4L),
                                       target_idx = c(2L, 5L, 2L, 8L, 12L, 15L),
                                       score = 1:6))
    expect_true(validObject(ms))
    expect_true(length(ms) == 10)
    expect_output(show(ms), "5 matched")
    expect_true(is(query(ms), "Spectra"))
    expect_true(is(target(ms), "Spectra"))
    expect_equal(whichQuery(ms), c(1L, 2L, 4L))
    expect_equal(whichTarget(ms), c(2L, 5L, 8L, 12L, 15L))
})

test_that(".subset_matches_nodim and [ works", {
    ms <- MatchedSpectra()
    expect_error(.subset_matches_nodim(ms, 1), "out-of-bounds")
    ms <- MatchedSpectra(
        sp1, sp2, matches = data.frame(query_idx = c(1L, 1L, 2L, 4L, 4L, 4L),
                                       target_idx = c(2L, 5L, 2L, 8L, 12L, 15L),
                                       score = 1:6))
    res <- .subset_matches_nodim(ms, 2)
    expect_true(length(res) == 1)
    expect_equal(res@matches$query_idx, 1L)
    expect_equal(res@matches$target_idx, ms@matches$target_idx[3])
    expect_equal(res@matches$score, ms@matches$score[3])
    expect_equal(res@query, ms@query[2])
    expect_equal(res@target, ms@target)

    ## no matches
    res <- .subset_matches_nodim(ms, c(10, 3))
    expect_true(length(res) == 2)
    expect_true(validObject(res))
    expect_true(nrow(res@matches) == 0)
    expect_equal(res@query, ms@query[c(10, 3)])
    expect_equal(res@target, ms@target)

    expect_error(.subset_matches_nodim(ms, 12), "out-of-bounds")

    res <- .subset_matches_nodim(ms, c(2, 4))
    expect_equal(res@query, ms@query[c(2, 4)])
    expect_true(length(res) == 2)
    expect_equal(res@matches$score, 3:6)

    ## arbitrary order
    res <- .subset_matches_nodim(ms, c(3, 2, 1, 1, 6, 9))
    expect_equal(query(res), query(ms)[c(3, 2, 1, 1, 6, 9)])
    expect_equal(target(res), target(ms))
    expect_equal(res@matches$query_idx, c(2L, 3L, 3L, 4L, 4L))
    expect_equal(res@matches$score, c(3, 1, 2, 1, 2))

    res <- ms[]
    expect_equal(res, ms)
    res <- ms[c(FALSE, TRUE)]
    expect_equal(res@matches$query_idx, 1L)
    expect_equal(res@matches$target_idx, 2L)
    expect_equal(res@matches$score, 3)

    ## All works even after subsetting and pruning.
    ms <- MatchedSpectra(
        sp1, sp2, matches = data.frame(query_idx = c(1L, 1L, 2L, 4L, 4L, 4L),
                                       target_idx = c(2L, 5L, 2L, 8L, 12L, 15L),
                                       score = 1:6))
    res <- ms[whichQuery(ms)]
    expect_equal(res$spectrum_id, c("a", "a", "b", "d", "d", "d"))
    res <- pruneTarget(res)
})

test_that("$,MatchedSpectra works", {
    ms <- MatchedSpectra()
    expect_equal(ms$msLevel, integer())
    expect_equal(ms$target_msLevel, integer())

    ms <- MatchedSpectra(
        sp1, sp2, matches = data.frame(query_idx = c(1L, 1L, 2L, 4L, 4L, 4L),
                                       target_idx = c(2L, 5L, 2L, 8L, 12L, 15L),
                                       score = 1:6))
    res <- ms$rtime
    expect_equal(res, c(1, 1, 2, 3, 4, 4, 4, 5, 6, 7, 8, 9, 10))
    res <- ms$target_rtime
    expect_equal(res, c(2, 5, 2, NA, 8, 2, 5, NA, NA, NA, NA, NA, NA))
    res <- ms$score
    expect_equal(res, c(1, 2, 3, NA, 4, 5, 6, NA, NA, NA, NA, NA, NA))

    ## A MatchedSpectra with no matching target spectra
    ms <- MatchedSpectra(sp1, sp2, matches = data.frame(query_idx = integer(),
                                                        target_idx = integer(),
                                                        score = numeric()))
    res <- ms$rtime
    expect_equal(res, 1:10)
    res <- ms$target_rtime
    expect_true(all(is.na(res)))
    res <- ms$score
    expect_true(all(is.na(res)))
})

test_that("spectraData,MatchedSpectra works", {
    ms <- MatchedSpectra()
    res <- spectraData(ms)
    expect_true(is(res, "DataFrame"))
    expect_true(nrow(res) == 0)
    expect_equal(colnames(res), spectraVariables(ms))
    expect_equal(colnames(ms), colnames(res))

    ms <- MatchedSpectra(
        sp1, sp2, matches = data.frame(query_idx = c(1L, 1L, 2L, 4L, 4L, 4L),
                                       target_idx = c(2L, 5L, 2L, 8L, 12L, 15L),
                                       score = 1:6))
    res <- spectraData(ms)
    expect_equal(res$rtime, c(1, 1, 2, 3, 4, 4, 4, 5, 6, 7, 8, 9, 10))
    expect_equal(res$target_rtime,
                 c(2, 5, 2, NA, 8, 2, 5, NA, NA, NA, NA, NA, NA))
    expect_equal(res$score, c(1, 2, 3, NA, 4, 5, 6, NA, NA, NA, NA, NA, NA))

    expect_error(spectraData(ms, columns = "other"), "other not available")


    ## Only query spectra variables
    res <- spectraData(ms, columns = c("rtime", "spectrum_id", "msLevel"))
    expect_equal(colnames(res), c("rtime", "spectrum_id", "msLevel"))
    expect_equal(res$rtime, c(1, 1, 2, 3, 4, 4, 4, 5, 6, 7, 8, 9, 10))

    ## Only target spectra variables
    res <- spectraData(ms, columns = c("target_rtime", "target_spectrum_id"))
    expect_equal(colnames(res), c("target_rtime", "target_spectrum_id"))
    expect_equal(res$target_rtime,
                 c(2, 5, 2, NA, 8, 2, 5, NA, NA, NA, NA, NA, NA))

    ## Only matches
    res <- spectraData(ms, columns = c("score"))
    expect_equal(colnames(res), c("score"))
    expect_equal(res$score,
                 c(1, 2, 3, NA, 4, 5, 6, NA, NA, NA, NA, NA, NA))

    ## A MatchedSpectra with no matching target spectra
    ms <- MatchedSpectra(sp1, sp2, matches = data.frame(query_idx = integer(),
                                                        target_idx = integer(),
                                                        score = numeric()))
    res <- spectraData(ms)
    expect_equal(res$rtime, 1:10)
    expect_true(all(is.na(res$target_rtime)))

    ## Only query spectra variables
    res <- spectraData(ms, columns = c("rtime", "spectrum_id", "msLevel"))
    expect_equal(colnames(res), c("rtime", "spectrum_id", "msLevel"))
    expect_equal(res$rtime, 1:10)

    ## Only target spectra variables
    res <- spectraData(ms, columns = c("target_rtime", "target_spectrum_id"))
    expect_true(nrow(res) == 10)
    expect_equal(colnames(res), c("target_rtime", "target_spectrum_id"))
    expect_true(all(is.na(res$target_rtime)))

    ## Only matches
    res <- spectraData(ms, columns = c("score"))
    expect_equal(colnames(res), c("score"))
    expect_true(all(is.na(res$score)))
})

test_that("pruneTarget,MatchedSpectra works", {
    ms <- MatchedSpectra()
    res <- pruneTarget(ms)
    expect_true(is(res, "MatchedSpectra"))
    expect_equal(res, ms)

    ms <- MatchedSpectra(
        sp1, sp2, matches = data.frame(query_idx = c(1L, 1L, 2L, 4L, 4L, 4L),
                                       target_idx = c(2L, 5L, 2L, 8L, 12L, 15L),
                                       score = 1:6))
    res <- pruneTarget(ms)
    expect_equal(spectraData(res), spectraData(ms))
    expect_true(length(res@target) < length(ms@target))

    ms <- MatchedSpectra(sp1, sp2, matches = data.frame(query_idx = integer(),
                                                        target_idx = integer(),
                                                        score = numeric()))
    res <- pruneTarget(ms)
    expect_equal(spectraData(res), spectraData(ms))
    expect_true(length(res@target) < length(ms@target))
})

test_that("plotSpectraMirror throws an error", {
    ms <- MatchedSpectra()
    expect_error(plotSpectraMirror(ms), "Length")

    ms <- MatchedSpectra(sp1, sp2, matches = data.frame(query_idx = integer(),
                                                        target_idx = integer(),
                                                        score = numeric()))
    plotSpectraMirror(ms[1])
    plotSpectraMirror(ms[1], scalePeaks = TRUE)

    ## test the helper function
    dots <- list(ppm = 15)
    pl <- as.list(param <- CompareSpectraParam())
    expect_equal(.res_setting(dots = dots, pl, "ppm", 20), 15)
    expect_equal(.res_setting(list(), pl, "ppm", 20), 5)
    expect_equal(.res_setting(list(), list(), "ppm", 20), 20)
})

test_that("addProcessing works", {
    ms <- MatchedSpectra(
        sp1, sp2, matches = data.frame(query_idx = c(1L, 1L, 2L, 4L, 4L, 4L),
                                       target_idx = c(2L, 5L, 2L, 8L, 12L, 15L),
                                       score = 1:6))
    res <- addProcessing(ms)
    expect_equal(res, ms)
    FUN <- function(x, ...) {
        x - 4
    }
    res <- addProcessing(ms, FUN = FUN)
    expect_equal(res@query@processingQueue, list(ProcessingStep(FUN)))
    expect_equal(res@target@processingQueue, list(ProcessingStep(FUN)))
})

test_that("setBackend,MatchedSpectra works", {
    res <- matchSpectra(pest_ms2, minimb, param = CompareSpectraParam())
    expect_s4_class(query(res)@backend, "MsBackendMzR")
    expect_s4_class(target(res)@backend, "MsBackendDataFrame")

    res <- setBackend(res, MsBackendDataFrame())
    expect_s4_class(query(res)@backend, "MsBackendDataFrame")
    expect_s4_class(target(res)@backend, "MsBackendDataFrame")
})

q1 <- data.frame(col1 = 1:5, col2 = 6:10)
t1 <- data.frame(col1 = 11:16, col2 = 17:22)
q2 <- list("A", c(1,2), 1L, list(), data.frame())
t2 <- list("A", c(1,2), 1L, list(), data.frame(), matrix(1:3))
library(SummarizedExperiment)
q3 <- SummarizedExperiment(assays = data.frame(matrix(NA, 5, 2)),
                           rowData = q1,
                           colData = data.frame(cD1 = c(NA, NA),
                                                cD2 = c(NA, NA)))
library(QFeatures)
q4 <- QFeatures(list(a1 = q3, a2 = q3[4:5]))
t3 <- t4 <- t1

m <- data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                target_idx = c(2L, 2L, 3L, 4L, 5L),
                score = seq(0.5, 0.9, by = 0.1))

test_that("validator work", {
    ## .validate_matches_format
    res <- .validate_matches_format(4)
    expect_match(res, "data.frame")
    res <- .validate_matches_format(data.frame())
    expect_match(res, "Not all ")
    tmp <- data.frame(query_idx = 1:3, target_idx = 1:6, score = 12)
    res <- .validate_matches_format(tmp)
    expect_equal(res, NULL)
    tmp$query_idx <- "a"
    res <- .validate_matches_format(tmp)
    expect_match(res, "query_idx")
    tmp$query_idx <- 1:6
    tmp$target_idx <- "a"
    res <- .validate_matches_format(tmp)
    expect_match(res, "target_idx")

    ## .validate_matches_content
    tmp$target_idx <- c(1:3, 1:3)
    res <- .validate_matches_content(tmp, 3, 3)
    expect_match(res, "query_idx")
    res <- .validate_matches_content(tmp, 9, 2)
    expect_match(res, "target_idx")
    res <- .validate_matches_content(tmp, 10, 10)
    expect_equal(res, NULL)

    ## .validate_qt
    q <-  matrix(rep(1,4),2,2)
    res <- .validate_qt(q)
    expect_match(res, "no column names")
    colnames(q) <- c("col1", "col2")
    res <- .validate_qt(q)
    expect_equal(res, NULL)
    q <- array(rep(1,8), dim = c(2,2,2))
    res <- .validate_qt(q)
    expect_match(res, "unsupported dimensions ")

})

test_that("Matched works", {
    mo <- Matched()
    expect_true(validObject(mo))
    expect_output(show(mo), "Matched")
    expect_equal(whichQuery(mo), integer())
    expect_equal(whichTarget(mo), integer())
    expect_equal(scoreVariables(mo), "score")
    expect_equal(queryVariables(mo), "query")
    expect_equal(targetVariables(mo), "target")

    ## With data. query and target of type data.frame
    mo <- Matched(query = q1, target = t1,
                  matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                       target_idx = c(2L, 2L, 3L, 4L, 5L),
                                       score = seq(0.5, 0.9, by = 0.1)))
    expect_true(validObject(mo))
    expect_true(length(mo) == 5)
    expect_output(show(mo), "4 matched")
    expect_true(is(query(mo), "data.frame"))
    expect_true(is(target(mo), "data.frame"))
    expect_equal(whichQuery(mo), c(1L, 2L, 5L))
    expect_equal(whichTarget(mo), c(2L, 3L, 4L, 5L))
    expect_equal(scoreVariables(mo), c("score"))
    expect_equal(queryVariables(mo), colnames(q1))
    expect_equal(targetVariables(mo), paste0("target_", colnames(t1)))

    ## With data. query and target of type list
    mo <- Matched(query = q2, target = t2,
                  matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                       target_idx = c(2L, 2L, 3L, 4L, 5L),
                                       score = seq(0.5, 0.9, by = 0.1)))
    expect_true(validObject(mo))
    expect_true(length(mo) == 5)
    expect_output(show(mo), "4 matched")
    expect_true(is(query(mo), "list"))
    expect_true(is(target(mo), "list"))
    expect_equal(whichQuery(mo), c(1L, 2L, 5L))
    expect_equal(whichTarget(mo), c(2L, 3L, 4L, 5L))
    expect_equal(scoreVariables(mo), "score")
    expect_equal(queryVariables(mo), "query")
    expect_equal(targetVariables(mo), "target")

    ## With data. query SummarizedExperiment and target data.frame
    mo <- Matched(query = q3, target = t3,
                  matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                       target_idx = c(2L, 2L, 3L, 4L, 5L),
                                       score = seq(0.5, 0.9, by = 0.1)))
    expect_true(validObject(mo))
    expect_true(length(mo) == 5)
    expect_output(show(mo), "4 matched")
    expect_true(is(query(mo), "SummarizedExperiment"))
    expect_true(is(target(mo), "data.frame"))
    expect_equal(whichQuery(mo), c(1L, 2L, 5L))
    expect_equal(whichTarget(mo), c(2L, 3L, 4L, 5L))
    expect_equal(scoreVariables(mo), "score")
    expect_equal(queryVariables(mo), colnames(rowData(q3)))
    expect_equal(targetVariables(mo), paste0("target_", colnames(t3)))
})

test_that(".subset_matches_nodim and [ works", {
    mo <- Matched()
    expect_error(.subset_matches_nodim(mo, 1), "out-of-bounds")

    #### query and target: data.frames
    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))

    res <- .subset_matches_nodim(mo, 2)
    expect_true(length(res) == 1)
    expect_equal(res@matches$query_idx, c(1L, 1L, 1L))
    expect_equal(res@matches$target_idx, mo@matches$target_idx[c(2, 3, 4)])
    expect_equal(res@matches$score, mo@matches$score[c(2, 3, 4)])
    expect_equal(res@query, mo@query[2, ])
    expect_equal(res@target, mo@target)

    ## No matching
    res <- .subset_matches_nodim(mo, c(3, 4))
    expect_true(validObject(res))
    expect_true(nrow(res@matches) == 0)
    expect_equal(res@query, mo@query[c(3, 4), ])

    expect_error(.subset_matches_nodim(mo, 12), "out-of-bounds")

    res <- .subset_matches_nodim(mo, c(1, 5))
    expect_equal(res@query, mo@query[c(1, 5), ])
    expect_true(length(res) == 2)
    expect_equal(res@matches$score, mo@matches$score[c(1, 5)])

    ## arbitrary order
    res <- .subset_matches_nodim(mo, c(3, 2, 1, 1, 4))
    expect_equal(query(res), query(mo)[c(3, 2, 1, 1, 4), ])
    expect_equal(target(res), target(mo))
    expect_equal(res@matches$query_idx, c(2L, 2L, 2L, 3L, 4L))
    expect_equal(res@matches$score, c(0.6, 0.7, 0.8, 0.5, 0.5))

    res <- mo[]
    expect_equal(res, mo)
    res <- mo[c(TRUE, FALSE)]
    expect_equal(res@matches$query_idx, 1L)
    expect_equal(res@matches$target_idx, 2L)
    expect_equal(res@matches$score, 0.5)

    ## All works even after subsetting and pruning.
    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))

    res <- mo[whichQuery(mo)]
    expect_equal(query(res), q1[whichQuery(mo), ])
    res <- pruneTarget(res)
    expect_equal(target(res), t1[c(2, 3, 4, 5), ])
    expect_equal(res@matches$target_idx, c(1L, 1L, 2L, 3L, 4L))

    #### query and target: lists
    mo <- Matched(
        q2, t2, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    res <- .subset_matches_nodim(mo, 2)
    expect_true(length(res) == 1)
    expect_equal(res@matches$query_idx, c(1L, 1L, 1L))
    expect_equal(res@matches$target_idx, mo@matches$target_idx[c(2, 3, 4)])
    expect_equal(res@matches$score, mo@matches$score[c(2, 3, 4)])
    expect_equal(res@query, mo@query[2])
    expect_equal(res@target, mo@target)

    expect_error(.subset_matches_nodim(mo, 12), "out-of-bounds")

    res <- .subset_matches_nodim(mo, c(1, 5))
    expect_equal(res@query, mo@query[c(1, 5)])
    expect_true(length(res) == 2)
    expect_equal(res@matches$score, mo@matches$score[c(1, 5)])

    ## arbitrary order
    res <- .subset_matches_nodim(mo, c(3, 2, 1, 1, 4))
    expect_equal(query(res), query(mo)[c(3, 2, 1, 1, 4)])
    expect_equal(target(res), target(mo))
    expect_equal(res@matches$query_idx, c(2L, 2L, 2L, 3L, 4L))
    expect_equal(res@matches$score, c(0.6, 0.7, 0.8, 0.5, 0.5))

    res <- mo[]
    expect_equal(res, mo)
    res <- mo[c(TRUE, FALSE)]
    expect_equal(res@matches$query_idx, 1L)
    expect_equal(res@matches$target_idx, 2L)
    expect_equal(res@matches$score, 0.5)

    ## All works even after subsetting and pruning.
    mo <- Matched(
        q2, t2, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))

    res <- mo[whichQuery(mo)]
    expect_equal(query(res), q2[whichQuery(mo)])
    res <- pruneTarget(res)
    expect_equal(target(res), t2[c(2, 3, 4, 5)])
    expect_equal(res@matches$target_idx, c(1L, 1L, 2L, 3L, 4L))

    #### query SummarizedExperiment and target data.frame
    mo <- Matched(
        q3, t3, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))

    res <- .subset_matches_nodim(mo, 2)
    expect_true(length(res) == 1)
    expect_equal(res@matches$query_idx, c(1L, 1L, 1L))
    expect_equal(res@matches$target_idx, mo@matches$target_idx[c(2, 3, 4)])
    expect_equal(res@matches$score, mo@matches$score[c(2, 3, 4)])
    expect_equal(res@query, mo@query[2, ])
    expect_equal(res@target, mo@target)


    expect_error(.subset_matches_nodim(mo, 12), "out-of-bounds")

    res <- .subset_matches_nodim(mo, c(1, 5))
    expect_equal(res@query, mo@query[c(1, 5), ])
    expect_true(length(res) == 2)
    expect_equal(res@matches$score, mo@matches$score[c(1, 5)])

    ## arbitrary order
    res <- .subset_matches_nodim(mo, c(3, 2, 1, 1, 4))
    expect_equal(query(res), query(mo)[c(3, 2, 1, 1, 4), ])
    expect_equal(target(res), target(mo))
    expect_equal(res@matches$query_idx, c(2L, 2L, 2L, 3L, 4L))
    expect_equal(res@matches$score, c(0.6, 0.7, 0.8, 0.5, 0.5))

    res <- mo[]
    expect_equal(res, mo)
    res <- mo[c(TRUE, FALSE)]
    expect_equal(res@matches$query_idx, 1L)
    expect_equal(res@matches$target_idx, 2L)
    expect_equal(res@matches$score, 0.5)

    ## All works even after subsetting and pruning.
    mo <- Matched(
        q3, t3, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))

    res <- mo[whichQuery(mo)]
    expect_equal(query(res), q3[whichQuery(mo), ])
    res <- pruneTarget(res)
    expect_equal(target(res), t3[c(2, 3, 4, 5), ])
    expect_equal(res@matches$target_idx, c(1L, 1L, 2L, 3L, 4L))
})

test_that(".fill_index works", {
    res <- .fill_index(1:5, integer())
    expect_equal(res, 1:5)

    res <- .fill_index(1:20, c(2, 4, 4, 4, 6, 9, 10))
    expect_equal(res, c(1, 2, 3, 4, 4, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13,
                        14, 15, 16, 17, 18, 19, 20))
})

test_that("$,Matched and .dollar works", {
    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    expect_equal(mo$score, c(0.5, 0.6, 0.7, 0.8, NA, NA, 0.9))
    expect_equal(mo$score, .dollar(mo@query, mo@target, mo@matches, "score"))
    expect_equal(mo$col1, q1[c(1, 2, 2, 2, 3, 4, 5), "col1"])
    expect_equal(mo$col1, .dollar(mo@query, mo@target, mo@matches, "col1"))
    expect_equal(mo$target_col1, t1[c(2,2,3,4, NA, NA, 5), "col1"])
    expect_equal(mo$target_col1,
                 .dollar(mo@query, mo@target, mo@matches, "target_col1"))

    expect_error(.dollar(mo@query, mo@target, mo@matches, "no_col"),
                 "not available")

    mo <- Matched(
        q2, t2, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    expect_equal(mo$score, c(0.5, 0.6, 0.7, 0.8, NA, NA, 0.9))
    expect_equal(mo$query, q2[c(1, 2, 2, 2, 3, 4, 5)])
    tmp <- t2[c(2,2,3,4, NA, NA, 5)]
    tmp[5:6] <- NA
    expect_equal(mo$target, tmp)

    mo <- Matched(
        q2, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    expect_equal(mo$score, c(0.5, 0.6, 0.7, 0.8, NA, NA, 0.9))
    expect_equal(mo$query, q2[c(1, 2, 2, 2, 3, 4, 5)])
    expect_equal(mo$target_col1, t1[c(2,2,3,4, NA, NA, 5), "col1"])

    mo <- Matched(
        q1, t2, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    expect_equal(mo$score, c(0.5, 0.6, 0.7, 0.8, NA, NA, 0.9))
    expect_equal(mo$col1, q1[c(1, 2, 2, 2, 3, 4, 5), "col1"])
    tmp <- t2[c(2,2,3,4, NA, NA, 5)]
    tmp[5:6] <- NA
    expect_equal(mo$target, tmp)

    mo <- Matched(
        q3, t3, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    expect_equal(mo$score, c(0.5, 0.6, 0.7, 0.8, NA, NA, 0.9))
    expect_equal(mo$col1, rowData(q3)[c(1, 2, 2, 2, 3, 4, 5), "col1"])
    expect_equal(mo$target_col1, t3[c(2,2,3,4, NA, NA, 5), "col1"])
})

test_that(".dollar works with unordered @matches", {
    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    ## Both query and target idx are unsorted.
    mo2 <- mo
    mo2@matches <- mo@matches[c(2, 1, 3, 5, 4), ]
    ## query
    expect_equal(mo$col1, mo2$col1)
    expect_equal(mo$col2, mo2$col2)
    ## matches
    expect_equal(mo$score, mo2$score)
    ## target
    expect_equal(mo$target_col1, mo2$target_col1)
    expect_equal(mo$target_col2, mo2$target_col2)

    ## query idx is sorted, target idx not.
    mo2 <- mo
    mo2@target <- mo@target[c(4, 5, 1, 2, 3, 6), ]
    mo2@matches$target_idx <- c(4L, 4L, 5L, 1L, 2L)

    ## matches
    expect_equal(mo$score, mo2$score)
    ## target
    expect_equal(mo$target_col1, mo2$target_col1)
    expect_equal(mo$target_col2, mo2$target_col2)
})

test_that("colnames,Matched works", {
    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    expect_equal(colnames(mo), c(colnames(q1),
                                 paste0("target_", colnames(t1)), "score"))
    mo <- Matched(
        q2, t2, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    expect_equal(colnames(mo), c("query", "target", "score"))

    mo <- Matched(
        q1, t2, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    expect_equal(colnames(mo), c(colnames(q1), "target", "score"))

    mo <- Matched(
        q2, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    expect_equal(colnames(mo), c("query",
                                 paste0("target_", colnames(t1)), "score"))

    mo <- Matched(
        q3, t3, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    expect_equal(colnames(mo), c(colnames(rowData(q3)),
                                 paste0("target_", colnames(t3)), "score"))
})

test_that("matchedData,Matched works", {

    mo <- Matched()
    res <- matchedData(mo)
    expect_true(is(res, "DataFrame"))
    expect_true(nrow(res) == 0)
    expect_equal(colnames(res), colnames(mo))

    #### query and target data.frames
    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    res <- matchedData(mo)
    expect_equal(res$col2, mo$col2)
    expect_equal(res$target_col2, mo$target_col2)
    expect_equal(res$score, mo$score)

    expect_error(matchedData(mo, columns = "other"), "other not available")

    ## Only query object variables
    res <- matchedData(mo, columns = c("col1", "col2"))
    expect_equal(colnames(res), c("col1", "col2"))
    expect_equal(res$col1, mo$col1)

    ## Only target object variables
    res <- matchedData(mo, columns = c("target_col1", "target_col2"))
    expect_equal(colnames(res), c("target_col1", "target_col2"))
    expect_equal(res$target_col1, mo$target_col1)

    ## Only matches
    res <- matchedData(mo, columns = c("score"))
    expect_equal(colnames(res), c("score"))
    expect_equal(res$score, mo$score)

    ## A Matched with no matching target elements (query and target data.frames)
    mo <- Matched(q1, t1, matches = data.frame(query_idx = integer(),
                                               target_idx = integer(),
                                               score = numeric()))
    res <- matchedData(mo)
    expect_equal(res$col2, mo$col2)
    expect_equal(res$target_col2, mo$target_col2)
    expect_equal(res$score, mo$score)

    ## Only query object variables
    res <- matchedData(mo, columns = c("col1", "col2"))
    expect_equal(colnames(res), c("col1", "col2"))
    expect_equal(res$col1, mo$col1)

    ## Only target object variables
    res <- matchedData(mo, columns = c("target_col1", "target_col2"))
    expect_equal(colnames(res), c("target_col1", "target_col2"))
    expect_equal(res$target_col1, mo$target_col1)

    ## Only matches
    res <- matchedData(mo, columns = c("score"))
    expect_equal(colnames(res), c("score"))
    expect_equal(res$score, mo$score)

    #### query data.frame and target list
    mo <- Matched(
        q1, t2, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    res <- matchedData(mo)
    expect_equal(res$col2, mo$col2)
    expect_equal(res$target, I(mo$target))
    expect_equal(res$score, mo$score)

    ## Only query object variables
    res <- matchedData(mo, columns = c("col1", "col2"))
    expect_equal(colnames(res), c("col1", "col2"))
    expect_equal(res$col1, mo$col1)

    ## Only target object variables
    res <- matchedData(mo, columns = c("target"))
    expect_equal(colnames(res), c("target"))
    expect_equal(res$target, mo$target)

    ## Only matches
    res <- matchedData(mo, columns = c("score"))
    expect_equal(colnames(res), c("score"))
    expect_equal(res$score, mo$score)

    #### A Matched with no matching target elements: query list and target
    #### data.frame
    mo <- Matched(q2, t1, matches = data.frame(query_idx = integer(),
                                               target_idx = integer(),
                                               score = numeric()))
    res <- matchedData(mo)
    expect_equal(res$query, I(mo$query))
    expect_equal(res$target_col1, mo$target_col1)
    expect_equal(res$score, mo$score)

    ## Only query object variables
    res <- matchedData(mo, columns = c("query"))
    expect_equal(colnames(res), c("query"))
    expect_equal(res$query, mo$query)
    expect_equal(res$query, q2)

    ## Only target object variables
    res <- matchedData(mo, columns = c("target_col1", "target_col2"))
    expect_equal(colnames(res), c("target_col1", "target_col2"))
    expect_equal(res$target_col1, mo$target_col1)

    ## Only matches
    res <- matchedData(mo, columns = c("score"))
    expect_equal(colnames(res), c("score"))
    expect_equal(res$score, mo$score)

    #### query SummarizedExoeriment and target data.frame
    mo <- Matched(
        q3, t3, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    res <- matchedData(mo)
    expect_equal(res$col2, mo$col2)
    expect_equal(res$target_col2, mo$target_col2)
    expect_equal(res$score, mo$score)

    expect_error(matchedData(mo, columns = "other"), "other not available")

    ## Only query object variables
    res <- matchedData(mo, columns = c("col1", "col2"))
    expect_equal(colnames(res), c("col1", "col2"))
    expect_equal(res$col1, mo$col1)

    ## Only target object variables
    res <- matchedData(mo, columns = c("target_col1", "target_col2"))
    expect_equal(colnames(res), c("target_col1", "target_col2"))
    expect_equal(res$target_col1, mo$target_col1)

    ## Only matches
    res <- matchedData(mo, columns = c("score"))
    expect_equal(colnames(res), c("score"))
    expect_equal(res$score, mo$score)

    ## A MatchedSummarizedExperiment with no matching target elements
    mo <- Matched(q3, t3, matches = data.frame(query_idx = integer(),
                                               target_idx = integer(),
                                               score = numeric()))
    res <- matchedData(mo)
    expect_equal(res$col2, mo$col2)
    expect_equal(res$target_col2, mo$target_col2)
    expect_equal(res$score, mo$score)

    ## Only query object variables
    res <- matchedData(mo, columns = c("col1", "col2"))
    expect_equal(colnames(res), c("col1", "col2"))
    expect_equal(res$col1, mo$col1)

    ## Only target object variables
    res <- matchedData(mo, columns = c("target_col1", "target_col2"))
    expect_equal(colnames(res), c("target_col1", "target_col2"))
    expect_equal(res$target_col1, mo$target_col1)

    ## Only matches
    res <- matchedData(mo, columns = c("score"))
    expect_equal(colnames(res), c("score"))
    expect_equal(res$score, mo$score)
})

test_that("pruneTarget,Matched works", {
    mo <- Matched()
    res <- pruneTarget(mo)
    expect_true(is(res, "Matched"))
    expect_equal(res, mo)

    #### query and target data.frames
    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    res <- pruneTarget(mo)
    expect_equal(matchedData(res), matchedData(mo))
    expect_true(nrow(res@target) < nrow(mo@target))

    mo <- Matched(q1, t1, matches = data.frame(query_idx = integer(),
                                               target_idx = integer(),
                                               score = numeric()))
    res <- pruneTarget(mo)
    expect_equal(matchedData(res), matchedData(mo))
    expect_true(nrow(res@target) < nrow(mo@target))

    #### query data.frame and target list
    mo <- Matched(
        q1, t2, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    res <- pruneTarget(mo)
    expect_equal(matchedData(res), matchedData(mo))
    expect_true(length(res@target) < length(mo@target))

    mo <- Matched(q1, t2, matches = data.frame(query_idx = integer(),
                                               target_idx = integer(),
                                               score = numeric()))
    res <- pruneTarget(mo)
    expect_equal(matchedData(res), matchedData(mo))
    expect_true(length(res@target) < length(mo@target))

    #### query SummarizedExperiment and target data.frame
    mo <- Matched(
        q3, t3, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    res <- pruneTarget(mo)
    expect_equal(matchedData(res), matchedData(mo))
    expect_true(nrow(res@target) < nrow(mo@target))

    mo <- Matched(q3, t3, matches = data.frame(query_idx = integer(),
                                               target_idx = integer(),
                                               score = numeric()))
    res <- pruneTarget(mo)
    expect_equal(matchedData(res), matchedData(mo))
    expect_true(nrow(res@target) < nrow(mo@target))
})

test_that("filterMatches,Matched works", {
    #### query and target data.frames
    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    ## out of bounds indexes
    expect_error(filterMatches(mo, index = c(1, 10)), "out-of-bounds")
    ## no index : every match is removed
    idxs <- integer(0)
    mosub <- filterMatches(mo, index = idxs)
    expect_equal(mosub@matches, mo@matches[idxs, ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
    expect_s4_class(mosub@metadata[[length(mosub@metadata)]],
                    "SelectMatchesParam")
    ## in range indexes
    idxs <- c(1, 3, 5)
    mosub <- filterMatches(mo, index = idxs)
    expect_equal(mosub@matches, mo@matches[idxs, ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
    mosub <- filterMatches(mo, index = idxs, keep = FALSE)
    expect_equal(mosub@matches, mo@matches[-idxs, ])
    ## keep matches based on query and target input values
    queryValue <- c(q1[mo@matches[idxs, "query_idx"], "col1"], -1, - 2)
    targetValue <- c(t1[mo@matches[idxs, "target_idx"], "col2"], -2, -1)
    mosub <- filterMatches(mo, queryValue = queryValue,
                         targetValue = targetValue,
                         queryColname = "col1", targetColname = "col2")
    expect_equal(mosub@matches, mo@matches[idxs, ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
    mosub <- filterMatches(mo, queryValue = queryValue,
                           targetValue = targetValue, queryColname = "col1",
                           targetColname = "col2", keep = FALSE)
    expect_equal(mosub@matches, mo@matches[-idxs, ])
    ## no matches corresponding to the input values
    mosub <- filterMatches(mo, queryValue = queryValue + 100,
                         targetValue = targetValue + 100,
                         queryColname = "col1", targetColname = "col2")
    expect_equal(mosub@matches, data.frame(query_idx = integer(),
                                           target_idx = integer(),
                                           score = numeric()))

    #### query and target vectors (multiple matches corresponding to a given
    #### couple of query and target input values)
    mo <- Matched(
        c(1, 1, 2, 3, 4), c("A", "A", "A", "C", "D"),
        matches = data.frame(query_idx = c(1L, 2L, 4L, 5L),
                             target_idx = c(1L, 2L, 4L, 5L),
                             score = seq(0.5, 0.8, by = 0.1)))
    queryValue <- c(1, 3)
    targetValue <- c("A", "C")
    mosub <- filterMatches(mo, queryValue = queryValue,
                         targetValue = targetValue)
    expect_equal(mosub@matches, mo@matches[c(1, 2, 3), ])

    #### query SummarizedExperiment and target data.frame
    mo <- Matched(
        q3, t3, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    ## out of bounds indexes
    expect_error(filterMatches(mo, index = c(1, 10)), "out-of-bounds")
    ## no index : every match is removed
    idxs <- integer(0)
    mosub <- filterMatches(mo, index = idxs)
    expect_equal(mosub@matches, mo@matches[idxs, ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
    ## in range indexes
    idxs <- c(1, 3, 5)
    mosub <- filterMatches(mo, index = idxs)
    expect_equal(mosub@matches, mo@matches[idxs, ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
    ## keep matches based on query and target input values
    queryValue <- c(rowData(q3)[mo@matches[idxs, "query_idx"], "col1"], -1, - 2)
    targetValue <- c(t3[mo@matches[idxs, "target_idx"], "col2"], -2, -1)
    mosub <- filterMatches(mo, queryValue = queryValue,
                           targetValue = targetValue,
                           queryColname = "col1", targetColname = "col2")
    expect_equal(mosub@matches, mo@matches[idxs, ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
    ## no matches corresponding to the input values
    mosub <- filterMatches(mo, queryValue = queryValue + 100,
                           targetValue = targetValue + 100,
                           queryColname = "col1", targetColname = "col2")
    expect_equal(mosub@matches, data.frame(query_idx = integer(),
                                           target_idx = integer(),
                                           score = numeric()))

})

test_that("filterMatches,Matched,SelectMatchesParam works", {
    #### query and target data.frames
    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    ## out of bounds indexes
    expect_error(filterMatches(mo, SelectMatchesParam(index = c(1L, 10L))),
                 "out-of-bounds")
    ## no index : every match is removed
    idxs <- integer(0)
    mosub <- filterMatches(mo, SelectMatchesParam(index = idxs))
    expect_equal(mosub@matches, mo@matches[idxs, ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
    expect_s4_class(mosub@metadata[[length(mosub@metadata)]],
                    "SelectMatchesParam")
    ## in range indexes
    idxs <- c(1L, 3L, 5L)
    mosub <- filterMatches(mo, SelectMatchesParam(index = idxs))
    expect_equal(mosub@matches, mo@matches[idxs, ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
    mosub_old <- filterMatches(mo, index = idxs)
    expect_equal(mosub@matches, mosub_old@matches)
    mosub <- filterMatches(mo, SelectMatchesParam(index = idxs, keep = FALSE))
    expect_equal(mosub@matches, mo@matches[-idxs, ])
    ## keep matches based on query and target input values
    queryValue <- c(q1[mo@matches[idxs, "query_idx"], "col1"], -1, - 2)
    targetValue <- c(t1[mo@matches[idxs, "target_idx"], "col2"], -2, -1)
    mosub <- filterMatches(mo, SelectMatchesParam(queryValue = queryValue,
                                                  targetValue = targetValue,
                                                  queryColname = "col1",
                                                  targetColname = "col2"))
    expect_equal(mosub@matches, mo@matches[idxs, ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
    mosub <- filterMatches(mo, SelectMatchesParam(queryValue = queryValue,
                                                  targetValue = targetValue,
                                                  queryColname = "col1",
                                                  targetColname = "col2",
                                                  keep = FALSE))
    expect_equal(mosub@matches, mo@matches[-idxs, ])
    ## no matches corresponding to the input values
    mosub <- filterMatches(mo,
                           SelectMatchesParam(queryValue = queryValue + 100,
                                              targetValue = targetValue + 100,
                                              queryColname = "col1",
                                              targetColname = "col2"))
    expect_equal(mosub@matches, data.frame(query_idx = integer(),
                                           target_idx = integer(),
                                           score = numeric()))

    #### query and target vectors (multiple matches corresponding to a given
    #### couple of query and target input values)
    mo <- Matched(
        c(1, 1, 2, 3, 4), c("A", "A", "A", "C", "D"),
        matches = data.frame(query_idx = c(1L, 2L, 4L, 5L),
                             target_idx = c(1L, 2L, 4L, 5L),
                             score = seq(0.5, 0.8, by = 0.1)))
    queryValue <- c(1, 3)
    targetValue <- c("A", "C")
    mosub <- filterMatches(mo, SelectMatchesParam(queryValue = queryValue,
                                                  targetValue = targetValue))
    expect_equal(mosub@matches, mo@matches[c(1, 2, 3), ])

    #### query SummarizedExperiment and target data.frame
    mo <- Matched(
        q3, t3, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    ## out of bounds indexes
    expect_error(filterMatches(mo, SelectMatchesParam(index = c(1L, 10L))),
                 "out-of-bounds")
    ## no index : every match is removed
    idxs <- integer(0)
    mosub <- filterMatches(mo, SelectMatchesParam(index = idxs))
    expect_equal(mosub@matches, mo@matches[idxs, ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
    ## in range indexes
    idxs <- c(1L, 3L, 5L)
    mosub <- filterMatches(mo, SelectMatchesParam(index = idxs))
    expect_equal(mosub@matches, mo@matches[idxs, ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
    ## keep matches based on query and target input values
    queryValue <- c(rowData(q3)[mo@matches[idxs, "query_idx"], "col1"], -1, - 2)
    targetValue <- c(t3[mo@matches[idxs, "target_idx"], "col2"], -2, -1)
    mosub <- filterMatches(mo, SelectMatchesParam(queryValue = queryValue,
                                                  targetValue = targetValue,
                                                  queryColname = "col1",
                                                  targetColname = "col2"))
    expect_equal(mosub@matches, mo@matches[idxs, ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
    ## no matches corresponding to the input values
    mosub <- filterMatches(mo,
                           SelectMatchesParam(queryValue = queryValue + 100,
                                              targetValue = targetValue + 100,
                                              queryColname = "col1",
                                              targetColname = "col2"))
    expect_equal(mosub@matches, data.frame(query_idx = integer(),
                                           target_idx = integer(),
                                           score = numeric()))
})

test_that("filterMatches,Matched,TopRankedMatchesParam works", {
    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = c(4, 4, 1, 3, 1)))
    mosub <- filterMatches(mo, TopRankedMatchesParam())
    expect_equal(mosub@matches, mo@matches[c(1L, 3L, 5L), ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
    expect_s4_class(mosub@metadata[[length(mosub@metadata)]],
                    "TopRankedMatchesParam")

    mosub <- filterMatches(mo, TopRankedMatchesParam(n = 2L))
    expect_equal(mosub@matches, mo@matches[c(1L, 3L, 4L, 5L), ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))

    mosub <- filterMatches(mo, TopRankedMatchesParam(n = 10L))
    expect_equal(mosub@matches, mo@matches)
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))

    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = c(4, 4, 1, 3, 1),
                                     score_rt = c(1, -1, -1, -2, 1)))

    mosub <- filterMatches(mo, TopRankedMatchesParam())
    expect_equal(mosub@matches, mo@matches[c(1L, 3L, 5L), ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))

    mosub <- filterMatches(mo, TopRankedMatchesParam(n = 2L))
    expect_equal(mosub@matches, mo@matches[c(1L, 2L, 3L, 5L), ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))

    mosub <- filterMatches(mo, TopRankedMatchesParam(n = 10L))
    expect_equal(mosub@matches, mo@matches)
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))

    q1 <- data.frame(col1 = 1:5, col2 = 6:10)
    t2 <- 11:16
    mo <- Matched(q1, t2,
                  matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                       target_idx = c(2L, 2L, 3L, 4L, 5L),
                                       score = seq(0.5, 0.9, by = 0.1)))
    res <- filterMatches(mo, TopRankedMatchesParam(n = 1L))
    expect_equal(res@matches$query_idx, c(1, 2, 5))
    expect_equal(res@matches$target_idx, c(2, 2, 5))
    expect_equal(res@matches$score, c(0.5, 0.6, 0.9))

    res <- filterMatches(mo, TopRankedMatchesParam(n = 1L, decreasing = TRUE))
    expect_equal(res@matches$query_idx, c(1, 2, 5))
    expect_equal(res@matches$target_idx, c(2, 4, 5))
    expect_equal(res@matches$score, c(0.5, 0.8, 0.9))
})


test_that("filterMatches,Matched,ScoreThresholdParam works", {
    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = c(4, 4, 1, 3, 1)))

    mosub <- filterMatches(mo, ScoreThresholdParam())
    expect_equal(mosub@matches, mo@matches[integer(0), ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
    expect_s4_class(mosub@metadata[[length(mosub@metadata)]],
                    "ScoreThresholdParam")

    mosub <- filterMatches(mo, ScoreThresholdParam(threshold = 2.5))
    expect_equal(mosub@matches, mo@matches[c(3L, 5L), ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))

    mosub <- filterMatches(mo, ScoreThresholdParam(threshold = 2.5,
                                                   above = TRUE))
    expect_equal(mosub@matches, mo@matches[c(1L, 2L, 4L), ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))

    expect_error(filterMatches(mo, ScoreThresholdParam(threshold = 2.5,
                                                       column = "notpresent")),
                 "variable not present")

    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = c(4, 4, 1, 3, 1),
                                     score_rt = c(1, -1, -1, -2, 1)))

    mosub <- filterMatches(mo, ScoreThresholdParam(threshold = 0,
                                                   column = "score_rt"))
    expect_equal(mosub@matches, mo@matches[c(2L, 3L, 4L), ])
    expect_equal(query(mosub), query(mo))
    expect_equal(target(mosub), target(mo))
})

test_that("addMatches,Matched works", {
    #### query and target data.frames
    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    moadd <- addMatches(mo, queryValue = q1[c(3, 5), "col1"],
                        targetValue = t1[c(1, 6), "col2"],
                        queryColname = "col1", targetColname = "col2",
                        score = data.frame(score = c(1, 1.1)), isIndex = FALSE)
    expect_equal(moadd@matches$query_idx, c(mo@matches$query_idx, 3, 5))
    expect_equal(moadd@matches$target_idx, c(mo@matches$target_idx, 1, 6))
    expect_equal(moadd@matches$score, c(mo@matches$score, 1, 1.1))
    expect_equal(query(moadd), query(mo))
    expect_equal(target(moadd), target(mo))

    moadd <- addMatches(mo, queryValue = c(3L, 5L),
                        targetValue = c(1L, 6L),
                        queryColname = "col1", targetColname = "col2",
                        score = data.frame(score = c(1, 1.1)), isIndex = TRUE)
    expect_equal(moadd@matches$query_idx, c(mo@matches$query_idx, 3, 5))
    expect_equal(moadd@matches$target_idx, c(mo@matches$target_idx, 1, 6))
    expect_equal(moadd@matches$score, c(mo@matches$score, 1, 1.1))
    expect_equal(query(moadd), query(mo))
    expect_equal(target(moadd), target(mo))

    #### query vector and target data.frame
    mo <- Matched(q1[, "col1"], t1,
                  matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                       target_idx = c(2L, 2L, 3L, 4L, 5L),
                                       score = seq(0.5, 0.9, by = 0.1)))
    moadd <- addMatches(mo, queryValue = c(q1[c(3, 5), "col1"], 50),
                        targetValue = c(t1[c(1, 6), "col2"], 50),
                        targetColname = "col2",
                        score = data.frame(score = c(1, 1.1, 2)))
    expect_equal(moadd@matches$query_idx, c(mo@matches$query_idx, 3, 5))
    expect_equal(moadd@matches$target_idx, c(mo@matches$target_idx, 1, 6))
    expect_equal(moadd@matches$score, c(mo@matches$score, 1, 1.1))
    expect_equal(query(moadd), query(mo))
    expect_equal(target(moadd), target(mo))

    moadd <- addMatches(mo, queryValue = c(3L, 5L),
                        targetValue = c(1L, 6L), targetColname = "col2",
                        score = data.frame(score = c(1, 1.1)), isIndex = TRUE)
    expect_equal(moadd@matches$query_idx, c(mo@matches$query_idx, 3, 5))
    expect_equal(moadd@matches$target_idx, c(mo@matches$target_idx, 1, 6))
    expect_equal(moadd@matches$score, c(mo@matches$score, 1, 1.1))
    expect_equal(query(moadd), query(mo))
    expect_equal(target(moadd), target(mo))

    #### query SummarizedExperiment and target data.frame
    mo <- Matched(
        query = q3, target = t3,
        matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                             target_idx = c(2L, 2L, 3L, 4L, 5L),
                             score = seq(0.5, 0.9, by = 0.1)))

    moadd <- addMatches(mo, queryValue = 2L, targetValue = 1L, isIndex = TRUE,
                        score = data.frame(score = 100))
    expect_true(matches(moadd)$query_idx[6] == 2)
    expect_true(matches(moadd)$target_idx[6] == 1)
    expect_true(matches(moadd)$score[6] == 100)
})

test_that(".extract_elements works", {
    a <- data.frame(a = 1:5)
    expect_equal(.extract_elements(a, 3)[, 1], 3)
    expect_equal(.extract_elements(a, c(3, 6))[, 1], c(3, NA))
    expect_equal(.extract_elements(a, c(NA, 1))[, 1], c(NA, 1))

    a <- 1:5
    expect_equal(.extract_elements(a, 3), 3)
    expect_equal(.extract_elements(a, c(3, 6)), c(3, NA))
    expect_equal(.extract_elements(a, c(NA, 1)), c(NA, 1))

    a <- list(1, 2, 3, 4, 5)
    expect_equal(.extract_elements(a, 3), list(3))
    expect_equal(.extract_elements(a, c(3, 6)), list(3, NULL))
    expect_equal(.extract_elements(a, c(NA, 1)), list(NA, 1))
})

test_that(".cnt works", {
    t <- data.frame(a = 1:5, b = 2:6)
    expect_equal(.cnt(t), c("target_a", "target_b"))

    t <- 1:5
    expect_equal(.cnt(t), "target")

    t <- array(data = 1, dim = c(1, 1, 1))
    expect_error(.cnt(t), "unsupported")
})

test_that(".validate_assay works", {
    expect_identical(.validate_assay(q4, c("a1", "a2")),
                     "`queryAssay` must be `character(1)`")
    expect_identical(.validate_assay(q4, "a3", "target"),
                     "No assay in `target` with name \"a3\"")
    expect_null(.validate_assay(q4, "a1"))
})

test_that(".objectToMatch works", {
    expect_identical(.objectToMatch(q1), q1)
    expect_identical(.objectToMatch(q2), q2)
    expect_identical(.objectToMatch(q3), rowData(q3))
    expect_error(.objectToMatch(q4), "has to be provided")
    expect_error(.objectToMatch(q4, c("a1", "a2")), "must be `character")
    expect_error(.objectToMatch(q4, c("a3")), "with name")
    expect_identical(.objectToMatch(q4, c("a1")), rowData(q4[["a1"]]))
    expect_identical(.objectToMatch(q1, "assayname", "col2"), q1[, "col2"])
    expect_identical(.objectToMatch(q3, colnames = "col2"),
                     rowData(q3)[, "col2"])
    expect_identical(.objectToMatch(q4, "a1", "col2"),
                     rowData(q4[["a1"]])[, "col2"])
    expect_error(.objectToMatch(q1, "a1", c("col1", "col3")), "Missing column")
    expect_error(.objectToMatch(q3, "a1", c("col1", "col3")), "Missing column")
    expect_error(.objectToMatch(q4, "a1", c("col1", "col3")), "Missing column")

    expect_error(.objectToMatch(pest_ms2, colnames = c("other_col")),
                 "Missing spectra")
    expect_identical(.objectToMatch(pest_ms2), spectraData(pest_ms2))
    expect_identical(.objectToMatch(pest_ms2, colnames = c("rtime", "msLevel")),
                     spectraData(pest_ms2, c("rtime", "msLevel")))

})

test_that(".subset_qt works", {
    i <- c(1, 3)
    expect_identical(.subset_qt(q1, i = i), q1[i, ])
    expect_identical(.subset_qt(q2, i = i), q2[i])
    expect_identical(.subset_qt(q3, i = i), (q3)[i, ])
    res <- .subset_qt(q4, "a1", i = i)
    expect_is(res, "QFeatures")
    expect_equal(res[["a1"]], q4[["a1"]][i, ])

    expect_error(.subset_qt(q4, assay = "no_way", i = i),
                 "Invalid assay name")
})

test_that("SelectMatchesParam works", {
    res <- SelectMatchesParam()
    expect_true(is(res, "SelectMatchesParam"))

    expect_error(SelectMatchesParam(queryValue = "A", targetValue = 1:2),
                 "must have the same length")
    expect_error(SelectMatchesParam(queryColname = c("A", "b")),
                 "cannot be of length greater than 1")
    expect_error(SelectMatchesParam(targetColname = c("A", "b")),
                 "cannot be of length greater than 1")
    expect_error(SelectMatchesParam(index = c(1L, -2L)),
                 "must contain positive integers")
    expect_error(SelectMatchesParam(keep = rep(TRUE, 2)),
                 "must be a logical of length 1")
})

test_that("TopRankedMatchesParam works", {
    res <- TopRankedMatchesParam()
    expect_true(is(res, "TopRankedMatchesParam"))

    expect_error(TopRankedMatchesParam(n = c(2L, 3L)), "length 1")
    expect_error(TopRankedMatchesParam(n = -4L), "positive integer")
})

test_that("endoapply,ANY works", {
    res <- endoapply(1:3, function(z) z)
    expect_equal(res, 1:3)
})

test_that("endoapply,Matched works", {
    mo <- Matched(
        query = data.frame(val = 1:5, name = letters[1:5]),
        target = data.frame(val = 11:16, name = LETTERS[1:6]),
        matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                             target_idx = c(2L, 2L, 3L, 4L, 5L),
                             score = seq(0.5, 0.9, by = 0.1)))

    fun0 <- function(x, threshold) {
        x@matches <- x@matches[x@matches$score < threshold, ]
        x
    }

    res <- endoapply(mo, fun0, threshold = 0.75)
    expect_true(is(res, "Matched"))
    expect_identical(query(res), query(mo))
    expect_identical(target(res), target(mo))
    expect_true(all(res$score < 0.75, na.rm = TRUE))

    fun1 <- function(x, cmpds_subs) {
        x@matches <- x@matches[x@target[x@matches$target_idx, "name"]
                               %in% cmpds_subs, ]
        x
    }

    res <- endoapply(mo, fun1, cmpds_subs = c("A", "B", "D"))
    expect_identical(query(res), query(mo))
    expect_identical(target(res), target(mo))
    expect_true(all(res$target_name %in% c("A", "B", "D", NA)))

    fun2 <- function(x) {
        x@matches <- x@matches[which.min(x@matches$score), ]
        x
    }

    res <- endoapply(mo, fun2)
    expect_identical(query(res), query(mo))
    expect_identical(target(res), target(mo))
    expect_identical(res@matches$query_idx, mo@matches[c(1, 2, 5), "query_idx"])
    expect_identical(res@matches$target_idx,
                     mo@matches[c(1, 2, 5), "target_idx"])
    expect_identical(res@matches$score, mo@matches[c(1, 2, 5), "score"])
})

test_that("lapply,Matched works", {
    mo <- Matched(query = q1, target = t1,
                  matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                       target_idx = c(2L, 2L, 3L, 4L, 5L),
                                       score = seq(0.5, 0.9, by = 0.1)))

    fun0 <- function(x, threshold) {
        x@matches[x@matches$score < threshold, "score"]
    }

    res <- lapply(mo, fun0, threshold = 0.75)
    expect_true(is(res, "list"))
    expect_identical(length(res), length(mo))
    expect_true(all(unlist(res) < 0.75))
})

test_that("scoreVariables,Matched works", {

    expect_identical(scoreVariables(Matched()), "score")

    mo <- Matched(query = 1:5, target = 1:6,
                  matches = data.frame(query_idx = 1:5,
                                       target_idx = 1:5,
                                       score = 1:5,
                                       Score2 = 1:5,
                                       thirdscore = 1:5,
                                       other = letters[1:5]))
    expect_identical(scoreVariables(mo), c("score", "Score2", "thirdscore"))
})

test_that("targetIndex works", {
    a <- Matched()
    expect_identical(targetIndex(a), integer())
    a@matches <- data.frame(query_idx = c(1L, 1L, 2L),
                            target_idx = c(2L, 3L, 2L),
                            score = 1:3)
    expect_identical(targetIndex(a), c(2L, 3L, 2L))
    expect_error(targetIndex("t"), "'Matched'")
})

test_that("queryIndex works", {
    a <- Matched()
    expect_identical(queryIndex(a), integer())
    a@matches <- data.frame(query_idx = c(1L, 1L, 2L),
                            target_idx = c(2L, 3L, 2L),
                            score = 1:3)
    expect_identical(queryIndex(a), c(1L, 1L, 2L))
    expect_error(queryIndex(4), "'Matched'")
})

test_that("SingleMatchParam works", {
    res <- SingleMatchParam()
    expect_s4_class(res, "SingleMatchParam")
    expect_true(validObject(res))
    expect_error(SingleMatchParam(duplicates = "other"), "should be")
    res <- SingleMatchParam("closest", column = "other", decreasing = FALSE)
    expect_equal(res@duplicates, "closest")
    expect_equal(res@column, "other")
    expect_equal(res@decreasing, FALSE)
})

test_that("filterMatches,Matched,SingleMatchParam works", {
    a <- Matched()
    p <- SingleMatchParam(column = "ops")
    expect_error(filterMatches(a, p), "not found")

    a <- Matched(
        query = q1, target = t1,
        matches = data.frame(query_idx = integer(),
                             target_idx = integer(),
                             score = numeric()))
    res <- filterMatches(a, SingleMatchParam())
    expect_equal(res@query, a@query)
    expect_equal(res@target, a@target)
    expect_equal(res@matches, a@matches)

    ## target is data.frame
    a <- Matched(
        query = q1, target = t1,
        matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                             target_idx = c(2L, 2L, 3L, 4L, 5L),
                             score = seq(0.5, 0.9, by = 0.1)))
    p@duplicates <- "sum"
    p@column <- "score"
    expect_error(filterMatches(a, p), "has to be one")
    p <- SingleMatchParam()
    ## remove
    res <- filterMatches(a, p)
    expect_equal(anyDuplicated(res@matches$query_idx), 0L)
    expect_equal(res@matches$target_idx, c(2L, 5L))
    ## closest
    p <- SingleMatchParam(duplicates = "closest")
    res <- filterMatches(a, p)
    expect_equal(anyDuplicated(res@matches$query_idx), 0L)
    expect_equal(res@matches$target_idx, c(2L, 2L, 5L))
    ## top_ranked
    p <- SingleMatchParam(duplicates = "top_ranked", column = "target_col2",
                          decreasing = TRUE)
    res <- filterMatches(a, p)
    expect_equal(anyDuplicated(res@matches$query_idx), 0L)
    expect_equal(res@matches$target_idx, c(2L, 4L, 5L))

    p <- SingleMatchParam(duplicates = "top_ranked", column = "score",
                          decreasing = TRUE)
    res <- filterMatches(a, p)
    expect_equal(anyDuplicated(res@matches$query_idx), 0L)
    expect_equal(res@matches$target_idx, c(2L, 4L, 5L))

    ## target is vector
    a <- Matched(
        query = q1, target = c(100, 200, 300, 800),
        matches = data.frame(query_idx = c(1L, 1L, 1L, 3L, 3L, 3L, 3L),
                             target_idx = c(3L, 4L, 1L, 2L, 3L, 1L, 4L),
                             score = c(1, 2, 4, 3, 1, 2, 7))
    )
    ## remove
    p <- SingleMatchParam()
    res <- filterMatches(a, p)
    expect_equal(anyDuplicated(res@matches$query_idx), 0L)
    expect_equal(res@matches$target_idx, integer())
    ## closest
    p <- SingleMatchParam(duplicates = "closest")
    res <- filterMatches(a, p)
    expect_equal(anyDuplicated(res@matches$query_idx), 0L)
    expect_equal(res@matches$target_idx, c(3L, 3L))
    ## top_ranked
    p <- SingleMatchParam(duplicates = "top_ranked", column = "target",
                          decreasing = FALSE)
    res <- filterMatches(a, p)
    expect_equal(anyDuplicated(res@matches$query_idx), 0L)
    expect_equal(res@matches$target_idx, c(1L, 1L))

    ## target is SummarizedExperiment
    rowData(q3)$new_col <- seq_len(nrow(q1))
    a <- Matched(query = q1, target = q3,
                 matches = data.frame(
                     query_idx = c(1L, 1L, 1L, 2L, 2L, 3L, 3L, 5L),
                     target_idx = c(2L, 4L, 5L, 1L, 2L, 3L, 4L, 5L),
                     score = c(4, 1, 2, 3, 4, 3, 1, 9)))
    ## remove
    p <- SingleMatchParam()
    res <- filterMatches(a, p)
    expect_equal(anyDuplicated(res@matches$query_idx), 0L)
    expect_equal(res@matches$query_idx, 5L)
    expect_equal(res@matches$target_idx, 5L)

    ## closest
    p <- SingleMatchParam(duplicates = "closest")
    res <- filterMatches(a, p)
    expect_equal(anyDuplicated(res@matches$query_idx), 0L)
    expect_equal(res@matches$query_idx, c(1L, 2L, 3L, 5L))
    expect_equal(res@matches$target_idx, c(4L, 1L, 4L, 5L))

    ## top_ranked
    p <- SingleMatchParam(duplicates = "top_ranked", column = "target_new_col",
                          decreasing = TRUE)
    res <- filterMatches(a, p)
    expect_equal(anyDuplicated(res@matches$query_idx), 0L)
    expect_equal(res@matches$query_idx, c(1L, 2L, 3L, 5L))
    expect_equal(res@matches$target_idx, c(5L, 2L, 4L, 5L))
})

test_that(".findMatchesIdxs works", {
    mo <- Matched(
        q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                                     score = seq(0.5, 0.9, by = 0.1)))
    expect_error(.findMatchesIdxs(mo@query, mo@target, mo@matches,
                                  queryValue = 3, queryColname = "col1",
                                  targetValue = c(1, 2)),
                 "have the same length")
    expect_error(.findMatchesIdxs(mo@query, mo@target, mo@matches,
                                  queryValue = 1, targetValue = 2),
                 "must be provided")
    expect_error(.findMatchesIdxs(mo@query, mo@target, mo@matches,
                                  queryValue = 1, targetValue = 2,
                                  queryColname = "some"),
                 "not a column")
    expect_error(.findMatchesIdxs(mo@query, mo@target, mo@matches,
                                  queryValue = 1, targetValue = 2,
                                  queryColname = "col1"),
                 "must be provided")
    expect_error(.findMatchesIdxs(mo@query, mo@target, mo@matches,
                                  queryValue = 1, targetValue = 2,
                                  queryColname = "col1",
                                  targetColname = "some"),
                 "not a column")

})

test_that(".addMatches works", {
    ## Check errors. Rest is tested above in addMatches()
    m1 <- data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                     target_idx = c(2L, 2L, 3L, 4L, 5L),
                     score = seq(0.5, 0.9, by = 0.1))
    expect_error(MetaboAnnotation:::.addMatches(q1, t1, m1, score = c(3, 6, 10)),
                 "must have the same length")
    expect_error(MetaboAnnotation:::.addMatches(q1, t1, m1, score = "a"),
                 "or a numeric")
    expect_error(.addMatches(q1, t1, m1, score = c(3, 6, 10), isIndex = TRUE,
                             targetValue = c("a", "b", "c"), queryValue = 1:3),
                 "must be integer vectors")
    expect_error(.addMatches(q1, t1, m1, score = c(3, 6, 10), isIndex = TRUE,
                             targetValue = 1:3, queryValue = c(10L, 2L, 1L)),
                 "'queryValue' are out-of-bounds")
    expect_error(.addMatches(q1, t1, m1, score = c(3, 6, 10), isIndex = TRUE,
                             targetValue = c(3L, 29L, 1L), queryValue = 1:3),
                 "'targetValue' are out-of-bounds")
    expect_error(.addMatches(q1, t1, m1, score = c(3, 6, 10), isIndex = FALSE,
                             queryColname = "col4", queryValue = 1:3,
                             targetValue = 1:3),
                 "column of 'query'")
    expect_error(.addMatches(q1, t1, m1, score = c(3, 6, 10), isIndex = FALSE,
                             queryColname = "col1", targetColname = "col13",
                             queryValue = 1:3, targetValue = 1:3),
                 "column of 'target'")
})

#' @include AllGenerics.R

#' @title Compound Annotation Sources
#'
#' @aliases CompAnnotationSource-class
#'
#' @description
#'
#' `CompAnnotationSource`s (i.e. classes extending the base virtual
#' `CompAnnotationSource` class) define and provide access to a (potentially
#' remote) compound annotation resource. This aims to simplify the integration
#' of external annotation resources by automating the actual connection
#' (or data resource download) process from the user. In addition, since the
#' reference resource is not directly exposed to the user it allows integration
#' of annotation resources that do not allow access to the full data.
#'
#' Objects extending `CompAnnotationSource` available in this package are:
#'
#' - [CompDbSource()]: annotation source referencing an annotation source in the
#'   `[CompoundDb::CompDb()]` format ( from the `CompoundDb` Bioconductor
#'   package).
#'
#' Classes extending `CompAnnotationSource` need to implement the `matchSpectra`
#' method with parameters `query`, `target` and `param` where `query` is
#' the `Spectra` object with the (experimental) query spectra, `target` the
#' object extending the `CompAnnotationSource` and `param` the parameter object
#' defining the similarity calculation (e.g. [CompareSpectraParam()]. The method
#' is expected to return a [MatchedSpectra] object.
#'
#' `CompAnnotationSource` objects are not expected to contain any annotation
#' data. Access to the annotation data (in form of a `Spectra` object) is
#' suggested to be only established within the object's `matchSpectra` method.
#' This would also enable parallel processing of annotations as no e.g. database
#' connection would have to be shared across processes.
#'
#' @param query for `matchSpectra`: [Spectra::Spectra] object with the query
#'   spectra.
#'
#' @param target for `matchSpectra`: object extending [CompAnnotationSource]
#'   (such as [CompDbSource]) with the target (reference) spectra to compare
#'   `query` against.
#'
#' @param object A `CompAnnotationSource` object.
#'
#' @param param for `matchSpectra`: parameter object (such as
#'   [CompareSpectraParam]) defining the settings for the matching.
#'
#' @param x A `CompAnnotationSource` object.
#'
#' @param ... additional parameters passed to `matchSpectra`.
#'
#' @section Methods that need to be implemented:
#'
#' For an example implementation see [CompDbSource()].
#'
#' - `matchSpectra`: function to match experimental MS2 spectra against the
#'   annotation source. See [matchSpectra()] for parameters.
#'
#' - `metadata`: function to provide metadata on the annotation resource (host,
#'   source, version etc).
#'
#' - `show` (optional): method to provide general information on the data
#'   source.
#'
#' @author Johannes Rainer, Nir Shachaf
#'
#' @name CompAnnotationSource
NULL

#' @exportClass CompAnnotationSource
setClass(
    "CompAnnotationSource",
    contains = "VIRTUAL",
    slots = c(
        version = "character"),
    prototype = prototype(version = "0.1"))

#' @rdname CompAnnotationSource
setMethod("matchSpectra",
          signature(query = "Spectra", target = "CompAnnotationSource",
                    param = "Param"),
          function(query, target, param, ...)
              stop("Not implemented for ", class(target), "."))

#' @rdname CompAnnotationSource
setMethod("show", "CompAnnotationSource", function(object) {
    cat("Object of class", class(object)[1L], "\n")
})

#' @importMethodsFrom S4Vectors metadata
#'
#' @rdname CompAnnotationSource
setMethod("metadata", "CompAnnotationSource", function(x, ...) {
    stop("Not implemented for ", class(x), ".")
})

#' @title Internal page for hidden aliases
#'
#' @description
#'
#' For S4 methods that require a documentation entry but only clutter the index.
#'
#' @usage NULL
#' @format NULL
#' @keywords internal
#' @return Not applicable
#' @docType methods
hidden_aliases <- NULL

library(MetaboAnnotation)
library(msdata)
library(Spectra)
library(testthat)

fl <- system.file("TripleTOF-SWATH", "PestMix1_DDA.mzML", package = "msdata")
pest_ms2 <- filterMsLevel(Spectra(fl), 2L)
pest_ms2 <- pest_ms2[c(808, 809, 945:955)]
load(system.file("extdata", "minimb.RData", package = "MetaboAnnotation"))

test_check("MetaboAnnotation")

#' @include CompAnnotationSource.R

#' @title Compound Annotation Sources for `CompDb` databases
#'
#' @aliases CompDbSource-class
#'
#' @description
#'
#' `CompDbSource` objects represent references to [CompoundDb::CompDb]
#' database-backed annotation resources. Instances are expected to be
#' created with the dedicated construction functions such as
#' `MassBankSource` or the generic `CompDbSource`. The annotation data is
#' not stored within the object but will be accessed/loaded within the
#' object's `matchSpectra` method.
#'
#' New `CompDbSource` objects can be created using the functions:
#'
#' - `CompDbSource`: create a new `CompDbSource` object from an existing
#'   `CompDb` database. The (SQLite) database file (including the full path)
#'   needs to be provided with parameter `dbfile`.
#'
#' - `MassBankSource`: retrieves a `CompDb` database for the specified MassBank
#'   release from Bioconductor's online `AnnotationHub` (if it exists) and
#'   uses that. Note that `AnnotationHub` resources are cached locally and thus
#'   only downloaded the first time.
#'   The function has parameters `release` which allows to define the desired
#'   MassBank release (e.g. `release = "2021.03"` or `release = "2022.06"`)
#'   and `...` which allows to pass optional parameters to the `AnnotationHub`
#'   constructor function, such as `localHub = TRUE` to use only the cached
#'   data and avoid updating/retrieving updates from the internet.
#'
#' Other functions:
#'
#' - `metadata`: get metadata (information) on the annotation resource.
#'
#' @param dbfile `character(1)` with the database file (including the full
#'     path).
#'
#' @param object A `CompDbSource` object.
#'
#' @param release A `character(1)` defining the version/release of MassBank that
#'     should be used.
#'
#' @param x A `CompDbSource` object.
#'
#' @param ... For `CompDbSource`: ignored. For `MassBankSource`: optional
#'     parameters passed to the `AnnotationHub` constructor function.
#'
#' @name CompDbSource
#'
#' @author Johannes Rainer
#'
#' @examples
#'
#' ## Locate a CompDb SQLite database file. For this example we use the test
#' ## database from the `CompoundDb` package.
#' fl <- system.file("sql", "CompDb.MassBank.sql", package = "CompoundDb")
#' ann_src <- CompDbSource(fl)
#'
#' ## The object contains only the reference/link to the annotation resource.
#' ann_src
#'
#' ## Retrieve a CompDb with MassBank data for a certain MassBank release
#' mb_src <- MassBankSource("2021.03")
#' mb_src
NULL

#' @description
#'
#' It is better to NOT put neither a connection object nor the data itself into
#' the source object to allow parallel processing. The `matchSpectra` method
#' needs then to make sure to load/retrieve the data and connect to it.
#'
#' @author Johannes Rainer
#'
#' @noRd
#'
#' @exportClass CompDbSource
setClass(
    "CompDbSource",
    contains = "CompAnnotationSource",
    slots = c(dbfile = "character"),
    prototype = list(dbfile = character())
)

#' validator to check if the provided file name is indeed a CompDb database.
#'
#' @importFrom CompoundDb CompDb
#'
#' @noRd
.validate_dbfile <- function(x) {
    if (length(x))
        validObject(CompDb(x))
    else TRUE
}

setValidity("CompDbSource", function(object) {
    .validate_dbfile(object@dbfile)
})

#' @rdname CompDbSource
#'
#' @export
CompDbSource <- function(dbfile = character()) {
    new("CompDbSource", dbfile = dbfile)
}

#' @export
#'
#' @rdname CompDbSource
setMethod("metadata", "CompDbSource", function(x, ...) {
    db <- CompDb(x@dbfile)
    metadata(db)
})

#' @importFrom methods callNextMethod
#'
#' @rdname CompDbSource
setMethod("show", "CompDbSource", function(object) {
    callNextMethod()
    md <- metadata(object)
    cat("Metadata information:\n")
    cat(paste0("  - ", md[, 1], ": ", md[, 2], "\n"), sep = "")
})

#' @rdname CompareSpectraParam
#'
#' @importFrom Spectra MsBackendDataFrame
#'
#' @importClassesFrom Spectra MsBackendDataFrame
setMethod(
    "matchSpectra", signature(query = "Spectra", target = "CompDbSource",
                              param = "Param"),
    function(query, target, param, BPPARAM = BiocParallel::SerialParam(),
             addOriginalQueryIndex = TRUE) {
        ## connect to the database
        db <- CompDb(target@dbfile)
        ## get the Spectra from the source and call matchSpectra
        res <- matchSpectra(
            query, Spectra(db), param = param, BPPARAM = BPPARAM,
            addOriginalQueryIndex = addOriginalQueryIndex)
        ## keep only matching reference/target spectra  and change the
        ## backend to MsBackendDataFrame
        res <- pruneTarget(res)
        res@target <- setBackend(res@target, backend = MsBackendDataFrame())
        res
    })

#' @export
#'
#' @rdname CompDbSource
MassBankSource <- function(release = "2021.03", ...) {
    if (!.is_available_annotation_hub())
        stop("'MassBankSource' requires the 'AnnotationHub' package.\n",
             "Please install it with 'BiocManager::install(",
             "\"AnnotationHub\")' and try again.")
    ah <- AnnotationHub::AnnotationHub(...)
    res <- AnnotationHub::query(ah, c("MassBank", release))
    if (!length(res))
        stop("MassBank release \"", release, "\" not found in AnnotationHub")
    if (length(res) > 1)
        stop("Provided release is ambiguous: ", length(res), " data sets in ",
             "AnnotationHub match the provided release information.")
    fn <- unname(AnnotationHub::cache(res))
    CompDbSource(fn)
}

.is_available_annotation_hub <- function(x) {
    requireNamespace("AnnotationHub", quietly = TRUE)
}

test_that("CompDbSource works", {
    expect_error(new("CompDbSource", dbfile = tempfile()), "unable")
    fl <- system.file("extdata", "MS1_example.txt", package = "MetaboAnnotation")
    expect_error(new("CompDbSource", dbfile = fl), "database")

    res <- new("CompDbSource")
    expect_true(validObject(res))

    fl <- system.file("sql", "CompDb.MassBank.sql", package = "CompoundDb")
    res  <- new("CompDbSource", dbfile = fl)
    expect_true(validObject(res))

    expect_true(.validate_dbfile(fl))

    expect_output(show(res), "CompDbSource")
})

test_that("metadata,CompDbSource works", {
    fl <- system.file("sql", "CompDb.MassBank.sql", package = "CompoundDb")
    src  <- new("CompDbSource", dbfile = fl)
    res <- metadata(src)
    expect_true(is.data.frame(res))
})

test_that("matchSpectra,Spectra,CompDbSource works", {
    fl <- system.file("sql", "CompDb.MassBank.sql", package = "CompoundDb")
    src  <- new("CompDbSource", dbfile = fl)

    res <- matchSpectra(pest_ms2, src, param = CompareSpectraParam(),
                        addOriginalQueryIndex = FALSE)
    expect_s4_class(res, "MatchedSpectra")
    expect_equal(query(res), pest_ms2)
    expect_s4_class(target(res)@backend, "MsBackendDataFrame")
    expect_true(length(target(res)) == 0)

    library(CompoundDb)
    library(BiocParallel)
    register(SerialParam())
    if (file.exists(fl)) {
        qry <- Spectra(CompoundDb::CompDb(fl))[3]
        res <- matchSpectra(qry, src, param = CompareSpectraParam())
        expect_true(length(target(res)) == 4)
        expect_equal(MetaboAnnotation::matches(res)$target_idx, 1:4)
        expect_s4_class(target(res)@backend, "MsBackendDataFrame")
    }
})

test_that("MassBankSource works with AnnotationHub", {
    if (requireNamespace("AnnotationHub", quietly = TRUE)) {
        expect_error(MassBankSource(release = "other"), "not found")
        expect_error(MassBankSource(release = ""), "ambiguous")

        mb <- MassBankSource("2021.03")
        expect_s4_class(mb, "CompDbSource")
        expect_true(length(mb@dbfile) == 1L)
    }
    with_mocked_bindings(
        ".is_available_annotation_hub" = function() FALSE,
        code = expect_error(MassBankSource(), "requires")
    )
})

test_that(".is_available_annotation_hub works", {
    expect_true(.is_available_annotation_hub())
})

#' @title Create Standard Mixes from a Matrix of Standard Compounds
#'
#' @description
#'
#' The `createStandardMixes` function defines groups (mixes) of compounds
#' (standards) with dissimilar m/z values. The expected size of the groups can
#' be defined with parameters `max_nstd` and `min_nstd` and the minimum required
#' difference between m/z values within each group with parameter `min_diff`.
#' The group assignment will be reported in an additional column in the result
#' data frame.
#'
#' @param x `numeric` matrix with row names representing the compounds and
#' columns representing different adducts. Such a matrix with m/z values for
#' different adducts for compounds could e.g. be created with the
#' [MetaboCoreUtils::mass2mz()] function.
#'
#' @param max_nstd `numeric` number of maximum standards per group.
#'
#' @param min_nstd `numeric` number of minimum standards per group. Only
#' needed when using `iterativeRandomization = TRUE`.
#'
#' @param min_diff `numeric` Minimum difference for considering two values as
#' distinct.
#'
#' @param iterativeRandomization `logical` default `FALSE`. If set to `TRUE`,
#' `createStandardMixes` will randomly rearrange the rows of `x` until the user
#' inputs are satisfied.
#'
#' @return `data.frame` created by adding a column `group` to the input `x`
#' matrix, comprising the group number for each compound.
#'
#' @details
#' Users should be aware that because the function iterates through `x`, the
#' compounds at the bottom of the matrix are more complicated to group, and
#' there is a possibility that some compounds will not be grouped with others.
#' We advise specifyiong `iterativeRandomization = TRUE` even if it takes more
#' time.
#'
#' @examples
#'
#' ## Iterative grouping only
#' x <- matrix(c(135.0288, 157.0107, 184.0604, 206.0424, 265.1118, 287.0937,
#'               169.0356, 191.0176, 468.9809, 490.9628, 178.0532, 200.0352),
#'             ncol = 2, byrow = TRUE,
#'             dimnames = list(c("Malic Acid", "Pyridoxic Acid", "Thiamine",
#'                                 "Uric acid", "dUTP", "N-Formyl-L-methionine"),
#'                              c("adduct_1", "adduct_2")))
#' result <- createStandardMixes(x, max_nstd = 3, min_diff = 2)
#'
#' ## Randomize grouping
#' set.seed(123)
#' x <- matrix(c(349.0544, 371.0363, 325.0431, 347.0251, 581.0416, 603.0235,
#'               167.0564, 189.0383, 150.0583, 172.0403, 171.0053, 192.9872,
#'               130.0863, 152.0682, 768.1225, 790.1044),
#'             ncol = 2, byrow = TRUE,
#'             dimnames = list(c("IMP", "UMP", "UDP-glucuronate",
#'                                 "1-Methylxanthine", "Methionine",
#'                                 "Dihydroxyacetone phosphate",
#'                                 "Pipecolic acid", "CoA"),
#'                              c("[M+H]+", "[M+Na]+")))
#' result <- createStandardMixes(x, max_nstd = 4, min_nstd = 3, min_diff = 2,
#'                                iterativeRandomization = TRUE)
#'
#' @author Philippine Louail
#'
#' @export
#'
#' @rdname createStandardMixes
createStandardMixes <- function(x, max_nstd = 10,
                                min_nstd =5, min_diff = 2,
                                iterativeRandomization = FALSE) {
    if (!is.matrix(x))
        stop("The input `x` needs to be a matrix, try as.matrix(x)")
    if (!iterativeRandomization)
        output <- .iterative_grouping(x = x, max_nstd = max_nstd,
                                      min_diff = min_diff)
    else
        output <- .randomize_grouping(x = x, max_nstd = max_nstd,
                                      min_nstd = min_nstd, min_diff = min_diff)

    find_index <- function(name) {
        which(vapply(output, function(X) any(X %in% name), logical(1)))
        }
    x <- as.data.frame(x)
    x$group <- vapply(rownames(x), find_index, integer(1))

    x
}

#'
#' @noRd
#'

.iterative_grouping <- function(x, max_nstd, min_diff) {
    output <- vector("list")
    g <- 0

    while (nrow(x) > 1) {
        g <- g + 1
        i <- 1
        group <- row.names(x)[i]

        while (length(group) < max_nstd & i < nrow(x)) {
            i <- i + 1
            diff_table <- abs(outer(as.vector(x[group, ]), as.vector(x[i,]),
                                    "-"))

            if (all(diff_table > min_diff, na.rm= TRUE)) {
                group <- c(group, row.names(x)[i])
            }
        }
        x <- x[!(rownames(x) %in% group), , drop = FALSE]
        output[[g]] <- group
    }
    if (nrow(x))
        output[[g + 1]] <- row.names(x)

    output
}

#'
#' @noRd
#'
.randomize_grouping <- function(x,
                                max_nstd,
                                min_nstd,
                                min_diff) {
    n <- nrow(x)
    output <- vector("list")
    i <- 0
    while (length(output) == 0 || any(lengths(output) < min_nstd)) {

        i <- i +1
        x <- x[sample(n), , drop = FALSE]
        output <- .iterative_grouping(x, max_nstd = max_nstd,
                                      min_diff = min_diff)
        if (i > n*n)
            stop("all combination were tested, no possibility to fit your",
                 " input requirements")
    }
    output
}

#' define some variable for test
x <- data.frame(
    row.names = c("Malic Acid", "Pyridoxic Acid", "Thiamine", "Uric acid",
                "dUTP", "N-Formyl-L-methionine"),
    "adduct_1" = c(135.0288, 184.0604, 265.1118, 169.0356, 468.9809, 178.0532),
    "adduct_2" = c(157.0107, 206.0424, 287.0937, 191.0176, 490.9628, 200.0352)
)
x <- as.matrix(x)

#' Expected list to get for maximum of 3 standard per group
results <- list(c("Malic Acid", "Pyridoxic Acid", "Thiamine"),
                c("Uric acid", "dUTP", "N-Formyl-L-methionine"))

test_that("essentially that .iterative_grouping works", {
    expect_equal(.iterative_grouping(x, max_nstd = 3, min_diff = 2), results)
    expect_is(.iterative_grouping(x, max_nstd = 3, min_diff = 2), "list")

    res <- createStandardMixes(
        x, max_nstd = 4, min_nstd = 3,
        min_diff = 2, iterativeRandomization = FALSE)
    expect_true(is.data.frame(res))
    expect_equal(res$group, c(1, 1, 1, 1, 2, 2))
})

#' test for randomization
set.seed(123)
#' Create a matrix with compound names and ion masses
x <- matrix(c(349.0544, 371.0363, 325.0431, 347.0251, 581.0416, 603.0235,
              167.0564, 189.0383, 150.0583, 172.0403, 171.0053, 192.9872,
              130.0863, 152.0682, 768.1225, 790.1044),
            ncol = 2, byrow = TRUE)

rownames(x) <- c("IMP", "UMP", "UDP-glucuronate", "1-Methylxanthine",
                 "Methionine", "Dihydroxyacetone phosphate", "Pipecolic acid",
                 "CoA")
colnames(x) <- c("[M+H]+", "[M+Na]+")

#' run using `.iterative_grouping`
standard_groups <- .iterative_grouping(x, max_nstd = 4, min_diff = 2)

#' run using `.random_grouping`.
standard_groups_r <- .randomize_grouping(x, max_nstd = 4,
                                         min_nstd = 3,
                                         min_diff = 2)
min_nstd <- 3
test_that("randomization improves results", {
    expect_true(length(standard_groups) > length(standard_groups_r))
    expect_true(any(lengths(standard_groups_r) > min_nstd))
    expect_false(length(standard_groups_r) == 0 )
})

#check format of global function ouput
global_result <- createStandardMixes(x, max_nstd = 4,
                                     min_nstd = 3,
                                     min_diff = 2, iterativeRandomizati = TRUE)
test_that("out put format is right", {
    expect_true(is.data.frame(global_result))
    expect_equal(ncol(x)+1 , ncol(global_result))
    expect_equal(nrow(x), nrow(global_result))
})

test_that("createStandardMixes works", {
    expect_error(createStandardMixes(4), "matrix")
})

test_that(".randomize_grouping works", {
    x <- data.frame(
        row.names = c("Malic Acid", "Pyridoxic Acid", "Thiamine", "Uric acid",
                      "dUTP", "N-Formyl-L-methionine"),
        "adduct_1" = c(135.0288, 184.0604, 265.1118, 169.0356, 468.9809, 178.0532),
        "adduct_2" = c(157.0107, 206.0424, 287.0937, 191.0176, 490.9628, 200.0352)
    )
    x <- as.matrix(x)
    expect_error(.randomize_grouping(
        x, max_nstd = 4, min_nstd = 3, min_diff = 2),
        "all combination")
})

#' @importClassesFrom S4Vectors DataFrame
#' @importClassesFrom QFeatures QFeatures
#' @importClassesFrom SummarizedExperiment SummarizedExperiment
setClassUnion("data.frameOrSimilar", c("data.frame", "matrix", "DataFrame",
                                       "SummarizedExperiment", "QFeatures"))
setClassUnion("adductClass", c("character", "data.frame"))
setClassUnion("characterOrNumeric", c("character", "numeric"))

#' @noRd
setClass("ValueParam",
         slots = c(
             tolerance = "numeric",
             ppm = "numeric"),
         contains = "Param",
         prototype = prototype(
             tolerance = 0,
             ppm = 5),
         validity = function(object) {
             msg <- NULL
             if (length(object@tolerance) != 1 || object@tolerance < 0)
                 msg <- c("'tolerance' has to be a positive number of length 1")
             if (length(object@ppm) != 1 || object@ppm < 0)
                 msg <- c("'ppm' has to be a positive number of length 1")
             msg
         })

#' @rdname matchValues
#'
#' @importFrom methods new
#'
#' @export
ValueParam <- function(tolerance = 0, ppm = 5) {
    new("ValueParam", tolerance = tolerance, ppm = ppm)
}

#' @noRd
setClass("MzParam", contains = "ValueParam")

#' @rdname matchValues
#'
#' @importFrom methods new
#'
#' @export
MzParam <- function(tolerance = 0, ppm = 5) {
    new("MzParam", tolerance = tolerance, ppm = ppm)
}

#' @importFrom MetaboCoreUtils adductNames
#'
#' @noRd
setClass("Mass2MzParam",
         slots = c(
             targetAdducts = "adductClass"),
         contains = "ValueParam",
         prototype = prototype(
             targetAdducts = c("[M+H]+")),
         validity = function(object) {
             .valid_adduct(object@targetAdducts, "`targetAdducts`")
         })

#' @rdname matchValues
#'
#' @importFrom methods new
#'
#' @export
Mass2MzParam <- function(adducts = c("[M+H]+"), tolerance = 0, ppm = 5) {
    new("Mass2MzParam", targetAdducts = adducts, tolerance = tolerance,
        ppm = ppm)
}

#' @importFrom MetaboCoreUtils adductNames
#'
#' @noRd
setClass("Mass2MzRtParam",
         slots = c(
             toleranceRt = "numeric"),
         contains = "Mass2MzParam",
         prototype = prototype(
             toleranceRt = 0),
         validity = function(object) {
             msg <- NULL
             if (length(object@toleranceRt) != 1 || object@toleranceRt < 0)
                 msg <- "'toleranceRt' has to be a positive number of length 1"
             msg
         })

#' @rdname matchValues
#'
#' @importFrom methods new
#'
#' @export
Mass2MzRtParam <- function(adducts = c("[M+H]+"), tolerance = 0, ppm = 5,
                           toleranceRt = 0) {
    new("Mass2MzRtParam", targetAdducts = adducts, tolerance = tolerance,
        ppm = ppm, toleranceRt = toleranceRt)
}

#' @noRd
setClass("MzRtParam",
         slots = c(
             toleranceRt = "numeric"),
         contains = "MzParam",
         prototype = prototype(
             toleranceRt = 0),
         validity = function(object) {
             msg <- NULL
             if (length(object@toleranceRt) != 1 || object@toleranceRt < 0)
                 msg <- "'toleranceRt' has to be a positive number of length 1"
             msg
         })

#' @rdname matchValues
#'
#' @importFrom methods new
#'
#' @export
MzRtParam <- function(tolerance = 0, ppm = 0, toleranceRt = 0) {
    new("MzRtParam", tolerance = tolerance, ppm = ppm, toleranceRt = toleranceRt)
}

#' @importFrom MetaboCoreUtils adductNames
#'
#' @noRd
setClass("Mz2MassParam",
         slots = c(
             queryAdducts = "adductClass",
             targetAdducts = "adductClass"),
         contains = "ValueParam",
         prototype = prototype(
             queryAdducts = c("[M+H]+"),
             targetAdducts = c("[M-H]-")),
         validity = function(object) {
             c(.valid_adduct(object@queryAdducts, "`queryAdducts`"),
               .valid_adduct(object@targetAdducts, "`targetAdducts`"))
         })

#' @rdname matchValues
#'
#' @importFrom methods new
#'
#' @export
Mz2MassParam <- function(queryAdducts = c("[M+H]+"),
                         targetAdducts = c("[M-H]-"),
                         tolerance = 0, ppm = 5) {
    new("Mz2MassParam", queryAdducts = queryAdducts,
        targetAdducts = targetAdducts, tolerance = tolerance, ppm = ppm)
}

#' @noRd
setClass("Mz2MassRtParam",
         slots = c(
             toleranceRt = "numeric"),
         contains = "Mz2MassParam",
         prototype = prototype(
             toleranceRt = 0),
         validity = function(object) {
             msg <- NULL
             if (length(object@toleranceRt) != 1 || object@toleranceRt < 0)
                 msg <- "'toleranceRt' has to be a positive number of length 1"
             msg
         })

#' @rdname matchValues
#'
#' @importFrom methods new
#'
#' @export
Mz2MassRtParam <- function(queryAdducts = c("[M+H]+"),
                           targetAdducts = c("[M+H]+"),
                           tolerance = 0, ppm = 5, toleranceRt = 0) {
    new("Mz2MassRtParam", queryAdducts = queryAdducts,
        targetAdducts = targetAdducts, tolerance = tolerance, ppm = ppm,
        toleranceRt = toleranceRt)
}

#' @title Matching of numeric values
#'
#' @aliases matchMz
#'
#' @name matchValues
#'
#' @export matchMz
#'
#' @description
#'
#' The `matchValues` method matches elements from `query` with those in `target`
#' using different matching approaches depending on parameter `param`.
#' Generally, `query` is expected to contain MS experimental values
#' (m/z and possibly retention time) while `target` reference values. `query`
#' and `target` can be `numeric`, a two dimensional array (such as a
#' `data.frame`, `matrix` or `DataFrame`), a `SummarizedExperiment`
#' or a `QFeatures`, `target` can in addition be a [Spectra::Spectra()] object.
#' For `SummarizedExperiment`, the information for the matching is expected
#' to be in the object's `rowData`. For `QFeatures` matching is performed
#' for values present in the `rowData` of one of the object's assays (which
#' needs to be specified with the `assayQuery` parameter - if a `QFeatures`
#' is used as `target` the name of the assay needs to be specified with
#' parameter `assayTarget`). If `target` is a `Spectra` matching is performed
#' against spectra variables of this object and the respective variable names
#' need to be specified e.g. with `mzColname` and/or `rtColname`.
#' `matchMz` is an alias for `matchValues` to allow backward compatibility.
#'
#' Available `param` objects and corresponding matching approaches are:
#'
#' - `ValueParam`: generic matching between values in `query` and `target` given
#'   acceptable differences expressed in `ppm` and `tolerance`. If `query` or
#'   `target` are not numeric, parameter `valueColname` has to be used to
#'   specify the name of the column that contains the values to be matched.
#'   The function returns a [Matched()] object.
#'
#' - `MzParam`: match query m/z values against reference compounds for which
#'   also m/z are known. Matching is performed similarly to the `ValueParam`
#'   above. If `query` or `target` are not numeric, the column name containing
#'   the values to be compared must be defined with `matchValues`' parameter
#'   `mzColname`, which defaults to `"mz"`. `MzParam` parameters `tolerance`
#'   and `ppm` allow to define the maximal acceptable (constant or m/z relative)
#'   difference between query and target m/z values.
#'
#' - `MzRtParam`: match m/z **and** retention time values between `query` and
#'   `target`. Parameters `mzColname` and `rtColname` of the `matchValues`
#'   function allow to define the columns in `query` and `target` containing
#'   these values (defaulting to `c("mz", "mz")` and `c("rt", "rt")`,
#'   respectively). `MzRtParam` parameters `tolerance` and
#'   `ppm` have the same meaning as in `MzParam`; `MzRtParam` parameter
#'   `toleranceRt` allows to specify the maximal acceptable difference between
#'   query and target retention time values.
#'
#' - `Mass2MzParam`: match m/z values against reference compounds for
#'   which only the (exact) mass is known. Before matching, m/z values are
#'   calculated from the compounds masses in the *target* table using the
#'   adducts specified via `Mass2MzParam` `adducts` parameter (defaults to
#'   `adducts = "[M+H]+"`). After conversion of adduct masses to m/z values,
#'   matching is performed similarly to `MzParam` (i.e. the same parameters
#'   `ppm` and `tolerance` can be used). If `query` is not `numeric`,
#'   parameter `mzColname` of `matchValues` can be used to specify the column
#'   containing the query's m/z values (defaults to `"mz"`). If `target` is a
#'   is not `numeric`, parameter `massColname` can be used to define the
#'   column containing the reference compound's masses (defaults to
#'   `"exactmass"`).
#'
#' - `Mass2MzRtParam`: match m/z **and** retention time values against
#'   reference compounds for which the (exact) mass **and** retention time are
#'   known. Before matching, exact masses in `target` are converted to m/z
#'   values as for `Mass2MzParam`. Matching is then performed similarly to
#'   `MzRtParam`, i.e. m/z and retention times of entities are compared. With
#'   `matchValues`' parameters `mzColname`, `rtColname` and `massColname` the
#'   columns containing m/z values (in `query`), retention time values (in
#'   `query` and `target`) and exact masses (in `target`) can be specified.
#'
#' - `Mz2MassParam`: input values for `query` and `target` are expected to be
#'   m/z values but matching is performed on exact masses calculated from these
#'   (based on the provided adduct definitions). In detail, m/z values in
#'   `query` are first converted to masses with the [MetaboCoreUtils::mz2mass()]
#'   function based on the adducts defined with `queryAdducts` (defaults to
#'   `"[M+H]+"`). The same is done for m/z values in `target` (adducts can
#'   be defined with `targetAdducts` which defaults to `"[M-H-]").
#'   Matching is then performed on these converted values similarly to
#'   `ValueParam`. If `query` or `target` are not numeric, the column
#'   containing the m/z values can be specified with `matchValues`'
#'   parameter `mzColname` (defaults to `"mz"`).
#'
#' - `Mz2MassRtParam`: same as `Mz2MassParam` but with additional comparison of
#'   retention times between `query` and `target`. Parameters `rtColname` and
#'   `mzColname` of `matchValues` allow to specify which columns contain the
#'   retention times and m/z values, respectively.
#'
#' @param adducts for `Mass2MzParam` or `Mass2MzRtParam`: either `character`
#'     with adduct names from [MetaboCoreUtils::adducts()] or `data.frame` with
#'     a custom adduct definition. This parameter is used to calculate m/z from
#'     target compounds' masses. Custom adduct definitions can be passed to the
#'     adduct parameter in form of a `data.frame`. This `data.frame` is expected
#'     to have columns `"mass_add"` and `"mass_multi"` defining the *additive*
#'     and *multiplicative* part of the calculation. See
#'     [MetaboCoreUtils::adducts()] for the expected format or use
#'     `MetaboCoreUtils::adductNames("positive")` and
#'     `MetaboCoreUtils::adductNames("negative")` for valid adduct names.
#'
#' @param massColname `character(1)` with the name of the column in `target`
#'     containing the mass of compounds. To be used when `param` is
#'     `Mass2MzParam` or `Mass2MzRtParam` (and target is not already `numeric`
#'     with the masses). Defaults to `massColname = "exactmass"`.
#'
#' @param mzColname `character` specifying the name(s) of the column(s) in
#'     `query` or/and `target`with the m/z values. If one among `query` and
#'     `target` is `numeric` (and therefore there is no need to specify the
#'     column name) or `query` is not `numeric` and `param` is `Mass2MzParam`
#'     or `Mass2MzRtParam` (and therefore the name of the column with m/z needs
#'     only to be specified for `query`) then `mzColname` is expected to be
#'     `character(1)`. If both `query` and `target` are not numeric `mzColname`
#'     is expected to be `character(2)` (or `character(1)` and in this last case
#'     the two column names are assumed to be the same). If not specified the
#'     assumed default name for columns with m/z values is `"mz"`. If `target`
#'     is a [Spectra::Spectra()] object, the name of the spectra variable
#'     that should be used for the matching needs to be specified
#'     with `mzColname`.
#'
#' @param param parameter object defining the matching approach and containing
#'     the settings for that approach. See description above for details.
#'
#' @param ppm for any `param` object: `numeric(1)` defining the maximal
#'     acceptable m/z-dependent (or mass-dependent for `Mz2MassParam`)
#'     difference (in parts-per-million) in m/z values to consider them to
#'     be *matching*.
#'
#' @param query feature table containing information on MS1 features. Can be
#'     a `numeric`, `data.frame`, `DataFrame`, `matrix`, `SummarizedExperiment`
#'     or `QFeatures`. It is expected to contain m/z values and can contain also
#'     other variables. Matchings based on both m/z and retention time can be
#'     performed when a column with retention times is present in both `query`
#'     and `target`.
#'
#' @param queryAdducts for `Mz2MassParam`. Adducts used to derive mass
#'     values from query m/z values. The expected format is the same as that
#'     for parameter `adducts`.
#'
#' @param queryAssay `character(1)` specifying the name of the assay of the
#'     provided `QFeatures` that should be used for the matching (values from
#'     this assay's `rowData` will be used for matching). Only used if `query`
#'     is an instance of a `QFeatures` object.
#'
#' @param rtColname `character(2)` with the name of the column containing
#'     the compounds retention times in `query` and the name for the one in
#'     `target`. It can also be `character(1)` if the two names are the same.
#'     To be used when `param` is `MzRtParam` or `Mass2MzRtParam`.
#'     Defaults to `rtColname = c("rt", "rt")`. If `target` is a
#'     [Spectra::Spectra()] object, the name of the spectra variable that
#'     should be used for the matching needs to be specified with `mzColname`.
#'
#' @param target compound table with metabolites to compare against. The
#'     expected types are the same as those for `query`.
#'
#' @param targetAdducts for `Mz2MassParam`. Adducts used to derive mass
#'     values from target m/z values. The expected format is the same as that
#'     for parameter `adducts`.
#'
#' @param targetAssay `character(1)` specifying the name of the assay of the
#'     provided `QFeatures` that should be used for the matching (values from
#'     this assay's `rowData` will be used for matching). Only used if `target`
#'     is an instance of a `QFeatures` object.
#'
#' @param tolerance for any `param` object: `numeric(1)` defining the maximal
#'     acceptable absolute difference in m/z (or in mass for `Mz2MassParam`)
#'     to consider them *matching*.
#'
#' @param toleranceRt for `Mass2MzRtParam` or `MzRtParam`: `numeric(1)`
#'     defining the maximal acceptable absolute difference in retention time
#'     values to consider them them *matching*.
#'
#' @param valueColname `character` specifying the name of the column in
#'     `query` or/and the one in `target`with the desired values for the
#'     matching. This parameter should only be used when `param` is
#'     `valueParam` and in this case it must be provided (unless both `query`
#'     and `target` are `numeric`). It can be `character(1)` or `character(2)`
#'     in a similar way to `mzColname`.
#'
#' @param ... currently ignored.
#'
#' @return [Matched] object representing the result.
#'
#' Depending on the `param` object different *scores* representing the quality
#' of the match are provided. This comprises absolute as well as relative
#' differences (column/variables `"score"` and `"ppm_error"` respectively).
#' If `param` is a `Mz2MassParam`, `"score"` and `"ppm_error"` represent
#' differences of the compared masses (calculated from the provided m/z values).
#' If `param` an `MzParam`, `MzRtParam`, `Mass2MzParam` or `Mass2MzRtParam`,
#' `"score"` and `"ppm_error"` represent absolute and relative differences of
#' m/z values.
#' Additionally, if `param` is either an `MzRtParam` or `Mass2MzRtParam`
#' differences between query and target retention times for each matched
#' element is available in the column/variable `"score_rt"` in the returned
#' `Matched` object.
#' Negative values of `"score"` (or `"score_rt"`) indicate that the m/z or mass
#' (or retention time) of the query element is smaller than that of the target
#' element.
#'
#' @author Andrea Vicini, Michael Witting
#'
#' @seealso [matchSpectra] or [CompareSpectraParam()] for spectra data matching
#'
#' @examples
#'
#' library(MetaboCoreUtils)
#' ## Create a simple "target/reference" compound table
#' target_df <- data.frame(
#'    name = c("Tryptophan", "Leucine", "Isoleucine"),
#'    formula = c("C11H12N2O2", "C6H13NO2", "C6H13NO2"),
#'    exactmass = c(204.089878, 131.094629, 131.094629)
#' )
#'
#' ## Create a "feature" table with m/z of features. We calculate m/z for
#' ## certain adducts of some of the compounds in the reference table.
#' fts <- data.frame(
#'     feature_id = c("FT001", "FT002", "FT003"),
#'     mz = c(mass2mz(204.089878, "[M+H]+"),
#'            mass2mz(131.094629, "[M+H]+"),
#'            mass2mz(204.089878, "[M+Na]+") + 1e-6))
#'
#' ## Define the parameters for the matching
#' parm <- Mass2MzParam(
#'     adducts = c("[M+H]+", "[M+Na]+"),
#'     tolerance = 0,
#'     ppm = 20)
#' res <- matchValues(fts, target_df, parm)
#' res
#'
#' ## List the available variables/columns
#' colnames(res)
#'
#' ## feature_id and mz are from the query data frame, while target_name,
#' ## target_formula and target_exactmass are from the query object (columns
#' ## from the target object have a prefix *target_* added to the original
#' ## column name. Columns adduct, score and ppm_error represent the results
#' ## of the matching: adduct the adduct/ion of the original compound for which
#' ## the m/z matches, score the absolute difference of the query and target
#' ## m/z and ppm_error the relative difference in m/z values.
#'
#' ## Get the full matching result:
#' matchedData(res)
#'
#' ## We have thus matches of FT002 to two different compounds (but with the
#' ## same mass).
#'
#' ## Individual columns can also be accessed with the $ operator:
#' res$feature_id
#' res$target_name
#' res$ppm_error
#'
#'
#' ## We repeat the matching requiring an exact match
#' parm <- Mass2MzParam(
#'     adducts = c("[M+H]+", "[M+Na]+"),
#'     tolerance = 0,
#'     ppm = 0)
#' res <- matchValues(fts, target_df, parm)
#' res
#'
#' matchedData(res)
#'
#' ## The last feature could thus not be matched to any compound.
#'
#' ## At last we use also different adduct definitions.
#' parm <- Mass2MzParam(
#'     adducts = c("[M+K]+", "[M+Li]+"),
#'     tolerance = 0,
#'     ppm = 20)
#' res <- matchValues(fts, target_df, parm)
#' res
#'
#' matchedData(res)
#'
#' ## No matches were found.
#'
#' ## We can also match a "feature" table with a target data.frame taking into
#' ## account both m/z and retention time values.
#' target_df <- data.frame(
#'   name = c("Tryptophan", "Leucine", "Isoleucine"),
#'   formula = c("C11H12N2O2", "C6H13NO2", "C6H13NO2"),
#'   exactmass = c(204.089878, 131.094629, 131.094629),
#'   rt = c(150, 140, 140)
#' )
#'
#' fts <- data.frame(
#'   feature_id = c("FT001", "FT002", "FT003"),
#'   mz = c(mass2mz(204.089878, "[M+H]+"),
#'          mass2mz(131.094629, "[M+H]+"),
#'          mass2mz(204.089878, "[M+Na]+") + 1e-6),
#'   rt = c(150, 140, 150.1)
#' )
#'
#' ## Define the parameters for the matching
#' parm <- Mass2MzRtParam(
#'   adducts = c("[M+H]+", "[M+Na]+"),
#'   tolerance = 0,
#'   ppm = 20,
#'   toleranceRt = 0)
#'
#' res <- matchValues(fts, target_df, parm)
#' res
#'
#' ## Get the full matching result:
#' matchedData(res)
#'
#' ## FT003 could not be matched to any compound, FT002 was matched to two
#' ## different compounds (but with the same mass).
#'
#' ## We repeat the matching allowing a positive tolerance for the matches
#' ## between rt values
#'
#' ## Define the parameters for the matching
#' parm <- Mass2MzRtParam(
#'   adducts = c("[M+H]+", "[M+Na]+"),
#'   tolerance = 0,
#'   ppm = 20,
#'   toleranceRt = 0.1)
#'
#' res <- matchValues(fts, target_df, parm)
#' res
#'
#' ## Get the full matching result:
#' matchedData(res)
#'
#' ## Also FT003 was matched in this case
#'
#' ## It is also possible to match directly m/z values
#' mz1 <- c(12, 343, 23, 231)
#' mz2 <- mz1 + rnorm(4, sd = 0.001)
#'
#' res <- matchValues(mz1, mz2, MzParam(tolerance = 0.001))
#'
#' matchedData(res)
#'
#' ## Matching with a SummarizedExperiment or a QFeatures work analogously,
#' ## only that the matching is performed on the object's `rowData`.
#'
#' ## Below we create a simple SummarizedExperiment with some random assay data.
#' ## Note that results from a data preprocessing with the `xcms` package could
#' ## be extracted as a `SummarizedExperiment` with the `quantify` method from
#' ## the `xcms` package.
#' library(SummarizedExperiment)
#' se <- SummarizedExperiment(
#'     assays = matrix(rnorm(12), nrow = 3, ncol = 4,
#'     dimnames = list(NULL, c("A", "B", "C", "D"))),
#'     rowData = fts)
#'
#' ## We can now perform the matching of this SummarizedExperiment against the
#' ## target_df as before.
#' res <- matchValues(se, target_df,
#'     param = Mass2MzParam(adducts = c("[M+H]+", "[M+Na]+"),
#'         tolerance = 0, ppm = 20))
#' res
#'
#' ## Getting the available columns
#' colnames(res)
#'
#' ## The query columns represent the columns of the object's `rowData`
#' rowData(se)
#'
#' ## matchedData also returns the query object's rowData along with the
#' ## matching entries in the target object.
#' matchedData(res)
#'
#' ## While `query` will return the full SummarizedExperiment.
#' query(res)
#'
#' ## To illustrate use with a QFeatures object we first create a simple
#' ## QFeatures object with two assays, `"ions"` representing the full feature
#' ## data.frame and `"compounds"` a subset of it.
#' library(QFeatures)
#' qf <- QFeatures(list(ions = se, compounds = se[2,]))
#'
#' ## We can perform the same matching as before, but need to specify which of
#' ## the assays in the QFeatures should be used for the matching. Below we
#' ## perform the matching using the "ions" assay.
#' res <- matchValues(qf, target_df, queryAssay = "ions",
#'     param = Mass2MzParam(adducts = c("[M+H]+", "[M+Na]+"),
#'         tolerance = 0, ppm = 20))
#' res
#'
#' ## colnames returns now the colnames of the `rowData` of the `"ions"` assay.
#' colnames(res)
#'
#' matchedData(res)
NULL

#' @rdname matchValues
#'
#' @export
setGeneric("matchValues", function(query, target, param, ...)
    standardGeneric("matchValues"))

matchMz <- matchValues

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "numeric",
                        target = "numeric",
                        param = "ValueParam"),
          function(query, target, param) {
              target_ <- data.frame(index = seq_along(target), value = target)
              queryl <- length(query)
              res <- vector("list", queryl)
              for (i in seq_len(queryl)) {
                  res[[i]] <- .getMatches(i, query[i], target = target_,
                                          tolerance = param@tolerance,
                                          ppm = param@ppm)
              }
              Matched(query = query, target = target,
                      matches = do.call(rbind, res),
                      metadata = list(param = param))
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "numeric",
                        target = "data.frameOrSimilar",
                        param = "ValueParam"),
          function(query, target, param, valueColname = character(),
                   targetAssay = character()) {
              if(!length(valueColname))
                  stop("`valueColname` has to be provided.")
              target_ <- .objectToMatch(target, targetAssay, valueColname)
              res <- matchValues(query, target_, param)
              res@target <- target
              res@targetAssay <- targetAssay
              res
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "data.frameOrSimilar",
                        target = "numeric",
                        param = "ValueParam"),
          function(query, target, param, valueColname = character(),
                   queryAssay = character()) {
              if(!length(valueColname))
                  stop("`valueColname` has to be provided.")
              query_ <- .objectToMatch(query, queryAssay, valueColname)
              res <- matchValues(query_, target, param)
              res@query <- query
              res@queryAssay <- queryAssay
              res
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "data.frameOrSimilar",
                        target = "data.frameOrSimilar",
                        param = "ValueParam"),
          function(query, target, param, valueColname = character(),
                   queryAssay = character(), targetAssay = character()) {
              if(!length(valueColname))
                  stop("`valueColname` has to be provided.")
              if(length(valueColname) == 1)
                  valueColname <- rep(valueColname, 2)
              query_ <- .objectToMatch(query, queryAssay, valueColname[1])
              target_ <- .objectToMatch(target, targetAssay, valueColname[2])
              res <- matchValues(query_, target_, param)
              res@query <- query
              res@queryAssay <- queryAssay
              res@target <- target
              res@targetAssay <- targetAssay
              res
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "numeric",
                        target = "numeric",
                        param = "Mass2MzParam"),
          function(query, target, param) {
              target_mz <- .mass_to_mz_df(target, param@targetAdducts)
              queryl <- length(query)
              res <- vector("list", queryl)
              for (i in seq_len(queryl)) {
                  res[[i]] <- .getMatches(i, query[i], target = target_mz,
                                          tolerance = param@tolerance,
                                          ppm = param@ppm)
              }
              Matched(query = query, target = target,
                      matches = do.call(rbind, res),
                      metadata = list(param = param))
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "numeric",
                        target = "data.frameOrSimilar",
                        param = "Mass2MzParam"),
          function(query, target, param, massColname = "exactmass",
                   targetAssay = character()) {
              target_ <- .objectToMatch(target, targetAssay, massColname)
              res <- matchValues(query, target_, param)
              res@target <- target
              res@targetAssay <- targetAssay
              res
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "data.frameOrSimilar",
                        target = "numeric",
                        param = "Mass2MzParam"),
          function(query, target, param, mzColname = "mz",
                   queryAssay = character()) {
              query_ <- .objectToMatch(query, queryAssay, mzColname)
              res <- matchValues(query_, target, param)
              res@query <- query
              res@queryAssay <- queryAssay
              res
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "data.frameOrSimilar",
                        target = "data.frameOrSimilar",
                        param = "Mass2MzParam"),
          function(query, target, param, mzColname = "mz",
                   massColname = "exactmass", queryAssay = character(0),
                   targetAssay = character(0)) {
              query_ <- .objectToMatch(query, queryAssay, mzColname)
              target_ <- .objectToMatch(target, targetAssay, massColname)
              res <- matchValues(query_, target_, param)
              res@query <- query
              res@queryAssay <- queryAssay
              res@target <- target
              res@targetAssay <- targetAssay
              res
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "numeric",
                        target = "data.frameOrSimilar",
                        param = "MzParam"),
          function(query, target, param, mzColname = "mz",
                   targetAssay = character()) {
              target_ <- .objectToMatch(target, targetAssay, mzColname)
              res <- matchValues(query, target_, param)
              res@target <- target
              res@targetAssay <- targetAssay
              res
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "numeric",
                        target = "Spectra",
                        param = "MzParam"),
          function(query, target, param, mzColname = "mz",
                   targetAssay = character()) {
              if (mzColname %in% peaksVariables(target))
                  stop("Matching against the peak variable \"", mzColname,
                       "\" is not supported. Please use on of the available ",
                       "spectra variables ('spectraVariables(target)').")
              target_ <- .objectToMatch(target, targetAssay, mzColname)
              res <- matchValues(query, target_, param)
              res@target <- target
              res@targetAssay <- targetAssay
              res
          })


#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "data.frameOrSimilar",
                        target = "numeric",
                        param = "MzParam"),
          function(query, target, param, mzColname = "mz",
                   queryAssay = character()) {
              query_ <- .objectToMatch(query, queryAssay, mzColname)
              res <- matchValues(query_, target, param)
              res@query <- query
              res@queryAssay <- queryAssay
              res
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "data.frameOrSimilar",
                        target = "data.frameOrSimilar",
                        param = "MzParam"),
          function(query, target, param, mzColname = c("mz", "mz"),
                   queryAssay = character(), targetAssay = character()) {
              if(length(mzColname) == 1)
                  mzColname <- rep(mzColname, 2)
              query_ <- .objectToMatch(query, queryAssay, mzColname[1])
              target_ <- .objectToMatch(target, targetAssay, mzColname[2])
              res <- matchValues(query_, target_, param)
              res@query <- query
              res@queryAssay <- queryAssay
              res@target <- target
              res@targetAssay <- targetAssay
              res
          })

#' @rdname matchValues
#'
#' @importMethodsFrom Spectra peaksVariables
setMethod("matchValues",
          signature = c(query = "data.frameOrSimilar",
                        target = "Spectra",
                        param = "MzParam"),
          function(query, target, param, mzColname = c("mz", "mz"),
                   queryAssay = character(), targetAssay = character()) {
              if(length(mzColname) == 1)
                  mzColname <- rep(mzColname, 2)
              if (mzColname[2L] %in% peaksVariables(target))
                  stop("Matching against the peak variable \"", mzColname[2L],
                       "\" is not supported. Please use on of the available ",
                       "spectra variables ('spectraVariables(target)').")
              query_ <- .objectToMatch(query, queryAssay, mzColname[1])
              target_ <- .objectToMatch(target, targetAssay, mzColname[2L])
              res <- matchValues(query_, target_, param)
              res@query <- query
              res@queryAssay <- queryAssay
              res@target <- target
              res@targetAssay <- targetAssay
              res
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "data.frameOrSimilar",
                        target = "data.frameOrSimilar",
                        param = "Mass2MzRtParam"),
          function(query, target, param, massColname = "exactmass",
                   mzColname = "mz", rtColname = c("rt", "rt"),
                   queryAssay = character(), targetAssay = character()) {
              if(length(rtColname) == 1)
                  rtColname <- rep(rtColname, 2)
              query_ <- .objectToMatch(query, queryAssay,
                                       c(mzColname, rtColname[1]))
              target_ <- .objectToMatch(target, targetAssay,
                                        c(massColname, rtColname[2]))
              target_mz <- .mass_to_mz_df(target_[, massColname],
                                          param@targetAdducts)
              target_mz$rt <- rep(target_[, rtColname[2]],
                                  .nelements(param@targetAdducts))
              queryl <- nrow(query_)
              matches <- vector("list", queryl)
              query_mz <- query_[, mzColname]
              query_rt <- query_[, rtColname[1L]]
              for (i in seq_len(queryl)) {
                  matches[[i]] <-
                      .getMatchesMzRt(i, query_mz[i], query_rt[i],
                                      target = target_mz,
                                      tolerance = param@tolerance,
                                      ppm = param@ppm,
                                      toleranceRt =param@toleranceRt)
              }
              Matched(query = query, target = target,
                      matches = do.call(rbind, matches),
                      queryAssay = queryAssay, targetAssay = targetAssay,
                      metadata = list(param = param))
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "data.frameOrSimilar",
                        target = "data.frameOrSimilar",
                        param = "MzRtParam"),
          function(query, target, param, mzColname = c("mz", "mz"),
                   rtColname = c("rt", "rt"), queryAssay = character(),
                   targetAssay = character()) {
              if(length(mzColname) == 1)
                  mzColname <- rep(mzColname, 2)
              if(length(rtColname) == 1)
                  rtColname <- rep(rtColname, 2)
              query_ <- .objectToMatch(query, queryAssay,
                                       c(mzColname[1], rtColname[1]))
              target_ <- .objectToMatch(target, targetAssay,
                                        c(mzColname[2], rtColname[2]))
              target_mz <- data.frame(index = seq_len(nrow(target_)),
                                      mz = target_[, 1], rt = target_[, 2])
              queryl <- nrow(query_)
              matches <- vector("list", queryl)
              query_mz <- query_[, 1]
              query_rt <- query_[, 2]
              for (i in seq_len(queryl)) {
                  matches[[i]] <-
                      .getMatchesMzRt(i, query_mz[i],
                                      query_rt[i],
                                      target = target_mz,
                                      tolerance = param@tolerance,
                                      ppm = param@ppm,
                                      toleranceRt = param@toleranceRt)
              }
              Matched(query = query, target = target,
                      matches = do.call(rbind, matches),
                      queryAssay = queryAssay, targetAssay = targetAssay,
                      metadata = list(param = param))
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "data.frameOrSimilar",
                        target = "Spectra",
                        param = "MzRtParam"),
          function(query, target, param, mzColname = c("mz", "mz"),
                   rtColname = c("rt", "rt"), queryAssay = character(),
                   targetAssay = character()) {
              if(length(mzColname) == 1)
                  mzColname <- rep(mzColname, 2)
              if(length(rtColname) == 1)
                  rtColname <- rep(rtColname, 2)
              if (mzColname[2L] %in% peaksVariables(target))
                  stop("Matching against the peak variable \"", mzColname[2L],
                       "\" is not supported. Please use on of the available ",
                       "spectra variables ('spectraVariables(target)').")
              target_ <- .objectToMatch(target, targetAssay,
                                        c(mzColname[2L], rtColname[2L]))
              res <- matchValues(query = query, target = target_, param = param,
                                 mzColname = mzColname, rtColname = rtColname,
                                 queryAssay = queryAssay,
                                 targetAssay = targetAssay)
              res@target <- target
              res@targetAssay <- targetAssay
              res
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "numeric",
                        target = "numeric",
                        param = "Mz2MassParam"),
          function(query, target, param) {
              query_mass <- .mz_to_mass_df(query, param@queryAdducts)
              target_mass<- .mz_to_mass_df(target, param@targetAdducts)
              queryl <- nrow(query_mass)
              matches <- vector("list", queryl)
              for(i in seq_len(queryl)) {
                  matches[[i]] <- .getMatches(query_mass$index[i],
                                              query_mass$mass[i],
                                              target = target_mass,
                                              tolerance = param@tolerance,
                                              ppm = param@ppm)
                  if (nrow(matches[[i]]))
                      matches[[i]]$query_adduct <- query_mass$adduct[i]
                  else matches[[i]]$query_adduct <- character()
              }
              matches <- do.call(rbind, matches)
              colnames(matches)[3] <- "target_adduct"
              Matched(query = query, target = target,
                      matches = matches[, c(1, 2, 6, 3, 4, 5)],
                      metadata = list(param = param))

          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "numeric",
                        target = "data.frameOrSimilar",
                        param = "Mz2MassParam"),
          function(query, target, param, mzColname = "mz",
                   targetAssay = character()) {
              target_ <- .objectToMatch(target, targetAssay, mzColname)
              res <- matchValues(query, target_, param)
              res@target <- target
              res@targetAssay <- targetAssay
              res
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "data.frameOrSimilar",
                        target = "numeric",
                        param = "Mz2MassParam"),
          function(query, target, param, mzColname = "mz",
                   queryAssay = character()) {
              query_ <- .objectToMatch(query, queryAssay, mzColname)
              res <- matchValues(query_, target, param)
              res@query <- query
              res@queryAssay <- queryAssay
              res
          })
#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "data.frameOrSimilar",
                        target = "data.frameOrSimilar",
                        param = "Mz2MassParam"),
          function(query, target, param, mzColname = c("mz", "mz"),
                   queryAssay = character(), targetAssay = character()) {
              if(length(mzColname) == 1)
                  mzColname <- rep(mzColname, 2)
              query_ <- .objectToMatch(query, queryAssay, mzColname[1])
              target_ <- .objectToMatch(target, targetAssay, mzColname[2])
              res <- matchValues(query_, target_, param)
              res@query <- query
              res@queryAssay <- queryAssay
              res@target <- target
              res@targetAssay <- targetAssay
              res
          })

#' @rdname matchValues
setMethod("matchValues",
          signature = c(query = "data.frameOrSimilar",
                        target = "data.frameOrSimilar",
                        param = "Mz2MassRtParam"),
          function(query, target, param, mzColname = c("mz", "mz"),
                   rtColname = c("rt", "rt"), queryAssay = character(),
                   targetAssay = character()) {
              if(length(mzColname) == 1)
                  mzColname <- rep(mzColname, 2)
              if(length(rtColname) == 1)
                  rtColname <- rep(rtColname, 2)
              query_ <- .objectToMatch(query, queryAssay,
                                       c(mzColname[1], rtColname[1]))
              target_ <- .objectToMatch(target, targetAssay,
                                        c(mzColname[2], rtColname[2 ]))
              query_mass <- .mz_to_mass_df(query_[, mzColname[1]],
                                           param@queryAdducts)
              query_mass$rt <- rep(query_[, rtColname[1]],
                                   .nelements(param@queryAdducts))
              target_mass<- .mz_to_mass_df(target_[, mzColname[2]],
                                           param@targetAdducts)
              target_mass$rt <- rep(target_[, rtColname[2]],
                                    .nelements(param@targetAdducts))
              queryl <- nrow(query_mass)
              matches <- vector("list", queryl)
              for (i in seq_len(queryl)) {
                  matches[[i]] <-
                      .getMatchesMzRt(query_mass$index[i],
                                      query_mass$mass[i],
                                      query_mass$rt[i],
                                      target = target_mass,
                                      tolerance = param@tolerance,
                                      ppm = param@ppm,
                                      toleranceRt = param@toleranceRt)
                  if (nrow(matches[[i]]))
                      matches[[i]]$query_adduct <- query_mass$adduct[i]
                  else matches[[i]]$query_adduct <- character()
              }
              matches <- do.call(rbind, matches)
              colnames(matches)[3] <- "target_adduct"
              Matched(query = query, target = target,
                      matches = matches[, c(1, 2, 7, 3, 4, 5, 6)],
                      queryAssay = queryAssay, targetAssay = targetAssay,
                      metadata = list(param = param))
          })

#' @author Andrea Vicini
#'
#' @param queryIndex `integer(1)` with the index of the query.
#'
#' @param queryMz `numeric(1)` with the m/z of the query.
#'
#' @param target `data.frame` with columns `"index"`, `"mz"` and optionally
#'     `"adduct"`.
#'
#' @noRd
.getMatches <- function(queryIndex, queryMz, target, tolerance, ppm){
    diffs <- queryMz - target[, 2]
    absdiffs <- abs(diffs)
    cls <- which(absdiffs <= (tolerance + ppm(queryMz, ppm)))
    if ("adduct" %in% colnames(target)){
        if (length(cls))
            data.frame(query_idx = queryIndex,
                       target_idx = target$index[cls],
                       adduct = target$adduct[cls],
                       score = diffs[cls],
                       ppm_error = absdiffs[cls] / target[cls, 2] * 10^6)
        else data.frame(query_idx = integer(),
                        target_idx = integer(),
                        adduct = character(),
                        score = numeric(),
                        ppm_error = numeric())
    } else {
        if (length(cls))
            data.frame(query_idx = queryIndex,
                       target_idx = target$index[cls],
                       score = diffs[cls],
                       ppm_error = absdiffs[cls] / target[cls, 2] * 10^6)
        else data.frame(query_idx = integer(),
                        target_idx = integer(),
                        score = numeric(),
                        ppm_error = numeric())
    }
}

#' @noRd
.getMatchesMzRt <- function(queryIndex, queryMz, queryRt, target, tolerance,
                            ppm, toleranceRt){
    diffs_rt <- queryRt - target$rt
    cls_rt <- which(abs(diffs_rt) <= toleranceRt)
    diffs <- queryMz - target[cls_rt, 2]
    absdiffs <- abs(diffs)
    cls <- which(absdiffs <= (tolerance + ppm(queryMz, ppm)))
    if ("adduct" %in% colnames(target)){
        if (length(cls))
            data.frame(query_idx = queryIndex,
                       target_idx = target$index[cls_rt[cls]],
                       adduct = target$adduct[cls_rt[cls]],
                       score = diffs[cls],
                       ppm_error = absdiffs[cls] /
                           target[cls_rt[cls], 2] * 10^6,
                       score_rt = diffs_rt[cls_rt[cls]])
        else data.frame(query_idx = integer(),
                        target_idx = integer(),
                        adduct = character(),
                        score = numeric(),
                        ppm_error = numeric(),
                        score_rt = numeric())
    } else {
        if (length(cls))
            data.frame(query_idx = queryIndex,
                       target_idx = target$index[cls_rt[cls]],
                       score = diffs[cls],
                       ppm_error = absdiffs[cls] /
                           target[cls_rt[cls], 2] * 10^6,
                       score_rt = diffs_rt[cls_rt[cls]])
        else data.frame(query_idx = integer(),
                        target_idx = integer(),
                        score = numeric(),
                        ppm_error = numeric(),
                        score_rt = numeric())
    }
}

#' creates a `data.frame` with columns `"index"`, `"adduct"` and `"mz"` from
#' provided mass values `x` and adduct definition `adducts`. `"index"` is the
#' index of the mass in the input vector `x`.
#'
#' @importFrom MetaboCoreUtils mass2mz
#'
#' @noRd
.mass_to_mz_df <- function(x, adducts) {
    mz <- mass2mz(x, adducts)
    data.frame(index = rep(seq_along(x), .nelements(adducts)),
               mz = as.numeric(mz),
               adduct = rep(colnames(mz), each = length(x)))
}

#' creates a `data.frame` with columns `"index"`, `"adduct"` and `"mass"` from
#' provided m/z values `x` and adduct definition `adducts`. `"index"` is the
#' index of the m/z in the input vector `x`.
#'
#' @importFrom MetaboCoreUtils mz2mass
#'
#' @noRd
.mz_to_mass_df <- function(x, adducts) {
    mass <- mz2mass(x, adducts)
    data.frame(index = rep(seq_along(x), .nelements(adducts)),
               mass = as.numeric(mass),
               adduct = rep(colnames(mass), each = length(x)))
}

.valid_adduct <- function(adducts, name = "`adducts`") {
    msg <- NULL
    if (is(adducts, "data.frame")) {
        if(any(!c("mass_add", "mass_multi") %in% colnames(adducts)))
            msg <- paste0("Columns \"mass_add\" and \"mass_multi\" must be ",
                          "present when ", name, " is a data.frame")
    } else {
        if (!all(adducts %in% c(adductNames("positive"),
                                adductNames("negative"))))
            msg <- paste0("Unknown adducts in ", name, " please check",
                          " MetaboCoreUtils for valid adducts")
    }
    msg
}

test_that("ValueParam works", {
    res <- ValueParam()
    expect_true(is(res, "ValueParam"))

    expect_error(ValueParam(tolerance = 1:3), "positive number")
    expect_error(ValueParam(ppm = -4), "positive number")
})

test_that("Mass2MzParam works", {
    res <- Mass2MzParam()
    expect_true(is(res, "Mass2MzParam"))

    expect_error(Mass2MzParam(tolerance = 1:3), "positive number")
    expect_error(Mass2MzParam(ppm = -4), "positive number")
    expect_error(Mass2MzParam(adducts = c("[M+H]+", "adduct2")), "Unknown")

    adds <- data.frame(mass_add = c(1, 2, 3), mass_multi = c(1, 2, 0.5))
    rownames(adds) <- c("a", "b", "c")
    res <- Mass2MzParam(adducts = adds)
    expect_true(is(res, "Mass2MzParam"))
})

test_that("Mass2MzRtParam works", {
    res <- Mass2MzRtParam()
    expect_true(is(res, "Mass2MzRtParam"))

    expect_error(Mass2MzRtParam(tolerance = 1:3), "positive number")
    expect_error(Mass2MzRtParam(toleranceRt = -1), "positive number")
    expect_error(Mass2MzRtParam(ppm = -4), "positive number")
    expect_error(Mass2MzRtParam(adducts = c("[M+H]+", "adduct2")), "Unknown")
})

test_that("MzParam works", {
    res <- MzParam()
    expect_true(is(res, "MzParam"))

    expect_error(MzParam(tolerance = 1:3), "positive number")
    expect_error(MzParam(ppm = -4), "positive number")
})

test_that("MzRtParam works", {
    res <- MzRtParam()
    expect_true(is(res, "MzRtParam"))

    expect_error(MzRtParam(tolerance = 1:3), "positive number")
    expect_error(MzRtParam(toleranceRt = -1), "positive number")
    expect_error(MzRtParam(ppm = -4), "positive number")
})

test_that("Mz2MassParam works", {
    res <- Mz2MassParam()
    expect_true(is(res, "Mz2MassParam"))

    expect_error(Mz2MassParam(tolerance = 1:3), "positive number")
    expect_error(Mz2MassParam(ppm = -4), "positive number")

    adds <- data.frame(mass_add = c(1, 2, 3), mass_multi = c(1, 2, 0.5))
    rownames(adds) <- c("a", "b", "c")
    res <- Mz2MassParam(queryAdducts = adds)
    expect_true(is(res, "Mz2MassParam"))
})

test_that("Mz2MassRtParam works", {
    res <- Mz2MassRtParam()
    expect_true(is(res, "Mz2MassRtParam"))

    expect_error(Mz2MassRtParam(tolerance = 1:3), "positive number")
    expect_error(Mz2MassRtParam(ppm = -4), "positive number")
    expect_error(Mz2MassRtParam(toleranceRt = -1), "positive number")

    adds <- data.frame(mass_add = c(1, 2, 3), mass_multi = c(1, 2, 0.5))
    rownames(adds) <- c("a", "b", "c")
    res <- Mz2MassRtParam(queryAdducts = adds)
    expect_true(is(res, "Mz2MassRtParam"))
})

test_that("matchValues, ValueParam works", {
    qry <- data.frame(value = c(150, 170, 179))
    trgt <- data.frame(value = seq(110, 200, 10))
    par <- ValueParam(tolerance = 0)

    ## numeric, data.frame
    expect_error(matchValues(qry[, "value"], trgt, par), "provided")
    expect_error(matchValues(qry[, "value"], trgt, par, valueColname = "some"),
                 "Missing")
    res <- matchValues(qry[, "value"], trgt, par, valueColname = "value")
    expect_s4_class(res, "Matched")
    expect_equal(res@metadata$param, par)
    expect_equal(res@query, qry[, "value"])
    expect_equal(res@target, trgt)

    ## data.frame, numeric
    expect_error(matchValues(qry, trgt[, "value"], par), "provided")
    expect_error(matchValues(qry, trgt[, "value"], par, valueColname = "some"),
                 "Missing")
    res <- matchValues(qry, trgt[, "value"], par, valueColname = "value")
    expect_s4_class(res, "Matched")
    expect_equal(res@metadata$param, par)
    expect_equal(res@query, qry)
    expect_equal(res@target, trgt[, "value"])

    ## data.frame, data.frame
    expect_error(matchValues(qry, trgt, par),
                 "`valueColname` has to be provided.")
    expect_error(matchValues(qry, trgt, par,
                         valueColname = c("notcolumn", "value")), "Missing")
    expect_error(matchValues(qry, trgt, par,
                         valueColname = c("value", "notcolumn")), "Missing")
    res <- matchValues(qry, trgt, par, valueColname = "value")
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt)
    expect_equal(res@matches$query_idx, c(1, 2))
    expect_equal(res@matches$target_idx, c(5, 7))
    expect_equal(res@matches$score, c(0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0))

    # no matches
    res <- matchValues(qry + 0.1, trgt, par, valueColname = "value")
    expect_true(is(res, "Matched"))
    expect_equal(query(res), qry + 0.1)
    expect_equal(target(res), trgt)
    expect_true(nrow(res@matches) == 0)

    # positive tolerance
    par <- ValueParam(tolerance = 10)
    res <- matchValues(qry, trgt, par, valueColname = "value")
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt)
    expect_equal(res@matches$query_idx, c(1, 1, 1, 2, 2, 2, 3, 3))
    expect_equal(res@matches$target_idx, c(4, 5, 6, 6, 7, 8, 7, 8))
    expect_equal(res@matches$score, c(10, 0, -10, 10, 0, -10, 9, -1))
    expect_equal(res@matches$ppm_error, c(10, 0, 10, 10, 0, 10, 9, 1) /
                     c(140, 150, 160, 160, 170, 180, 170, 180) * 10^6)

    ## data.frame, SummarizedExperiment
    trgt_se <- SummarizedExperiment(
        assays = data.frame(matrix(NA, 10, 2,
                                   dimnames = list(NULL, c("A", "B")))),
        rowData = trgt)
    par <- ValueParam(tolerance = 0)

    expect_error(matchValues(qry, trgt_se, par),
                 "`valueColname` has to be provided.")
    expect_error(matchValues(qry, trgt_se, par,
                             valueColname = c("notcolumn", "value")), "Missing")
    expect_error(matchValues(qry, trgt_se, par,
                             valueColname = c("value", "notcolumn")), "Missing")
    res <- matchValues(qry, trgt_se, par, valueColname = "value")
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt_se)
    expect_equal(res@matches$query_idx, c(1, 2))
    expect_equal(res@matches$target_idx, c(5, 7))
    expect_equal(res@matches$score, c(0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0))

    # no matches
    res <- matchValues(qry + 0.1, trgt_se, par, valueColname = "value")
    expect_true(is(res, "Matched"))
    expect_equal(query(res), qry + 0.1)
    expect_equal(target(res), trgt_se)
    expect_true(nrow(res@matches) == 0)

    # positive tolerance
    par <- ValueParam(tolerance = 10)
    res <- matchValues(qry, trgt_se, par, valueColname = "value")
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt_se)
    expect_equal(res@matches$query_idx, c(1, 1, 1, 2, 2, 2, 3, 3))
    expect_equal(res@matches$target_idx, c(4, 5, 6, 6, 7, 8, 7, 8))
    expect_equal(res@matches$score, c(10, 0, -10, 10, 0, -10, 9, -1))
    expect_equal(res@matches$ppm_error, c(10, 0, 10, 10, 0, 10, 9, 1) /
                     c(140, 150, 160, 160, 170, 180, 170, 180) * 10^6)

    ## data.frame, QFeatures
    trgt_qf <- QFeatures(list(a1 = trgt_se))
    par <- ValueParam(tolerance = 0)

    expect_error(matchValues(qry, trgt_qf, par), "`valueColname` has to be")
    expect_error(matchValues(qry, trgt_qf, par,
                             valueColname = c("notcolumn", "value")), "Missing")
    expect_error(matchValues(qry, trgt_qf, par, targetAssay = "a3",
                             valueColname = "value"), "No assay")
    expect_error(matchValues(qry, trgt_qf, par, targetAssay = "a1",
                             valueColname = c("value", "notcolumn")), "Missing")
    res <- matchValues(qry, trgt_qf, par, targetAssay = "a1",
                       valueColname = "value")
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt_qf)
    expect_equal(res@matches$query_idx, c(1, 2))
    expect_equal(res@matches$target_idx, c(5, 7))
    expect_equal(res@matches$score, c(0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0))

    # no matches
    res <- matchValues(qry + 0.1, trgt_qf, par, valueColname = "value",
                       targetAssay = "a1")
    expect_true(is(res, "Matched"))
    expect_equal(query(res), qry + 0.1)
    expect_equal(target(res), trgt_qf)
    expect_true(nrow(res@matches) == 0)

    # positive tolerance
    par <- ValueParam(tolerance = 10)
    res <- matchValues(qry, trgt_qf, par, valueColname = "value",
                       targetAssay = "a1")
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt_qf)
    expect_equal(res@matches$query_idx, c(1, 1, 1, 2, 2, 2, 3, 3))
    expect_equal(res@matches$target_idx, c(4, 5, 6, 6, 7, 8, 7, 8))
    expect_equal(res@matches$score, c(10, 0, -10, 10, 0, -10, 9, -1))
    expect_equal(res@matches$ppm_error, c(10, 0, 10, 10, 0, 10, 9, 1) /
                     c(140, 150, 160, 160, 170, 180, 170, 180) * 10^6)
})

test_that("matchValues,Mass2MzParam works", {

    cmpds <- data.frame(
        name = c("Tryptophan", "Leucine", "Isoleucine"),
        formula = c("C11H12N2O2", "C6H13NO2", "C6H13NO2"),
        exactmass = c(204.089878, 131.094629, 131.094629)
    )
    adducts <- c("[M+H]+", "[M+Na]+")
    x <- data.frame(
        mz = c(mass2mz(204.089878, "[M+H]+"),
               mass2mz(131.094629, "[M+H]+"),
               mass2mz(204.089878, "[M+Na]+") + 1e-6)
    )

    par <- Mass2MzParam(adducts = adducts, tolerance = 0, ppm = 20)

    ## data.frame, data.frame

    # zero tolerance, positive ppm
    expect_error(matchValues(x, cmpds, par, mzColname = "mzz"), "Missing")
    res <- matchValues(x, cmpds, par)
    expect_equal(query(res), x)
    expect_equal(target(res), cmpds)
    expect_equal(res@matches$query_idx, c(1, 2, 2, 3))
    expect_equal(res@matches$target_idx, c(1, 2, 3, 1))
    expect_equal(res@matches$score, c(0, 0, 0, 1e-6))
    expect_equal(res@matches$ppm_error,
                 c(0, 0, 0, 1 / mass2mz(204.089878, "[M+Na]+")))

    # zero tolerance and ppm
    par <- Mass2MzParam(adducts = adducts, tolerance = 0, ppm = 0)
    res <- matchValues(x, cmpds, par)
    expect_equal(query(res), x)
    expect_equal(target(res), cmpds)
    expect_equal(res@matches$query_idx, c(1, 2, 2))
    expect_equal(res@matches$target_idx, c(1, 2, 3))
    expect_equal(res@matches$score, c(0, 0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0, 0))

    # no matches
    adducts <- c("[M+Li]+", "[M+K]+")
    par <- Mass2MzParam(adducts = adducts, tolerance = 0, ppm = 20)
    res <- matchValues(x, cmpds, par)
    expect_true(is(res, "Matched"))
    expect_equal(query(res), x)
    expect_equal(target(res), cmpds)
    expect_true(nrow(res@matches) == 0)

    # with custom adduct definition, zero tolerance, positive ppm
    adducts <- data.frame(mass_add = c(1, 3), mass_multi = c(1, 0.5))
    rownames(adducts) <- c("a", "b")
    x <- data.frame(
        mz = c(mass2mz(204.089878, adducts["a", ]),
               mass2mz(131.094629, adducts["a", ]),
               mass2mz(204.089878, adducts["b", ]) - 2e-6)
    )

    par <- Mass2MzParam(adducts = adducts, tolerance = 0, ppm = 20)
    res <- matchValues(x, cmpds, par)
    expect_equal(query(res), x)
    expect_equal(target(res), cmpds)
    expect_equal(res@matches$query_idx, c(1, 2, 2, 3))
    expect_equal(res@matches$target_idx, c(1, 2, 3, 1))
    expect_equal(res@matches$score, c(0, 0, 0, -2e-6))
    expect_equal(res@matches$ppm_error,
                 c(0, 0, 0, 2 / mass2mz(204.089878, adducts["b", ])))

    # with custom adduct definition, zero tolerance, positive ppm
    par <- Mass2MzParam(adducts = adducts, tolerance = 0, ppm = 0)
    res <- matchValues(x, cmpds, par)
    expect_equal(query(res), x)
    expect_equal(target(res), cmpds)
    expect_equal(res@matches$query_idx, c(1, 2, 2))
    expect_equal(res@matches$target_idx, c(1, 2, 3))
    expect_equal(res@matches$score, c(0, 0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0, 0))

    expect_error(matchValues(x, cmpds, par, massColname = "other"), "other")

    ## numeric, data.frame
    res <- matchValues(x$mz, cmpds, par)
    expect_equal(query(res), x$mz)
    expect_equal(target(res), cmpds)
    expect_equal(res@matches$query_idx, c(1, 2, 2))
    expect_equal(res@matches$target_idx, c(1, 2, 3))
    expect_equal(res@matches$score, c(0, 0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0, 0))

    expect_error(matchValues(x$mz, cmpds, par, massColname = "other"), "other")

    ## data.frame, numeric
    res <- matchValues(x, cmpds$exactmass, par)
    expect_equal(query(res), x)
    expect_equal(target(res), cmpds$exactmass)
    expect_equal(res@matches$query_idx, c(1, 2, 2))
    expect_equal(res@matches$target_idx, c(1, 2, 3))
    expect_equal(res@matches$score, c(0, 0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0, 0))

    expect_error(matchValues(x, cmpds$exactmass, par, mzColname = "other"),
                 "other")

    # different adducts, no match
    adducts2 <- data.frame(mass_add = c(2, 4), mass_multi = c(2, 1))
    par <- Mass2MzParam(adducts = adducts2, tolerance = 0, ppm = 20)
    res <- matchValues(x, cmpds, par)
    expect_true(is(res, "Matched"))
    expect_equal(query(res), x)
    expect_equal(target(res), cmpds)
    expect_true(nrow(res@matches) == 0)

    ## SummarizedExperiment, data.frame
    par <- Mass2MzParam(adducts = adducts, tolerance = 0, ppm = 0)
    cmpds_se <- SummarizedExperiment(
        assays = data.frame(matrix(NA, 3, 2)),
        rowData = cmpds,
        colData = data.frame(cD1 = c(NA, NA),
                             cD2 = c(NA, NA)))
    res <- matchValues(x, cmpds_se, par)
    expect_equal(query(res), x)
    expect_equal(target(res), cmpds_se)
    expect_equal(res@matches$query_idx, c(1, 2, 2))
    expect_equal(res@matches$target_idx, c(1, 2, 3))
    expect_equal(res@matches$score, c(0, 0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0, 0))

    ## data.frame, QFeatures
    cmpds_qf <- QFeatures(list(a1 = cmpds_se))
    res <- matchValues(x, cmpds_qf, par, targetAssay = "a1")
    expect_equal(query(res), x)
    expect_equal(target(res), cmpds_qf)
    expect_equal(res@matches$query_idx, c(1, 2, 2))
    expect_equal(res@matches$target_idx, c(1, 2, 3))
    expect_equal(res@matches$score, c(0, 0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0, 0))
})

test_that(".getMatches works", {
    trgt <- data.frame(index = c(1, 2, 3, 1, 2, 3),
                       mz = c(1, 1.2, 1.21, 1.2, 2, 2.1),
                       adduct = c("A", "A", "A", "B", "B", "B"))
    trgt <- trgt[order(trgt$mz), ]

    res <- .getMatches(3, 1.2, trgt, tolerance = 0, ppm = 0)

    expect_true(is.data.frame(res))
    expect_equal(res$query_idx, c(3, 3))
    expect_equal(res$target_idx, c(2, 1))

    res <- .getMatches(4, 4, trgt, tolerance = 0, ppm = 0)
    expect_true(is.data.frame(res))
    expect_true(nrow(res) == 0)

    res <- .getMatches(3, 1.2, trgt, tolerance = 0.1, ppm = 0)
    expect_equal(res$target_idx, c(2, 1, 3))
})

test_that("matchValues,Mass2MzRtParam works", {

    cmpds <- data.frame(
        name = c("Tryptophan", "Leucine", "Isoleucine"),
        formula = c("C11H12N2O2", "C6H13NO2", "C6H13NO2"),
        exactmass = c(204.089878, 131.094629, 131.094629),
        rt = c(150, 140, 140)
    )
    adducts <- c("[M+H]+", "[M+Na]+")
    x <- data.frame(
        mz = c(mass2mz(204.089878, "[M+H]+"),
               mass2mz(131.094629, "[M+H]+"),
               mass2mz(204.089878, "[M+Na]+") + 1e-6),
        rt = c(150, 140, 150.1)
    )

    par <- Mass2MzRtParam(adducts = adducts, tolerance = 0, ppm = 20,
                          toleranceRt = 0)

    ## data.frame, data.frame
    res <- matchValues(x, cmpds, par, rtColname = "rt")
    expect_equal(query(res), x)
    expect_error(matchValues(x, cmpds, par, rtColname = c("r", "rt")), "Missin")
    expect_error(matchValues(x, cmpds, par, rtColname = c("rt", "r")), "Missin")
    expect_error(matchValues(x, cmpds, par, mzColname = "r"), "Missin")
    expect_error(matchValues(x, cmpds, par, massColname = "r"), "Missin")

    res <- matchValues(x, cmpds, par)
    expect_equal(query(res), x)
    expect_equal(target(res), cmpds)
    expect_equal(res@matches$query_idx, c(1, 2, 2))
    expect_equal(res@matches$target_idx, c(1, 2, 3))
    expect_equal(res@matches$score, c(0, 0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0, 0))
    expect_equal(res@matches$score_rt, c(0, 0, 0))

    par <- Mass2MzRtParam(adducts = adducts, tolerance = 0, ppm = 20,
                          toleranceRt = 0.2)
    res <- matchValues(x, cmpds, par)
    expect_equal(query(res), x)
    expect_equal(target(res), cmpds)
    expect_equal(res@matches$query_idx, c(1, 2, 2, 3))
    expect_equal(res@matches$target_idx, c(1, 2, 3, 1))
    expect_equal(res@matches$score, c(0, 0, 0, 1e-6))
    expect_equal(res@matches$ppm_error,
                 c(0, 0, 0, 1 / mass2mz(204.089878, "[M+Na]+")))
    expect_equal(res@matches$score_rt, c(0, 0, 0, 0.1))

    par <- Mass2MzRtParam(adducts = adducts, tolerance = 0, ppm = 0,
                          toleranceRt = 0.2)
    res <- matchValues(x, cmpds, par)
    expect_equal(query(res), x)
    expect_equal(target(res), cmpds)
    expect_equal(res@matches$query_idx, c(1, 2, 2))
    expect_equal(res@matches$target_idx, c(1, 2, 3))
    expect_equal(res@matches$score, c(0, 0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0, 0))
    expect_equal(res@matches$score_rt, c(0, 0, 0))

    ## no matches
    adducts2 <- c("[M+Li]+", "[M+K]+")
    par <- Mass2MzRtParam(adducts = adducts2, tolerance = 0, ppm = 20,
                          toleranceRt = 0.2)
    res <- matchValues(x, cmpds, par)
    expect_true(is(res, "Matched"))
    expect_equal(query(res), x)
    expect_equal(target(res), cmpds)
    expect_true(nrow(res@matches) == 0)

    ## QFeatures, SummarizedExperiment
    par <- Mass2MzRtParam(adducts = adducts, tolerance = 0, ppm = 20,
                          toleranceRt = 0)

    x_qf <- QFeatures(list(a1 = SummarizedExperiment(assays = data.frame(1:3),
                                                     rowData = x,
                                                     colData = data.frame(1))))
    cmpds_se <- SummarizedExperiment(assays = data.frame(2:4), rowData = cmpds,
                                     colData = data.frame(2))
    res <- matchValues(x_qf, cmpds_se, par, queryAssay = "a1")
    expect_equal(query(res), x_qf)
    expect_equal(target(res), cmpds_se)
    expect_equal(res@matches$query_idx, c(1, 2, 2))
    expect_equal(res@matches$target_idx, c(1, 2, 3))
    expect_equal(res@matches$score, c(0, 0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0, 0))
    expect_equal(res@matches$score_rt, c(0, 0, 0))


})

test_that(".getMatchesMzRt works", {
    trgt <- data.frame(index = rep(1:7, 2),
                       mz = c(11:17, 12:18),
                       adduct = c(rep("A", 7), rep("B", 7)),
                       rt = rep(21:27, 2))
    trgt <- trgt[order(trgt$mz), ]

    res <- .getMatchesMzRt(queryIndex = 3, queryMz = 13, queryRt = 23,
                           target = trgt, tolerance = 0, ppm = 0,
                           toleranceRt = 0)

    expect_true(is.data.frame(res))
    expect_equal(res$query_idx, c(3))
    expect_equal(res$target_idx, c(3))

    res <- .getMatchesMzRt(queryIndex = 3, queryMz = 13, queryRt = 24,
                           target = trgt, tolerance = 0, ppm = 0,
                           toleranceRt = 0)
    expect_true(is.data.frame(res))
    expect_true(nrow(res) == 0)

    res <- .getMatchesMzRt(queryIndex = 3, queryMz = 13, queryRt = 24,
                           target = trgt, tolerance = 0, ppm = 0,
                           toleranceRt = 1)
    expect_true(is.data.frame(res))
    expect_equal(res$query_idx, c(3))
    expect_equal(res$target_idx, c(3))

    res <- .getMatchesMzRt(queryIndex = 3, queryMz = 13, queryRt = 23,
                           target = trgt, tolerance = 1, ppm = 0,
                           toleranceRt = 1)
    expect_true(is.data.frame(res))
    expect_equal(res$query_idx, c(3, 3, 3, 3, 3))
    expect_equal(res$target_idx, c(2, 3, 2, 4, 3))
    expect_equal(res$adduct, c("A", "A", "B", "A", "B"))
    expect_equal(res$score, c(1, 0, 0, -1, -1))
    expect_equal(res$ppm_error, c(1, 0, 0, 1, 1) / c(12, 13, 13, 14, 14) * 10^6)
    expect_equal(res$score_rt, c(1, 0, 1, -1, 0))
})

test_that(".mass_to_mz_df works", {
    mass <- c(4, 2, 5, 10)
    adds <- MetaboCoreUtils::adductNames()
    res <- .mass_to_mz_df(mass, adducts = adds)
    expect_true(is.data.frame(res))
    expect_equal(colnames(res), c("index", "mz", "adduct"))
    expect_equal(res$index, rep(1:4, length(adds)))
    expect_equal(res$adduct, rep(adds, each = 4))

    adds <- MetaboCoreUtils::adducts()
    res1 <- .mass_to_mz_df(mass, adducts = adds)
    expect_true(is.data.frame(res))
    expect_equal(colnames(res), c("index", "mz", "adduct"))
    expect_equal(res$index, rep(1:4, nrow(adds)))
    expect_equal(res$adduct, rep(adds$name, each = 4))
})

test_that("matchValues, MzParam works", {
    qry <- data.frame(mz = c(150, 170, 179))
    trgt <- data.frame(mz = seq(110, 200, 10))
    par <- MzParam(tolerance = 0)

    ## numeric, data.frame
    expect_error(matchValues(qry$mz, trgt, par, mzColname = "mzz"), "Missing")
    res <- matchValues(qry$mz, trgt, par)
    expect_s4_class(res, "Matched")
    expect_equal(query(res), qry$mz)
    expect_equal(target(res), trgt)
    expect_equal(res@metadata$param, par)

    ## data.frame, numeric
    expect_error(matchValues(qry, trgt$mz, par, mzColname = "mzz"), "Missing")
    res <- matchValues(qry, trgt$mz, par)
    expect_s4_class(res, "Matched")
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt$mz)
    expect_equal(res@metadata$param, par)

    ## data.frame, data.frame
    res <- matchValues(qry, trgt, par, mzColname = "mz")
    expect_s4_class(res, "Matched")
    expect_error(matchValues(qry, trgt, par, mzColname = c("mzz", "mz")),
                 "Missing")
    expect_error(matchValues(qry, trgt, par, mzColname = c("mz", "mzz")),
                 "Missing")

    res <- matchValues(qry, trgt, par)
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt)
    expect_equal(res@matches$query_idx, c(1, 2))
    expect_equal(res@matches$target_idx, c(5, 7))
    expect_equal(res@matches$score, c(0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0))

    ## no matches
    res <- matchValues(qry + 0.1, trgt, par)
    expect_true(is(res, "Matched"))
    expect_equal(query(res), qry + 0.1)
    expect_equal(target(res), trgt)
    expect_true(nrow(res@matches) == 0)

    # positive tolerance
    par <- MzParam(tolerance = 10)
    res <- matchValues(qry, trgt, par)
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt)
    expect_equal(res@matches$query_idx, c(1, 1, 1, 2, 2, 2, 3, 3))
    expect_equal(res@matches$target_idx, c(4, 5, 6, 6, 7, 8, 7, 8))
    expect_equal(res@matches$score, c(10, 0, -10, 10, 0, -10, 9, -1))
    expect_equal(res@matches$ppm_error, c(10, 0, 10, 10, 0, 10, 9, 1) /
                     c(140, 150, 160, 160, 170, 180, 170, 180) * 10^6)

    ## numeric, QFeatures
    qry <- qry$mz
    trgt <- QFeatures(list(a1 = SummarizedExperiment(assays = data.frame(1:10),
                                                     rowData = trgt,
                                                     colData = data.frame(1))))

    par <- MzParam()
    res <- matchValues(qry, trgt, par, mzColname = "mz", targetAssay = "a1")
    expect_s4_class(res, "Matched")
    expect_error(matchValues(qry, trgt, par, mzColname = c("mzz", "mz"),
                             targetAssay = "a1"), "Missing")
    expect_error(matchValues(qry, trgt, par, mzColname = c("mz", "mzz"),
                             targetAssay = "a1"), "Missing")

    res <- matchValues(qry, trgt, par, targetAssay = "a1")
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt)
    expect_equal(res@matches$query_idx, c(1, 2))
    expect_equal(res@matches$target_idx, c(5, 7))
    expect_equal(res@matches$score, c(0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0))

    ## no matches
    res <- matchValues(qry + 0.1, trgt, par, targetAssay = "a1")
    expect_true(is(res, "Matched"))
    expect_equal(query(res), qry + 0.1)
    expect_equal(target(res), trgt)
    expect_true(nrow(res@matches) == 0)

    # positive tolerance
    par <- MzParam(tolerance = 10)
    res <- matchValues(qry, trgt, par, targetAssay = "a1")
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt)
    expect_equal(res@matches$query_idx, c(1, 1, 1, 2, 2, 2, 3, 3))
    expect_equal(res@matches$target_idx, c(4, 5, 6, 6, 7, 8, 7, 8))
    expect_equal(res@matches$score, c(10, 0, -10, 10, 0, -10, 9, -1))
    expect_equal(res@matches$ppm_error, c(10, 0, 10, 10, 0, 10, 9, 1) /
                     c(140, 150, 160, 160, 170, 180, 170, 180) * 10^6)
})

test_that("matchValues, MzRtParam works", {

    qry <- data.frame(mz = c(13, 14.1, 17, 18), rt = c(23, 24, 26.8, 23))
    trgt <- data.frame(mz = 11:17, rt = 21:27)

    par <- MzRtParam(tolerance = 0, ppm = 0, toleranceRt = 0)

    ## data.frame, data.frame
    res <- matchValues(qry, trgt, par, rtColname = "rt")
    expect_equal(query(res), qry)
    expect_error(matchValues(qry, trgt, par, mzColname = c("r", "mz")), "Mis")
    expect_error(matchValues(qry, trgt, par, mzColname = c("mz", "m")), "Mis")
    expect_error(matchValues(qry, trgt, par, rtColname = c("r", "rt")), "Mis")
    expect_error(matchValues(qry, trgt, par, rtColname = c("rt", "r")), "Mis")

    res <- matchValues(qry, trgt, par, mzColname = "mz")
    expect_equal(query(res), qry)
    expect_error(matchValues(qry, trgt, par, mzColname = c("r", "mz")), "Mis")
    expect_error(matchValues(qry, trgt, par, mzColname = c("mz", "m")), "Mis")
    expect_error(matchValues(qry, trgt, par, rtColname = c("r", "rt")), "Mis")
    expect_error(matchValues(qry, trgt, par, rtColname = c("rt", "r")), "Mis")

    res <- matchValues(qry, trgt, par)
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt)
    expect_equal(res@matches$query_idx, c(1))
    expect_equal(res@matches$target_idx, c(3))
    expect_equal(res@matches$score, c(0))
    expect_equal(res@matches$ppm_error, c(0))
    expect_equal(res@matches$score_rt, c(0))

    par <- MzRtParam(tolerance = 0.1, ppm = 0, toleranceRt = 0)
    res <- matchValues(qry, trgt, par)
    expect_equal(res@matches$query_idx, c(1, 2))
    expect_equal(res@matches$target_idx, c(3, 4))
    expect_equal(res@matches$score, c(0, 0.1))
    expect_equal(res@matches$ppm_error, c(0, 0.1 / 14 * 10^6))
    expect_equal(res@matches$score_rt, c(0, 0))

    par <- MzRtParam(tolerance = 0.1, ppm = 0, toleranceRt = 0.2)
    res <- matchValues(qry, trgt, par)
    expect_equal(res@matches$query_idx, c(1, 2, 3))
    expect_equal(res@matches$target_idx, c(3, 4, 7))
    expect_equal(res@matches$score, c(0, 0.1, 0))
    expect_equal(res@matches$ppm_error, c(0, 0.1 / 14 * 10^6, 0))
    expect_equal(res@matches$score_rt, c(0, 0, -0.2))

    # no matches
    res <- matchValues(qry + 0.5, trgt, par)
    expect_true(is(res, "Matched"))
    expect_equal(query(res), qry + 0.5)
    expect_equal(target(res), trgt)
    expect_true(nrow(res@matches) == 0)

    ## `SummarizedExperiment`, data.frame
    library(SummarizedExperiment)
    qry <- SummarizedExperiment(
        assays = data.frame(matrix(NA, 4, 2)),
        colData = data.frame(cD1 = c(NA, NA), cD2 = c(NA, NA)),
        rowData = data.frame(mz = c(13, 14.1, 17, 18),
                             rt = c(23, 24, 26.8, 23)))
    trgt <- data.frame(mz = 11:17, rt = 21:27)

    par <- MzRtParam(tolerance = 0, ppm = 0, toleranceRt = 0)
    res <- matchValues(qry, trgt, par)
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt)
    expect_equal(res@matches$query_idx, c(1))
    expect_equal(res@matches$target_idx, c(3))
    expect_equal(res@matches$score, c(0))
    expect_equal(res@matches$ppm_error, c(0))
    expect_equal(res@matches$score_rt, c(0))

    par <- MzRtParam(tolerance = 0.1, ppm = 0, toleranceRt = 0)
    res <- matchValues(qry, trgt, par)
    expect_equal(res@matches$query_idx, c(1, 2))
    expect_equal(res@matches$target_idx, c(3, 4))
    expect_equal(res@matches$score, c(0, 0.1))
    expect_equal(res@matches$ppm_error, c(0, 0.1 / 14 * 10^6))
    expect_equal(res@matches$score_rt, c(0, 0))

    par <- MzRtParam(tolerance = 0.1, ppm = 0, toleranceRt = 0.2)
    res <- matchValues(qry, trgt, par)
    expect_equal(res@matches$query_idx, c(1, 2, 3))
    expect_equal(res@matches$target_idx, c(3, 4, 7))
    expect_equal(res@matches$score, c(0, 0.1, 0))
    expect_equal(res@matches$ppm_error, c(0, 0.1 / 14 * 10^6, 0))
    expect_equal(res@matches$score_rt, c(0, 0, -0.2))

    # no matches
    res <- matchValues(qry , trgt + 0.5, par)
    expect_true(is(res, "Matched"))
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt + 0.5)
    expect_true(nrow(res@matches) == 0)

    ## QFeatures
    ions <- data.frame(id = c("a", "b", "c", "d", "e", "f"),
                       mz = c(3, 4, 3, 5, 5, 6),
                       rt = c(4, 4, 8, 8, 9, 10))
    cmps <- data.frame(ID = c("AB", "CD", "EF"),
                       mz = c(3.5, 4, 5.5),
                       rt = c(4, 8, 9.5))
    trgt <- data.frame(id = c("X", "Y", "Z"),
                       mz = c(3.5, 8, 5.5),
                       rt = c(4.5, 8.7, 10))
    ions <- SummarizedExperiment(matrix(rnorm(12), nrow = 6, ncol = 2,
                                        dimnames = list(NULL, c("a", "b"))),
                                 rowData = ions)
    cmps <- SummarizedExperiment(matrix(rnorm(6), nrow = 3, ncol = 2,
                                        dimnames = list(NULL, c("a", "b"))),
                                 rowData = cmps)
    qf <- QFeatures(list(ions = ions, compounds = cmps))
    prm <- MzRtParam(tolerance = 0.5, toleranceRt = 0.5)
    res <- matchValues(qf, trgt, queryAssay = "ions",
                       param = prm)
    expect_true(all(colnames(rowData(ions)) %in% colnames(res)))
    expect_equal(res$score, c(-0.5, 0.5, NA, NA, NA, 0.5))
    expect_equal(res$score_rt, c(-0.5, -0.5, NA, NA, NA, 0))
    expect_error(matchValues(qf, trgt, param = prm), "queryAssay")
    expect_equal(matchedData(res), matchedData(matchValues(ions, trgt, prm)))

    res <- matchValues(qf, trgt, queryAssay = "compounds",
                       param = prm)
    expect_equal(res$score, c(0, NA, 0))
    expect_equal(res$score_rt, c(-0.5, NA, -0.5))
    expect_equal(matchedData(res), matchedData(matchValues(cmps, trgt, prm)))

})

test_that("matchValues, Mz2MassParam works", {

    m <- c(200, 300)
    qry <- c(100, as.numeric(mass2mz(m, c("[M+H]+", "[M+K]+"))) + c(0, 0, 0, 5))
    trgt <- c(mass2mz(m, "[M-H]-"), 400, 500)

    par <- Mz2MassParam(queryAdducts = c("[M+H]+", "[M+K]+"),
                        targetAdducts = "[M-H]-")

    ## numeric, data.frame
    expect_error(
        matchMz(qry, data.frame(mz = trgt), par, mzColname = "m"), "Mis")
    res <- matchMz(qry, data.frame(mz = trgt), par)
    expect_s4_class(res, "Matched")
    expect_equal(query(res), qry)
    expect_equal(target(res), data.frame(mz = trgt))
    expect_equal(res@metadata$param, par)
    expect_true(is(matchedData(res), "DataFrame"))

    ## data.frame, numeric
    expect_error(
        matchMz(data.frame(mz = qry), trgt, par, mzColname = "m"), "Mis")
    res <- matchMz(data.frame(mz = qry), trgt, par)
    expect_s4_class(res, "Matched")
    expect_equal(query(res), data.frame(mz = qry))
    expect_equal(target(res), trgt)
    expect_equal(res@metadata$param, par)

    a <- data.frame(mz = qry)
    b <- data.frame(mz = trgt)
    ## data.frame, data.frame
    expect_error(matchMz(a, b, par, mzColname = c("m", "mz")), "Missing")
    expect_error(matchMz(a, b, par, mzColname = c("mz", "m")), "Missing")

    res <- matchValues(qry, trgt, par)
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt)
    expect_equal(res@matches$query_idx, c(2, 3, 4))
    expect_equal(res@matches$target_idx, c(1, 2, 1))
    expect_equal(res@matches$query_adduct, c("[M+H]+", "[M+H]+", "[M+K]+"))
    expect_equal(res@matches$target_adduct, rep("[M-H]-", 3))
    expect_equal(res@matches$score, c(0, 0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0, 0))

    res <- matchValues(a, b, par, mzColname = "mz")
    expect_equal(query(res), a)
    expect_equal(target(res), b)
    expect_equal(res@matches$query_idx, c(2, 3, 4))
    expect_equal(res@matches$target_idx, c(1, 2, 1))

    ## no matches
    res <- matchValues(qry + 0.1, trgt, par)
    expect_true(is(res, "Matched"))
    expect_equal(query(res), qry + 0.1)
    expect_equal(target(res), trgt)
    expect_true(nrow(res@matches) == 0)

    # positive tolerance
    par <- Mz2MassParam(queryAdducts = c("[M+H]+", "[M+K]+"),
                        targetAdducts = "[M-H]-",
                        tolerance = 10)
    res <- matchValues(qry, trgt, par)
    expect_equal(query(res), qry)
    expect_equal(target(res), trgt)
    expect_equal(res@matches$query_idx, c(2, 3, 4, 5))
    expect_equal(res@matches$target_idx, c(1, 2, 1, 2))
    expect_equal(res@matches$target_adduct, rep("[M-H]-", 4))
    expect_equal(res@matches$query_adduct, c(rep("[M+H]+", 2),
                                             rep("[M+K]+", 2)))
    expect_equal(res@matches$score, c(0, 0, 0, 5))
    expect_equal(res@matches$ppm_error, c(0, 0, 0, 5 / 300 * 10^6))

    # `query` and `target` data.frames
    qry_df <- data.frame(mzq = qry, other = 1)
    trgt_df <- data.frame(other = "b", mzt = trgt)
    res <- matchValues(qry_df, trgt_df, par, mzColname = c("mzq", "mzt"))
    expect_equal(query(res), qry_df)
    expect_equal(target(res), trgt_df)
    expect_equal(res@matches$query_idx, c(2, 3, 4, 5))
    expect_equal(res@matches$target_adduct, rep("[M-H]-", 4))
    expect_equal(res@matches$query_adduct, c(rep("[M+H]+", 2),
                                             rep("[M+K]+", 2)))
    expect_equal(res@matches$score, c(0, 0, 0, 5))
    expect_equal(res@matches$ppm_error, c(0, 0, 0, 5 / 300 * 10^6))

    par <- Mz2MassParam(queryAdducts = "[M]+", targetAdducts = "[M]+")
    res <- matchValues(qry, qry, par)
    expect_equal(res@matches$query_idx, 1:5)
    expect_equal(res@matches$target_idx, 1:5)

    res_2 <- matchValues(qry, qry, ValueParam())
    expect_equal(res_2@matches$query_idx, 1:5)
    expect_equal(res_2@matches$target_idx, 1:5)

    # QFeatures, data.frame
    qry_se <- SummarizedExperiment(assays = data.frame(1:5), rowData = qry_df,
                                   colData = data.frame(a = 1, b = 2))
    qry_qf <- QFeatures(list(a1 = qry_se))
    par <- Mz2MassParam(queryAdducts = c("[M+H]+", "[M+K]+"),
                        targetAdducts = "[M-H]-",
                        tolerance = 10)
    res <- matchValues(qry_df, trgt_df, par, mzColname = c("mzq", "mzt"))
    expect_equal(query(res), qry_df)
    expect_equal(target(res), trgt_df)
    expect_equal(res@matches$query_idx, c(2, 3, 4, 5))
    expect_equal(res@matches$target_adduct, rep("[M-H]-", 4))
    expect_equal(res@matches$query_adduct, c(rep("[M+H]+", 2),
                                             rep("[M+K]+", 2)))
    expect_equal(res@matches$score, c(0, 0, 0, 5))
    expect_equal(res@matches$ppm_error, c(0, 0, 0, 5 / 300 * 10^6))

    par <- Mz2MassParam(queryAdducts = "[M]+", targetAdducts = "[M]+")
    res <- matchValues(qry, qry, par)
    expect_equal(res@matches$query_idx, 1:5)
    expect_equal(res@matches$target_idx, 1:5)

    res_2 <- matchValues(qry, qry, ValueParam())
    expect_equal(res_2@matches$query_idx, 1:5)
    expect_equal(res_2@matches$target_idx, 1:5)

    ## numeric, numeric
    a <- 1:10
    b <- 1:5
    res <- matchValues(a, b, Mz2MassParam(queryAdducts = "[M]+",
                                          targetAdducts = "[M]-"))
    expect_equal(res@matches$target_idx, 1:5)
    expect_equal(res@matches$query_idx, 1:5)
    expect_true(all(res@matches$query_adduct == "[M]+"))
    expect_true(all(res@matches$target_adduct == "[M]-"))
    expect_s4_class(matchedData(res), "DataFrame")
    expect_equal(res$query, a)
    expect_equal(res$target, c(b, NA, NA, NA, NA, NA))
})

test_that("matchValues, Mz2MassRtParam works", {

    ## `query` and `target` data.frames
    m <- c(200, 300)
    qry <- c(100, as.numeric(mass2mz(m, c("[M+H]+", "[M+K]+"))) + c(0, 0, 0, 5))
    trgt <- c(mass2mz(m, "[M-H]-"), 400, 500)
    qry_df <- data.frame(mz = qry, rt = c(50, 100, 151, 100, 150))
    trgt_df <- data.frame(mz = trgt, rt = c(100, 150, 200, 250))

    par <- Mz2MassRtParam(queryAdducts = c("[M+H]+", "[M+K]+"),
                          targetAdducts = "[M-H]-")
    res <- matchValues(qry_df, trgt_df, par, mzColname = "mz")
    expect_s4_class(res, "Matched")
    res <- matchValues(qry_df, trgt_df, par, rtColname = "rt")
    expect_s4_class(res, "Matched")

    expect_error(
        matchValues(qry_df, trgt_df, par, mzColname = c("m", "mz")), "Mis")
    expect_error(
        matchValues(qry_df, trgt_df, par, mzColname = c("mz", "m")), "Mis")
    expect_error(
        matchValues(qry_df, trgt_df, par, rtColname = c("r", "rt")), "Mis")
    expect_error(
        matchValues(qry_df, trgt_df, par, rtColname = c("rt", "r")), "Mis")

    res <- matchValues(qry_df, trgt_df, par)
    expect_equal(query(res), qry_df)
    expect_equal(target(res), trgt_df)
    expect_equal(res@matches$query_idx, c(2, 4))
    expect_equal(res@matches$target_idx, c(1, 1))
    expect_equal(res@matches$query_adduct, c("[M+H]+", "[M+K]+"))
    expect_equal(res@matches$target_adduct, rep("[M-H]-", 2))
    expect_equal(res@matches$score, c(0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0))
    expect_equal(res@matches$score_rt, c(0, 0))
    md <- matchedData(res)
    expect_s4_class(md, "DataFrame")
    expect_true(all(c("mz", "rt", "target_mz", "target_rt",
                      "query_adduct", "target_adduct") %in% colnames(md)))

    # positive toleranceRt
    par <- Mz2MassRtParam(queryAdducts = c("[M+H]+", "[M+K]+"),
                          targetAdducts = "[M-H]-",
                          toleranceRt = 2)
    res <- matchValues(qry_df, trgt_df, par)
    expect_equal(query(res), qry_df)
    expect_equal(target(res), trgt_df)
    expect_equal(res@matches$query_idx, c(2, 3, 4))
    expect_equal(res@matches$target_idx, c(1, 2, 1))
    expect_equal(res@matches$query_adduct, c("[M+H]+", "[M+H]+", "[M+K]+"))
    expect_equal(res@matches$target_adduct, rep("[M-H]-", 3))
    expect_equal(res@matches$score, c(0, 0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0, 0))
    expect_equal(res@matches$score_rt, c(0, 1, 0))

    # positive tolerance and toleranceRt
    par <- Mz2MassRtParam(queryAdducts = c("[M+H]+", "[M+K]+"),
                          targetAdducts = "[M-H]-", tolerance = 10,
                          toleranceRt = 2)
    res <- matchValues(qry_df, trgt_df, par)
    expect_equal(query(res), qry_df)
    expect_equal(target(res), trgt_df)
    expect_equal(res@matches$query_idx, c(2, 3, 4, 5))
    expect_equal(res@matches$target_idx, c(1, 2, 1, 2))
    expect_equal(res@matches$query_adduct, c(rep("[M+H]+", 2),
                                             rep("[M+K]+", 2)))
    expect_equal(res@matches$target_adduct, rep("[M-H]-", 4))
    expect_equal(res@matches$score, c(0, 0, 0, 5))
    expect_equal(res@matches$ppm_error, c(0, 0, 0, 5 / 300 * 10^6))
    expect_equal(res@matches$score_rt, c(0, 1, 0, 0))

    ## SummarizedExperiment, QFeatures
    qry_se <- SummarizedExperiment(assays = data.frame(1:5), rowData = qry_df,
                                   colData = data.frame(1))
    trgt_se <- SummarizedExperiment(assays = data.frame(1:4), rowData = trgt_df,
                         colData = data.frame(1))
    trgt_qf <- QFeatures(list(a1 = trgt_se))
    par <- Mz2MassRtParam(queryAdducts = c("[M+H]+", "[M+K]+"),
                          targetAdducts = "[M-H]-")

    res <- matchValues(qry_se, trgt_qf, par, mzColname = "mz",
                       targetAssay = "a1")
    expect_s4_class(res, "Matched")
    res <- matchValues(qry_se, trgt_qf, par, rtColname = "rt",
                       targetAssay = "a1")
    expect_s4_class(res, "Matched")

    expect_error(matchValues(qry_se, trgt_qf, par, mzColname = c("m", "mz"),
                             targetAssay = "a1"), "Mis")
    expect_error(matchValues(qry_se, trgt_qf, par, mzColname = c("mz", "m"),
                             targetAssay = "a1"), "Mis")
    expect_error(matchValues(qry_se, trgt_qf, par, rtColname = c("r", "rt"),
                             targetAssay = "a1"), "Mis")
    expect_error(matchValues(qry_se, trgt_qf, par, rtColname = c("rt", "r"),
                             targetAssay = "a1"), "Mis")

    res <- matchValues(qry_se, trgt_qf, par, targetAssay = "a1")
    expect_equal(query(res), qry_se)
    expect_equal(target(res), trgt_qf)
    expect_equal(res@matches$query_idx, c(2, 4))
    expect_equal(res@matches$target_idx, c(1, 1))
    expect_equal(res@matches$query_adduct, c("[M+H]+", "[M+K]+"))
    expect_equal(res@matches$target_adduct, rep("[M-H]-", 2))
    expect_equal(res@matches$score, c(0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0))
    expect_equal(res@matches$score_rt, c(0, 0))
    md <- matchedData(res)
    expect_s4_class(md, "DataFrame")
    expect_true(all(c("mz", "rt", "target_mz", "target_rt",
                      "query_adduct", "target_adduct") %in% colnames(md)))

    # positive toleranceRt
    par <- Mz2MassRtParam(queryAdducts = c("[M+H]+", "[M+K]+"),
                          targetAdducts = "[M-H]-",
                          toleranceRt = 2)
    res <- matchValues(qry_se, trgt_qf, par, targetAssay = "a1")
    expect_equal(query(res), qry_se)
    expect_equal(target(res), trgt_qf)
    expect_equal(res@matches$query_idx, c(2, 3, 4))
    expect_equal(res@matches$target_idx, c(1, 2, 1))
    expect_equal(res@matches$query_adduct, c("[M+H]+", "[M+H]+", "[M+K]+"))
    expect_equal(res@matches$target_adduct, rep("[M-H]-", 3))
    expect_equal(res@matches$score, c(0, 0, 0))
    expect_equal(res@matches$ppm_error, c(0, 0, 0))
    expect_equal(res@matches$score_rt, c(0, 1, 0))

    # positive tolerance and toleranceRt
    par <- Mz2MassRtParam(queryAdducts = c("[M+H]+", "[M+K]+"),
                          targetAdducts = "[M-H]-", tolerance = 10,
                          toleranceRt = 2)
    res <- matchValues(qry_se, trgt_qf, par, targetAssay = "a1")
    expect_equal(query(res), qry_se)
    expect_equal(target(res), trgt_qf)
    expect_equal(res@matches$query_idx, c(2, 3, 4, 5))
    expect_equal(res@matches$target_idx, c(1, 2, 1, 2))
    expect_equal(res@matches$query_adduct, c(rep("[M+H]+", 2),
                                             rep("[M+K]+", 2)))
    expect_equal(res@matches$target_adduct, rep("[M-H]-", 4))
    expect_equal(res@matches$score, c(0, 0, 0, 5))
    expect_equal(res@matches$ppm_error, c(0, 0, 0, 5 / 300 * 10^6))
    expect_equal(res@matches$score_rt, c(0, 1, 0, 0))

    # no matches
    trgt_df$rt <- trgt_df$rt + 5
    res <- matchValues(qry_df, trgt_df, par)
    expect_true(is(res, "Matched"))
    expect_equal(query(res), qry_df)
    expect_equal(target(res), trgt_df)
    expect_true(nrow(res@matches) == 0)
})

test_that(".valid_adduct works", {
    expect_equal(.valid_adduct(c("A", "[M+H]+"), name = "`targetAdducts`"),
                 paste("Unknown adducts in `targetAdducts` please check",
                       "MetaboCoreUtils for valid adducts"))
    expect_null(.valid_adduct(c("[M+H]+")))
    expect_equal(.valid_adduct(data.frame(a = 1, b = 2)),
                 paste("Columns \"mass_add\" and \"mass_multi\" must be",
                       "present when `adducts` is a data.frame"))
    expect_null(.valid_adduct(data.frame(mass_add = 1, mass_multi = 2)))
})

test_that("matchValues,data.frame,Spectra,MzRtParam works", {
    df <- data.frame(mz = c(279.09, 200, 224.08, 100),
                     rt = c(379, 200, 378, 100))
    expect_error(matchValues(df, pest_ms2, MzRtParam(),
                             mzColname = c("mz", "mz")), "peak variable")
    expect_error(matchValues(df, pest_ms2, MzRtParam(),
                             mzColname = c("mz", "other")), "in target")
    expect_error(matchValues(df, pest_ms2, mzColname = c("mz", "precursorMz"),
                             MzRtParam(tolerance = 0.1,
                                       toleranceRt = 3)), "in target")
    res <- matchValues(df, pest_ms2, mzColname = c("mz", "precursorMz"),
                       MzRtParam(tolerance = 0.1, toleranceRt = 3),
                       rtColname = c("rt", "rtime"))
    expect_true(validObject(res))
    expect_equal(whichQuery(res), c(1L, 3L))
    expect_equal(whichTarget(res), c(4L, 9L, 7L, 11L))
    expect_equal(query(res), df)
    expect_equal(target(res), pest_ms2)

    res <- matchValues(df, pest_ms2, mzColname = c("mz", "precursorMz"),
                       rtColname = c("rt", "rtime"), param = MzRtParam())
    expect_equal(whichQuery(res), integer())
    expect_equal(whichTarget(res), integer())

    colnames(df) <- c("precursorMz", "rtime")
    res <- matchValues(df, pest_ms2, mzColname = c("precursorMz"),
                       MzRtParam(tolerance = 0.1, toleranceRt = 3),
                       rtColname = c("rtime"))
    expect_true(validObject(res))
    expect_equal(whichQuery(res), c(1L, 3L))
    expect_equal(whichTarget(res), c(4L, 9L, 7L, 11L))
    expect_equal(query(res), df)
    expect_equal(target(res), pest_ms2)

})

test_that("matchValues,data.frame,Spectra,MzParam works", {
    df <- data.frame(mz = c(279.09, 200, 224.08, 100),
                     rt = c(379, 200, 378, 100))
    expect_error(matchValues(df, pest_ms2, MzParam(),
                             mzColname = c("mz")), "peak variable")
    expect_error(matchValues(df, pest_ms2, MzParam(),
                             mzColname = c("mz", "other")), "in target")

    res <- matchValues(df, pest_ms2, MzParam(tolerance = 0.01),
                       mzColname = c("mz", "precursorMz"))

    expect_true(validObject(res))
    expect_equal(whichQuery(res), c(1L, 3L))
    expect_equal(whichTarget(res), c(4L, 9L, 7L, 11L))
    expect_equal(query(res), df)
    expect_equal(target(res), pest_ms2)
})

test_that("matchValues,numeric,Spectra,MzParam works", {
    mzs <- c(200, 400, 224.08, 124)

    expect_error(matchValues(mzs, pest_ms2, MzParam(),
                             mzColname = c("mz")), "peak variable")
    expect_error(matchValues(mzs, pest_ms2, MzParam(),
                             mzColname = c("other")), "in target")

    res <- matchValues(mzs, pest_ms2, MzParam(tolerance = 0.01),
                       mzColname = c("precursorMz"))

    expect_true(validObject(res))
    expect_equal(whichQuery(res), c(3L))
    expect_equal(whichTarget(res), c(7L, 11L))
    expect_equal(query(res), mzs)
    expect_equal(target(res), pest_ms2)
})

#' @rdname Matched
#'
#' @exportMethod addMatches
setGeneric("addMatches", function(object, ...)
  standardGeneric("addMatches"))

#' @rdname Matched
#'
#' @exportMethod endoapply
setGeneric("endoapply", function(X, FUN, ...)
    standardGeneric("endoapply"))

#' @rdname Matched
#'
#' @exportMethod filterMatches
setGeneric("filterMatches", function(object, param, ...)
  standardGeneric("filterMatches"))

#' @rdname Matched
#'
#' @exportMethod matchedData
setGeneric("matchedData", function(object, ...)
    standardGeneric("matchedData"))

#' @title Spectral matching
#'
#' @description
#'
#' The `matchSpectra` method matches (compares) spectra from `query` with those
#' from `target` based on settings specified with `param` and returns the result
#' from this as a [MatchedSpectra] object.
#'
#' @param query [Spectra::Spectra] object with the (experimental) spectra.
#'
#' @param target MS data to compare against. Can be another [Spectra::Spectra].
#'
#' @param param parameter object containing the settings for the matching (e.g.
#'     eventual prefiltering settings, cut-off value for similarity above which
#'     spectra are considered matching etc).
#'
#' @param ... optional parameters.
#'
#' @return a [MatchedSpectra] object with the spectra matching results.
#'
#' @author Johannes Rainer
#'
#' @seealso [CompareSpectraParam()] for the comparison between
#'     [Spectra::Spectra] objects.
#'
#' @export
setGeneric("matchSpectra", function(query, target, param, ...)
           standardGeneric("matchSpectra"))

#' @rdname Matched
#'
#' @exportMethod queryVariables
setGeneric("queryVariables", function(object, ...)
    standardGeneric("queryVariables"))

#' @rdname Matched
#'
#' @exportMethod targetVariables
setGeneric("targetVariables", function(object, ...)
    standardGeneric("targetVariables"))

#' @title Matching MS Spectra against a reference
#'
#' @aliases CompareSpectraParam-class MatchForwardReverseParam-class
#'
#' @name CompareSpectraParam
#'
#' @description
#'
#' `matchSpectra` compares experimental (*query*) MS2 spectra against
#' reference (*target*) MS2 spectra and reports matches with a similarity that
#' passing a specified threshold. The function performs the similarity
#' calculation between each query spectrum against each target spectrum.
#' Parameters `query` and `target` can be used to define the query and target
#' spectra, respectively, while parameter `param` allows to define and configure
#' the similarity calculation and matching condition. Parameter `query` takes
#' a [Spectra::Spectra] object while `target` can be either a
#' [Spectra::Spectra] object, a [CompoundDb::CompDb] (reference library)
#' object defined in the `CompoundDb` package or
#' a [CompAnnotationSource] (e.g. a [CompDbSource()])
#' with the reference or connection information to a supported annotation
#' resource).
#'
#' Some notes on performance and information on parallel processing are
#' provided in the vignette.
#'
#' Currently supported parameter objects defining the matching are:
#'
#' - `CompareSpectraParam`: the *generic* parameter object allowing to set all
#'   settings for the [Spectra::compareSpectra()] call that is used to
#'   perform the similarity calculation.
#'   This includes `MAPFUN` and `FUN` defining the peak-mapping and similarity
#'   calculation functions and `ppm` and `tolerance` to define an acceptable
#'   difference between m/z values of the compared peaks. Parameter
#'   `matchedPeaksCount` is also passed to `compareSpectra()` and, if set to
#'   `TRUE` (default is `FALSE`) will report the number of peaks defined to be
#'   *matching* by the `MAPFUN`.
#'   Additional parameters to the `compareSpectra` call can be passed along
#'   with `...`. See the help of [Spectra::Spectra()] for more information on
#'   these parameters. Importantly, if *msentropy* or a GNPS-like similarity
#'   calculation is used, `MAPFUN` should be selected accordingly (see section
#'   *Using alternative spectra similarity functions* in the package vignette
#'   for more information).
#'   By default, parameters `ppm` and `tolerance` are passed to the similarity
#'   calculation function, but if this function uses different parameters
#'   (e.g., `msentropy_similarity()` uses `ms2_tolerance_in_ppm` instead of
#'   `ppm`), these should be submitted to the `CompareSpectraParam()` function
#'   throught the `...` parameter.
#'   Parameters `requirePrecursor` (default `TRUE`) and `requirePrecursorPeak`
#'   (default `FALSE`) allow to pre-filter the target spectra prior to the
#'   actual similarity calculation for each individual query spectrum.
#'   Parameters `ppm` and `tolerance` are also used to define the maximal
#'   acceptable difference in precursor m/z if `requirePrecursor` or
#'   `requirePrecursorPeak` are set to `TRUE`.
#'   Target spectra can also be pre-filtered based on
#'   retention time if parameter `toleranceRt` is set to a value different than
#'   the default `toleranceRt = Inf`. Only target spectra with a retention time
#'   within the query's retention time +/- (`toleranceRt` + `percentRt`% of the
#'   query's retention time) are considered. Note that while for `ppm` and
#'   `tolerance` only a single value is accepted, `toleranceRt` and `percentRt`
#'   can be also of length equal to the number of query spectra hence allowing
#'   to define different rt boundaries for each query spectrum.
#'   While these pre-filters can considerably improve performance, it should be
#'   noted that no matches will be found between query and target spectra with
#'   missing values in the considered variable (precursor m/z or retention
#'   time). For target spectra without retention times (such as for `Spectra`
#'   from a public reference database such as MassBank) the default
#'   `toleranceRt = Inf` should thus be used.
#'   Finally, parameter `THRESHFUN` allows to define a function to be applied to
#'   the similarity scores to define which matches to report. See below for more
#'   details.
#'
#' - `MatchForwardReverseParam`: performs spectra matching as with
#'   `CompareSpectraParam` but reports, similar to MS-DIAL, also the *reverse*
#'   similarity score and the *presence ratio*. Please refer to the
#'   documentation of `CompareSpectraParam` for explanation of the parameters.
#'   With `MatchForwardReverseParam`, the matching of query
#'   spectra to target spectra is performed by considering all peaks from the
#'   query and all peaks from the target (reference) spectrum (i.e. *forward*
#'   matching using an *outer join*-based peak matching strategy). For matching
#'   spectra also the *reverse* similarity is calculated considering only peaks
#'   present in the target (reference) spectrum (i.e. using a *right join*-based
#'   peak matching). This is reported as spectra variable `"reverse_score"`.
#'   In addition, the ratio between the number of matched peaks and the total
#'   number of peaks in the target (reference) spectra is reported as the
#'   *presence ratio* (spectra variable `"presence_ratio"`) and the total
#'   number of matched peaks as `"matched_peaks_count"`. See examples below
#'   for details. Parameter `THRESHFUN_REVERSE` allows to define an additional
#'   *threshold function* to filter matches. If `THRESHFUN_REVERSE` is defined
#'   only matches with a spectra similarity fulfilling both `THRESHFUN` **and**
#'   `THRESHFUN_REVERSE` are returned. With the default
#'   `THRESHFUN_REVERSE = NULL` all matches passing `THRESHFUN` are reported.
#'
#' @param addOriginalQueryIndex for `matchSpectra()`: `logical(1)` whether an
#'   additional spectra variable `".original_query_index"` should be added to
#'   the `query` `Spectra` object providing the index of the spectrum in this
#'   originally provided object. This spectra variable can be useful to link
#'   back to the original `Spectra` object if the `MatchedSpectra` object gets
#'   subsetted/processed.
#'
#' @param BPPARAM for `matchSpectra`: parallel processing setup (see the
#'   `BiocParallel` package for more information). Parallel processing is
#'   disabled by default (with the default setting `BPPARAM = SerialParam()`).
#'
#' @param FUN `function` used to calculate similarity between spectra. Defaults
#'   for `CompareSpectraParam` to [MsCoreUtils::ndotproduct()]. See
#'   [MsCoreUtils::ndotproduct()] for details.
#'
#' @param MAPFUN `function` used to map peaks between the compared spectra.
#'    Defaults for `CompareSpectraParam` to [Spectra::joinPeaks()]. See
#'    [Spectra::compareSpectra()] for details.
#'
#' @param matchedPeaksCount `logical(1)` for `CompareSpectraParam()`: whether
#'    also the number of matching peaks should be reported (in column
#'    `"matched_peaks_count"`). This number represents the number of peaks
#'    reported *matching* by the `MAPFUN`.
#'
#' @param param for `matchSpectra`: parameter object (such as
#'   `CompareSpectraParam`) defining the settings for the matching.
#'
#' @param percentRt `numeric` of length 1 or equal to the number of query
#'   spectra defining the maximal accepted relative difference in retention
#'   time between query and target spectra expressed in percentage of the query
#'   rt. For `percentRt = 10`, similarities are defined between the query
#'   spectrum and all target spectra with a retention time within query rt
#'   +/- 10% of the query. By default (with `toleranceRt = Inf`) the retention
#'   time-based filter is not considered. Thus, to consider the `percentRt`
#'   parameter, `toleranceRt` should be set to a value different than that.
#'   See help of `CompareSpectraParam` above for more information.
#'
#' @param ppm `numeric(1)` for a relative, m/z-dependent, maximal accepted
#'   difference between m/z values. This will be used in `compareSpectra` as
#'   well as for eventual precursor m/z matching.
#'
#' @param query for `matchSpectra`: [Spectra::Spectra] object with the query
#'   spectra.
#'
#' @param requirePrecursor `logical(1)` whether only target spectra are
#'   considered in the similarity calculation with a precursor m/z that matches
#'   the precursor m/z of the query spectrum (considering also `ppm` and
#'   `tolerance`). With `requirePrecursor = TRUE` (the default) the function
#'   will complete much faster, but will not find any hits for target (or query
#'   spectra) with missing precursor m/z. It is suggested to check first the
#'   availability of the precursor m/z in `target` and `query`.
#'
#' @param requirePrecursorPeak `logical(1)` whether only target spectra will be
#'   considered in the spectra similarity calculation that have a peak with an
#'   m/z matching the precursor m/z of the query spectrum. Defaults to
#'   `requirePrecursorPeak = FALSE`. It is suggested to check first the
#'   availability of the precursor m/z in `query`, as no match will be reported
#'   for query spectra with missing precursor m/z.
#'
#' @param rtColname `character(2)` with the name of the spectra variable
#'     containing the retention time information for compounds to be used in
#'     retention time matching (only used if `toleranceRt` is not `Inf`).
#'     It can also be `character(1)` if the two names are the same.
#'     Defaults to `rtColname = c("rtime", "rtime")`.
#'
#' @param target for `matchSpectra`: [Spectra::Spectra], [CompoundDb::CompDb]
#'   or object extending [CompAnnotationSource] (such as [CompDbSource]) with
#'   the target (reference) spectra to compare `query` against.
#'
#' @param tolerance `numeric(1)` for an absolute maximal accepted difference
#'   between m/z values. This will be used in `compareSpectra` as well as for
#'   eventual precursor m/z matching.
#'
#' @param toleranceRt `numeric` of length 1 or equal to the number of query
#'   spectra defining the maximal accepted (absolute) difference in retention
#'   time between query and target spectra. By default
#'   (with `toleranceRt = Inf`) the retention time-based filter is not
#'   considered. See help of `CompareSpectraParam` above for more
#'   information.
#'
#' @param THRESHFUN `function` applied to the similarity score to define which
#'   target spectra are considered *matching*. Defaults to
#'   `THRESHFUN = function(x) which(x >= 0.7)` hence selects
#'   all target spectra matching a query spectrum with a similarity higher or
#'   equal than `0.7`. Any function that takes a numeric vector with similarity
#'   scores from the comparison of a query spectrum with all target spectra (as
#'   returned by [Spectra::compareSpectra()]) as input and returns a
#'   `logical` vector (same dimensions as the similarity scores) or an integer
#'   with the matches is supported.
#'
#' @param THRESHFUN_REVERSE for `MatchForwardReverseParam`: optional additional
#'   *thresholding function* to filter the results on the reverse score. If
#'   specified the same format than `THRESHFUN` is expected.
#'
#' @param ... for `CompareSpectraParam`: additional parameters passed along
#'   to the [Spectra::compareSpectra()] call, including eventual additional
#'   parameters of the selected mapping or similarity calculation functions.
#'
#' @author Johannes Rainer, Michael Witting
#'
#' @importClassesFrom ProtGenerics Param
#'
#' @importClassesFrom Spectra Spectra
#'
#' @importFrom MsCoreUtils ndotproduct
#'
#' @importFrom Spectra joinPeaks
#'
#' @rdname CompareSpectraParam
#'
#' @return `matchSpectra` returns a [MatchedSpectra()] object with the matching
#'   results. If `target` is a `CompAnnotationSource` only matching target
#'   spectra will be reported.
#'
#'   Constructor functions return an instance of the class.
#'
#' @exportClass CompareSpectraParam
#'
#' @examples
#'
#' library(Spectra)
#' library(msdata)
#' fl <- system.file("TripleTOF-SWATH", "PestMix1_DDA.mzML", package = "msdata")
#' pest_ms2 <- filterMsLevel(Spectra(fl), 2L)
#'
#' ## subset to selected spectra.
#' pest_ms2 <- pest_ms2[c(808, 809, 945:955)]
#'
#' ## Load a small example MassBank data set
#' load(system.file("extdata", "minimb.RData", package = "MetaboAnnotation"))
#'
#' ## Match spectra with the default similarity score (normalized dot product)
#' csp <- CompareSpectraParam(requirePrecursor = TRUE, ppm = 10)
#' mtches <- matchSpectra(pest_ms2, minimb, csp)
#'
#' mtches
#'
#' ## Are there any matching spectra for the first query spectrum?
#' mtches[1]
#' ## No
#'
#' ## And for the second query spectrum?
#' mtches[2]
#' ## The second query spectrum matches 4 target spectra. The scores for these
#' ## matches are:
#' mtches[2]$score
#'
#' ## To access the score for the full data set
#' mtches$score
#'
#' ## Below we use a THRESHFUN that returns for each query spectrum the (first)
#' ## best matching target spectrum.
#' csp <- CompareSpectraParam(requirePrecursor = FALSE, ppm = 10,
#'     THRESHFUN = function(x) which.max(x))
#' mtches <- matchSpectra(pest_ms2, minimb, csp)
#' mtches
#'
#' ## Each of the query spectra is matched to one target spectrum
#' length(mtches)
#' matches(mtches)
#'
#' ## Match spectra considering also measured retention times. This requires
#' ## that both query and target spectra have non-missing retention times.
#' rtime(pest_ms2)
#' rtime(minimb)
#'
#' ## Target spectra don't have retention times. Below we artificially set
#' ## retention times to show how an additional retention time filter would
#' ## work.
#' rtime(minimb) <- rep(361, length(minimb))
#'
#' ## Matching spectra requiring a matching precursor m/z and the difference
#' ## of retention times between query and target spectra to be <= 2 seconds.
#' csp <- CompareSpectraParam(requirePrecursor = TRUE, ppm = 10,
#'     toleranceRt = 2)
#' mtches <- matchSpectra(pest_ms2, minimb, csp)
#' mtches
#' matches(mtches)
#'
#' ## Note that parameter `rtColname` can be used to define different spectra
#' ## variables with retention time information (such as retention indices etc).
#'
#' ## A `CompDb` compound annotation database could also be used with
#' ## parameter `target`. Below we load the test `CompDb` database from the
#' ## `CompoundDb` Bioconductor package.
#' library(CompoundDb)
#' fl <- system.file("sql", "CompDb.MassBank.sql", package = "CompoundDb")
#' cdb <- CompDb(fl)
#' res <- matchSpectra(pest_ms2, cdb, CompareSpectraParam())
#'
#' ## We do however not find any matches since the used compound annotation
#' ## database contains only a very small subset of the MassBank.
#' res
#'
#' ## As `target` we have now however the MS2 spectra data from the compound
#' ## annotation database
#' target(res)
#'
#' ## See the package vignette for details, descriptions and more examples,
#' ## also on how to retrieve e.g. MassBank reference databases from
#' ## Bioconductor's AnnotationHub.
NULL

setClass("CompareSpectraParam",
         slots = c(
             MAPFUN = "function",
             tolerance = "numeric",
             ppm = "numeric",
             FUN = "function",
             dots = "list",
             requirePrecursor = "logical",
             requirePrecursorPeak = "logical",
             THRESHFUN = "function",
             toleranceRt = "numeric",
             percentRt = "numeric",
             matchedPeaksCount = "logical"
             ),
         contains = "Param",
         prototype = prototype(
             MAPFUN = joinPeaks,
             tolerance = 0,
             ppm = 5,
             FUN = MsCoreUtils::ndotproduct,
             dots = list(),
             requirePrecursor = TRUE,
             requirePrecursorPeak = FALSE,
             THRESHFUN = function(x) which(x >= 0.7),
             toleranceRt = Inf,
             percentRt = 0,
             matchedPeaksCount = FALSE
         ),
         validity = function(object) {
             msg <- NULL
             if (length(object@tolerance) > 1 || object@tolerance < 0)
                 msg <- c("'tolerance' has to be a positive number of length 1")
             if (length(object@ppm) > 1 || object@ppm < 0)
                 msg <- c("'ppm' has to be positive number of length 1")
             msg <- c(msg, .valid_threshfun(object@THRESHFUN))
             if (any(object@toleranceRt < 0))
                 msg <- c("'toleranceRt' has to be positive")
             if (any(object@percentRt < 0))
                 msg <- c("'percentRt' has to be positive")
             msg
         })

#' @importFrom methods .hasSlot
.matched_peaks_count <- function(x) {
    if (.hasSlot(x, "matchedPeaksCount"))
        x@matchedPeaksCount
    else FALSE
}

setClass("MatchForwardReverseParam",
         slots = c(THRESHFUN_REVERSE = "ANY"),
         contains = "CompareSpectraParam",
         prototype = prototype(
             THRESHFUN_REVERSE = NULL
         ),
         validity = function(object) {
             .valid_threshfun(object@THRESHFUN_REVERSE, "THRESHFUN_REVERSE")
         })

#' @rdname CompareSpectraParam
#'
#' @importMethodsFrom Spectra compareSpectra spectraNames<- precursorMz
#'
#' @importFrom MsCoreUtils ppm
#'
#' @importFrom BiocParallel bplapply
#'
#' @importFrom methods new
#'
#' @export
CompareSpectraParam <- function(MAPFUN = joinPeaks, tolerance = 0, ppm = 5,
                                FUN = MsCoreUtils::ndotproduct,
                                requirePrecursor = TRUE,
                                requirePrecursorPeak = FALSE,
                                THRESHFUN = function(x) which(x >= 0.7),
                                toleranceRt = Inf, percentRt = 0,
                                matchedPeaksCount = FALSE, ...) {
    new("CompareSpectraParam", MAPFUN = force(MAPFUN), tolerance = tolerance,
        ppm = ppm, FUN = force(FUN), requirePrecursor = requirePrecursor[1L],
        requirePrecursorPeak = requirePrecursorPeak[1L],
        THRESHFUN = force(THRESHFUN), toleranceRt = toleranceRt,
        percentRt = percentRt, matchedPeaksCount = matchedPeaksCount,
        dots = list(...))
}

#' @rdname CompareSpectraParam
#'
#' @export
MatchForwardReverseParam <- function(MAPFUN = joinPeaks, tolerance = 0, ppm = 5,
                                     FUN = MsCoreUtils::ndotproduct,
                                     requirePrecursor = TRUE,
                                     requirePrecursorPeak = FALSE,
                                     THRESHFUN = function(x) which(x >= 0.7),
                                     THRESHFUN_REVERSE = NULL,
                                     toleranceRt = Inf, percentRt = 0, ...) {
    dots <- list(...)
    if (any(names(dots) == "type")) {
        warning("Specifying a join type with parameter 'type' is not ",
                "supported. Will ignore parameter 'type'")
        dots$type <- NULL
    }
    new("MatchForwardReverseParam", MAPFUN = MAPFUN, tolerance = tolerance,
        ppm = ppm, FUN = FUN, requirePrecursor = requirePrecursor[1L],
        requirePrecursorPeak = requirePrecursorPeak[1L],
        THRESHFUN = THRESHFUN, THRESHFUN_REVERSE = THRESHFUN_REVERSE,
        toleranceRt = toleranceRt, percentRt = percentRt, dots = dots)
}

.valid_threshfun <- function(x, variable = "THRESHFUN") {
    msg <- NULL
    if (length(x)) {
        if (!is.function(x))
            return(paste0(variable, " needs to be a function."))
        res <- x(1:10)
        if (!is.logical(res) & !is.integer(res))
            return(paste0(variable,
                          " needs to return integer or logical vector."))
        if (is.logical(res) && length(res) != 10)
            return(paste0(variable, " needs to return a logical vector with",
                          " the same length than target spectra."))
        if (is.integer(res) && any(res < 1 | res > 10))
            return(paste0(variable, " needs to return an integer vector with ",
                          "values between 1 and the length of target spectra."))
    }
    msg
}

.compare_spectra_parms_list <- function(x) {
    c(list(MAPFUN = x@MAPFUN, tolerance = x@tolerance, ppm = x@ppm,
           FUN = x@FUN), x@dots)
}

#' @importMethodsFrom Spectra spectraNames containsMz filterPrecursorMzRange
#'
#' @importMethodsFrom Spectra backendBpparam
#'
#' @rdname CompareSpectraParam
#'
#' @export
setMethod(
    "matchSpectra",
    signature(query = "Spectra", target = "Spectra",
              param = "CompareSpectraParam"),
    function(query, target, param, rtColname = c("rtime", "rtime"),
             BPPARAM = BiocParallel::SerialParam(),
             addOriginalQueryIndex = TRUE) {
        BPPARAM <- .check_bpparam(query, target, BPPARAM)
        if (addOriginalQueryIndex) {
            if (any(spectraVariables(query) == ".original_query_index"))
                warning("Overwriting already present spectra variable ",
                        "\".original_query_index\"")
            query$.original_query_index <- seq_along(query)
        }
        if (length(query) == 1 || param@requirePrecursor ||
            param@requirePrecursorPeak || any(is.finite(param@toleranceRt)) ||
            any(param@percentRt != 0))
            .match_spectra(query, target, param, rtColname = rtColname, BPPARAM)
        else .match_spectra_without_precursor(query, target, param)
    })

#' Returns SerialParam if any of the backends does not support parallel
#' processing.
#'
#' @noRd
.check_bpparam <- function(query, target, BPPARAM) {
    BPPARAM <- backendBpparam(query, BPPARAM)
    if (!is(BPPARAM, "SerialParam"))
        BPPARAM <- backendBpparam(target, BPPARAM)
    BPPARAM
}

#' @importClassesFrom CompoundDb CompDb
#'
#' @rdname CompareSpectraParam
#'
#' @export
setMethod(
    "matchSpectra", signature(query = "Spectra", target = "CompDb",
                              param = "Param"),
    function(query, target, param, rtColname = c("rtime", "rtime"),
             BPPARAM = BiocParallel::SerialParam(),
             addOriginalQueryIndex = TRUE) {
        matchSpectra(query, Spectra(target), param = param,
                     rtColname = rtColname, BPPARAM = BPPARAM,
                     addOriginalQueryIndex = addOriginalQueryIndex)
    })

.match_spectra <- function(query, target, param,
                           rtColname = c("rtime", "rtime"), BPPARAM) {
    if (length(rtColname) != 2)
        rtColname <- rep(rtColname[1L], 2)
    parms <- .compare_spectra_parms_list(param)
    queryl <- length(query)
    toleranceRt <- param@toleranceRt
    percentRt <- param@percentRt
    if (length(toleranceRt) == 1L)
        toleranceRt <- rep(toleranceRt, queryl)
    if (length(percentRt) == 1L)
        percentRt <- rep(percentRt, queryl)
    if (length(percentRt) != queryl || length(toleranceRt) != queryl)
        stop("Length of 'toleranceRt' and 'percentRt' has to be either 1 or ",
             "equal to the number of query spectra")
    if (is.null(spectraNames(target)))
        spectraNames(target) <- seq_along(target)
    snames <- spectraNames(target)
    maps <- bplapply(seq_along(query), .get_matches_spectra,
                     query = query,
                     target = target,
                     parlist = parms,
                     THRESHFUN = param@THRESHFUN,
                     precMz = param@requirePrecursor,
                     precMzPeak = param@requirePrecursorPeak,
                     matchedPeaksCount = .matched_peaks_count(param),
                     toleranceRt = toleranceRt,
                     percentRt = percentRt,
                     query_rt_col = rtColname[1L],
                     target_rt_col = rtColname[2L],
                     sn = snames, BPPARAM = BPPARAM)
    maps <- do.call(rbind.data.frame, maps)
    if (!nrow(maps))
        maps <- data.frame(query_idx = integer(),
                           target_idx = integer(),
                           score = numeric())
    .matched_spectra(query = query, target = target, matches = maps,
                     metadata = list(param = param), validate = FALSE)
}

#' This version does not use a loop but performs the comparison all in one. It
#' can thus only be performed if no precursor check/filter is used.
#'
#' @noRd
.match_spectra_without_precursor <- function(query, target, param) {
    parlist <- .compare_spectra_parms_list(param)
    mpc <- .matched_peaks_count(param)
    cor <- do.call(compareSpectra,
                   c(list(x = query, y = target, matchedPeaksCount = mpc),
                     parlist))
    if (mpc) {
        pc <- cor[, , 2L]
        cor <- cor[, , 1L]
    }
    res <- lapply(seq_len(nrow(cor)), function(i) param@THRESHFUN(cor[i, ]))
    if (is.logical(res[[1L]]))
        res <- lapply(res, which)
    tidx <- unlist(res, use.names = FALSE)
    if (length(tidx)) {
        qidx <- rep(seq_along(res), lengths(res))
        maps <- data.frame(query_idx = qidx, target_idx = tidx,
                           score = cor[cbind(qidx, tidx)])
        if (mpc)
            maps$matched_peaks_count <- pc[cbind(qidx, tidx)]
    }
    else maps <- data.frame(query_idx = integer(), target_idx = integer(),
                            score = numeric())
    .matched_spectra(query = query, target = target, matches = maps,
                     metadata = list(param = param), validate = FALSE)
}

#' @importMethodsFrom Spectra filterRt rtime spectraData
#'
#' @importFrom MsCoreUtils between
#'
#' @return `data.frame` with matches or `NULL` (if not matches passed the
#'     filter).
#'
#' @noRd
.get_matches_spectra <- function(i, query, target, parlist, THRESHFUN, precMz,
                                 precMzPeak, toleranceRt = Inf,
                                 percentRt = 0, sn, query_rt_col = "rtime",
                                 target_rt_col = "rtime",
                                 matchedPeaksCount = FALSE) {
    qi <- query[i]
    if (is.finite(toleranceRt[i])) {
        rt_qi <- spectraData(qi, query_rt_col)[, 1L]
        if (toleranceRt[i] == 0 && percentRt[i] == 0)
            trt <- 0.0000001
        else
            trt <- toleranceRt[i] + rt_qi * percentRt[i] / 100
        if (target_rt_col != "rtime") {
            rt_t <- spectraData(target, target_rt_col)[, 1L]
            target <- target[between(rt_t, rt_qi + c(-trt, trt))]
        } else target <- filterRt(target, rt = rt_qi + c(-trt, trt))
    }
    if (precMz) {
        pmz <- precursorMz(qi)
        pmz <- pmz + c(-1, 1) * (ppm(pmz, parlist$ppm) + parlist$tolerance)
        target <- filterPrecursorMzRange(target, mz = pmz)
    }
    if (precMzPeak)
        target <- target[containsMz(target, mz = precursorMz(qi),
                                ppm = parlist$ppm,
                                tolerance = parlist$tolerance)]
    if (!length(target))
        return(NULL)
    cor <- base::do.call(compareSpectra,
                         c(list(x = qi, y = target, SIMPLIFY = FALSE,
                                matchedPeaksCount = matchedPeaksCount),
                           parlist))
    if (matchedPeaksCount) {
    ## Handle matched peaks count.
        keep <- THRESHFUN(as.vector(cor[, , 1L]))
        if (is.logical(keep))
            keep <- which(keep)
        kl <- length(keep)
        if (kl)
            data.frame(query_idx = rep(i, kl),
                       target_idx = base::match(spectraNames(target)[keep], sn),
                       score = as.vector(cor[, , 1L])[keep],
                       matched_peaks_count = as.integer(
                           as.vector(cor[, , 2L])[keep]))
        else NULL
    } else {
        keep <- THRESHFUN(as.vector(cor))
        if (is.logical(keep))
            keep <- which(keep)
        kl <- length(keep)
        if (kl)
            data.frame(query_idx = rep(i, kl),
                       target_idx = base::match(spectraNames(target)[keep], sn),
                       score = cor[keep])
        else NULL
    }
}

#' @rdname CompareSpectraParam
#'
#' @importFrom methods as
#'
#' @importMethodsFrom Spectra peaksData
#'
#' @export
setMethod(
    "matchSpectra",
    signature(query = "Spectra", target = "Spectra",
              param = "MatchForwardReverseParam"),
    function(query, target, param, rtColname = c("rtime", "rtime"),
             BPPARAM = BiocParallel::SerialParam(),
             addOriginalQueryIndex = TRUE) {
        res <- matchSpectra(query, target, as(param, "CompareSpectraParam"),
                            rtColname = rtColname, BPPARAM = BPPARAM,
                            addOriginalQueryIndex = addOriginalQueryIndex)
        ## Loop over the matches and assign additional stuff...
        nm <- nrow(res@matches)
        res@matches$reverse_score <- rep(NA_real_, nm)
        res@matches$presence_ratio <- rep(NA_real_, nm)
        res@matches$matched_peaks_count <- rep(NA_real_, nm)
        parms_rv <- .compare_spectra_parms_list(param)
        parms_rv$type <- "right"
        query_pmz <- precursorMz(query)
        target_pmz <- precursorMz(target)
        for (i in seq_len(nm)) {
            spl <- c(
                list(x = peaksData(query[res@matches$query_idx[i]])[[1L]],
                     y = peaksData(target[res@matches$target_idx[i]])[[1L]],
                     xPrecursorMz = query_pmz[res@matches$query_idx[i]],
                     yPrecursorMz = target_pmz[res@matches$target_idx[i]]),
                parms_rv)
            map <- do.call(param@MAPFUN, spl)
            spl$x <- map$x
            spl$y <- map$y
            cor <- do.call(param@FUN, spl)
            res@matches$reverse_score[i] <- cor
            nmatched <- sum(!is.na(map$x[, 1L]))
            res@matches$presence_ratio[i] <- nmatched /
                nrow(map$y)
            res@matches$matched_peaks_count[i] <- nmatched
        }
        if (length(param@THRESHFUN_REVERSE))
            res@matches <- res@matches[param@THRESHFUN_REVERSE(
                                                 res@matches$reverse_score), ,
                                       drop = FALSE]
        res@metadata$param <- param
        res
    })

test_that("CompareSpectraParam works", {
    res <- CompareSpectraParam()
    expect_true(is(res, "CompareSpectraParam"))

    expect_error(CompareSpectraParam(tolerance = 1:3), "positive number")
    expect_error(CompareSpectraParam(ppm = -4), "positive number")
    expect_error(CompareSpectraParam(toleranceRt = -1), "positive")
    expect_error(CompareSpectraParam(percentRt = c(-1, 1, 1)), "positive")

    res <- CompareSpectraParam(other_param = 5, b = 3)
    expect_equal(res@dots, list(other_param = 5, b = 3))

    res <- .compare_spectra_parms_list(res)
    expect_true(is.list(res))
    expect_equal(res$other_param, 5L)
    expect_equal(res$b, 3L)
    expect_equal(res$ppm, 5)
})

test_that(".valid_threshfun works", {
    res <- .valid_threshfun(function(x) which(x > 0.4))
    expect_true(length(res) == 0)
    res <- .valid_threshfun(function(x) x > 0.4)
    expect_true(length(res) == 0)
    res <- .valid_threshfun(NULL)
    expect_true(length(res) == 0)

    res <- .valid_threshfun("f")
    expect_match(res, "function")

    res <- .valid_threshfun(function(z) z / 3)
    expect_match(res, "integer")

    res <- .valid_threshfun(function(z) -1)
    expect_match(res, "integer")

    res <- .valid_threshfun(function(z) TRUE)
    expect_match(res, "logical")

    expect_match(.valid_threshfun(function(z) -1L), "values between 1 and")
})

test_that(".compare_spectra, .compare_spectra_without_precursor work", {
    prm <- CompareSpectraParam(requirePrecursor = FALSE,
                               THRESHFUN = function(x) x > 0.3)

    res <- .match_spectra(pest_ms2, minimb, prm, BPPARAM = SerialParam())
    res_2 <- .match_spectra_without_precursor(pest_ms2, minimb, prm)
    expect_equal(res@matches, res_2@matches)

    res_3 <- .match_spectra(pest_ms2, minimb, prm, BPPARAM = SerialParam(),
                            rtColname = "rtime")
    expect_equal(res@matches, res_3@matches)

    ## returns integer.
    prm <- CompareSpectraParam(requirePrecursor = FALSE,
                               THRESHFUN = function(x) which.max(x))
    res <- .match_spectra(pest_ms2, minimb, prm, BPPARAM = SerialParam())
    res_2 <- .match_spectra_without_precursor(pest_ms2, minimb, prm)
    expect_equal(res@matches, res_2@matches)

    ## no matches
    prm <- CompareSpectraParam(requirePrecursor = FALSE,
                               THRESHFUN = function (x) which(x > 20))
    res <- .match_spectra(pest_ms2, minimb, prm, BPPARAM = SerialParam())
    res_2 <- .match_spectra_without_precursor(pest_ms2, minimb, prm)
    expect_equal(res@matches, res_2@matches)
})

test_that(".get_matches_spectra, matchSpectra,CompareSpectraParam works", {
    csp <- CompareSpectraParam(
        requirePrecursor = FALSE,
        THRESHFUN = function(x) which.max(x))
    res <- matchSpectra(pest_ms2, minimb, csp)
    expect_equal(res@matches$query_idx, 1:13)
    expect_equal(length(unique(res$target_spectrum_id)), 11)
    expect_true(any(spectraVariables(res) == ".original_query_index"))
    expect_equal(res@query$.original_query_index, seq_along(pest_ms2))
    expect_warning(matchSpectra(res@query, minimb, csp), "Overwriting")

    res_2 <- matchSpectra(
        pest_ms2, minimb, CompareSpectraParam(
                              requirePrecursor = FALSE,
                              THRESHFUN = function(x) which.max(x),
                              matchedPeaksCount = TRUE))
    expect_equal(res@matches$query_idx, res_2@matches$query_idx)
    expect_equal(res@matches$target_idx, res_2@matches$target_idx)
    expect_equal(res@matches$score, res_2@matches$score)
    expect_equal(colnames(res_2@matches), c("query_idx", "target_idx", "score",
                                            "matched_peaks_count"))

    res_3 <- matchSpectra(
        pest_ms2, minimb, CompareSpectraParam(
                              requirePrecursor = TRUE,
                              THRESHFUN = function(x) which.max(x),
                              matchedPeaksCount = TRUE))
    expect_equal(colnames(res_3@matches), c("query_idx", "target_idx", "score",
                                            "matched_peaks_count"))

    res <- matchSpectra(pest_ms2, minimb, csp, addOriginalQueryIndex = FALSE)
    expect_false(any(spectraVariables(res) == ".original_query_index"))

    mb2 <- minimb
    spectraNames(mb2) <- seq_along(mb2)
    res <- .get_matches_spectra(
        2, pest_ms2, mb2,
        .compare_spectra_parms_list(csp),
        csp@THRESHFUN, precMz = csp@requirePrecursor,
        precMzPeak = csp@requirePrecursorPeak, sn = spectraNames(mb2))
    expect_true(is.data.frame(res))
    expect_equal(res$query_idx, 2L)
    ## With matchedPeaksCount
    res_2 <- .get_matches_spectra(
        2, pest_ms2, mb2,
        .compare_spectra_parms_list(csp),
        csp@THRESHFUN, precMz = csp@requirePrecursor,
        precMzPeak = csp@requirePrecursorPeak, sn = spectraNames(mb2),
        matchedPeaksCount = TRUE
        )
    expect_true(is.data.frame(res_2))
    expect_equal(res_2$query_idx, 2L)
    expect_equal(colnames(res_2),
                 c("query_idx", "target_idx", "score", "matched_peaks_count"))
    expect_equal(res$score, res_2$score)
    expect_equal(res$target_idx, res_2$target_idx)
    expect_true(res_2$matched_peaks_count > 0)

    res <- .get_matches_spectra(
        2, pest_ms2, mb2, .compare_spectra_parms_list(csp),
        csp@THRESHFUN, precMz = FALSE, precMzPeak = TRUE,
        sn = spectraNames(mb2))
    expect_true(is.data.frame(res))
    expect_equal(res$query_idx, 2L)
    expect_equal(res$target_idx, 73L)

    csp <- CompareSpectraParam(requirePrecursor = TRUE,
                               THRESHFUN = function(x) which.max(x))
    res <- matchSpectra(pest_ms2, minimb, csp)
    expect_equal(res@matches$query_idx, c(2, 4, 6, 8, 9))
    expect_true(anyDuplicated(res@matches$query_idx) == 0)

    csp <- CompareSpectraParam(requirePrecursor = TRUE)
    res <- matchSpectra(pest_ms2, minimb, csp)
    expect_equal(unique(res@matches$query_idx), c(2, 4, 6, 8))
    expect_true(anyDuplicated(res@matches$query_idx) == 2)

    csp <- CompareSpectraParam(toleranceRt = 10)
    res <- matchSpectra(pest_ms2, minimb, csp)
    expect_true(nrow(MetaboAnnotation::matches(res)) == 0)

    mb2 <- minimb
    mb2$rtime <- 350
    csp <- CompareSpectraParam(toleranceRt = Inf)
    res <- matchSpectra(pest_ms2, mb2, csp)
    csp <- CompareSpectraParam(toleranceRt = 100)
    res_2 <- matchSpectra(pest_ms2, mb2, csp)
    expect_equal(MetaboAnnotation::matches(res),
                 MetaboAnnotation::matches(res_2))
    mb2$rtime <- 361
    csp <- CompareSpectraParam(toleranceRt = 1)
    res <- matchSpectra(pest_ms2, mb2, csp)
    expect_true(all(res@matches$query_idx == 2))
    expect_equal(res@matches$target_idx, c(70, 73, 75))

    csp <- CompareSpectraParam(toleranceRt = c(1, 4))
    expect_error(matchSpectra(pest_ms2, mb2, csp), "equal to the number")
    csp <- CompareSpectraParam(toleranceRt = c(1, 2, 10, 0, 0, 0, 4,
                                               Inf, 4, 0, 0, 0, 0))
    res <- matchSpectra(pest_ms2, mb2, csp)
    expect_equal(res@matches$target_idx, c(70, 73, 75, 47, 51, 53, 59))

    csp <- CompareSpectraParam(percentRt = 3, toleranceRt = 0)
    res <- matchSpectra(pest_ms2, mb2, csp)
    expect_true(all(res@matches$query_idx == 2))
    expect_equal(res@matches$target_idx, c(70, 73, 75))

    csp <- CompareSpectraParam(percentRt = c(5, 3, 0, 0.1, 0.1, 0.1, 0,
                                             10, 0.1, 0, 0.1, 0, 0.1),
                               toleranceRt = 0)
    res <- matchSpectra(pest_ms2, mb2, csp)
    expect_equal(res@matches$target_idx, c(70, 73, 75, 47, 51, 53, 59))

    ## Alternative retention time column.
    csp <- CompareSpectraParam()
    mb2 <- minimb
    mb2$rtime <- 361
    spectraNames(mb2) <- seq_along(mb2)
    res <- .get_matches_spectra(
        2, pest_ms2, mb2,
        .compare_spectra_parms_list(csp),
        csp@THRESHFUN, precMz = csp@requirePrecursor,
        precMzPeak = csp@requirePrecursorPeak, sn = spectraNames(mb2),
        toleranceRt = rep(1, length(pest_ms2)),
        percentRt = rep(0, length(pest_ms2)))
    expect_equal(res$query_idx, c(2, 2, 2))
    expect_equal(res$target_idx, c(70, 73, 75))

    expect_error(.get_matches_spectra(
        2, pest_ms2, mb2,
        .compare_spectra_parms_list(csp),
        csp@THRESHFUN, precMz = csp@requirePrecursor,
        precMzPeak = csp@requirePrecursorPeak, sn = spectraNames(mb2),
        toleranceRt = rep(1, length(pest_ms2)),
        percentRt = rep(0, length(pest_ms2)),
        query_rt_col = "other"), "not available")

    expect_error(.get_matches_spectra(
        2, pest_ms2, mb2,
        .compare_spectra_parms_list(csp),
        csp@THRESHFUN, precMz = csp@requirePrecursor,
        precMzPeak = csp@requirePrecursorPeak, sn = spectraNames(mb2),
        toleranceRt = rep(1, length(pest_ms2)),
        percentRt = rep(0, length(pest_ms2)),
        target_rt_col = "other"), "not available")

    pest_ms2$other <- pest_ms2$rtime
    mb2$other <- mb2$rtime
    mb2$rtime <- 900

    res <- .get_matches_spectra(
        2, pest_ms2, mb2,
        .compare_spectra_parms_list(csp),
        csp@THRESHFUN, precMz = csp@requirePrecursor,
        precMzPeak = csp@requirePrecursorPeak, sn = spectraNames(mb2),
        toleranceRt = rep(1, length(pest_ms2)),
        percentRt = rep(0, length(pest_ms2)),
        query_rt_col = "other", target_rt_col = "other")
    expect_equal(res$query_idx, c(2, 2, 2))
    expect_equal(res$target_idx, c(70, 73, 75))
})

test_that("MatchForwardReverseParam works", {
    res <- MatchForwardReverseParam()
    expect_true(is(res, "MatchForwardReverseParam"))
    expect_equal(res@FUN, MsCoreUtils::ndotproduct)
    expect_equal(res@requirePrecursor, TRUE)

    expect_warning(res <- MatchForwardReverseParam(type = "left"), "supported")
    expect_true(!any(names(res@dots) == "type"))
})

test_that("matchSpectra,MatchForwardReverseParam works", {
    mp <- MatchForwardReverseParam(requirePrecursor = FALSE,
                                   THRESHFUN = function(x) which.max(x))
    res <- matchSpectra(pest_ms2, minimb, mp)
    expect_equal(res@matches$query_idx, 1:13)
    expect_equal(colnames(res@matches), c("query_idx", "target_idx", "score",
                                          "reverse_score", "presence_ratio",
                                          "matched_peaks_count"))
    expect_true(any(spectraVariables(res) == ".original_query_index"))
    expect_equal(query(res)$.original_query_index, seq_along(pest_ms2))

    mp <- MatchForwardReverseParam(requirePrecursor = TRUE,
                                   THRESHFUN = function(x) which.max(x))
    res <- matchSpectra(pest_ms2, minimb, mp)
    expect_equal(res@matches$query_idx, c(2, 4, 6, 8, 9))
    expect_true(anyDuplicated(res@matches$query_idx) == 0)
    expect_true(all(res$reverse_score > res$score, na.rm = TRUE))

    mp <- MatchForwardReverseParam(requirePrecursor = TRUE)
    res <- matchSpectra(pest_ms2, minimb, mp)
    expect_equal(unique(res@matches$query_idx), c(2, 4, 6, 8))

    mp <- MatchForwardReverseParam(requirePrecursor = TRUE,
                                   THRESHFUN_REVERSE = function(z) z > 0.9)
    res_2 <- matchSpectra(pest_ms2, minimb, mp)
    expect_true(all(res_2@matches$reverse_score > 0.9))
    expect_equal(unique(res_2@matches$query_idx), c(2, 4, 8))

    a <- data.frame(
        precursorMz = c(623.1618, 609.1825)
    )
    a$mz <- list(
        c(300.0276, 315.0511),
        c(242.0585, 286.0483, 301.0718)
    )
    a$intensity <- list(
        c(20, 100),
        c(1, 10, 100)
    )
    s <- Spectra(a)
    res <- matchSpectra(
        s[1], s[2],
        param =  MatchForwardReverseParam(
            requirePrecursor = FALSE, tolerance = 0.01,
            THRESHFUN = function(x) which(x >= 0),
            MAPFUN = joinPeaksGnps, FUN = MsCoreUtils::gnps))
    expect_equal(res@matches$matched_peaks_count, 2)
    ## Check that res@matches is not all 0

    res <- matchSpectra(
        s[1], s[2],
        param =  MatchForwardReverseParam(
            requirePrecursor = FALSE, tolerance = 1.1,
            THRESHFUN = function(x) which(x >= 0)))
    expect_equal(res@matches$matched_peaks_count, 1)
    expect_equal(res@matches$presence_ratio, 1/3)
})

test_that("matchSpectra,Spectra,CompDb works", {
    cdb <- new("CompDb")

    res <- matchSpectra(pest_ms2, cdb, CompareSpectraParam())
    expect_s4_class(res, "MatchedSpectra")
    expect_equal(length(res), 13)
    expect_s4_class(target(res), "Spectra")
    expect_equal(length(target(res)), 0)

    fl <- system.file("sql", "CompDb.MassBank.sql", package = "CompoundDb")
    cdb <- CompoundDb::CompDb(fl)
    res <- matchSpectra(pest_ms2, cdb, CompareSpectraParam())
    expect_s4_class(res, "MatchedSpectra")
    expect_equal(length(res), 13)
    expect_s4_class(target(res), "Spectra")
    expect_equal(length(target(res)), 70)

    expect_error(
        matchSpectra(pest_ms2, cdb, CompareSpectraParam(toleranceRt = 1),
                     rtColname = c("other", "other")), "not available")
})

test_that(".check_bpparam works", {
    res <- .check_bpparam(pest_ms2, minimb, SnowParam())
    expect_s4_class(res, "SnowParam")
})

test_that(".matched_peaks_count works", {
    p <- CompareSpectraParam()
    expect_false(.matched_peaks_count(p))
    p@matchedPeaksCount <- TRUE
    expect_true(.matched_peaks_count(p))
})

test_that("CompAnnotationSource works", {
    setClass("DummySource",
             contains = "CompAnnotationSource")
    d <- new("DummySource")
    expect_true(inherits(d, "CompAnnotationSource"))
    expect_equal(d@version, "0.1")
    expect_output(show(d), "DummySource")
})

test_that("matchSpectra,CompAnnotationSource works", {
    setClass("DummySource",
             contains = "CompAnnotationSource")
    d <- new("DummySource")
    expect_error(matchSpectra(pest_ms2, d, param = CompareSpectraParam()),
                 "Not implemented")
})

test_that("metadata,CompAnnotationSource works", {
    setClass("DummySource",
             contains = "CompAnnotationSource")
    d <- new("DummySource")
    expect_error(metadata(d), "Not implemented")
})

---
title: "Annotation of MS-based Metabolomics Data"
package: MetaboAnnotation
output:
  BiocStyle::html_document:
    toc_float: true
vignette: >
  %\VignetteIndexEntry{Annotation of MS-based Metabolomics Data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteKeywords{Mass Spectrometry, MS, MSMS, Metabolomics, Infrastructure, Quantitative }
  %\VignetteEncoding{UTF-8}
  %\VignetteDepends{Spectra,BiocStyle,msdata,MetaboAnnotation,SummarizedExperiment,QFeatures,AnnotationHub,microbenchmark,mzR,msentropy}
bibliography: references.bib
---

```{r style, echo = FALSE, results = 'asis'}
library(BiocStyle)
BiocStyle::markdown()
```

**Package**: `r BiocStyle::Biocpkg("MetaboAnnotation")`<br />
**Authors**: `r packageDescription("MetaboAnnotation")[["Author"]] `<br />
**Compiled**: `r date()`

```{r message = FALSE, warning = FALSE, echo = FALSE}
library(AnnotationHub)
library(MetaboAnnotation)
library(Spectra)
```


# Introduction

The `r Biocpkg("MetaboAnnotation")` package defines high-level user
functionality to support and facilitate annotation of MS-based metabolomics data
[@rainer_modular_2022].


# Installation

The package can be installed with the *BiocManager* package. To
install `BiocManager` use `install.packages("BiocManager")` and, after that,
`BiocManager::install("MetaboAnnotation")` to install this package.


# General description

*MetaboAnnotation* provides a set of *matching* functions that allow comparison
(and matching) between *query* and *target* entities. These entities can be
chemical formulas, numeric values (e.g. m/z or retention times) or fragment
spectra. The available matching functions are:

- `matchFormula()`: to match chemical formulas.
- `matchSpectra()`: to match fragment spectra.
- `matchValues()` (formerly `matchMz()`): to match numerical values (m/z,
  masses, retention times etc).

For each of these matching functions *parameter* objects are available that
allow different types or matching algorithms. Refer to the help pages for a
detailed listing of these (e.g. `?matchFormula`, `?matchSpectra` or
`?matchValues`). As a result, a `Matched` (or `MatchedSpectra`) object is
returned which streamlines and simplifies handling of the potential one-to-many
(or one-to-none) matching.


# Example use cases

The following sections illustrate example use cases of the functionality
provided by the *MetaboAnnotation* package.

```{r, message = FALSE}
library(MetaboAnnotation)
```


## Matching of m/z values

In this section a simple matching of feature m/z values against theoretical m/z
values is performed. This is the lowest level of confidence in metabolite
annotation. However, it gives ideas about potential metabolites that can be
analyzed in further downstream experiments and analyses.

The following example loads the feature table from a lipidomics experiments and
matches the measured m/z values against reference masses from LipidMaps. Below
we use a `data.frame` as *reference* database, but a `CompDb` compound database
instance (as created by the `r Biocpkg("CompoundDb")` package) would also be
supported.

```{r, message = FALSE}
ms1_features <- read.table(system.file("extdata", "MS1_example.txt",
                                       package = "MetaboAnnotation"),
                           header = TRUE, sep = "\t")
head(ms1_features)
target_df <- read.table(system.file("extdata", "LipidMaps_CompDB.txt",
                                    package = "MetaboAnnotation"),
                        header = TRUE, sep = "\t")
head(target_df)
```

For reference (target) compounds we have only the mass available. We need to
convert this mass to m/z values in order to match the m/z values from the
features (i.e. the query m/z values) against them. For this we need to define
the *most likely* ions/adducts that would be generated from the compounds based
on the ionization used in the experiment. We assume the most abundant adducts
from the compounds being `"[M+H]+"` and `"[M+Na]+`. We next perform the matching
with the `matchValues()` function providing the query and target data as well as
a parameter object (in our case a `Mass2MzParam`) with the settings for the
matching. With the `Mass2MzParam`, the mass or target compounds get first
converted to m/z values, based on the defined adducts, and these are then
matched against the query m/z values (i.e. the m/z values for the features). To
get a full list of supported adducts the `MetaboCoreUtils::adductNames(polarity
= "positive")` or `MetaboCoreUtils::adductNames(polarity = "negative")` can be
used). Note also, to keep the runtime of this vignette short, we match only the
first 100 features.

```{r}
parm <- Mass2MzParam(adducts = c("[M+H]+", "[M+Na]+"),
                           tolerance = 0.005, ppm = 0)

matched_features <- matchValues(ms1_features[1:100, ], target_df, parm)
matched_features
```

From the tested 100 features 55 were matched against at least one target
compound (all matches are against a single compound). The result object (of type
`Matched`) contains the full query data frame and target data frames as well as
the matching information. We can access the original query data with `query()`
and the original target data with `target()` function:

```{r}
head(query(matched_features))
head(target(matched_features))
```

Functions `whichQuery()` and `whichTarget()` can be used to identify the rows in
the query and target data that could be matched:

```{r}
whichQuery(matched_features)
whichTarget(matched_features)
```

The `colnames` function can be used to evaluate which variables/columns are
available in the `Matched` object.

```{r}
colnames(matched_features)
```

These are all columns from the `query`, all columns from the `target` (the
prefix `"target_"` is added to the original column names in `target`) and
information on the matching result (in this case columns `"adduct"`, `"score"`
and `"ppm_error"`).

We can extract the full matching table with `matchedData()`. This returns a
`DataFrame` with all rows in *query* the corresponding matches in *target* along
with the matching adduct (column `"adduct"`) and the difference in m/z (column
`"score"` for absolute differences and `"ppm_error"` for the m/z relative
differences). Note that if a row in *query* matches multiple elements in
*target*, this row will be duplicated in the `DataFrame` returned by
`matchedData()`.  For rows that can not be matched `NA` values are reported.

```{r}
matchedData(matched_features)
```

Individual columns can be simply extracted with the `$` operator:

```{r}
matched_features$target_name
```

`NA` is reported for query entries for which no match was found. See also the
help page for `?Matched` for more details and information. In addition to the
matching of query m/z against target exact masses as described above it would
also be possible to match directly query m/z against target m/z values by using
the `MzParam` instead of the `Mass2MzParam`.


## Matching of m/z and retention time values

If expected retention time values were available for the target compounds, an
annotation with higher confidence could be performed with `matchValues()` and a
`Mass2MzRtParam` parameter object. To illustrate this we randomly assign
retention times from query features to the target compounds adding also 2
seconds difference. In a real use case the target `data.frame` would contain
masses (or m/z values) for standards along with the retention times when ions of
these standards were measured on the same LC-MS setup from which the query data
derives.

Below we subset our data table with the MS1 features to the first 100 rows (to
keep the runtime of the vignette short).

```{r}
ms1_subset <- ms1_features[1:100, ]
head(ms1_subset)
```

The table contains thus retention times of the features in a column named
`"rtime"`.

Next we randomly assign retention times of the features to compounds in our
target data adding a deviation of 2 seconds. As described above, in a real use
case retention times are supposed to be determined by measuring the compounds
with the same LC-MS setup.

```{r}
set.seed(123)
target_df$rtime <- sample(ms1_subset$rtime,
                          nrow(target_df), replace = TRUE) + 2
```

We have now retention times available for both the query and the target data and
can thus perform a matching based on m/z **and** retention times. We use the
`Mass2MzRtParam` which allows us to specify (as for the
`Mass2MzParam`) the expected adducts, the maximal acceptable m/z relative
and absolute deviation as well as the maximal acceptable (absolute) difference
in retention times. We use the settings from the previous section and allow a
difference of 10 seconds in retention times. The retention times are provided in
columns named `"rtime"` which is different from the default (`"rt"`). We thus
specify the name of the column containing the retention times with parameter
`rtColname`.

```{r}
parm <- Mass2MzRtParam(adducts = c("[M+H]+", "[M+Na]+"),
                       tolerance = 0.005, ppm = 0,
                       toleranceRt = 10)
matched_features <- matchValues(ms1_subset, target_df, param = parm,
                                rtColname = "rtime")
matched_features
```

Less features were matched based on m/z and retention times.

```{r}
matchedData(matched_features)[whichQuery(matched_features), ]
```


## Matching of `SummarizedExperiment` or `QFeatures` objects

Results from LC-MS preprocessing (e.g. by the `r BiocStyle::Biocpkg("xcms")`
package) or generally metabolomics results might be best represented and bundled
as `SummarizedExperiment` or `QFeatures` objects (from the same-named
Bioconductor packages). A `XCMSnExp` preprocessing result from `xcms` can for
example be converted to a `SummarizedExperiment` using the `quantify()` method
from the `xcms` package. The feature definitions (i.e. their m/z and retention
time values) will then be stored in the object's `rowData()` while the assay
(the numerical matrix) will contain the feature abundances across all
samples. Such `SummarizedExperiment` objects can be simply passed as `query`
objects to the `matchValues()` method. To illustrate this, we create below a
simple `SummarizedExperiment` using the `ms1_features` data frame from the
example above as `rowData` and adding a `matrix` with random values as assay.

```{r}
library(SummarizedExperiment)

se <- SummarizedExperiment(
    assays = matrix(rnorm(nrow(ms1_features) * 4), ncol = 4,
                    dimnames = list(NULL, c("A", "B", "C", "D"))),
    rowData = ms1_features)
```

We can now use the same `matchValues()` call as before to perform the
matching. Matching will be performed on the object's `rowData`, i.e. each
row/element of the `SummarizedExperiment` will be matched against the target
using e.g. m/z values available in columns of the object's `rowData`:

```{r}
parm <- Mass2MzParam(adducts = c("[M+H]+", "[M+Na]+"),
                     tolerance = 0.005, ppm = 0)
matched_features <- matchValues(se, target_df, param = parm)
matched_features
```

As `query`, the result contains the full `SummarizedExperiment`, but
`colnames()` and `matchedData()` will access the respective information from the
`rowData` of this `SummarizedExperiment`:

```{r}
colnames(matched_features)
matchedData(matched_features)
```

Subsetting the result object, to e.g. just matched elements will also subset the
`SummarizedExperiment`.

```{r}
matched_sub <- matched_features[whichQuery(matched_features)]
MetaboAnnotation::query(matched_sub)
```

A `QFeatures` object is essentially a container for several
`SummarizedExperiment` objects which rows (features) are related with each
other. Such an object could thus for example contain the full feature data from
an LC-MS experiment as one assay and a compounded feature data in which data
from ions of the same compound are aggregated as an additional
assay. Below we create such an object using our `SummarizedExperiment` as an
assay of name `"features"`. For now we don't add any additional assay to that
`QFeatures`, thus, the object contains only this single data set.

```{r, warning = FALSE}
library(QFeatures)
qf <- QFeatures(list(features = se))
qf
```

`matchValues()` supports also matching of `QFeatures` objects but the user
needs to define the assay which should be used for the matching with the
`queryAssay` parameter.

```{r}
matched_qf <- matchValues(qf, target_df, param = parm, queryAssay = "features")
matched_qf
```

`colnames()` and `matchedData()` allow to access the `rowData` of the
`SummarizedExperiment` stored in the `QFeatures`' `"features"` assay:

```{r}
colnames(matched_qf)
matchedData(matched_qf)
```


## Matching of MS/MS spectra

In this section we match experimental MS/MS spectra against reference
spectra. This can also be performed with functions from the
`r BiocStyle::Biocpkg("Spectra")` package (see
[SpectraTutorials](https://jorainer.github.io/SpectraTutorials/), but the
functions and concepts used here are more suitable to the *end user* as they
simplify the handling of the spectra matching results.

Below we load spectra from a file from a reversed-phase (DDA) LC-MS/MS run of
the Agilent Pesticide mix. With `filterMsLevel()` we subset the data set to only
MS2 spectra. To reduce processing time of the example we further subset the
`Spectra` to a small set of selected MS2 spectra. In addition we assign *feature
identifiers* to each spectrum (again, for this example these are arbitrary IDs,
but in a *real* data analysis such identifiers could indicate to which LC-MS
feature these spectra belong).

```{r, message = FALSE}
library(Spectra)
library(msdata)
fl <- system.file("TripleTOF-SWATH", "PestMix1_DDA.mzML", package = "msdata")
pest_ms2 <- filterMsLevel(Spectra(fl), 2L)
## subset to selected spectra.
pest_ms2 <- pest_ms2[c(808, 809, 945:955)]
## assign arbitrary *feature IDs* to each spectrum.
pest_ms2$feature_id <- c("FT001", "FT001", "FT002", "FT003", "FT003", "FT003",
                         "FT004", "FT004", "FT004", "FT005", "FT005", "FT006",
                         "FT006")
## assign also *spectra IDs* to each
pest_ms2$spectrum_id <- paste0("sp_", seq_along(pest_ms2))
pest_ms2
```

This `Spectra` should now represent MS2 spectra associated
with LC-MS features from an untargeted LC-MS/MS experiment that we would like to
annotate by matching them against a spectral reference library.

We thus load below a `Spectra` object that represents MS2 data from a very small
subset of [MassBank](https://massbank.eu/MassBank/) release *2021.03*. This
small `Spectra` object is provided within this package but it would be possible
to use any other `Spectra` object with reference fragment spectra instead (see
also the [SpectraTutorials](https://jorainer.github.io/SpectraTutorials/)
workshop). As an alternative, it would also be possible to use a `CompDb` object
representing a compound annotation database (defined in the
`r Biocpkg("CompoundDb")` package) with parameter `target`. See the
`matchSpectra()` help page or section *Query against multiple reference
databases* below for more details and options to retrieve such annotation
resources from Bioconductor's `r Biocpkg("AnnotationHub")`.

```{r}
load(system.file("extdata", "minimb.RData", package = "MetaboAnnotation"))
minimb
```

We can now use the `matchSpectra()` function to match each of our experimental
*query* spectra against the *target* (reference) spectra. Settings for this
matching can be defined with a dedicated *param* object. We use below the
`CompareSpectraParam` that uses the `compareSpectra()` function from the
`Spectra` package to calculate similarities between each query spectrum and all
target spectra. `CompareSpectraParam` allows to set all individual settings for
the `compareSpectra()` call with parameters `MAPFUN`, `ppm`, `tolerance` and
`FUN` (see the help on `compareSpectra()` in the `r Biocpkg("Spectra")` package
for more details).
In addition, we can *pre-filter* the target spectra for each individual query
spectrum to speed-up the calculations. By setting `requirePrecursor = TRUE` we
compare below each query spectrum only to target spectra with matching
precursor m/z (accepting a deviation defined by parameters `ppm` and
`tolerance`). By default, `matchSpectra()` with `CompareSpectraParam`
considers spectra with a similarity score higher than 0.7 as *matching* and
these are thus reported. As an additional configuration we set
`matchedPeaksCount = TRUE` to report the number of matching peaks between the
compared (and matched) spectra.

```{r}
csp <- CompareSpectraParam(requirePrecursor = TRUE, ppm = 10,
                           matchedPeaksCount = TRUE)
mtches <- matchSpectra(pest_ms2, minimb, param = csp)
mtches
```

The results are reported as a `MatchedSpectra` object which represents the
matching results for all query spectra. This type of object contains all query
spectra, all target spectra, the matching information and the parameter object
with the settings of the matching. The object can be subset to results of
selected **query** spectra using the `[` method. The length of the result object
matches thus the number of query spectra:

```{r}
length(pest_ms2)
length(mtches)
```

To subset to the results for the first query spectrum:

```{r}
mtches[1]
```

In this case, for the first query spectrum, no match was found among the target
spectra. Below we subset the `MatchedSpectra` to results for the second query
spectrum:

```{r}
mtches[2]
```

The second query spectrum could be matched to 4 target spectra. The matching
between query and target spectra can be n:m, i.e. each query spectrum can match
no or multiple target spectra and each target spectrum can be matched to none,
one or multiple query spectra.

Data (spectra variables of either the query and/or the target spectra) can be
extracted from the result object with the `spectraData()` function or with `$`
(similar to a `Spectra` object). The `spectraVariables` function can be used to
list all available spectra variables in the result object:

```{r}
spectraVariables(mtches)
```

This lists the spectra variables from both the *query* **and** the *target*
spectra, with the prefix `"target_"` being used for spectra variable names of
the target spectra. Spectra variable `"score"` contains the similarity score. If
`matchedPeaksCount = TRUE` was used, a column `"matched_peaks_count"` is
available that lists the number of matching peaks between the compared spectra.

Note that by default also an additional column `".original_query_index"` is
added to the `query` `Spectra` object by the `matchSpectra()` function, that
enables an easier mapping of results to the *original* query object used as
input, in particular, if the `MatchedSpectra` object gets further subset. As the
name says, this column contains for each query spectrum the index in the
original `Spectra` object provided with the `query` parameter.

We could thus use `$target_compound_name` to extract the compound name of the
matching target spectra for the second query spectrum:

```{r}
mtches[2]$target_compound_name
```

The same information can also be extracted on the *full* `MatchedSpectra`.
Below we use `$spectrum_id` to extract the query spectra identifiers we added
above from the full result object.

```{r}
mtches$spectrum_id
```

We added this column manually to the query object before the `matchSpectra()`
call, but the automatically added spectra variable `".original_query_index"`
would provide the same information:

```{r}
mtches$.original_query_index
```

And the respective values in the query object:

```{r}
query(mtches)$.original_query_index
```

Because of the n:m mapping between query and target spectra, the number of
values returned by `$` (or `spectraData`) can be larger than the total number of
query spectra. Also in the example above, some of the spectra IDs are present
more than once in the result returned by `$spectrum_id`. The respective spectra
could be matched to more than one target spectrum (based on our settings) and
hence their IDs are reported multiple times. Both `spectraData` and `$` for
`MatchedSpectra` use a *left join* strategy to report/return values: a value
(row) is reported for each query spectrum (even if it does **not** match any
target spectrum) with eventually duplicated values (rows) if the query spectrum
matches more than one target spectrum (each value for a query spectrum is
repeated as many times as it matches target spectra). To illustrate this we
use below the `spectraData()` function to extract specific data from our
result object, i.e. the spectrum and feature IDs for the query spectra we
defined above, the MS2 spectra similarity score, and the target spectra's ID and
compound name.

```{r}
mtches_df <- spectraData(mtches, columns = c("spectrum_id", "feature_id",
                                             "score", "target_spectrum_id",
                                             "target_compound_name"))
as.data.frame(mtches_df)
```

Using the `plotSpectraMirror()` function we can visualize the matching results
for one query spectrum. Note also that an interactive, `shiny`-based, validation
of matching results is available with the `validateMatchedSpectra()`
function. Below we call this function to show all matches for the second
spectrum.

```{r}
plotSpectraMirror(mtches[2])
```

Not unexpectedly, the peak intensities of query and target spectra are on
different scales. While this was no problem for the similarity calculation (the
normalized dot-product which is used by default is independent of the absolute
peak values) it is not ideal for visualization. Thus, we apply below a simple
scaling function to both the query and target spectra and plot the
spectra again afterwards (see the help for `addProcessing()` in the `Spectra`
package for more details on spectra data manipulations). This function will
replace the absolute spectra intensities with intensities relative to the
maximum intensity of each spectrum. Note that functions for `addProcessing()`
should include (like in the example below) the `...` parameter.

```{r}
scale_int <- function(x, ...) {
    x[, "intensity"] <- x[, "intensity"] / max(x[, "intensity"], na.rm = TRUE)
    x
}
mtches <- addProcessing(mtches, scale_int)
plotSpectraMirror(mtches[2])
```

The query spectrum seems to nicely match the identified target spectra. Below we
extract the compound name of the target spectra for this second query spectrum.

```{r}
mtches[2]$target_compound_name
```

As alternative to the `CompareSpectraParam` we could also use the
`MatchForwardReverseParam` with `matchSpectra()`. This has the same settings and
performs the same spectra similarity search than `CompareSpectraParam`, but
reports in addition (similar to MS-DIAL) to the (*forward*) similarity score
also the *reverse* spectra similarity score as well as the *presence ratio* for
matching spectra. While the default *forward* score is calculated considering
all peaks from the query and the target spectrum (the peak mapping is performed
using an *outer join* strategy), the *reverse score* is calculated only on peaks
that are present in the target spectrum and the matching peaks from the query
spectrum (the peak mapping is performed using a *right join* strategy). The
*presence ratio* is the ratio between the number of mapped peaks between the
query and the target spectrum and the total number of peaks in the target
spectrum. These values are available as spectra variables `"reverse_score"` and
`"presence_ratio"` in the result object). Below we perform the same spectra
matching as above, but using the `MatchForwardReverseParam`.

```{r}
mp <- MatchForwardReverseParam(requirePrecursor = TRUE, ppm = 10)
mtches <- matchSpectra(pest_ms2, minimb, param = mp)
mtches
```

Below we extract the query and target spectra IDs, the compound name and all
scores.

```{r}
as.data.frame(
    spectraData(mtches, c("spectrum_id", "target_spectrum_id",
                          "target_compound_name", "score", "reverse_score",
                          "presence_ratio")))
```

In these examples we matched query spectra only to target spectra if their
precursor m/z is ~ equal and reported only matches with a similarity higher than
0.7. `CompareSpectraParam`, through its parameter `THRESHFUN` would however also
allow other types of analyses. We could for example also report the *best
matching* target spectrum for each query spectrum, independently of whether the
similarity score is higher than a certain threshold. Below we perform such an
analysis defining a `THRESHFUN` that selects always the best match.

```{r}
select_top_match <- function(x) {
    which.max(x)
}
csp2 <- CompareSpectraParam(ppm = 10, requirePrecursor = FALSE,
                            THRESHFUN = select_top_match)
mtches <- matchSpectra(pest_ms2, minimb, param = csp2)
res <- spectraData(mtches, columns = c("spectrum_id", "target_spectrum_id",
                                       "target_compound_name", "score"))
as.data.frame(res)
```

Note that this whole example would work on any `Spectra` object with MS2
spectra. Such objects could also be extracted from an `xcms`-based LC-MS/MS data
analysis with the `chromPeaksSpectra()` or `featureSpectra()` functions from the
`r Biocpkg("xcms")` package. Note also that retention times could in addition be
considered in the matching by selecting a non-infinite value for the
`toleranceRt` of any of the parameter classes. By default this uses the
retention times provided by the query and target spectra (i.e. spectra variable
`"rtime"`) but it is also possible to specify any other spectra variable for the
additional retention time matching (e.g. retention indices instead of times)
using the `rtColname` parameter of the `matchSpectra(0` function (see
`?matchSpectra` help page for more information).

Matches can be also further validated using an interactive Shiny app by calling
`validateMatchedSpectra()` on the `MatchedSpectra` object. Individual matches
can be set to TRUE or FALSE in this app. By closing the app via the Save & Close
button a filtered `MatchedSpectra` is returned, containing only matches manually
validated.


### Using alternative spectra similarity functions

The previous examples used the default spectra similarity scoring function used
by `compareSpectra()` which is the normalized dot product
`MsCoreUtils::ndotproduct()`. It is however also possible to use other
similarity functions, either those defined in the `r Biocpkg("MsCoreUtils")`
package (see `?MsCoreUtils::distance` for a list of available options), or
functions defined by other, external, packages. Here it is however very
important to carefully evaluate how this similarity function works and which
parameters are needed/defined. As an example we use the spectral entropy
similarity function `msentropy_similarity()` from the
[*msentropy*](https://cran.r-project.org/web/packages/msentropy/index.html)
package. This function performs both, the mapping between the fragment peaks of
the compared spectra **and** the similarity calculation. It is therefore crucial
to **disable** the default mapping of fragment spectra by setting `MAPFUN =
joinPeaksNone` in the `CompareSpectraParam()` call. In addition,
`msentropy_similarity()` uses different parameters and parameter names to
configure the ppm and tolerance for the mapping. These parameters are
`ms2_tolerance_in_ppm` (instead of `ppm`) and `ms2_tolerance_in_da` (for
`tolerance`). To perform the spectra matching using the *msentropy* similarity
score we thus configure the `CompareSpectraParam` as shown below:

- disable the default mapping of fragment peaks: use `MAPFUN = joinPeaksNone`
  instead of the default `MAPFUN = joinPeaks`.
- use `FUN = msentropy_similarity`: this will use the respective function from
  the *msentropy* function that performs peak mapping (eventual spectra
  cleaning) and similarity calculation. See `?msentropy_similarity` for more
  details and parameters.
- define the ppm tolerance for the mapping of peaks between compared spectra
  using `ms2_tolerance_in_ppm`.
- define the absolute tolerance for the mapping of peaks between spectra using
  `ms2_tolerance_in_da` - setting this to `-1` has the same effect as `tolerance
  = 0` has with the default `joinPeaks()` mapping function.

```{r, warning = FALSE}
library(msentropy)
csp <- CompareSpectraParam(requirePrecursor = TRUE, ppm = 10,
                           matchedPeaksCount = TRUE,
                           MAPFUN = joinPeaksNone,
                           FUN = msentropy_similarity,
                           ms2_tolerance_in_ppm = 10,
                           ms2_tolerance_in_da = -1)
mtches <- matchSpectra(pest_ms2, minimb, param = csp)
mtches
```

Note that we still configured the `ppm` and `tolerance` parameters, as these
are used for the similarity between the spectra's precursor m/z. Similarly, to
use GNPS-like similarity calculation, `MAPFUN = joinPeaksGnps` and `FUN =
MsCoreUtils::gnps` would need to be used.


### Query against multiple reference databases

Getting access to reference spectra can sometimes be a little cumbersome since
it might involve lookup and download of specific resources or eventual
conversion of these into a format suitable for import. `MetaboAnnotation`
provides *compound annotation sources* to simplify this process. These
annotation source objects represent references (links) to annotation resources
and can be used in the `matchSpectra()` call to define the targed/reference
spectra. The annotation source object takes then care, upon request, of
retrieving the annotation data or connecting to the annotation resources.

Also, *compound annotation sources* can be combined to allow
matching query spectra against multiple reference libraries in a single call.

At present `MetaboAnnotation` supports the following types of *compound
annotation sources* (i.e. objects extending `CompAnnotationSource`):

- Annotation resources that provide their data as a `CompDb` database (defined
  by the `r BiocStyle::Biocpkg("CompoundDb")`) package. These are supported by
  the `CompDbSource` class.

- Annotation resources for which a dedicated `MsBackend` backend is available
  hence supporting to access the data *via* a `Spectra` object. These are
  supported by the `SpectraDbSource` class.

Various helper functions, specific for the annotation resource, are available to
create such annotation source objects:

- `CompDbSource`: creates a compound annotation source object from the provided
  `CompDb` SQLite data base file. This function can be used to integrate an
  existing (locally available) `CompDb` annotation database into an annotation
  workflow.

- `MassBankSource`: creates a annotation source object for a specific MassBank
  release. The desired release can be specified with the `release` parameter
  (e.g. `release = "2021.03"` or `release = "2022.06"`). The function will then
  download the respective annotation database from Bioconductor's
  `r Biocpkg("AnnotationHub")`.

In the example below we create a annotation source for MassBank release
*2022.06*. This call will lookup the requested version in Biocondutor's (online)
`AnnotationHub` and download the data. Subsequent requests for the same
annotation resource will load the locally cached version instead. Upcoming
MassBank database releases will be added to `AnnotationHub` after their official
release and all previous releases will be available as well.

```{r, message = FALSE}
mbank <- MassBankSource("2022.06")
mbank
```

We can now use that annotation source object in the `matchSpectra()` call to
compare the experimental spectra from the previous examples against that release
of MassBank.

```{r}
res <- matchSpectra(
    pest_ms2, mbank,
    param = CompareSpectraParam(requirePrecursor = TRUE, ppm = 10))
res
```

The result object contains only the matching fragment spectra from the reference
database.

```{r}
target(res)
```

And the names of the compounds with matching fragment spectra.

```{r}
matchedData(res)$target_name
```


### Finding MS2 spectra for selected m/z and retention times

Sometimes it is needed to identify fragment spectra in a `Spectra` object for
selected (precursor) m/z values and retention times. An example would be if
compound quantification was performed with a LC-MS run and in a second LC-MS/MS
run (with the same chromatographic setup) fragment spectra of the same samples
were generated. From the first LC-MS data set *features* (or chromatographic
peaks) would be identified for which it would be necessary to retrieve fragment
spectra matching the m/z and retention times of these from the second, LC-MS/MS
data set (assuming that no big retention time shifts between the measurement
runs are expected). To illustrate this, we below first define a `data.frame`
that should represent a feature table such as defined by an analysis with the
`r Biocpkg("xcms")` package.

```{r}
fts <- data.frame(
    feature_id = c("FT001", "FT002", "FT003", "FT004", "FT005"),
    mzmed = c(313.43, 256.11, 224.08, 159.22, 224.08),
    rtmed = c(38.5, 379.1, 168.2, 48.2, 381.1))
```

We next match the features from this data frame against the `Spectra` object
using an `MzRtParam` to identify fragment spectra with their precursor m/z and
retention times matching (with some tolerance) the values from the features.

```{r}
fts_mtch <- matchValues(fts, pest_ms2, MzRtParam(ppm = 50, toleranceRt = 3),
                        mzColname = c("mzmed", "precursorMz"),
                        rtColname = c("rtmed", "rtime"))
fts_mtch
whichQuery(fts_mtch)
```

Thus, we found fragment spectra matching the m/z and retention times for the 2nd
and 5th feature. To extract the `Spectra` matching these features, it would be
best to first reduce the object to features with at least one matching fragment
spectrum. The indices of query elements (in our case features) with matches can
be returned using the `whichQuery()` function. We use these below to subset our
matched result keeping only features for which matches were found:

```{r}
fts_mtched <- fts_mtch[whichQuery(fts_mtch)]
fts_mtched
```

The feature IDs for the matched spectra can be extracted using:

```{r}
fts_mtched$feature_id
```

We next need to extract the matching fragment spectra from the `target`
`Spectra` object. Here we use the `targetIndex()` function, that returns the
indices of the target spectra that were matched to the query.

```{r}
targetIndex(fts_mtched)
```

We extract thus next the fragment spectra matching at least one feature:

```{r}
fts_ms2 <- target(fts_mtched)[targetIndex(fts_mtched)]
fts_ms2
```

While we have now the spectra, we can't relate them (yet) to the features we
used as `query`. Extracting the `"feature_id"` column using the `$` function
from the the matched object would however return, for each match (since we
restricted the matched object to contain only features with matches) the feature
ID (provided in the original data frame). We can thus add this information as an
additional spectra variable to our `Spectra` object:

```{r}
fts_ms2$feature_id <- fts_mtched$feature_id
```

Be aware that extracting the `"feature_id"` column from the matched object
**before** restricting to features with matches would also return the values for
features for which no MS2 spectrum was found:

```{r}
fts_mtch$feature_id
```

Without the initial subsetting of the matched object to features with at least
one matching spectra, the extraction would be a bit more complicated:

```{r}
fts_ms2 <- target(fts_mtch)[targetIndex(fts_mtch)]
fts_ms2$feature_id <- query(fts_mtch)$feature_id[queryIndex(fts_mtch)]
fts_ms2$feature_id
```

This `Spectra` could next be used to match the fragment spectra from the
experiment to e.g. a reference database and with the assigned spectra variable
`"feature_id"` it would allow to map the results back to the quantified feature
matrix from the LC-MS run.


### Performance and parallel processing

Pre-filtering the target spectra based on similar precursor m/z (using
`requirePrecursor = TRUE` generally speeds up the call because a spectra
comparison needs only to be performed on subsets of target spectra. Performance
of the `matchSpectra()` function depends however also on the backend used for
the query and target `Spectra`. For some backends the peaks data (i.e. m/z and
intensity values) might not be already loaded into memory and hence spectra
comparisons might be slower because that data needs to be first loaded. As an
example, for `Spectra` objects, such as our `pest_ms2` variable, that use the
`MsBackendMzR`backend, the peaks data needs to be loaded from the raw data files
before the spectra similarity scores can be calculated. Changing the backend to
an in-memory data representation before `matchSpectra()` can thus improve the
performance (at the cost of a higher memory demand).

Below we change the backends of the `pest_ms2` and `minimb` objects to
`MsBackendMemory` which keeps all data (spectra and peaks data) in memory and we
compare the performance against the originally used `MsBackendMzR` (for
`pest_ms2`) and `MsBackendDataFrame` (for `minimb`).

```{r}
pest_ms2_mem <- setBackend(pest_ms2, MsBackendMemory())
minimb_mem <- setBackend(minimb, MsBackendMemory())
csp <- CompareSpectraParam(requirePrecursor = TRUE, ppm = 10,
                           matchedPeaksCount = TRUE)

library(microbenchmark)
microbenchmark(compareSpectra(pest_ms2, minimb, param = csp),
               compareSpectra(pest_ms2_mem, minimb_mem, param = csp),
               times = 5)
```

There is a considerable performance gain by using the `MsBackendMemory` over the
two other backends, that comes however at the cost of a higher memory
demand. Thus, for large data sets (or reference libraries) this might not be an
option. See also [issue
#93](https://github.com/rformassspectrometry/MetaboAnnotation/issues/93) in the
`MetaboAnnotation` github repository for more benchmarks and information on
performance of `matchSpectra()`.

If for `target` a `Spectra` using a SQL database-based backend is used (such as
a `MsBackendMassbankSql`, `MsBackendCompDb` or `MsBackendSql`) and spectra
matching is performed with `requirePrecursorMz = TRUE`, simply *caching* the
precursor m/z values of all target spectra in memory improves the performance of
`matchSpectra` considerably. This can be easily done with e.g.
`target_sps$precursorMz <- precursorMz(target_sps)` where `target_sps` is the
`Spectra` object that uses one of the above mentioned backends. With this call
all precursor m/z values will be cached within `target_sps` and any
`precursorMz(target_sps)` call (which is used by `matchSpectra()` to select the
candidate spectra against which to compare a query spectrum) will not require
a separate SQL call.

Parallel processing can also improve performance, but might not be possible for
all backends. In particular, backends based on SQL databases don't allow
parallel processing because the database connection can not be shared across
different processes.


# Utility functions

*MetaboAnnotation* provides also other utility functions not directly related to
the annotation process. These are presented in this section.

## Creating mixes of standard compounds

The function `createStandardMixes()` allows for grouping of standard compounds
with a minimum difference in m/z based on user input.

```{r}
library(MetaboCoreUtils)
```

### Input format

As an example here I will extract a list of a 100 standard compounds with their
formula from a tab delimited text file provided with the package. Such files
could also be imported from an xlsx sheet using the *readxl* package.

```{r}
standard <- read.table(system.file("extdata", "Standard_list_example.txt",
                               package = "MetaboAnnotation"),
                   header = TRUE, sep = "\t", quote = "")
```

We will use functions from the MetaboCoreUtil package to get the mass of each
compounds and the m/z for the adducts wanted.

```{r}
#' Calculate mass based on formula of compounds
standard$mass <- calculateMass(standard$formula)

#' Create input for function
#' Calculate charge for 2 adducts
standard_charged <- mass2mz(standard$mass, adduct = c("[M+H]+", "[M+Na]+"))

#' have compounds names as rownames
rownames(standard_charged) <- standard[ , 1]

#' ensure the input `x` is a matrix
if (!is.matrix(standard_charged))
    standard_charged <- as.matrix(standard_charged, drop = FALSE)
```

The input table for the createStandardMixes should thus look like the one shown
below, i.e. should be a numeric matrix with each row representing one compound.
Columns are expected to contain m/z values for different adducts of that
compound. Importantly, the row names of the matrix should represent the (unique)
compound names (or any other unique identifier for the compound).

```{r}
standard_charged
```

### Using the function

The `createStandardMixes()` function organizes given compounds in such a way
that each compound is placed in a group where all ions (adducts) have a m/z
difference exceeding a user-defined threshold (default: `min_diff = 2`).  In
this initial example, we aim to group only a subset of our compound list and
execute the function with default parameters:

```{r}
group_no_randomization <- createStandardMixes(standard_charged[1:20,])
group_no_randomization
```

Let's see the number of compounds per group:

```{r}
table(group_no_randomization$group)
```

The grouping here worked perfectly, but let's now use the entire compound list
and run with the default parameter again:

```{r}
group_no_randomization <- createStandardMixes(standard_charged)
table(group_no_randomization$group)
```

This time we can see that the grouping is less ideal.
In this case we can switch the `iterativeRandomization = TRUE`.

```{r}
group_with_ramdomization <- createStandardMixes(standard_charged,
                                                iterativeRandomization = TRUE)

table(group_with_ramdomization$group)
```

Changing `iterativeRandomization =` from the default `FALSE` to `TRUE` enables
the randomization of input `x` rows until it fits the `min_nstd` parameter. If
the list of compounds is very long or the requirement is hard to fit, this
function can take a bit longer if `iterativeRandomization =` is set to `TRUE.`

What if we want groups of a maximum of 20 and a minimum of 15 compounds, and
with a minimum difference of 2 m/z between compounds of the same group? If you
want to know more about the parameters of this function, look at
`?createStandardMixes`.

```{r}
set.seed(123)
group_with_ramdomization <- createStandardMixes(standard_charged,
                                                max_nstd = 15,
                                                min_nstd = 10,
                                                min_diff = 2,
                                                iterativeRandomization = TRUE)

table(group_with_ramdomization$group)
```

Great ! these groups look good; we can now export. As the function already
returns a `data.frame`, you can directly save is as an Excel file using
`write_xlsx()` from the *writexl* R package or as below in text format that can
also be open in Excel.

```{r eval=FALSE}
write.table(group_with_ramdomization,
           file = "standard_mixes.txt", sep = "\t", quote = FALSE)
```

# Session information {-}

```{r sessioninfo, echo=FALSE}
sessionInfo()
```

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/matchValues.R
\name{ValueParam}
\alias{ValueParam}
\alias{MzParam}
\alias{Mass2MzParam}
\alias{Mass2MzRtParam}
\alias{MzRtParam}
\alias{Mz2MassParam}
\alias{Mz2MassRtParam}
\alias{matchValues}
\alias{matchMz}
\alias{matchValues,numeric,numeric,ValueParam-method}
\alias{matchValues,numeric,data.frameOrSimilar,ValueParam-method}
\alias{matchValues,data.frameOrSimilar,numeric,ValueParam-method}
\alias{matchValues,data.frameOrSimilar,data.frameOrSimilar,ValueParam-method}
\alias{matchValues,numeric,numeric,Mass2MzParam-method}
\alias{matchValues,numeric,data.frameOrSimilar,Mass2MzParam-method}
\alias{matchValues,data.frameOrSimilar,numeric,Mass2MzParam-method}
\alias{matchValues,data.frameOrSimilar,data.frameOrSimilar,Mass2MzParam-method}
\alias{matchValues,numeric,data.frameOrSimilar,MzParam-method}
\alias{matchValues,numeric,Spectra,MzParam-method}
\alias{matchValues,data.frameOrSimilar,numeric,MzParam-method}
\alias{matchValues,data.frameOrSimilar,data.frameOrSimilar,MzParam-method}
\alias{matchValues,data.frameOrSimilar,Spectra,MzParam-method}
\alias{matchValues,data.frameOrSimilar,data.frameOrSimilar,Mass2MzRtParam-method}
\alias{matchValues,data.frameOrSimilar,data.frameOrSimilar,MzRtParam-method}
\alias{matchValues,data.frameOrSimilar,Spectra,MzRtParam-method}
\alias{matchValues,numeric,numeric,Mz2MassParam-method}
\alias{matchValues,numeric,data.frameOrSimilar,Mz2MassParam-method}
\alias{matchValues,data.frameOrSimilar,numeric,Mz2MassParam-method}
\alias{matchValues,data.frameOrSimilar,data.frameOrSimilar,Mz2MassParam-method}
\alias{matchValues,data.frameOrSimilar,data.frameOrSimilar,Mz2MassRtParam-method}
\title{Matching of numeric values}
\usage{
ValueParam(tolerance = 0, ppm = 5)

MzParam(tolerance = 0, ppm = 5)

Mass2MzParam(adducts = c("[M+H]+"), tolerance = 0, ppm = 5)

Mass2MzRtParam(adducts = c("[M+H]+"), tolerance = 0, ppm = 5, toleranceRt = 0)

MzRtParam(tolerance = 0, ppm = 0, toleranceRt = 0)

Mz2MassParam(
  queryAdducts = c("[M+H]+"),
  targetAdducts = c("[M-H]-"),
  tolerance = 0,
  ppm = 5
)

Mz2MassRtParam(
  queryAdducts = c("[M+H]+"),
  targetAdducts = c("[M+H]+"),
  tolerance = 0,
  ppm = 5,
  toleranceRt = 0
)

matchValues(query, target, param, ...)

\S4method{matchValues}{numeric,numeric,ValueParam}(query, target, param)

\S4method{matchValues}{numeric,data.frameOrSimilar,ValueParam}(
  query,
  target,
  param,
  valueColname = character(),
  targetAssay = character()
)

\S4method{matchValues}{data.frameOrSimilar,numeric,ValueParam}(
  query,
  target,
  param,
  valueColname = character(),
  queryAssay = character()
)

\S4method{matchValues}{data.frameOrSimilar,data.frameOrSimilar,ValueParam}(
  query,
  target,
  param,
  valueColname = character(),
  queryAssay = character(),
  targetAssay = character()
)

\S4method{matchValues}{numeric,numeric,Mass2MzParam}(query, target, param)

\S4method{matchValues}{numeric,data.frameOrSimilar,Mass2MzParam}(
  query,
  target,
  param,
  massColname = "exactmass",
  targetAssay = character()
)

\S4method{matchValues}{data.frameOrSimilar,numeric,Mass2MzParam}(query, target, param, mzColname = "mz", queryAssay = character())

\S4method{matchValues}{data.frameOrSimilar,data.frameOrSimilar,Mass2MzParam}(
  query,
  target,
  param,
  mzColname = "mz",
  massColname = "exactmass",
  queryAssay = character(0),
  targetAssay = character(0)
)

\S4method{matchValues}{numeric,data.frameOrSimilar,MzParam}(query, target, param, mzColname = "mz", targetAssay = character())

\S4method{matchValues}{numeric,Spectra,MzParam}(query, target, param, mzColname = "mz", targetAssay = character())

\S4method{matchValues}{data.frameOrSimilar,numeric,MzParam}(query, target, param, mzColname = "mz", queryAssay = character())

\S4method{matchValues}{data.frameOrSimilar,data.frameOrSimilar,MzParam}(
  query,
  target,
  param,
  mzColname = c("mz", "mz"),
  queryAssay = character(),
  targetAssay = character()
)

\S4method{matchValues}{data.frameOrSimilar,Spectra,MzParam}(
  query,
  target,
  param,
  mzColname = c("mz", "mz"),
  queryAssay = character(),
  targetAssay = character()
)

\S4method{matchValues}{data.frameOrSimilar,data.frameOrSimilar,Mass2MzRtParam}(
  query,
  target,
  param,
  massColname = "exactmass",
  mzColname = "mz",
  rtColname = c("rt", "rt"),
  queryAssay = character(),
  targetAssay = character()
)

\S4method{matchValues}{data.frameOrSimilar,data.frameOrSimilar,MzRtParam}(
  query,
  target,
  param,
  mzColname = c("mz", "mz"),
  rtColname = c("rt", "rt"),
  queryAssay = character(),
  targetAssay = character()
)

\S4method{matchValues}{data.frameOrSimilar,Spectra,MzRtParam}(
  query,
  target,
  param,
  mzColname = c("mz", "mz"),
  rtColname = c("rt", "rt"),
  queryAssay = character(),
  targetAssay = character()
)

\S4method{matchValues}{numeric,numeric,Mz2MassParam}(query, target, param)

\S4method{matchValues}{numeric,data.frameOrSimilar,Mz2MassParam}(query, target, param, mzColname = "mz", targetAssay = character())

\S4method{matchValues}{data.frameOrSimilar,numeric,Mz2MassParam}(query, target, param, mzColname = "mz", queryAssay = character())

\S4method{matchValues}{data.frameOrSimilar,data.frameOrSimilar,Mz2MassParam}(
  query,
  target,
  param,
  mzColname = c("mz", "mz"),
  queryAssay = character(),
  targetAssay = character()
)

\S4method{matchValues}{data.frameOrSimilar,data.frameOrSimilar,Mz2MassRtParam}(
  query,
  target,
  param,
  mzColname = c("mz", "mz"),
  rtColname = c("rt", "rt"),
  queryAssay = character(),
  targetAssay = character()
)
}
\arguments{
\item{tolerance}{for any \code{param} object: \code{numeric(1)} defining the maximal
acceptable absolute difference in m/z (or in mass for \code{Mz2MassParam})
to consider them \emph{matching}.}

\item{ppm}{for any \code{param} object: \code{numeric(1)} defining the maximal
acceptable m/z-dependent (or mass-dependent for \code{Mz2MassParam})
difference (in parts-per-million) in m/z values to consider them to
be \emph{matching}.}

\item{adducts}{for \code{Mass2MzParam} or \code{Mass2MzRtParam}: either \code{character}
with adduct names from \code{\link[MetaboCoreUtils:adductNames]{MetaboCoreUtils::adducts()}} or \code{data.frame} with
a custom adduct definition. This parameter is used to calculate m/z from
target compounds' masses. Custom adduct definitions can be passed to the
adduct parameter in form of a \code{data.frame}. This \code{data.frame} is expected
to have columns \code{"mass_add"} and \code{"mass_multi"} defining the \emph{additive}
and \emph{multiplicative} part of the calculation. See
\code{\link[MetaboCoreUtils:adductNames]{MetaboCoreUtils::adducts()}} for the expected format or use
\code{MetaboCoreUtils::adductNames("positive")} and
\code{MetaboCoreUtils::adductNames("negative")} for valid adduct names.}

\item{toleranceRt}{for \code{Mass2MzRtParam} or \code{MzRtParam}: \code{numeric(1)}
defining the maximal acceptable absolute difference in retention time
values to consider them them \emph{matching}.}

\item{queryAdducts}{for \code{Mz2MassParam}. Adducts used to derive mass
values from query m/z values. The expected format is the same as that
for parameter \code{adducts}.}

\item{targetAdducts}{for \code{Mz2MassParam}. Adducts used to derive mass
values from target m/z values. The expected format is the same as that
for parameter \code{adducts}.}

\item{query}{feature table containing information on MS1 features. Can be
a \code{numeric}, \code{data.frame}, \code{DataFrame}, \code{matrix}, \code{SummarizedExperiment}
or \code{QFeatures}. It is expected to contain m/z values and can contain also
other variables. Matchings based on both m/z and retention time can be
performed when a column with retention times is present in both \code{query}
and \code{target}.}

\item{target}{compound table with metabolites to compare against. The
expected types are the same as those for \code{query}.}

\item{param}{parameter object defining the matching approach and containing
the settings for that approach. See description above for details.}

\item{...}{currently ignored.}

\item{valueColname}{\code{character} specifying the name of the column in
\code{query} or/and the one in \code{target}with the desired values for the
matching. This parameter should only be used when \code{param} is
\code{valueParam} and in this case it must be provided (unless both \code{query}
and \code{target} are \code{numeric}). It can be \code{character(1)} or \code{character(2)}
in a similar way to \code{mzColname}.}

\item{targetAssay}{\code{character(1)} specifying the name of the assay of the
provided \code{QFeatures} that should be used for the matching (values from
this assay's \code{rowData} will be used for matching). Only used if \code{target}
is an instance of a \code{QFeatures} object.}

\item{queryAssay}{\code{character(1)} specifying the name of the assay of the
provided \code{QFeatures} that should be used for the matching (values from
this assay's \code{rowData} will be used for matching). Only used if \code{query}
is an instance of a \code{QFeatures} object.}

\item{massColname}{\code{character(1)} with the name of the column in \code{target}
containing the mass of compounds. To be used when \code{param} is
\code{Mass2MzParam} or \code{Mass2MzRtParam} (and target is not already \code{numeric}
with the masses). Defaults to \code{massColname = "exactmass"}.}

\item{mzColname}{\code{character} specifying the name(s) of the column(s) in
\code{query} or/and \code{target}with the m/z values. If one among \code{query} and
\code{target} is \code{numeric} (and therefore there is no need to specify the
column name) or \code{query} is not \code{numeric} and \code{param} is \code{Mass2MzParam}
or \code{Mass2MzRtParam} (and therefore the name of the column with m/z needs
only to be specified for \code{query}) then \code{mzColname} is expected to be
\code{character(1)}. If both \code{query} and \code{target} are not numeric \code{mzColname}
is expected to be \code{character(2)} (or \code{character(1)} and in this last case
the two column names are assumed to be the same). If not specified the
assumed default name for columns with m/z values is \code{"mz"}. If \code{target}
is a \code{\link[Spectra:Spectra]{Spectra::Spectra()}} object, the name of the spectra variable
that should be used for the matching needs to be specified
with \code{mzColname}.}

\item{rtColname}{\code{character(2)} with the name of the column containing
the compounds retention times in \code{query} and the name for the one in
\code{target}. It can also be \code{character(1)} if the two names are the same.
To be used when \code{param} is \code{MzRtParam} or \code{Mass2MzRtParam}.
Defaults to \code{rtColname = c("rt", "rt")}. If \code{target} is a
\code{\link[Spectra:Spectra]{Spectra::Spectra()}} object, the name of the spectra variable that
should be used for the matching needs to be specified with \code{mzColname}.}
}
\value{
\link{Matched} object representing the result.

Depending on the \code{param} object different \emph{scores} representing the quality
of the match are provided. This comprises absolute as well as relative
differences (column/variables \code{"score"} and \code{"ppm_error"} respectively).
If \code{param} is a \code{Mz2MassParam}, \code{"score"} and \code{"ppm_error"} represent
differences of the compared masses (calculated from the provided m/z values).
If \code{param} an \code{MzParam}, \code{MzRtParam}, \code{Mass2MzParam} or \code{Mass2MzRtParam},
\code{"score"} and \code{"ppm_error"} represent absolute and relative differences of
m/z values.
Additionally, if \code{param} is either an \code{MzRtParam} or \code{Mass2MzRtParam}
differences between query and target retention times for each matched
element is available in the column/variable \code{"score_rt"} in the returned
\code{Matched} object.
Negative values of \code{"score"} (or \code{"score_rt"}) indicate that the m/z or mass
(or retention time) of the query element is smaller than that of the target
element.
}
\description{
The \code{matchValues} method matches elements from \code{query} with those in \code{target}
using different matching approaches depending on parameter \code{param}.
Generally, \code{query} is expected to contain MS experimental values
(m/z and possibly retention time) while \code{target} reference values. \code{query}
and \code{target} can be \code{numeric}, a two dimensional array (such as a
\code{data.frame}, \code{matrix} or \code{DataFrame}), a \code{SummarizedExperiment}
or a \code{QFeatures}, \code{target} can in addition be a \code{\link[Spectra:Spectra]{Spectra::Spectra()}} object.
For \code{SummarizedExperiment}, the information for the matching is expected
to be in the object's \code{rowData}. For \code{QFeatures} matching is performed
for values present in the \code{rowData} of one of the object's assays (which
needs to be specified with the \code{assayQuery} parameter - if a \code{QFeatures}
is used as \code{target} the name of the assay needs to be specified with
parameter \code{assayTarget}). If \code{target} is a \code{Spectra} matching is performed
against spectra variables of this object and the respective variable names
need to be specified e.g. with \code{mzColname} and/or \code{rtColname}.
\code{matchMz} is an alias for \code{matchValues} to allow backward compatibility.

Available \code{param} objects and corresponding matching approaches are:
\itemize{
\item \code{ValueParam}: generic matching between values in \code{query} and \code{target} given
acceptable differences expressed in \code{ppm} and \code{tolerance}. If \code{query} or
\code{target} are not numeric, parameter \code{valueColname} has to be used to
specify the name of the column that contains the values to be matched.
The function returns a \code{\link[=Matched]{Matched()}} object.
\item \code{MzParam}: match query m/z values against reference compounds for which
also m/z are known. Matching is performed similarly to the \code{ValueParam}
above. If \code{query} or \code{target} are not numeric, the column name containing
the values to be compared must be defined with \code{matchValues}' parameter
\code{mzColname}, which defaults to \code{"mz"}. \code{MzParam} parameters \code{tolerance}
and \code{ppm} allow to define the maximal acceptable (constant or m/z relative)
difference between query and target m/z values.
\item \code{MzRtParam}: match m/z \strong{and} retention time values between \code{query} and
\code{target}. Parameters \code{mzColname} and \code{rtColname} of the \code{matchValues}
function allow to define the columns in \code{query} and \code{target} containing
these values (defaulting to \code{c("mz", "mz")} and \code{c("rt", "rt")},
respectively). \code{MzRtParam} parameters \code{tolerance} and
\code{ppm} have the same meaning as in \code{MzParam}; \code{MzRtParam} parameter
\code{toleranceRt} allows to specify the maximal acceptable difference between
query and target retention time values.
\item \code{Mass2MzParam}: match m/z values against reference compounds for
which only the (exact) mass is known. Before matching, m/z values are
calculated from the compounds masses in the \emph{target} table using the
adducts specified via \code{Mass2MzParam} \code{adducts} parameter (defaults to
\code{adducts = "[M+H]+"}). After conversion of adduct masses to m/z values,
matching is performed similarly to \code{MzParam} (i.e. the same parameters
\code{ppm} and \code{tolerance} can be used). If \code{query} is not \code{numeric},
parameter \code{mzColname} of \code{matchValues} can be used to specify the column
containing the query's m/z values (defaults to \code{"mz"}). If \code{target} is a
is not \code{numeric}, parameter \code{massColname} can be used to define the
column containing the reference compound's masses (defaults to
\code{"exactmass"}).
\item \code{Mass2MzRtParam}: match m/z \strong{and} retention time values against
reference compounds for which the (exact) mass \strong{and} retention time are
known. Before matching, exact masses in \code{target} are converted to m/z
values as for \code{Mass2MzParam}. Matching is then performed similarly to
\code{MzRtParam}, i.e. m/z and retention times of entities are compared. With
\code{matchValues}' parameters \code{mzColname}, \code{rtColname} and \code{massColname} the
columns containing m/z values (in \code{query}), retention time values (in
\code{query} and \code{target}) and exact masses (in \code{target}) can be specified.
\item \code{Mz2MassParam}: input values for \code{query} and \code{target} are expected to be
m/z values but matching is performed on exact masses calculated from these
(based on the provided adduct definitions). In detail, m/z values in
\code{query} are first converted to masses with the \code{\link[MetaboCoreUtils:mz2mass]{MetaboCoreUtils::mz2mass()}}
function based on the adducts defined with \code{queryAdducts} (defaults to
\code{"[M+H]+"}). The same is done for m/z values in \code{target} (adducts can
be defined with \code{targetAdducts} which defaults to \verb{"[M-H-]"). Matching is then performed on these converted values similarly to }ValueParam\verb{. If }query\code{or}target\verb{are not numeric, the column containing the m/z values can be specified with}matchValues\verb{' parameter }mzColname\verb{(defaults to}"mz"`).
\item \code{Mz2MassRtParam}: same as \code{Mz2MassParam} but with additional comparison of
retention times between \code{query} and \code{target}. Parameters \code{rtColname} and
\code{mzColname} of \code{matchValues} allow to specify which columns contain the
retention times and m/z values, respectively.
}
}
\examples{

library(MetaboCoreUtils)
## Create a simple "target/reference" compound table
target_df <- data.frame(
   name = c("Tryptophan", "Leucine", "Isoleucine"),
   formula = c("C11H12N2O2", "C6H13NO2", "C6H13NO2"),
   exactmass = c(204.089878, 131.094629, 131.094629)
)

## Create a "feature" table with m/z of features. We calculate m/z for
## certain adducts of some of the compounds in the reference table.
fts <- data.frame(
    feature_id = c("FT001", "FT002", "FT003"),
    mz = c(mass2mz(204.089878, "[M+H]+"),
           mass2mz(131.094629, "[M+H]+"),
           mass2mz(204.089878, "[M+Na]+") + 1e-6))

## Define the parameters for the matching
parm <- Mass2MzParam(
    adducts = c("[M+H]+", "[M+Na]+"),
    tolerance = 0,
    ppm = 20)
res <- matchValues(fts, target_df, parm)
res

## List the available variables/columns
colnames(res)

## feature_id and mz are from the query data frame, while target_name,
## target_formula and target_exactmass are from the query object (columns
## from the target object have a prefix *target_* added to the original
## column name. Columns adduct, score and ppm_error represent the results
## of the matching: adduct the adduct/ion of the original compound for which
## the m/z matches, score the absolute difference of the query and target
## m/z and ppm_error the relative difference in m/z values.

## Get the full matching result:
matchedData(res)

## We have thus matches of FT002 to two different compounds (but with the
## same mass).

## Individual columns can also be accessed with the $ operator:
res$feature_id
res$target_name
res$ppm_error


## We repeat the matching requiring an exact match
parm <- Mass2MzParam(
    adducts = c("[M+H]+", "[M+Na]+"),
    tolerance = 0,
    ppm = 0)
res <- matchValues(fts, target_df, parm)
res

matchedData(res)

## The last feature could thus not be matched to any compound.

## At last we use also different adduct definitions.
parm <- Mass2MzParam(
    adducts = c("[M+K]+", "[M+Li]+"),
    tolerance = 0,
    ppm = 20)
res <- matchValues(fts, target_df, parm)
res

matchedData(res)

## No matches were found.

## We can also match a "feature" table with a target data.frame taking into
## account both m/z and retention time values.
target_df <- data.frame(
  name = c("Tryptophan", "Leucine", "Isoleucine"),
  formula = c("C11H12N2O2", "C6H13NO2", "C6H13NO2"),
  exactmass = c(204.089878, 131.094629, 131.094629),
  rt = c(150, 140, 140)
)

fts <- data.frame(
  feature_id = c("FT001", "FT002", "FT003"),
  mz = c(mass2mz(204.089878, "[M+H]+"),
         mass2mz(131.094629, "[M+H]+"),
         mass2mz(204.089878, "[M+Na]+") + 1e-6),
  rt = c(150, 140, 150.1)
)

## Define the parameters for the matching
parm <- Mass2MzRtParam(
  adducts = c("[M+H]+", "[M+Na]+"),
  tolerance = 0,
  ppm = 20,
  toleranceRt = 0)

res <- matchValues(fts, target_df, parm)
res

## Get the full matching result:
matchedData(res)

## FT003 could not be matched to any compound, FT002 was matched to two
## different compounds (but with the same mass).

## We repeat the matching allowing a positive tolerance for the matches
## between rt values

## Define the parameters for the matching
parm <- Mass2MzRtParam(
  adducts = c("[M+H]+", "[M+Na]+"),
  tolerance = 0,
  ppm = 20,
  toleranceRt = 0.1)

res <- matchValues(fts, target_df, parm)
res

## Get the full matching result:
matchedData(res)

## Also FT003 was matched in this case

## It is also possible to match directly m/z values
mz1 <- c(12, 343, 23, 231)
mz2 <- mz1 + rnorm(4, sd = 0.001)

res <- matchValues(mz1, mz2, MzParam(tolerance = 0.001))

matchedData(res)

## Matching with a SummarizedExperiment or a QFeatures work analogously,
## only that the matching is performed on the object's `rowData`.

## Below we create a simple SummarizedExperiment with some random assay data.
## Note that results from a data preprocessing with the `xcms` package could
## be extracted as a `SummarizedExperiment` with the `quantify` method from
## the `xcms` package.
library(SummarizedExperiment)
se <- SummarizedExperiment(
    assays = matrix(rnorm(12), nrow = 3, ncol = 4,
    dimnames = list(NULL, c("A", "B", "C", "D"))),
    rowData = fts)

## We can now perform the matching of this SummarizedExperiment against the
## target_df as before.
res <- matchValues(se, target_df,
    param = Mass2MzParam(adducts = c("[M+H]+", "[M+Na]+"),
        tolerance = 0, ppm = 20))
res

## Getting the available columns
colnames(res)

## The query columns represent the columns of the object's `rowData`
rowData(se)

## matchedData also returns the query object's rowData along with the
## matching entries in the target object.
matchedData(res)

## While `query` will return the full SummarizedExperiment.
query(res)

## To illustrate use with a QFeatures object we first create a simple
## QFeatures object with two assays, `"ions"` representing the full feature
## data.frame and `"compounds"` a subset of it.
library(QFeatures)
qf <- QFeatures(list(ions = se, compounds = se[2,]))

## We can perform the same matching as before, but need to specify which of
## the assays in the QFeatures should be used for the matching. Below we
## perform the matching using the "ions" assay.
res <- matchValues(qf, target_df, queryAssay = "ions",
    param = Mass2MzParam(adducts = c("[M+H]+", "[M+Na]+"),
        tolerance = 0, ppm = 20))
res

## colnames returns now the colnames of the `rowData` of the `"ions"` assay.
colnames(res)

matchedData(res)
}
\seealso{
\link{matchSpectra} or \code{\link[=CompareSpectraParam]{CompareSpectraParam()}} for spectra data matching
}
\author{
Andrea Vicini, Michael Witting
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validateMatchedSpectra.R
\name{validateMatchedSpectra}
\alias{validateMatchedSpectra}
\title{Validating MatchedSpectra}
\usage{
validateMatchedSpectra(object)
}
\arguments{
\item{object}{A non-empty instance of class \code{MatchedSpectra}.}
}
\value{
A \code{MatchedSpectra} with validated results.
}
\description{
The \code{validateMatchedSpectra()} function opens a simple shiny application
that allows to browse results stored in a \code{MatchedSpectra} object and to
\emph{validate} the presented matches. For each query spectrum a table with
matched target spectra are shown (if available) and an interactive mirror
plot is generated. Valid matches can be selected using a check box which is
displayed below the mirror plot. Upon pushing the "Save & Close"
button the app is closed and a filtered \code{MatchedSpectra} is returned,
containing only \emph{validated} matches.

Note that column \code{"query_index_"} and \code{"target_index_"} are temporarily
added to the query and target \code{Spectra} object to display them in the
interactive graphics for easier identification of the compared spectra.
}
\examples{

library(Spectra)
## Load test data
fl <- system.file("TripleTOF-SWATH", "PestMix1_DDA.mzML", package = "msdata")
pest_ms2 <- filterMsLevel(Spectra(fl), 2L)
pest_ms2 <- pest_ms2[c(808, 809, 945:955)]
load(system.file("extdata", "minimb.RData", package = "MetaboAnnotation"))

## Normalize intensities and match spectra
csp <- CompareSpectraParam(requirePrecursor = TRUE,
                           THRESHFUN = function(x) x >= 0.7)
norm_int <- function(x) {
    x[, "intensity"] <- x[, "intensity"] / max(x[, "intensity"]) * 100
    x
}
ms <- matchSpectra(addProcessing(pest_ms2, norm_int),
                   addProcessing(minimb, norm_int), csp)

## validate matches using the shiny app. Note: the call is only executed
## in interactive mode.
if (interactive()) {
    res <- validateMatchedSpectra(ms)
}
}
\author{
Carolin Huber, Michael Witting, Johannes Rainer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CompAnnotationSource.R
\name{CompAnnotationSource}
\alias{CompAnnotationSource}
\alias{CompAnnotationSource-class}
\alias{matchSpectra,Spectra,CompAnnotationSource,Param-method}
\alias{show,CompAnnotationSource-method}
\alias{metadata,CompAnnotationSource-method}
\title{Compound Annotation Sources}
\usage{
\S4method{matchSpectra}{Spectra,CompAnnotationSource,Param}(query, target, param, ...)

\S4method{show}{CompAnnotationSource}(object)

\S4method{metadata}{CompAnnotationSource}(x, ...)
}
\arguments{
\item{query}{for \code{matchSpectra}: \link[Spectra:Spectra]{Spectra::Spectra} object with the query
spectra.}

\item{target}{for \code{matchSpectra}: object extending \link{CompAnnotationSource}
(such as \link{CompDbSource}) with the target (reference) spectra to compare
\code{query} against.}

\item{param}{for \code{matchSpectra}: parameter object (such as
\link{CompareSpectraParam}) defining the settings for the matching.}

\item{...}{additional parameters passed to \code{matchSpectra}.}

\item{object}{A \code{CompAnnotationSource} object.}

\item{x}{A \code{CompAnnotationSource} object.}
}
\description{
\code{CompAnnotationSource}s (i.e. classes extending the base virtual
\code{CompAnnotationSource} class) define and provide access to a (potentially
remote) compound annotation resource. This aims to simplify the integration
of external annotation resources by automating the actual connection
(or data resource download) process from the user. In addition, since the
reference resource is not directly exposed to the user it allows integration
of annotation resources that do not allow access to the full data.

Objects extending \code{CompAnnotationSource} available in this package are:
\itemize{
\item \code{\link[=CompDbSource]{CompDbSource()}}: annotation source referencing an annotation source in the
\verb{[CompoundDb::CompDb()]} format ( from the \code{CompoundDb} Bioconductor
package).
}

Classes extending \code{CompAnnotationSource} need to implement the \code{matchSpectra}
method with parameters \code{query}, \code{target} and \code{param} where \code{query} is
the \code{Spectra} object with the (experimental) query spectra, \code{target} the
object extending the \code{CompAnnotationSource} and \code{param} the parameter object
defining the similarity calculation (e.g. \code{\link[=CompareSpectraParam]{CompareSpectraParam()}}. The method
is expected to return a \link{MatchedSpectra} object.

\code{CompAnnotationSource} objects are not expected to contain any annotation
data. Access to the annotation data (in form of a \code{Spectra} object) is
suggested to be only established within the object's \code{matchSpectra} method.
This would also enable parallel processing of annotations as no e.g. database
connection would have to be shared across processes.
}
\section{Methods that need to be implemented}{


For an example implementation see \code{\link[=CompDbSource]{CompDbSource()}}.
\itemize{
\item \code{matchSpectra}: function to match experimental MS2 spectra against the
annotation source. See \code{\link[=matchSpectra]{matchSpectra()}} for parameters.
\item \code{metadata}: function to provide metadata on the annotation resource (host,
source, version etc).
\item \code{show} (optional): method to provide general information on the data
source.
}
}

\author{
Johannes Rainer, Nir Shachaf
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/matchFormula.R
\name{matchFormula}
\alias{matchFormula}
\alias{matchFormula,character,character-method}
\alias{matchFormula,data.frameOrSimilar,data.frameOrSimilar-method}
\alias{matchFormula,character,data.frameOrSimilar-method}
\alias{matchFormula,data.frameOrSimilar,character-method}
\title{Chemical Formula Matching}
\usage{
matchFormula(query, target, ...)

\S4method{matchFormula}{character,character}(query, target, BPPARAM = SerialParam())

\S4method{matchFormula}{data.frameOrSimilar,data.frameOrSimilar}(
  query,
  target,
  formulaColname = c("formula", "formula"),
  BPPARAM = SerialParam()
)

\S4method{matchFormula}{character,data.frameOrSimilar}(
  query,
  target,
  formulaColname = "formula",
  BPPARAM = SerialParam()
)

\S4method{matchFormula}{data.frameOrSimilar,character}(
  query,
  target,
  formulaColname = "formula",
  BPPARAM = SerialParam()
)
}
\arguments{
\item{query}{\code{character} or \code{data.frame} with chemical formulas to search.}

\item{target}{\code{character} or \code{data.frame} with chemical formulas to compare
against.}

\item{...}{currently ignored}

\item{BPPARAM}{parallel processing setup. See \code{BiocParallel::bpparam()} for
details.}

\item{formulaColname}{\code{character} with the name of the column containing
chemical formulas. Can be of length 1 if both \code{query} and \code{target} are
\code{data.frame}s and the name of the column with chemical formulas is the
same for both. If different columns are used, \code{formulaColname[1]} can be
used to define the column name in \code{query} and \code{formulaColname[2]} the
one of \code{target}.}
}
\value{
\link{Matched} object representing the result.
}
\description{
The \code{matchFormula} method matches chemical formulas from different inputs
(parameter \code{query} and \code{target}). Before comparison all formulas are
normalized using \code{\link[MetaboCoreUtils:standardizeFormula]{MetaboCoreUtils::standardizeFormula()}}. Inputs can be
either a \code{character} or \code{data.frame} containing a column with formulas.
In case of \code{data.frame}s parameter \code{formulaColname} needs to be used to
specify the name of the column containing the chemical formulas.
}
\examples{

## input formula
query <- c("H12C6O6", "C11H12O2", "HN3")
target <- c("HCl", "C2H4O", "C6H12O6")

query_df <- data.frame(
    formula = c("H12C6O6", "C11H12O2", "HN3"),
    name = c("A", "B", "C")
)
target_df <- data.frame(
    formula = c("HCl", "C2H4O", "C6H12O6"),
    name = c("D", "E", "F")
)

## character vs character
matches <- matchFormula(query, target)
matchedData(matches)

## data.frame vs data.frame
matches <- matchFormula(query_df, target_df)
matchedData(matches)
## data.frame vs character
matches <- matchFormula(query_df, target)
matchedData(matches)
## character vs data.frame
matches <- matchFormula(query, target_df)
matchedData(matches)
}
\author{
Michael Witting
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CompDbSource.R, R/matchSpectra.R
\name{matchSpectra,Spectra,CompDbSource,Param-method}
\alias{matchSpectra,Spectra,CompDbSource,Param-method}
\alias{CompareSpectraParam}
\alias{CompareSpectraParam-class}
\alias{MatchForwardReverseParam-class}
\alias{MatchForwardReverseParam}
\alias{matchSpectra,Spectra,Spectra,CompareSpectraParam-method}
\alias{matchSpectra,Spectra,CompDb,Param-method}
\alias{matchSpectra,Spectra,Spectra,MatchForwardReverseParam-method}
\title{Matching MS Spectra against a reference}
\usage{
\S4method{matchSpectra}{Spectra,CompDbSource,Param}(
  query,
  target,
  param,
  BPPARAM = BiocParallel::SerialParam(),
  addOriginalQueryIndex = TRUE
)

CompareSpectraParam(
  MAPFUN = joinPeaks,
  tolerance = 0,
  ppm = 5,
  FUN = MsCoreUtils::ndotproduct,
  requirePrecursor = TRUE,
  requirePrecursorPeak = FALSE,
  THRESHFUN = function(x) which(x >= 0.7),
  toleranceRt = Inf,
  percentRt = 0,
  matchedPeaksCount = FALSE,
  ...
)

MatchForwardReverseParam(
  MAPFUN = joinPeaks,
  tolerance = 0,
  ppm = 5,
  FUN = MsCoreUtils::ndotproduct,
  requirePrecursor = TRUE,
  requirePrecursorPeak = FALSE,
  THRESHFUN = function(x) which(x >= 0.7),
  THRESHFUN_REVERSE = NULL,
  toleranceRt = Inf,
  percentRt = 0,
  ...
)

\S4method{matchSpectra}{Spectra,Spectra,CompareSpectraParam}(
  query,
  target,
  param,
  rtColname = c("rtime", "rtime"),
  BPPARAM = BiocParallel::SerialParam(),
  addOriginalQueryIndex = TRUE
)

\S4method{matchSpectra}{Spectra,CompDb,Param}(
  query,
  target,
  param,
  rtColname = c("rtime", "rtime"),
  BPPARAM = BiocParallel::SerialParam(),
  addOriginalQueryIndex = TRUE
)

\S4method{matchSpectra}{Spectra,Spectra,MatchForwardReverseParam}(
  query,
  target,
  param,
  rtColname = c("rtime", "rtime"),
  BPPARAM = BiocParallel::SerialParam(),
  addOriginalQueryIndex = TRUE
)
}
\arguments{
\item{query}{for \code{matchSpectra}: \link[Spectra:Spectra]{Spectra::Spectra} object with the query
spectra.}

\item{target}{for \code{matchSpectra}: \link[Spectra:Spectra]{Spectra::Spectra}, \link[CompoundDb:CompDb]{CompoundDb::CompDb}
or object extending \link{CompAnnotationSource} (such as \link{CompDbSource}) with
the target (reference) spectra to compare \code{query} against.}

\item{param}{for \code{matchSpectra}: parameter object (such as
\code{CompareSpectraParam}) defining the settings for the matching.}

\item{BPPARAM}{for \code{matchSpectra}: parallel processing setup (see the
\code{BiocParallel} package for more information). Parallel processing is
disabled by default (with the default setting \code{BPPARAM = SerialParam()}).}

\item{addOriginalQueryIndex}{for \code{matchSpectra()}: \code{logical(1)} whether an
additional spectra variable \code{".original_query_index"} should be added to
the \code{query} \code{Spectra} object providing the index of the spectrum in this
originally provided object. This spectra variable can be useful to link
back to the original \code{Spectra} object if the \code{MatchedSpectra} object gets
subsetted/processed.}

\item{MAPFUN}{\code{function} used to map peaks between the compared spectra.
Defaults for \code{CompareSpectraParam} to \code{\link[Spectra:joinPeaks]{Spectra::joinPeaks()}}. See
\code{\link[Spectra:compareSpectra]{Spectra::compareSpectra()}} for details.}

\item{tolerance}{\code{numeric(1)} for an absolute maximal accepted difference
between m/z values. This will be used in \code{compareSpectra} as well as for
eventual precursor m/z matching.}

\item{ppm}{\code{numeric(1)} for a relative, m/z-dependent, maximal accepted
difference between m/z values. This will be used in \code{compareSpectra} as
well as for eventual precursor m/z matching.}

\item{FUN}{\code{function} used to calculate similarity between spectra. Defaults
for \code{CompareSpectraParam} to \code{\link[MsCoreUtils:distance]{MsCoreUtils::ndotproduct()}}. See
\code{\link[MsCoreUtils:distance]{MsCoreUtils::ndotproduct()}} for details.}

\item{requirePrecursor}{\code{logical(1)} whether only target spectra are
considered in the similarity calculation with a precursor m/z that matches
the precursor m/z of the query spectrum (considering also \code{ppm} and
\code{tolerance}). With \code{requirePrecursor = TRUE} (the default) the function
will complete much faster, but will not find any hits for target (or query
spectra) with missing precursor m/z. It is suggested to check first the
availability of the precursor m/z in \code{target} and \code{query}.}

\item{requirePrecursorPeak}{\code{logical(1)} whether only target spectra will be
considered in the spectra similarity calculation that have a peak with an
m/z matching the precursor m/z of the query spectrum. Defaults to
\code{requirePrecursorPeak = FALSE}. It is suggested to check first the
availability of the precursor m/z in \code{query}, as no match will be reported
for query spectra with missing precursor m/z.}

\item{THRESHFUN}{\code{function} applied to the similarity score to define which
target spectra are considered \emph{matching}. Defaults to
\code{THRESHFUN = function(x) which(x >= 0.7)} hence selects
all target spectra matching a query spectrum with a similarity higher or
equal than \code{0.7}. Any function that takes a numeric vector with similarity
scores from the comparison of a query spectrum with all target spectra (as
returned by \code{\link[Spectra:compareSpectra]{Spectra::compareSpectra()}}) as input and returns a
\code{logical} vector (same dimensions as the similarity scores) or an integer
with the matches is supported.}

\item{toleranceRt}{\code{numeric} of length 1 or equal to the number of query
spectra defining the maximal accepted (absolute) difference in retention
time between query and target spectra. By default
(with \code{toleranceRt = Inf}) the retention time-based filter is not
considered. See help of \code{CompareSpectraParam} above for more
information.}

\item{percentRt}{\code{numeric} of length 1 or equal to the number of query
spectra defining the maximal accepted relative difference in retention
time between query and target spectra expressed in percentage of the query
rt. For \code{percentRt = 10}, similarities are defined between the query
spectrum and all target spectra with a retention time within query rt
+/- 10\% of the query. By default (with \code{toleranceRt = Inf}) the retention
time-based filter is not considered. Thus, to consider the \code{percentRt}
parameter, \code{toleranceRt} should be set to a value different than that.
See help of \code{CompareSpectraParam} above for more information.}

\item{matchedPeaksCount}{\code{logical(1)} for \code{CompareSpectraParam()}: whether
also the number of matching peaks should be reported (in column
\code{"matched_peaks_count"}). This number represents the number of peaks
reported \emph{matching} by the \code{MAPFUN}.}

\item{...}{for \code{CompareSpectraParam}: additional parameters passed along
to the \code{\link[Spectra:compareSpectra]{Spectra::compareSpectra()}} call, including eventual additional
parameters of the selected mapping or similarity calculation functions.}

\item{THRESHFUN_REVERSE}{for \code{MatchForwardReverseParam}: optional additional
\emph{thresholding function} to filter the results on the reverse score. If
specified the same format than \code{THRESHFUN} is expected.}

\item{rtColname}{\code{character(2)} with the name of the spectra variable
containing the retention time information for compounds to be used in
retention time matching (only used if \code{toleranceRt} is not \code{Inf}).
It can also be \code{character(1)} if the two names are the same.
Defaults to \code{rtColname = c("rtime", "rtime")}.}
}
\value{
\code{matchSpectra} returns a \code{\link[=MatchedSpectra]{MatchedSpectra()}} object with the matching
results. If \code{target} is a \code{CompAnnotationSource} only matching target
spectra will be reported.

Constructor functions return an instance of the class.
}
\description{
\code{matchSpectra} compares experimental (\emph{query}) MS2 spectra against
reference (\emph{target}) MS2 spectra and reports matches with a similarity that
passing a specified threshold. The function performs the similarity
calculation between each query spectrum against each target spectrum.
Parameters \code{query} and \code{target} can be used to define the query and target
spectra, respectively, while parameter \code{param} allows to define and configure
the similarity calculation and matching condition. Parameter \code{query} takes
a \link[Spectra:Spectra]{Spectra::Spectra} object while \code{target} can be either a
\link[Spectra:Spectra]{Spectra::Spectra} object, a \link[CompoundDb:CompDb]{CompoundDb::CompDb} (reference library)
object defined in the \code{CompoundDb} package or
a \link{CompAnnotationSource} (e.g. a \code{\link[=CompDbSource]{CompDbSource()}})
with the reference or connection information to a supported annotation
resource).

Some notes on performance and information on parallel processing are
provided in the vignette.

Currently supported parameter objects defining the matching are:
\itemize{
\item \code{CompareSpectraParam}: the \emph{generic} parameter object allowing to set all
settings for the \code{\link[Spectra:compareSpectra]{Spectra::compareSpectra()}} call that is used to
perform the similarity calculation.
This includes \code{MAPFUN} and \code{FUN} defining the peak-mapping and similarity
calculation functions and \code{ppm} and \code{tolerance} to define an acceptable
difference between m/z values of the compared peaks. Parameter
\code{matchedPeaksCount} is also passed to \code{compareSpectra()} and, if set to
\code{TRUE} (default is \code{FALSE}) will report the number of peaks defined to be
\emph{matching} by the \code{MAPFUN}.
Additional parameters to the \code{compareSpectra} call can be passed along
with \code{...}. See the help of \code{\link[Spectra:Spectra]{Spectra::Spectra()}} for more information on
these parameters. Importantly, if \emph{msentropy} or a GNPS-like similarity
calculation is used, \code{MAPFUN} should be selected accordingly (see section
\emph{Using alternative spectra similarity functions} in the package vignette
for more information).
By default, parameters \code{ppm} and \code{tolerance} are passed to the similarity
calculation function, but if this function uses different parameters
(e.g., \code{msentropy_similarity()} uses \code{ms2_tolerance_in_ppm} instead of
\code{ppm}), these should be submitted to the \code{CompareSpectraParam()} function
throught the \code{...} parameter.
Parameters \code{requirePrecursor} (default \code{TRUE}) and \code{requirePrecursorPeak}
(default \code{FALSE}) allow to pre-filter the target spectra prior to the
actual similarity calculation for each individual query spectrum.
Parameters \code{ppm} and \code{tolerance} are also used to define the maximal
acceptable difference in precursor m/z if \code{requirePrecursor} or
\code{requirePrecursorPeak} are set to \code{TRUE}.
Target spectra can also be pre-filtered based on
retention time if parameter \code{toleranceRt} is set to a value different than
the default \code{toleranceRt = Inf}. Only target spectra with a retention time
within the query's retention time +/- (\code{toleranceRt} + \code{percentRt}\% of the
query's retention time) are considered. Note that while for \code{ppm} and
\code{tolerance} only a single value is accepted, \code{toleranceRt} and \code{percentRt}
can be also of length equal to the number of query spectra hence allowing
to define different rt boundaries for each query spectrum.
While these pre-filters can considerably improve performance, it should be
noted that no matches will be found between query and target spectra with
missing values in the considered variable (precursor m/z or retention
time). For target spectra without retention times (such as for \code{Spectra}
from a public reference database such as MassBank) the default
\code{toleranceRt = Inf} should thus be used.
Finally, parameter \code{THRESHFUN} allows to define a function to be applied to
the similarity scores to define which matches to report. See below for more
details.
\item \code{MatchForwardReverseParam}: performs spectra matching as with
\code{CompareSpectraParam} but reports, similar to MS-DIAL, also the \emph{reverse}
similarity score and the \emph{presence ratio}. Please refer to the
documentation of \code{CompareSpectraParam} for explanation of the parameters.
With \code{MatchForwardReverseParam}, the matching of query
spectra to target spectra is performed by considering all peaks from the
query and all peaks from the target (reference) spectrum (i.e. \emph{forward}
matching using an \emph{outer join}-based peak matching strategy). For matching
spectra also the \emph{reverse} similarity is calculated considering only peaks
present in the target (reference) spectrum (i.e. using a \emph{right join}-based
peak matching). This is reported as spectra variable \code{"reverse_score"}.
In addition, the ratio between the number of matched peaks and the total
number of peaks in the target (reference) spectra is reported as the
\emph{presence ratio} (spectra variable \code{"presence_ratio"}) and the total
number of matched peaks as \code{"matched_peaks_count"}. See examples below
for details. Parameter \code{THRESHFUN_REVERSE} allows to define an additional
\emph{threshold function} to filter matches. If \code{THRESHFUN_REVERSE} is defined
only matches with a spectra similarity fulfilling both \code{THRESHFUN} \strong{and}
\code{THRESHFUN_REVERSE} are returned. With the default
\code{THRESHFUN_REVERSE = NULL} all matches passing \code{THRESHFUN} are reported.
}
}
\examples{

library(Spectra)
library(msdata)
fl <- system.file("TripleTOF-SWATH", "PestMix1_DDA.mzML", package = "msdata")
pest_ms2 <- filterMsLevel(Spectra(fl), 2L)

## subset to selected spectra.
pest_ms2 <- pest_ms2[c(808, 809, 945:955)]

## Load a small example MassBank data set
load(system.file("extdata", "minimb.RData", package = "MetaboAnnotation"))

## Match spectra with the default similarity score (normalized dot product)
csp <- CompareSpectraParam(requirePrecursor = TRUE, ppm = 10)
mtches <- matchSpectra(pest_ms2, minimb, csp)

mtches

## Are there any matching spectra for the first query spectrum?
mtches[1]
## No

## And for the second query spectrum?
mtches[2]
## The second query spectrum matches 4 target spectra. The scores for these
## matches are:
mtches[2]$score

## To access the score for the full data set
mtches$score

## Below we use a THRESHFUN that returns for each query spectrum the (first)
## best matching target spectrum.
csp <- CompareSpectraParam(requirePrecursor = FALSE, ppm = 10,
    THRESHFUN = function(x) which.max(x))
mtches <- matchSpectra(pest_ms2, minimb, csp)
mtches

## Each of the query spectra is matched to one target spectrum
length(mtches)
matches(mtches)

## Match spectra considering also measured retention times. This requires
## that both query and target spectra have non-missing retention times.
rtime(pest_ms2)
rtime(minimb)

## Target spectra don't have retention times. Below we artificially set
## retention times to show how an additional retention time filter would
## work.
rtime(minimb) <- rep(361, length(minimb))

## Matching spectra requiring a matching precursor m/z and the difference
## of retention times between query and target spectra to be <= 2 seconds.
csp <- CompareSpectraParam(requirePrecursor = TRUE, ppm = 10,
    toleranceRt = 2)
mtches <- matchSpectra(pest_ms2, minimb, csp)
mtches
matches(mtches)

## Note that parameter `rtColname` can be used to define different spectra
## variables with retention time information (such as retention indices etc).

## A `CompDb` compound annotation database could also be used with
## parameter `target`. Below we load the test `CompDb` database from the
## `CompoundDb` Bioconductor package.
library(CompoundDb)
fl <- system.file("sql", "CompDb.MassBank.sql", package = "CompoundDb")
cdb <- CompDb(fl)
res <- matchSpectra(pest_ms2, cdb, CompareSpectraParam())

## We do however not find any matches since the used compound annotation
## database contains only a very small subset of the MassBank.
res

## As `target` we have now however the MS2 spectra data from the compound
## annotation database
target(res)

## See the package vignette for details, descriptions and more examples,
## also on how to retrieve e.g. MassBank reference databases from
## Bioconductor's AnnotationHub.
}
\author{
Johannes Rainer, Michael Witting
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/Matched.R
\name{addMatches}
\alias{addMatches}
\alias{endoapply}
\alias{filterMatches}
\alias{matchedData}
\alias{queryVariables}
\alias{targetVariables}
\alias{Matched}
\alias{Matched-class}
\alias{[,Matched-method}
\alias{length,Matched-method}
\alias{show,Matched-method}
\alias{[,Matched,ANY,ANY,ANY-method}
\alias{matches}
\alias{target}
\alias{query,Matched-method}
\alias{targetIndex}
\alias{queryIndex}
\alias{whichTarget}
\alias{whichQuery}
\alias{$,Matched-method}
\alias{colnames,Matched-method}
\alias{scoreVariables}
\alias{queryVariables,Matched-method}
\alias{targetVariables,Matched-method}
\alias{matchedData,Matched-method}
\alias{pruneTarget}
\alias{filterMatches,Matched,missing-method}
\alias{SelectMatchesParam}
\alias{TopRankedMatchesParam}
\alias{ScoreThresholdParam}
\alias{filterMatches,Matched,SelectMatchesParam-method}
\alias{filterMatches,Matched,TopRankedMatchesParam-method}
\alias{filterMatches,Matched,ScoreThresholdParam-method}
\alias{SingleMatchParam}
\alias{filterMatches,Matched,SingleMatchParam-method}
\alias{addMatches,Matched-method}
\alias{endoapply,ANY-method}
\alias{endoapply,Matched-method}
\alias{lapply,Matched-method}
\title{Representation of generic objects matches}
\usage{
addMatches(object, ...)

endoapply(X, FUN, ...)

filterMatches(object, param, ...)

matchedData(object, ...)

queryVariables(object, ...)

targetVariables(object, ...)

Matched(
  query = list(),
  target = list(),
  matches = data.frame(query_idx = integer(), target_idx = integer(), score = numeric()),
  queryAssay = character(),
  targetAssay = character(),
  metadata = list()
)

\S4method{length}{Matched}(x)

\S4method{show}{Matched}(object)

\S4method{[}{Matched,ANY,ANY,ANY}(x, i, j, ..., drop = FALSE)

matches(object)

target(object)

\S4method{query}{Matched}(x, pattern, ...)

targetIndex(object)

queryIndex(object)

whichTarget(object)

whichQuery(object)

\S4method{$}{Matched}(x, name)

\S4method{colnames}{Matched}(x)

scoreVariables(object)

\S4method{queryVariables}{Matched}(object)

\S4method{targetVariables}{Matched}(object)

\S4method{matchedData}{Matched}(object, columns = colnames(object), ...)

pruneTarget(object)

\S4method{filterMatches}{Matched,missing}(
  object,
  queryValue = integer(),
  targetValue = integer(),
  queryColname = character(),
  targetColname = character(),
  index = integer(),
  keep = TRUE,
  ...
)

SelectMatchesParam(
  queryValue = numeric(),
  targetValue = numeric(),
  queryColname = character(),
  targetColname = character(),
  index = integer(),
  keep = TRUE
)

TopRankedMatchesParam(n = 1L, decreasing = FALSE)

ScoreThresholdParam(threshold = 0, above = FALSE, column = "score")

\S4method{filterMatches}{Matched,SelectMatchesParam}(object, param, ...)

\S4method{filterMatches}{Matched,TopRankedMatchesParam}(object, param, ...)

\S4method{filterMatches}{Matched,ScoreThresholdParam}(object, param, ...)

SingleMatchParam(
  duplicates = c("remove", "closest", "top_ranked"),
  column = "score",
  decreasing = TRUE
)

\S4method{filterMatches}{Matched,SingleMatchParam}(object, param, ...)

\S4method{addMatches}{Matched}(
  object,
  queryValue = integer(),
  targetValue = integer(),
  queryColname = character(),
  targetColname = character(),
  score = rep(NA_real_, length(queryValue)),
  isIndex = FALSE
)

\S4method{endoapply}{ANY}(X, FUN, ...)

\S4method{endoapply}{Matched}(X, FUN, ...)

\S4method{lapply}{Matched}(X, FUN, ...)
}
\arguments{
\item{object}{a \code{Matched} object.}

\item{...}{additional parameters.}

\item{X}{\code{Matched} object.}

\item{FUN}{for \code{lapply} and \code{endoapply}: user defined \code{function} that takes a
\code{Matched} object as a first parameter and possibly additional parameters
(that need to be provided in the \code{lapply} or \code{endoapply} call. For lapply
\code{FUN} can return any object while for \code{endoapply} it must return a
\code{Matched} object.}

\item{param}{for \code{filterMatches}: parameter object to select and customize
the filtering procedure.}

\item{query}{object with the query elements.}

\item{target}{object with the elements against which \code{query} has been
matched.}

\item{matches}{\code{data.frame} with columns \code{"query_idx"} (\code{integer}),
\code{"target_idx"} (\code{integer}) and \code{"score"} (\code{numeric}) representing the n:m
mapping of elements between the \code{query} and the \code{target} objects.}

\item{queryAssay}{\code{character} that needs to be specified when \code{query} is
a \code{QFeatures}. In this case, \code{queryAssay} is expected to be the name of
one of the assays in \code{query} (the one on which the matching was performed).}

\item{targetAssay}{\code{character} that needs to be specified when \code{target} is
a \code{QFeatures}. In this case, \code{targetAssay} is expected to be the name of
one of the assays in \code{target} (the one on which the matching was
performed).}

\item{metadata}{\code{list} with optional additional metadata.}

\item{x}{\code{Matched} object.}

\item{i}{\code{integer} or \code{logical} defining the \code{query} elements to keep.}

\item{j}{for \code{[}: ignored.}

\item{drop}{for \code{[}: ignored.}

\item{pattern}{for \code{query}: ignored.}

\item{name}{for \code{$}: the name of the column (or variable) to extract.}

\item{columns}{for \code{matchedData}: \code{character} vector with column names of
variables that should be extracted.}

\item{queryValue}{for \code{SelectMatchesParam}: vector of values to search for in
\code{query} (if \code{query} is 1-dimensional) or in column \code{queryColname} of
\code{query} (if \code{query} is 2-dimensional). For \code{addMatches}: either an index
in \code{query} or value in column \code{queryColname} of \code{query} defining (together
with \code{targetValue}) the pair of query and target elements for which a
match should be manually added. Lengths of \code{queryValue} and
\code{targetValue} have to match.}

\item{targetValue}{for \code{SelectMatchesParam}: vector of values to search for
in \code{target} (if \code{target} is 1-dimensional) or in column \code{targetColname} of
\code{target} (if \code{target} is 2-dimensional). For \code{addMatches}: either an
index in \code{target} or value in column \code{targetColname} of \code{target} defining
(together with \code{queryValue}) the pair of query and target elements for
which a match should be manually added. Lengths of \code{queryValue} and
\code{targetValue} have to match.}

\item{queryColname}{for \code{SelectMatchesParam}: if \code{query} is 2-dimensional it
represents the column of \code{query} against which elements of \code{queryValue}
are compared.}

\item{targetColname}{for \code{SelectMatchesParam}: if \code{query} is 2-dimensional it
represents the column of \code{target} against which elements of \code{targetValue}
are compared.}

\item{index}{for \code{SelectMatchesParam}: indices of the matches to keep (if
\code{keep = TRUE}) or to drop if (\code{keep = FALSE}).}

\item{keep}{for \code{SelectMatchesParam}: \code{logical}. If \code{keep = TRUE} the matches
are kept, if \code{keep = FALSE} they are removed.}

\item{n}{for \code{TopRankedMatchesParam}: \code{integer(1)} with number of best
ranked matches to keep for each \code{query} element.}

\item{decreasing}{for \code{TopRankedMatchesParam}: \code{logical(1)} whether scores
should be ordered increasing or decreasing. Defaults to
\code{decreasing = FALSE}.}

\item{threshold}{for \code{ScoreThresholdParam} : \code{numeric(1)} specifying the
threshold to consider for the filtering.}

\item{above}{for \code{ScoreThresholdParam} : \code{logical(1)} specifying whether
to keep matches above (\code{above = TRUE}) or below (\code{above = FALSE}) a certain
threshold.}

\item{column}{for \code{ScoreThresholdParam}: \code{character(1)} specifying the name
of the score variable to consider for the filtering (the default is
\code{column = "score"}). For \code{SingleMatchParam}: \code{character(1)} defining the
name of the column to be used for de-duplication. See description of
\code{SingleMatchParam} in the \emph{Filtering and subsetting} section for details.}

\item{duplicates}{for \code{SingleMatchParam}: \code{character(1)} defining the
\emph{de-duplication} strategy. See the description of \code{SingleMatchParam} in
the \emph{Filtering and subsetting} subsection for choices and details.}

\item{score}{for \code{addMatches}: \code{numeric} (same length than \code{queryValue}) or
\code{data.frame} (same number of rows than \code{queryValue}) specifying the scores
for the matches to add. If not specified, a \code{NA} will be used as score.}

\item{isIndex}{for \code{addMatches}: specifies if \code{queryValue} and
\code{targetValue} are expected to be vectors of indices.}
}
\value{
See individual method description above for details.
}
\description{
Matches between \emph{query} and \emph{target} generic objects can be represented by
the \code{Matched} object. By default, all data accessors work as
\emph{left joins} between the \emph{query} and the \emph{target} object, i.e. values are
returned for each \emph{query} object with eventual duplicated entries (values)
if the \emph{query} object matches more than one \emph{target} object. See also
\emph{Creation and subsetting} as well as \emph{Extracting data} sections below for
details and more information.

The \code{Matched} object allows to represent matches between one-dimensional
\code{query} and \code{target} objects (being e.g. \code{numeric} or \code{list}),
two-dimensional objects (\code{data.frame} or \code{matrix}) or more complex
structures such as \code{SummarizedExperiments} or \code{QFeatures}. Combinations of
all these different data types are also supported. Matches are represented
between elements of one-dimensional objects, or rows for two-dimensional
objects (including \code{SummarizedExperiment} or \code{QFeatures}). For
\code{\link[QFeatures:QFeatures-class]{QFeatures::QFeatures()}} objects matches to only one of the \emph{assays}
within the object is supported.
}
\section{Creation and general handling}{


\code{Matched} object is returned as result from the \code{\link[=matchValues]{matchValues()}} function.

Alternatively, \code{Matched} objects can also be created with the \code{Matched}
function providing the \code{query} and \code{target} objects as well as the \code{matches}
\code{data.frame} with two columns of integer indices defining which elements
from \emph{query} match which element from \emph{target}.
\itemize{
\item \code{addMatches}: add new matches to an existing object. Parameters
\code{queryValue} and \code{targetValue} allow to define which element(s) in
\code{query} and \code{target} should be considered matching. If \code{isIndex = TRUE},
both \code{queryValue} and \code{targetValue} are considered to be integer indices
identifying the matching elements in \code{query} and \code{target}, respectively.
Alternatively (with \code{isIndex = FALSE}) \code{queryValue} and \code{targetValue} can
be elements in columns \code{queryColname} or \code{targetColname} which can be used
to identify the matching elements. Note that in this case
\strong{only the first} matching pair is added. Parameter \code{score} allows to
provide the score for the match. It can be a numeric with the score or a
\code{data.frame} with additional information on the manually added matches. In
both cases its length (or number of rows) has to match the length of
\code{queryValue}. See examples below for more information.
\item \code{endoapply}: applies a user defined function \code{FUN} to each subset of
matches in a \code{Matched} object corresponding to a \code{query} element (i.e. for
each \code{x[i]} with \code{i} being 1 to \code{length(x)}). The results are then combined
in a single \code{Matched} object representing updated matches. Note that \code{FUN}
has to return a \code{Matched} object.
\item \code{lapply}: applies a user defined function \code{FUN} to each subset of
matches in a \code{Matched} object for each \code{query} element (i.e. to each \code{x[i]}
with \code{i} from \code{1} to \code{length(x)}). It returns a \code{list} of \code{length(object)}
elements where each element is the output of \code{FUN} applied to each subset
of matches.
}
}

\section{Filtering and subsetting}{

\itemize{
\item \code{[}: subset the object selecting \code{query} object elements to keep with
parameter \code{i}. The resulting object will contain all the matches
for the selected query elements. The \code{target} object will by default be
returned as-is.
\item \code{filterMatches}: filter matches in a \code{Matched} object using different
approaches depending on the class of \code{param}:
\itemize{
\item \code{ScoreThresholdParam}: keeps only the matches whose score is strictly
above or strictly below a certain threshold (respectively when parameter
\code{above = TRUE} and \code{above = FALSE}). The name of the column containing
the scores to be used for the filtering can be specified with parameter
\code{column}. The default for \code{column} is \code{"score"}. Such variable is present
in each \code{Matched} object. The name of other score variables (if present)
can be provided (the names of all score variables can be obtained with
\code{scoreVariables()} function). For example \code{column = "score_rt"} can be
used to filter matches based on retention time scores for \code{Matched}
objects returned by \code{\link[=matchValues]{matchValues()}} when \code{param} objects involving a
retention time comparison are used.
\item \code{SelectMatchesParam}: keeps or removes (respectively when parameter
\code{keep = TRUE} and \code{keep = FALSE}) matches corresponding to certain
indices or values of \code{query} and \code{target}. If \code{queryValue} and
\code{targetValue} are provided, matches for these value pairs are kept or
removed. Parameter index\verb{allows to filter matches providing their index in the [matches()] matrix. Note that}filterMatches\verb{removes only matches from the [matches()] matrix from the}Matched\verb{object but thus not alter the}query\code{or}target` in the object. See examples below for more
information.
\item \code{SingleMatchParam}: reduces matches to keep only (at most) a
single match per query. The deduplication strategy can be defined with
parameter \code{duplicates}:
\itemize{
\item \code{duplicates = "remove"}: all matches for query elements matching more
than one target element will be removed.
\item \code{duplicates = "closest"}: keep only the \emph{closest} match for each
query element. The closest match is defined by the value(s) of
\emph{score} (and eventually \emph{score_rt}, if present). The one match with
the smallest value for this (these) column(s) is retained. This is
equivalent to \code{TopRankedMatchesParam(n = 1L, decreasing = FALSE)}.
\item \code{duplicates = "top_ranked"}: select the \emph{best ranking} match for each
query element. Parameter \code{column} allows to specify the column by
which matches are ranked (use \code{targetVariables(object)} or
\code{scoreVariables(object)} to list possible columns). Parameter
\code{decreasing} allows to define whether the match with the highest
(\code{decreasing = TRUE}) or lowest (\code{decreasing = FALSE}) value in
\code{column} for each \emph{query} will be selected.
}
\item \code{TopRankedMatchesParam}: for each query element the matches are ranked
according to their score and only the \code{n} best of them are kept (if \code{n}
is larger than the number of matches for a given query element all the
matches are returned). For the ranking (ordering) R's \code{rank} function is
used on the absolute values of the scores (variable \code{"score"}), thus,
smaller score values (representing e.g. smaller differences between
expected and observed m/z values) are considered \emph{better}. By
setting parameter \code{decreasing = TRUE} matches can be ranked in decreasing
order (i.e. higher scores are ranked higher and are thus selected).
If besides variable \code{"score"} also variable \code{"score_rt"} is available in
the \code{Matched} object (which is the case for the \code{Matched} object
returned by \code{\link[=matchValues]{matchValues()}} for \code{param} objects involving a retention
time comparison), the ordering of the matches is based on the product of
the ranks of the two variables (ranking of retention time differences
is performed on the absolute value of \code{"score_rt"}). Thus, matches with
small (or, depending on parameter \code{decreasing}, large) values for
\code{"score"} \strong{and} \code{"score_rt"} are returned.
}
\item \code{pruneTarget}: \emph{cleans} the object by removing non-matched
\strong{target} elements.
}
}

\section{Extracting data}{

\itemize{
\item \code{$} extracts a single variable from the \code{Matched} \code{x}. The variables that
can be extracted can be listed using \code{colnames(x)}. These variables can
belong to \emph{query}, \emph{target} or be related to the matches (e.g. the
score of each match). If the \emph{query} (\emph{target}) object is two dimensional,
its columns can be extracted (prefix\code{ "target_"} is used for columns in the
\emph{target} object) otherwise if \emph{query} (\emph{target}) has only a single
dimension (e.g. is a \code{list} or a \code{character}) the whole object can be
extracted with \code{x$query} (\code{x$target}). More precisely, when
\emph{query} (\emph{target}) is a \code{SummarizedExperiment} the columns from
\code{rowData(query)} (rowData(\code{target})) are extracted; when \emph{query} (\emph{target})
is a \code{\link[QFeatures:QFeatures-class]{QFeatures::QFeatures()}} the columns from \code{rowData} of the assay
specified in the \code{queryAssay} (\code{targetAssay}) slot are extracted.
The matching scores
are available as \emph{variable} \code{"score"}. Similar to a left join between the
query and target elements, this function returns a value for each query
element, with eventual duplicated values for query elements matching more
than one target element. If variables from the target \code{data.frame} are
extracted, an \code{NA} is reported for the entries corresponding to \emph{query}
elements that don't match any target element. See examples below for
more details.
\item \code{length} returns the number of \strong{query} elements.
\item \code{matchedData} allows to extract multiple variables contained in the
\code{Matched} object as a \code{DataFrame}. Parameter \code{columns} allows to
define which columns (or variables) should be returned (defaults to
\code{columns = colnames(object)}). Each single column in the returned
\code{DataFrame} is constructed in the same way as in \code{$}. That is, like \code{$},
this function performs a \emph{left join} of variables from the \emph{query} and
\emph{target} objects returning all values for all \emph{query} elements
(eventually returning duplicated elements for query elements matching
multiple target elements) and the values for the target elements matched
to the respective query elements (or \code{NA} if the target element is not
matched to any query element).
\item \code{matches} returns a \code{data.frame} with the actual matching information with
columns \code{"query_idx"} (index of the element in \code{query}), \code{"target_idx"}
(index of the element in \code{target}) \code{"score"} (the score of the match) and
eventual additional columns.
\item \code{target} returns the \emph{target} object.
\item \code{targetIndex} returns the indices of the matched targets in the order they
are assigned to the query elements. The length of the returned \code{integer}
vector is equal to the total number of matches in the object. \code{targetIndex}
and \code{queryIndex} are aligned, i.e. each element in them represent a matched
query-target pair.
\item \code{query} returns the \emph{query} object.
\item \code{queryIndex} returns the indices of the query elements with matches to
target elements. The length of the returned \code{integer} vector is equal to
the total number of matches in the object. \code{targetIndex} and \code{queryIndex}
are aligned, i.e. each element in them represent a matched query-target
pair.
\item \code{queryVariables} returns the names of the variables (columns) in \emph{query}.
\item \code{scoreVariables} returns the names of the score variables stored in the
\code{Matched} object (precisely the names of the variables in \code{matches(object)}
containing the string "score" in their name ignoring the case).
\item \code{targetVariables} returns the names of the variables (columns) in \emph{target}
(prefixed with \code{"target_"}).
\item \code{whichTarget} returns an \code{integer} with the indices of the elements in
\emph{target} that match at least one element in \emph{query}.
\item \code{whichQuery} returns an \code{integer} with the indices of the elements in
\emph{query} that match at least one element in \emph{target}.
}
}

\examples{

## Creating a `Matched` object.
q1 <- data.frame(col1 = 1:5, col2 = 6:10)
t1 <- data.frame(col1 = 11:16, col2 = 17:22)
## Define matches between query row 1 with target row 2 and, query row 2
## with target rows 2,3,4 and query row 5 with target row 5.
mo <- Matched(
    q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                 target_idx = c(2L, 2L, 3L, 4L, 5L),
                                 score = seq(0.5, 0.9, by = 0.1)))
mo

## Which of the query elements (rows) match at least one target
## element (row)?
whichQuery(mo)

## Which target elements (rows) match at least one query element (row)?
whichTarget(mo)

## Extracting variable "col1" from query object .
mo$col1

## We have duplicated values for the entries of `col1` related to query
## elements (rows) matched to multiple rows of the target object). The
## value of `col1` is returned for each element (row) in the query.

## Extracting variable "col1" from target object. To access columns from
## target we have to prefix the name of the column by `"target_"`.
## Note that only values of `col1` for rows matching at least one query
## row are returned and an NA is reported for query rows without matching
## target rows.
mo$target_col1

## The 3rd and 4th query rows do not match any target row, thus `NA` is
## returned.

## `matchedData` can be used to extract all (or selected) columns
## from the object. Same as with `$`, a left join between the columns
## from the query and the target is performed. Below we extract selected
## columns from the object as a DataFrame.
res <- matchedData(mo, columns = c("col1", "col2", "target_col1",
                                   "target_col2"))
res
res$col1
res$target_col1

## With the `queryIndex` and `targetIndex` it is possible to extract the
## indices of the matched query-target pairs:
queryIndex(mo)
targetIndex(mo)

## Hence, the first match is between the query with index 1 to the target
## with index 2, then, query with index 2 is matched to target with index 2
## and so on.

## The example matched object contains all query and all target
## elements (rows). Below we subset the object keeping only query rows that
## are matched to at least one target row.
mo_sub <- mo[whichQuery(mo)]

## mo_sub contains now only 3 query rows:
nrow(query(mo_sub))

## while the original object contains all 5 query rows:
nrow(query(mo))

## Both objects contain however still the full target object:
nrow(target(mo))
nrow(target(mo_sub))

## With the `pruneTarget` we can however reduce also the target rows to
## only those that match at least one query row
mo_sub <- pruneTarget(mo_sub)
nrow(target(mo_sub))

########
## Creating a `Matched` object with a `data.frame` for `query` and a `vector`
## for `target`. The matches are specified in the same way as the example
## before.

q1 <- data.frame(col1 = 1:5, col2 = 6:10)
t2 <- 11:16
mo <- Matched(q1, t2, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
    target_idx = c(2L, 2L, 3L, 4L, 5L), score = seq(0.5, 0.9, by = 0.1)))

## *target* is a simple vector and has thus no columns. The matched values
## from target, if it does not have dimensions and hence column names, can
## be retrieved with `$target`
mo$target

## Note that in this case "target" is returned by the function `colnames`
colnames(mo)

## As before, we can extract all data as a `DataFrame`
res <- matchedData(mo)
res

## Note that the columns of the obtained `DataFrame` are the same as the
## corresponding vectors obtained with `$`
res$col1
res$target

## Also subsetting and pruning works in the same way as the example above.

mo_sub <- mo[whichQuery(mo)]

## mo_sub contains now only 3 query rows:
nrow(query(mo_sub))

## while the original object contains all 5 query rows:
nrow(query(mo))

## Both object contain however still the full target object:
length(target(mo))
length(target(mo_sub))

## Reducing the target elements to only those that match at least one query
## row
mo_sub <- pruneTarget(mo_sub)
length(target(mo_sub))

########
## Filtering `Matched` with `filterMatches`

## Inspecting the matches in `mo`:
mo$col1
mo$target

## We have thus target *12* matched to both query elements with values 1 and
## 2, and query element 2 is matching 3 target elements. Let's assume we want
## to resolve this multiple mappings to keep from them only the match between
## query 1 (column `"col1"` containing value `1`) with target 1 (value `12`)
## and query 2 (column `"col1"` containing value `2`) with target 2 (value
## `13`). In addition we also want to keep query element 5 (value `5` in
## column `"col1"`) with the target with value `15`:
mo_sub <- filterMatches(mo,
    SelectMatchesParam(queryValue = c(1, 2, 5), queryColname = "col1",
                       targetValue = c(12, 13, 15)))
matchedData(mo_sub)

## Alternatively to specifying the matches to filter with `queryValue` and
## `targetValue` it is also possible to specify directly the index of the
## match(es) in the `matches` `data.frame`:
matches(mo)

## To keep only matches like in the example above we could use:
mo_sub <- filterMatches(mo, SelectMatchesParam(index = c(1, 3, 5)))
matchedData(mo_sub)

## Note also that, instead of keeping the specified matches, it would be
## possible to remove them by setting `keep = FALSE`. Below we remove
## selected matches from the object:
mo_sub <- filterMatches(mo,
    SelectMatchesParam(queryValue = c(2, 2), queryColname = "col1",
                       targetValue = c(12, 14), keep = FALSE))
mo_sub$col1
mo_sub$target

## As alternative to *manually* selecting matches it is also possible to
## filter matches keeping only the *best matches* using the
## `TopRankedMatchesParam`. This will rank matches for each query based on
## their *score* value and select the best *n* matches with lowest score
## values (i.e. smallest difference in m/z values).
mo_sub <- filterMatches(mo, TopRankedMatchesParam(n = 1L))
matchedData(mo_sub)

## Additionally it is possible to select matches based on a threshold
## for their *score*. Below we keep matches with score below 0.75 (one
## could select matches with *score* greater than the threshold by setting
## `ScoreThresholdParam` parameter `above = TRUE`.
mo_sub <- filterMatches(mo, ScoreThresholdParam(threshold = 0.75))
matchedData(mo_sub)

########
## Selecting the best match for each `query` element with `endoapply`

## It is also possible to select for each `query` element the match with the
## lowest score using `endoapply`. We manually define a function to select
## the best match for each query and give it as input to `endoapply`
## together with the `Matched` object itself. We obtain the same results as
## in the `filterMatches` example above.

FUN <- function(x) {
    if(nrow(x@matches) > 1)
        x@matches <- x@matches[order(x@matches$score)[1], , drop = FALSE]
    x
}

mo_sub <- endoapply(mo, FUN)
matchedData(mo_sub)

########
## Adding matches using `addMatches`

## `addMatches` allows to manually add matches. Below we add a new match
## between the `query` element with a value of `1` in column `"col1"` and
## the target element with a value of `15`. Parameter `score` allows to
## assign a score value to the match.
mo_add <- addMatches(mo, queryValue = 1, queryColname = "col1",
    targetValue = 15, score = 1.40)
matchedData(mo_add)
## Matches are always sorted by `query`, thus, the new match is listed as
## second match.

## Alternatively, we can also provide a `data.frame` with parameter `score`
## which enables us to add additional information to the added match. Below
## we define the score and an additional column specifying that this match
## was added manually. This information will then also be available in the
## `matchedData`.
mo_add <- addMatches(mo, queryValue = 1, queryColname = "col1",
    targetValue = 15, score = data.frame(score = 5, manual = TRUE))
matchedData(mo_add)

## The match will get a score of NA if we're not providing any score.
mo_add <- addMatches(mo, queryValue = 1, queryColname = "col1",
    targetValue = 15)
matchedData(mo_add)

## Creating a `Matched` object with a `SummarizedExperiment` for `query` and
## a `vector` for `target`. The matches are specified in the same way as
## the example before.
library(SummarizedExperiment)
q1 <- SummarizedExperiment(
  assays = data.frame(matrix(NA, 5, 2)),
  rowData = data.frame(col1 = 1:5, col2 = 6:10),
  colData = data.frame(cD1 = c(NA, NA), cD2 = c(NA, NA)))
t1 <- data.frame(col1 = 11:16, col2 = 17:22)
## Define matches between row 1 in rowData(q1) with target row 2 and,
## rowData(q1) row 2 with target rows 2,3,4 and rowData(q1) row 5 with target
## row 5.
mo <- Matched(
    q1, t1, matches = data.frame(query_idx = c(1L, 2L, 2L, 2L, 5L),
                                target_idx = c(2L, 2L, 3L, 4L, 5L),
                                 score = seq(0.5, 0.9, by = 0.1)))
mo

## Which of the query elements (rows) match at least one target
## element (row)?
whichQuery(mo)

## Which target elements (rows) match at least one query element (row)?
whichTarget(mo)

## Extracting variable "col1" from rowData(q1).
mo$col1

## We have duplicated values for the entries of `col1` related to rows of
## rowData(q1) matched to multiple rows of the target data.frame t1. The
## value of `col1` is returned for each row in the rowData of query.

## Extracting variable "col1" from target object. To access columns from
## target we have to prefix the name of the column by `"target_"`.
## Note that only values of `col1` for rows matching at least one row in
## rowData of query are returned and an NA is reported for those without
## matching target rows.
mo$target_col1

## The 3rd and 4th query rows do not match any target row, thus `NA` is
## returned.

## `matchedData` can be used to extract all (or selected) columns
## from the object. Same as with `$`, a left join between the columns
## from the query and the target is performed. Below we extract selected
## columns from the object as a DataFrame.
res <- matchedData(mo, columns = c("col1", "col2", "target_col1",
                                  "target_col2"))
res
res$col1
res$target_col1

## The example `Matched` object contains all rows in the
## `rowData` of the `SummarizedExperiment` and all target rows. Below we
## subset the object keeping only rows that are matched to at least one
## target row.
mo_sub <- mo[whichQuery(mo)]

## mo_sub contains now a `SummarizedExperiment` with only 3 rows:
nrow(query(mo_sub))

## while the original object contains a `SummarizedExperiment` with all 5
## rows:
nrow(query(mo))

## Both objects contain however still the full target object:
nrow(target(mo))
nrow(target(mo_sub))

## With the `pruneTarget` we can however reduce also the target rows to
## only those that match at least one in the `rowData` of query
mo_sub <- pruneTarget(mo_sub)
nrow(target(mo_sub))
}
\seealso{
\code{\link[=MatchedSpectra]{MatchedSpectra()}} for matched \code{\link[Spectra:Spectra]{Spectra::Spectra()}} objects.
}
\author{
Andrea Vicini, Johannes Rainer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_standards.R
\name{createStandardMixes}
\alias{createStandardMixes}
\title{Create Standard Mixes from a Matrix of Standard Compounds}
\usage{
createStandardMixes(
  x,
  max_nstd = 10,
  min_nstd = 5,
  min_diff = 2,
  iterativeRandomization = FALSE
)
}
\arguments{
\item{x}{\code{numeric} matrix with row names representing the compounds and
columns representing different adducts. Such a matrix with m/z values for
different adducts for compounds could e.g. be created with the
\code{\link[MetaboCoreUtils:mass2mz]{MetaboCoreUtils::mass2mz()}} function.}

\item{max_nstd}{\code{numeric} number of maximum standards per group.}

\item{min_nstd}{\code{numeric} number of minimum standards per group. Only
needed when using \code{iterativeRandomization = TRUE}.}

\item{min_diff}{\code{numeric} Minimum difference for considering two values as
distinct.}

\item{iterativeRandomization}{\code{logical} default \code{FALSE}. If set to \code{TRUE},
\code{createStandardMixes} will randomly rearrange the rows of \code{x} until the user
inputs are satisfied.}
}
\value{
\code{data.frame} created by adding a column \code{group} to the input \code{x}
matrix, comprising the group number for each compound.
}
\description{
The \code{createStandardMixes} function defines groups (mixes) of compounds
(standards) with dissimilar m/z values. The expected size of the groups can
be defined with parameters \code{max_nstd} and \code{min_nstd} and the minimum required
difference between m/z values within each group with parameter \code{min_diff}.
The group assignment will be reported in an additional column in the result
data frame.
}
\details{
Users should be aware that because the function iterates through \code{x}, the
compounds at the bottom of the matrix are more complicated to group, and
there is a possibility that some compounds will not be grouped with others.
We advise specifyiong \code{iterativeRandomization = TRUE} even if it takes more
time.
}
\examples{

## Iterative grouping only
x <- matrix(c(135.0288, 157.0107, 184.0604, 206.0424, 265.1118, 287.0937,
              169.0356, 191.0176, 468.9809, 490.9628, 178.0532, 200.0352),
            ncol = 2, byrow = TRUE,
            dimnames = list(c("Malic Acid", "Pyridoxic Acid", "Thiamine",
                                "Uric acid", "dUTP", "N-Formyl-L-methionine"),
                             c("adduct_1", "adduct_2")))
result <- createStandardMixes(x, max_nstd = 3, min_diff = 2)

## Randomize grouping
set.seed(123)
x <- matrix(c(349.0544, 371.0363, 325.0431, 347.0251, 581.0416, 603.0235,
              167.0564, 189.0383, 150.0583, 172.0403, 171.0053, 192.9872,
              130.0863, 152.0682, 768.1225, 790.1044),
            ncol = 2, byrow = TRUE,
            dimnames = list(c("IMP", "UMP", "UDP-glucuronate",
                                "1-Methylxanthine", "Methionine",
                                "Dihydroxyacetone phosphate",
                                "Pipecolic acid", "CoA"),
                             c("[M+H]+", "[M+Na]+")))
result <- createStandardMixes(x, max_nstd = 4, min_nstd = 3, min_diff = 2,
                               iterativeRandomization = TRUE)

}
\author{
Philippine Louail
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hidden-aliases.R
\docType{methods}
\name{hidden_aliases}
\alias{hidden_aliases}
\title{Internal page for hidden aliases}
\value{
Not applicable
}
\description{
For S4 methods that require a documentation entry but only clutter the index.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MatchedSpectra.R
\name{MatchedSpectra}
\alias{MatchedSpectra}
\alias{MatchedSpectra-class}
\alias{spectraVariables,MatchedSpectra-method}
\alias{queryVariables,MatchedSpectra-method}
\alias{targetVariables,MatchedSpectra-method}
\alias{colnames,MatchedSpectra-method}
\alias{$,MatchedSpectra-method}
\alias{spectraData,MatchedSpectra-method}
\alias{matchedData,MatchedSpectra-method}
\alias{addProcessing,MatchedSpectra-method}
\alias{plotSpectraMirror,MatchedSpectra-method}
\alias{setBackend,MatchedSpectra,MsBackend-method}
\title{Representation of Spectra matches}
\usage{
MatchedSpectra(
  query = Spectra(),
  target = Spectra(),
  matches = data.frame(query_idx = integer(), target_idx = integer(), score = numeric())
)

\S4method{spectraVariables}{MatchedSpectra}(object)

\S4method{queryVariables}{MatchedSpectra}(object)

\S4method{targetVariables}{MatchedSpectra}(object)

\S4method{colnames}{MatchedSpectra}(x)

\S4method{$}{MatchedSpectra}(x, name)

\S4method{spectraData}{MatchedSpectra}(object, columns = spectraVariables(object))

\S4method{matchedData}{MatchedSpectra}(object, columns = spectraVariables(object), ...)

\S4method{addProcessing}{MatchedSpectra}(object, FUN, ..., spectraVariables = character())

\S4method{plotSpectraMirror}{MatchedSpectra}(
  x,
  xlab = "m/z",
  ylab = "intensity",
  main = "",
  scalePeaks = FALSE,
  ...
)

\S4method{setBackend}{MatchedSpectra,MsBackend}(object, backend, ...)
}
\arguments{
\item{query}{\code{Spectra} with the query spectra.}

\item{target}{\code{Spectra} with the spectra against which \code{query} has been
matched.}

\item{matches}{\code{data.frame} with columns \code{"query_idx"} (\code{integer}),
\code{"target_idx"} (\code{integer}) and \code{"score"} (\code{numeric}) representing the
\emph{n:m} mapping of elements between the \code{query} and the \code{target} \code{Spectra}.}

\item{object}{\code{MatchedSpectra} object.}

\item{x}{\code{MatchedSpectra} object.}

\item{name}{for \code{$}: the name of the spectra variable to extract.}

\item{columns}{for \code{spectraData}: \code{character} vector with spectra variable
names that should be extracted.}

\item{...}{for \code{addProcessing}: additional parameters for the function \code{FUN}.
For \code{plotSpectraMirror}: additional parameters passed to the plotting
functions.}

\item{FUN}{for \code{addProcessing}: function to be applied to the peak matrix
of each spectrum in \code{object}. See \code{\link[Spectra:Spectra]{Spectra::Spectra()}} for more details.}

\item{spectraVariables}{for \code{addProcessing}: \code{character} with additional
spectra variables that should be passed along to the function defined
with \code{FUN}. See \code{\link[Spectra:Spectra]{Spectra::Spectra()}} for details.}

\item{xlab}{for \code{plotSpectraMirror}: the label for the x-axis.}

\item{ylab}{for \code{plotSpectraMirror}: the label for the y-axis.}

\item{main}{for \code{plotSpectraMirror}: an optional title for each plot.}

\item{scalePeaks}{for \code{plotSpectraMirror}: \code{logical(1)} if peak intensities
(per spectrum) should be scaled to a total sum of one (per spectrum) prior
to plotting.}

\item{backend}{for \code{setBackend}: instance of an object extending
\link[Spectra:MsBackend]{Spectra::MsBackend}. See help for \code{\link[Spectra:Spectra]{Spectra::setBackend()}}
for more details.}
}
\value{
See individual method desciption above for details.
}
\description{
Matches between query and target spectra can be represented by the
\code{MatchedSpectra} object. Functions like the \code{\link[=matchSpectra]{matchSpectra()}} function will
return this type of object. By default, all data accessors work as
\emph{left joins} between the \emph{query} and the \emph{target} spectra, i.e. values are
returned for each \emph{query} spectrum with eventual duplicated entries (values)
if the query spectrum matches more than one target spectrum.
}
\section{Creation, subset and filtering}{


\code{MatchedSpectra} objects are the result object from the \code{\link[=matchSpectra]{matchSpectra()}}.
While generally not needed, \code{MatchedSpectra} objects can also be created
with the \code{MatchedSpectra} function providing the \code{query} and \code{target}
\code{Spectra} objects as well as a \code{data.frame} with the \emph{matches} between
query and target elements. This data frame is expected to have columns
\code{"query_idx"}, \code{"target_idx"} with the \code{integer} indices of query and
target objects that are \emph{matched} and a column \code{"score"} with a \code{numeric}
score for the match.

\code{MatchedSpectra} objects can be subset using:
\itemize{
\item \code{[} subset the \code{MatchedSpectra} selecting \code{query} spectra to keep with
parameter \code{i}. The \code{target} spectra will by default be returned as-is.
\item \code{pruneTarget} \emph{cleans} the \code{MatchedSpectra} object by removing non-matched
target spectra.
}

In addition, \code{MatchedSpectra} can be filtered with any of the filtering
approaches defined for \code{\link[=Matched]{Matched()}} objects: \code{\link[=SelectMatchesParam]{SelectMatchesParam()}},
\code{\link[=TopRankedMatchesParam]{TopRankedMatchesParam()}} or \code{\link[=ScoreThresholdParam]{ScoreThresholdParam()}}.
}

\section{Extracting data}{

\itemize{
\item \code{$} extracts a single spectra variable from the \code{MatchedSpectra} \code{x}. Use
\code{spectraVariables} to get all available spectra variables. Prefix
\code{"target_"} is used for spectra variables from the \emph{target} \code{Spectra}. The
matching scores are available as \emph{spectra variable} \code{"score"}.
Similar to a left join between the query and target spectra, this function
returns a value for each query spectrum with eventual duplicated values for
query spectra matching more than one target spectrum. If spectra variables
from the target spectra are extracted, an \code{NA} is reported for \emph{query}
spectra that don't match any target spectra. See examples below for more
details.
\item \code{length} returns the number of \strong{query} spectra.
\item \code{matchedData} same as \code{spectraData} below.
\item \code{query} returns the \emph{query} \code{Spectra}.
\item \code{queryVariables} returns the \code{spectraVariables} of \emph{query}.
\item \code{spectraData} returns spectra variables from the query and/or target
\code{Spectra} as a \code{DataFrame}. Parameter \code{columns} allows to define which
variables should be returned (defaults to
\code{columns = spectraVariables(object)}), spectra variable names of the target
spectra need to be prefixed with \code{target_} (e.g. \code{target_msLevel} to get
the MS level from target spectra). The score from the matching function is
returned as spectra variable \code{"score"}. Similar to \code{$}, this function
performs a \emph{left join} of spectra variables from the \emph{query} and \emph{target}
spectra returning all values for all query spectra (eventually returning
duplicated elements for query spectra matching multiple target spectra)
and the values for the target spectra matched to the respective query
spectra. See help on \code{$} above or examples below for details.
\item \code{spectraVariables} returns all available spectra variables in the \emph{query}
and \emph{target} spectra. The prefix \code{"target_"} is used to label spectra
variables of target spectra (e.g. the name of the spectra variable for the
MS level of target spectra is called \code{"target_msLevel"}).
\item \code{target} returns the \emph{target} \code{Spectra}.
\item \code{targetVariables} returns the \code{spectraVariables} of \emph{target} (prefixed
with \code{"target_"}).
\item \code{whichTarget} returns an \code{integer} with the indices of the spectra in
\emph{target} that match at least on spectrum in \emph{query}.
\item \code{whichQuery} returns an \code{integer} with the indices of the spectra in
\emph{query} that match at least on spectrum in \emph{target}.
}
}

\section{Data manipulation and plotting}{

\itemize{
\item \code{addProcessing}: add a processing step to both the \emph{query} and \emph{target}
\code{Spectra} in \code{object}. Additional parameters for \code{FUN} can be passed \emph{via}
\code{...}. See \code{addProcessing} documentation in \code{\link[Spectra:Spectra]{Spectra::Spectra()}} for more
information.
\item \code{plotSpectraMirror}: creates a mirror plot between the query and each
matching target spectrum. Can only be applied to a \code{MatchedSpectra} with a
single query spectrum. Setting parameter \code{scalePeaks = TRUE} will scale
the peak intensities per spectrum to a total sum of one for a better
graphical visualization. Additional plotting parameters can be passed
through \code{...}. The parameters \code{ppm} and \code{tolerance} can be used to
define the m/z tolerance for matching peaks between the query and target
spectra. If not provided by the user, the values from the \code{param}
object used to create the \code{MatchedSpectra} object are used; if these are
missing, the default values (\code{ppm =20} and \code{tolerance = 0}) are used.
\item \code{setBackend}: allows to change the \emph{backend} of both the query and target
\code{\link[Spectra:Spectra]{Spectra::Spectra()}} object. The function will return a \code{MatchedSpectra}
object with the query and target \code{Spectra} changed to the specified
\code{backend}, which can be any backend extending \link[Spectra:MsBackend]{Spectra::MsBackend}.
}
}

\examples{

## Creating a dummy MatchedSpectra object.
library(Spectra)
df1 <- DataFrame(
    msLevel = 2L, rtime = 1:10,
    spectrum_id = c("a", "b", "c", "d", "e", "f", "g", "h", "i", "j"))
df2 <- DataFrame(
    msLevel = 2L, rtime = rep(1:10, 20),
    spectrum_id = rep(c("A", "B", "C", "D", "E"), 20))
sp1 <- Spectra(df1)
sp2 <- Spectra(df2)
## Define matches between query spectrum 1 with target spectra 2 and 5,
## query spectrum 2 with target spectrum 2 and query spectrum 4 with target
## spectra 8, 12 and 15.
ms <- MatchedSpectra(
    sp1, sp2, matches = data.frame(query_idx = c(1L, 1L, 2L, 4L, 4L, 4L),
                                   target_idx = c(2L, 5L, 2L, 8L, 12L, 15L),
                                   score = 1:6))

## Which of the query spectra match at least one target spectrum?
whichQuery(ms)

## Extracting spectra variables: accessor methods for spectra variables act
## as "left joins", i.e. they return a value for each query spectrum, with
## eventually duplicated elements if one query spectrum matches more than
## one target spectrum.

## Which target spectrum matches at least one query spectrum?
whichTarget(ms)

## Extracting the retention times of the query spectra.
ms$rtime

## We have duplicated retention times for query spectrum 1 (matches 2 target
## spectra) and 4 (matches 3 target spectra). The retention time is returned
## for each query spectrum.

## Extracting retention times of the target spectra. Note that only retention
## times for target spectra matching at least one query spectrum are returned
## and an NA is reported for query spectra without matching target spectrum.
ms$target_rtime

## The first query spectrum matches target spectra 2 and 5, thus their
## retention times are returned as well as the retention time of the second
## target spectrum that matches also query spectrum 2. The 3rd query spectrum
## does match any target spectrum, thus `NA` is returned. Query spectrum 4
## matches target spectra 8, 12, and 15, thus the next reported retention
## times are those from these 3 target spectra. None of the remaining 6 query
## spectra matches any target spectra and thus `NA` is reported for each of
## them.

## With `queryIndex` and `targetIndex` it is possible to extract the indices
## of the matched query-index pairs
queryIndex(ms)
targetIndex(ms)

## The first match is between query index 1 and target index 2, the second
## match between query index 1 and target index 5 and so on.
## We could use these indices to extract a `Spectra` object containing only
## matched target spectra and assign a spectra variable with the indices of
## the query spectra
matched_target <- target(ms)[targetIndex(ms)]
matched_target$query_index <- queryIndex(ms)

## This `Spectra` object thus contains information from the matching, but
## is a *conventional* `Spectra` object that could be used for further
## analyses.

## `spectraData` can be used to extract all (or selected) spectra variables
## from the object. Same as with `$`, a left join between the specta
## variables from the query spectra and the target spectra is performed. The
## prefix `"target_"` is used to label the spectra variables from the target
## spectra. Below we extract selected spectra variables from the object.
res <- spectraData(ms, columns = c("rtime", "spectrum_id",
    "target_rtime", "target_spectrum_id"))
res
res$spectrum_id
res$target_spectrum_id

## Again, all values for query spectra are returned and for query spectra not
## matching any target spectrum NA is reported as value for the respecive
## variable.

## The example matched spectra object contains all query and all target
## spectra. Below we subset the object keeping only query spectra that are
## matched to at least one target spectrum.
ms_sub <- ms[whichQuery(ms)]

## ms_sub contains now only 3 query spectra:
length(query(ms_sub))

## while the original object contains all 10 query spectra:
length(query(ms))

## Both object contain however still the full target `Spectra`:
length(target(ms))
length(target(ms_sub))

## With the `pruneTarget` we can however reduce also the target spectra to
## only those that match at least one query spectrum
ms_sub <- pruneTarget(ms_sub)
length(target(ms_sub))
}
\seealso{
\code{\link[=Matched]{Matched()}} for additional functions available for \code{MatchedSpectra}.
}
\author{
Johannes Rainer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R
\name{matchSpectra}
\alias{matchSpectra}
\title{Spectral matching}
\usage{
matchSpectra(query, target, param, ...)
}
\arguments{
\item{query}{\link[Spectra:Spectra]{Spectra::Spectra} object with the (experimental) spectra.}

\item{target}{MS data to compare against. Can be another \link[Spectra:Spectra]{Spectra::Spectra}.}

\item{param}{parameter object containing the settings for the matching (e.g.
eventual prefiltering settings, cut-off value for similarity above which
spectra are considered matching etc).}

\item{...}{optional parameters.}
}
\value{
a \link{MatchedSpectra} object with the spectra matching results.
}
\description{
The \code{matchSpectra} method matches (compares) spectra from \code{query} with those
from \code{target} based on settings specified with \code{param} and returns the result
from this as a \link{MatchedSpectra} object.
}
\seealso{
\code{\link[=CompareSpectraParam]{CompareSpectraParam()}} for the comparison between
\link[Spectra:Spectra]{Spectra::Spectra} objects.
}
\author{
Johannes Rainer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CompDbSource.R
\name{CompDbSource}
\alias{CompDbSource}
\alias{CompDbSource-class}
\alias{metadata,CompDbSource-method}
\alias{show,CompDbSource-method}
\alias{MassBankSource}
\title{Compound Annotation Sources for \code{CompDb} databases}
\usage{
CompDbSource(dbfile = character())

\S4method{metadata}{CompDbSource}(x, ...)

\S4method{show}{CompDbSource}(object)

MassBankSource(release = "2021.03", ...)
}
\arguments{
\item{dbfile}{\code{character(1)} with the database file (including the full
path).}

\item{x}{A \code{CompDbSource} object.}

\item{...}{For \code{CompDbSource}: ignored. For \code{MassBankSource}: optional
parameters passed to the \code{AnnotationHub} constructor function.}

\item{object}{A \code{CompDbSource} object.}

\item{release}{A \code{character(1)} defining the version/release of MassBank that
should be used.}
}
\description{
\code{CompDbSource} objects represent references to \link[CompoundDb:CompDb]{CompoundDb::CompDb}
database-backed annotation resources. Instances are expected to be
created with the dedicated construction functions such as
\code{MassBankSource} or the generic \code{CompDbSource}. The annotation data is
not stored within the object but will be accessed/loaded within the
object's \code{matchSpectra} method.

New \code{CompDbSource} objects can be created using the functions:
\itemize{
\item \code{CompDbSource}: create a new \code{CompDbSource} object from an existing
\code{CompDb} database. The (SQLite) database file (including the full path)
needs to be provided with parameter \code{dbfile}.
\item \code{MassBankSource}: retrieves a \code{CompDb} database for the specified MassBank
release from Bioconductor's online \code{AnnotationHub} (if it exists) and
uses that. Note that \code{AnnotationHub} resources are cached locally and thus
only downloaded the first time.
The function has parameters \code{release} which allows to define the desired
MassBank release (e.g. \code{release = "2021.03"} or \code{release = "2022.06"})
and \code{...} which allows to pass optional parameters to the \code{AnnotationHub}
constructor function, such as \code{localHub = TRUE} to use only the cached
data and avoid updating/retrieving updates from the internet.
}

Other functions:
\itemize{
\item \code{metadata}: get metadata (information) on the annotation resource.
}
}
\examples{

## Locate a CompDb SQLite database file. For this example we use the test
## database from the `CompoundDb` package.
fl <- system.file("sql", "CompDb.MassBank.sql", package = "CompoundDb")
ann_src <- CompDbSource(fl)

## The object contains only the reference/link to the annotation resource.
ann_src

## Retrieve a CompDb with MassBank data for a certain MassBank release
mb_src <- MassBankSource("2021.03")
mb_src
}
\author{
Johannes Rainer
}
